#!/bin/sh

DEBUG=1
ISMPY_PATH=$HOME/.ismpy
PYTHON=python3
modules="Flask pyaml numpy matplotlib h5py Flask-Caching cheroot"

usage() {
    cat << EOF

usage: sh install_ismpy.sh [-d]

    -d causes install_ismpy.sh to uninstall ismpy if installed.

EOF
}

uninstall() {
    [ ! -d $ISMPY_PATH ] && echo "ismpy not installed" && return
    echo "uninstalling ismpy..."
    rm -rf $ISMPY_PATH
    echo "done"
}

[ $# -gt 1 -o "$1" = "-h" ] && usage && exit


which $PYTHON > /dev/null 2>&1
if [ $? -ne 0 ]; then
    echo "$PYTHON is not installed or not in the PATH, install it and re-try"
    exit 1
fi
 

if [ "$1" = "-d" ]; then
    uninstall
    exit
fi


if [ ! -d  $ISMPY_PATH ]; then
    echo "ismpy not installed"
    error=0
    echo "installing ismpy in $ISMPY_PATH..."
    # bug in python 3.5
    if [ $(expr "$($PYTHON --version)" : ".*3.5") -gt 0 ]; then
        echo "mitigating a bug in python3.5 venv..."
        $PYTHON -m venv $ISMPY_PATH
        $PYTHON -m venv --system-site-packages $ISMPY_PATH
    else
        $PYTHON -m venv --system-site-packages $ISMPY_PATH
    fi

    error=$(( $error + $? ))

    echo -n "installing ismpy modules..."
    for m in $modules; do
        printf "\n  %s" "installing $m..."
        if [ $DEBUG -eq 1 ]; then
            $ISMPY_PATH/bin/pip install $m
        else
            $ISMPY_PATH/bin/pip install $m > /dev/null 2>&1
        fi

        if [ $? -eq 0 ]; then
            printf "%s" "ok"
        else
            printf "%s" "nok"
            error=$(( $error + 1 ))
        fi

    done
    echo ""

    if [ $error -eq 0 ]; then
        printf "%s\n  %s\n\n" "you can now launch the ISM python software with the following python instance:" "$ISMPY_PATH/bin/$PYTHON"
    else
        echo "sorry, an error occured during the installation of ismpy"
    fi
fi

