# -*- coding : utf-8 -*-

import sys, os
import json
sys.path.append( "../server/" )
import base_test as test
import analyzer

   
class TestAnalyzer(test.Test):
  
  def __init__(self):    
    super(TestAnalyzer, self).__init__()     
    self.testFile = "test_c_1.hdf5"
    
  def doTests(self):
    self.testInstanciation()
    self.testFileLoading()
    self.testGetTempetaruresAvs()
    self.testGetReactions()
    self.testGetGasConditions()
    self.testGetSpecies()
    return self.getLogString()
    

  def baseTest(self, doTest, source=None) :
    if source is not None : 
      self.resultAppend("do test : %s "% source)    
    try : 
      doTest()
    except AssertionError as e:
      self.addProblem()
      self.resultAppend("assertion error : %s "%e)
    except Exception as e : 
      self.addProblem()
      self.resultAppend("error : %s "%e)       
    
    
  def testInstanciation(self):  

    def test1():
    # default directory  
      try : 
        analyzer.Analyzer('../data')
      except Exception as e:
        self.addProblem()
        self.resultAppend("assertion error : %s "%e)
      
    self.baseTest(test1, "testInstanciation")      
          
 
      
      
  def testFileLoading(self):  
    
    def test1() : 
      # load test file
      try : 
        test_analyzer = analyzer.Analyzer('../data')
        test_analyzer.loadHdf(self.testFile)
      except Exception as e:
        self.addProblem()
        self.resultAppend("assertion error : %s "%e)
        
    self.baseTest(test1, "testFileLoading")   
    
  def testGetTempetaruresAvs(self):
    
    def test1() :
      test_analyzer = analyzer.Analyzer('../data')
      test_analyzer.loadHdf(self.testFile)
      
      assert('avs' in test_analyzer.getTemperaturesAvs()), "avs are missing"
      assert('temperatures' in test_analyzer.getTemperaturesAvs()), "temperatures are missing"
      assert(len(test_analyzer.getTemperaturesAvs()['avs']) == len(test_analyzer.hdf.getTemperaturesAvs()['avs'])), "Length of av values is not correct, found : % should be : %s"%(len(test_analyzer.getAvs), len(test_analyzer.hdf.getAllAv))
      assert(test_analyzer.getTemperaturesAvs()['avs'][0] == 0), "av[0] is not correct : %s"%test_analyzer.getTemperaturesAvs()['avs'][0]
      
    self.baseTest(test1, "testGetTempetaruresAvs")      
    
  def testGetGasConditions(self):
    
    def test1() :
      test_analyzer = analyzer.Analyzer('../data')
      test_analyzer.loadHdf(self.testFile)
      data = test_analyzer.getGasConditions(0)
      
      for key, value in data.items() : 
        assert(key == value['name'] ), "error in property name : %s != %s"%(key, value['name'])
        assert(isinstance(value['value'], float) ), "property value should be a float : %s found"%(type(value['value']))
              
    self.baseTest(test1, "testGetGasConditions")    
    
  def testGetSpecies(self):
    
    def test1() :
      test_analyzer = analyzer.Analyzer('../data')
      test_analyzer.loadHdf(self.testFile)
      data = test_analyzer.getSpecies()
      
      assert(len(data) > 0), "species list is empty"
      assert('H2' in data), 'H2 species was not found in data file'
      
    self.baseTest(test1, "testGetSpecies")    
      
      
      
  def testGetReactions(self):
    
    def test1() :
      test_analyzer = analyzer.Analyzer('../data')
      test_analyzer.loadHdf(self.testFile)
      species =  test_analyzer.getSpecies()
      element = 'H2'
      result = test_analyzer.getReactions(element, 0)   
      
      for key in result.keys() :
        assert(key == element), "Searched element has not been found : %s should be %s"%(key, element)
        assert('abundances' in result[element]), "abundances are missing in result"
        assert('abundance' in result[element]), 'element abudance is missing in result'
        assert('abundance_unit' in result[element]), 'abundance unit is missing'
        assert('creation' in result[element]), 'list of formation reactions is missing'
        assert('destruction' in result[element]), 'list of destruction reactions id missing'

      for element in species:
        result = test_analyzer.getReactions(element, 0)   
        for key in result.keys() :
          assert( len(test_analyzer.hdf.getDestructionBySpecies(element)) == len(result[element]['destruction']) ), "destruction reactions are missing in analyzed data, found %s, should be %s" % (len(test_analyzer.hdf.getDestructionBySpecies(element)), len(result[element]['destruction']) )
          assert( len(test_analyzer.hdf.getFormationBySpecies(element)) == len(result[element]['creation']) ), "formation reactions are missing in analyzed data, found %s, should be %s" % (len(test_analyzer.hdf.getDestructionBySpecies(element)), len(result[element]['destruction']) )      
        
    self.baseTest(test1, "testGetReactions")      

