
"use strict";

let getSuccessObject = function(){
  return {"status" : "success"};
}

let getErrorObject = function(message){
  return {"status" : "error", "message" : message};
}


//
// Configuration variables
//
var config = {
  service_url : "http://localhost:12345/chemistry-analyzer",
  main_pos_var_step : 1
}

//
// Ajax functions calling server
//
var ajax = {
  //allSpecies : [],
  sessionId : null,   // session id, created by client and submitted to server
                                      // ok while local application


  getData : function(species, main_pos_var){
    //var spinner = null;
    var self = this;
    return new Promise(function(resolve, reject){
      let url = config.service_url+"/session/"+self.sessionId+"/mainposvar/"+main_pos_var+"/species/"+species;
      let req = new XMLHttpRequest();
      req.open("GET", url);

      if (page.isNewFile){
        mol_network.removeGraph();
        //page.showSpinner();
        page.isNewFile = false;
      }

      req.onload = function(){
        if(req.status == 200){
          let data = JSON.parse(req.response);
          mol_network.species = data;
          mol_network.updateGraph(species, page.thresholdInput.value/100, mol_network.current_threshold_type);

          page.enableButtons();
          page.toogleBackButton(function(){if(mol_network.species_history.length >= 2) return true; return false});
          page.updateSpeciesDisplay(species,
                                    data[species]["abundance"],
                                    data[species]["abundance_unit"]);

          let pos_var = parseInt(page.mainPosVarInput.value);
          page.updateInfosDisplay(  mol_network.nodes,
                                    data[species]["abundance"],
                                    data[species]["total_formation_rates"][pos_var],
                                    data[species]["total_destruction_rates"][pos_var],
                                    data[species]["total_rates_unit"]);

          if(mol_network.network_events === false){
            mol_network.network.on('hoverNode', function (hovered) {
              if(mol_network.nodes.get(hovered.node).group === node_groups.product || mol_network.nodes.get(hovered.node).group === node_groups.reactant)
                page.initReactionsTable();
                if( !(threshold_types.simple === mol_network.current_threshold_type && mol_network.current_threshold == 0)
                    && !(threshold_types.cumulated === mol_network.current_threshold_type && mol_network.current_threshold == 1)){
                  //all reactions displayed in reactions table but not on graph if threshold == 0 or 100
                  page.highlightReaction('reaction-'+hovered.node);
                }
            });
            mol_network.network.on('select', function (properties) {
              var species = null;
              if(properties.nodes.length == 1 ){ // click on a node
                species = mol_network.nodes.get(properties.nodes[0]).label;
              }else if(properties.edges.length == 1){ // click on an edge
                species = mol_network.edges.get(properties.edges[0]).label;
              }
              if(species != null && page.allSpecies[species] != undefined){ //no request for grain/photon ...
                page.goBackSpecies(species);
                self.getData(species, page.mainPosVarInput.value).then((success)=>{}, (error)=>{
                  alert(error.message);
                });
                self.getAbundances(species).then((success)=>{}, (error)=>{
                  alert(error.message);
                });
              }
            });

          }
          mol_network.network_events = true;
          page.clearSavedFiles();
          page.hideSpinner();
          resolve(getSuccessObject())
        }else{
          reject(getErrorObject("Error while querying server"))
        }
      };
      req.send();
    });
  },

  getGasConditions : function(main_pos_var){
    let self = this;
    return new Promise(function(resolve, reject){
      let url = config.service_url+"/session/"+self.sessionId+"/gasconditions/mainposvar/"+main_pos_var;
      let req = new XMLHttpRequest();
      req.open("GET", url);
      req.onload = function(){
        if(req.status == 200){
          let data = JSON.parse(req.response);
          var result = "<ul>";
          var others = [];
          var temperatures = [];
          var densities = [];
          for(var cond in data){
            if(cond.indexOf("nd") > -1 && cond !== "nd(e-)")  {
              densities.push(cond);
            }else if (cond.indexOf("Td") > -1){
              temperatures.push(cond);
            }else{
              others.push(cond);
            }
          }
          temperatures.sort();
          densities.sort();

          page.updateGasConditionsDisplay(datalist_to_html(others, data));
          page.updateGrainConditionsDisplay(datalist_to_html(temperatures, data), datalist_to_html(densities, data));
          resolve(getSuccessObject());
        }else{
          reject(getErrorObject(("Error while querying server")));
        }
      };
      req.send();
    });
  },

  getMainPosVars : function(){
    var self = this;
    return new Promise(function(resolve, reject){
      let url =  config.service_url+"/session/"+self.sessionId+"/allmainposvars";
      let req = new XMLHttpRequest();
      req.open("GET", url);
      page.showSpinner();
      if (page.isNewFile){
        mol_network.removeGraph();
        page.isNewFile = false;
      }
      req.onload = function(){
        if(req.status == 200){
          let data = JSON.parse(req.response);
          page.updateMainPosVarInput(data['main_pos_vars']);
          plot2d.plotTemperature(data['main_pos_vars'], data['temperatures'], config.main_pos_var_step);
          page.hideSpinner();
          resolve(getSuccessObject());
        }else{
          page.hideSpinner();
          reject(getErrorObject("Error while querying server"));
        }
      };
      req.send();
    });
  },

  saveNetwork : function(filename){
    var self = this;
    jQuery.ajax({
        type: "GET",
        url: config.service_url+"/session/"+self.sessionId+"/network",
        dataType: 'json',
        success: function(data){
          self.saveData(filename, data, page.displayNetworkFile);
        },
        error: function(jqXHR, textStatus, errorThrown ){
          alert("Error while querying server");
        }
    });
  },

  getSpecies : function(){
    var self = this;
    return new Promise(function(resolve, reject){
      let url = config.service_url+"/session/"+self.sessionId+"/allspecies";
      let req = new XMLHttpRequest();
      req.open("GET", url);
      req.onload = function(){
        if(req.status == 200){
          let data = JSON.parse(req.response);
          page.updateSpeciesList(data.sort())
          page.triggerEvent(page.speciesList, new Event('change'));
          resolve(getSuccessObject());
        }else{
          reject(getErrorObject("Error while querying server"));
        }
      };
      req.send();
    });
  },

  saveData : function(filename, data, success_callback){
    jQuery.ajax({
        type: "POST",
        url: config.service_url+"/savefile",
        dataType: 'json',
        data: { 'filename' : filename, 'string_data' : data},
        success: function(data){
          success_callback(config.service_url, filename);
        },
        error: function(jqXHR, textStatus, errorThrown ){
          alert("Error while writing file");
        }
    });
  },

  checkAndSaveData : function(filename, savefunction){
    var svfunction = savefunction;
    return new Promise(function(resolve, reject){
      let url = config.service_url+"/existfile/"+filename;
      let req = new XMLHttpRequest();
      req.open("GET", url);
      req.onload = function(){
        if(req.status == 200){
          // checks if file already exists
          let response = JSON.parse(req.response);
          var answer = false;
          if(response['status'] === true){
            answer = confirm ("File already exists. Overwrite ?");
          }
          //writes if not
          if(response['status'] === false || answer === true){
            svfunction();
          }
          resolve(getSuccessObject());
        }else{
          reject(getErrorObject("Error while checking file status"));
        }
      };
      req.send();
    });
  },

  saveRatesMainPosVarData : function(species, threshold, filename){
    let self = this;
    return new Promise(function(resolve, reject){
      page.showSpinner();
      let url = config.service_url+"/session/"+self.sessionId+"/saverates/species/"+species+"/threshold/"+threshold+"/filename/"+filename;
      let req = new XMLHttpRequest();
      req.open("GET", url);
      req.onload = function(){
        if(req.status == 200){
          let data = JSON.parse(req.response);
          page.hideSpinner();
          page.displayFormationDestructionFiles(config.service_url, filename, data['formation'], data['destruction']);
          resolve(getSuccessObject());
        }else{
          page.hideSpinner();
          resolve(getErrorObject("Error while writing file"));
        }
      };
      req.send();
    });
  },

  getAbundances : function(species){
    let self = this;
    return new Promise(function(resolve, reject){
      let url = config.service_url+"/session/"+self.sessionId+"/abundances/species/"+species;
      let req = new XMLHttpRequest();
      req.open("GET", url);
      req.onload = function(){
        if(req.status == 200){
          let data = JSON.parse(req.response);
          plot2d.plotDensity(page.allMainPosVars, data, config.main_pos_var_step, species);
          resolve(getSuccessObject);
        }else{
          resolve(getErrorObject("Error while querying server"));
        }
      };
      req.send();
    });
  },

  setMainPositionVar : function(){
    let self = this;
    return new Promise(function(resolve, reject){
      let url = config.service_url+"/session/"+self.sessionId+"/positionvarname";
      let req = new XMLHttpRequest();
      req.open("GET", url);
      req.onload = function(){
        if(req.status == 200){
          let data = JSON.parse(req.response);
          page.setMainPositionVarName(data['name'], data['unit']);
          resolve(getSuccessObject());
        }else{
          reject(getErrorObject("Error while querying server"))
        }
      };
      req.send();
    });
  },

  getFiles : function(){
    return new Promise(function(resolve, reject){
      let url = config.service_url+"/allfiles";
      let req = new XMLHttpRequest();
      req.open("GET", url);
      req.onload = function(){
        if(req.status == 200){
          let data = JSON.parse(req.response);
          page.updateFilesList(data);
          resolve(getSuccessObject());
        }else{
          reject(getErrorObject("Error while querying server"))
        }
      };
      req.send();
    });
  },

  loadFile : function(filename){
    let self = this;
    return new Promise(function(resolve, reject){
      if( self.sessionId != null ){
        let url = config.service_url+"/session/"+self.sessionId+"/changefile/file/"+filename;
        let req = new XMLHttpRequest();
        req.open("GET", url);
        req.onload = function(){
          if(req.status == 200){
            self.getMainPosVars().then((success)=>{}, (error)=>{
              alert(error.message);
            });
            self.getSpecies().then((success)=>{}, (error)=>{
              alert(error.message);
            });
            page.setModelName(filename);
            resolve(getSuccessObject());
          }else if(req.status == 204){
            reject(getErrorObject("File not found"));
          }else{
            let response = JSON.parse(req.responseText);
            if(response.status='error'){
              reject(getErrorObject(response.message));
            }else{
              reject(getErrorObject("Server error."));
            }
          }
        };
        req.send();
      }else{
        reject(getErrorObject("No session id"));
      }
    });
  },

  runSession : function(toexec){
    let self = this;
    return new Promise(function(resolve, reject){
      let url = config.service_url+"/opensession";
      let req = new XMLHttpRequest();
      req.open("GET", url);
      req.onload = function(){
        let response = req.response;
        if(req.status == 200){
          let data = JSON.parse(response);
          self.sessionId = data["sessionid"];
          toexec();
          resolve(getSuccessObject());
        }else{
          reject(getErrorObject("Unable to open a session. Is the server running ? "))
        }
      };
      req.send();
    });

  }
}
