"use strict";

/**
 * pad a string to reach desired size
 */
function pad(pad, str, padLeft) {
  if (typeof str === 'undefined')
    return pad;
  if (padLeft) {
    return (pad + str).slice(-pad.length);
  } else {
    return (str + pad).substring(0, pad.length);
  }
}

/**
 * Removes all children from a DOM element
 * @param {
 * } element 
 */
function removeAllChildren(element){
  if(element.hasChildNodes()){
    for(let i = element.childNodes.length - 1; i >= 0; i--){
      element.childNodes[i].remove();
    }
  }
}

var DOMUtil ={

  getTable : function(id, css_class){
    let table =  document.createElement("table");
    table.id = id;
    table.className = css_class;
    return table;
  },

  getTr : function(id, css_class){
    let tr = document.createElement("tr");
    tr.id = id;
    tr.className = css_class;
    return tr;
  },

  getTd : function(css_class, content){
    let td = document.createElement('td');
    td.className = css_class;
    td.innerHTML = content;
    return td;
  },

  getTh : function(css_class, content, colspan){
    let th = document.createElement('th');
    th.className = css_class;
    th.innerHTML = content;
    th.colSpan = colspan;
    return th;
  },

  getSpan : function(css_class, title, content){
    let span = document.createElement("span");
    span.className = css_class;
    span.innerHTML = content;
    span.title = title;
    return span;
  }


}

var formatter = {

  dataAsHtml : function(nodes, abundance, total_form_rate, total_dest_rate, rates_unit){
    let reactions = {formation : [], destruction : []};
    let node = null;
    let ids = nodes.getIds();
    let mainspecies = "";
    for (let i = 0, len = ids.length; i < len; i++ ){
        let tr = null; 
        node = nodes.get(ids[i]);
        if(node.metadata.reaction !== undefined && node.group !== node_groups.legend){
          tr = DOMUtil.getTr("reaction-"+ids[i], "");
          tr.appendChild(DOMUtil.getTd("left-align", this.formatReactionHtml(node.metadata.reaction)));
          tr.appendChild(DOMUtil.getTd("", str_to_exp_number(node.metadata.abundances[0], 2)));
          tr.appendChild(DOMUtil.getTd("", str_to_exp_number(node.metadata.abundances[1], 2)));
          tr.appendChild(DOMUtil.getTd("", str_to_exp_number(get_constant(node.metadata.rate, node.metadata.abundances), 2)));
          tr.appendChild(DOMUtil.getTd("", str_to_exp_number(node.metadata.rate, 2)));
          tr.appendChild(DOMUtil.getTd("", node.metadata.proportion));
        }

        // node group types : reactant, product, node_legend, mainspecies
        if(node.group === node_groups.product){
          reactions.destruction.push(tr);
        }

        if(node.group === node_groups.reactant){
          reactions.formation.push(tr);
        }

        if(node.group === "mainspecies"){
            mainspecies = node.label;
        }
    }

    let colnbr = 7;

    let tr_title1 = DOMUtil.getTr("", "reactions-table-tr first");
    tr_title1.appendChild(DOMUtil.getTh("fl-left", "Formation of "+mainspecies,parseInt(colnbr/2)), "");
    tr_title1.appendChild(DOMUtil.getTh("fl-right", "Total formation rate "+str_to_exp_number(total_form_rate, 2)+" "+rates_unit,parseInt(colnbr/2)), "");

    let tr_title2 = DOMUtil.getTr("", "reactions-table-tr second");
    tr_title2.appendChild(DOMUtil.getTh("left-align", "A + B &#8594; C + D", ""));
    tr_title2.appendChild(DOMUtil.getTh("", "n(A)<br/>(cm-3)", ""));
    tr_title2.appendChild(DOMUtil.getTh("", "n(B)<br/>(cm-3)", ""));

    let tr_title2_th = DOMUtil.getTh("", "", "");
    tr_title2_th = DOMUtil.getTh("", "", "").appendChild(DOMUtil.getSpan("", "reaction constant", "k"));
    tr_title2.appendChild(tr_title2_th);

    tr_title2.appendChild(DOMUtil.getTh("", "rate<br/>(cm-3 s-1)", ""));
    tr_title2.appendChild(DOMUtil.getTh("", "proportion", ""));

    let formation_table = DOMUtil.getTable("formation-table", "reactions-table");
    formation_table.appendChild(tr_title1);
    formation_table.appendChild(tr_title2);
    for (let i=0; i < reactions.formation.length; i++){
      formation_table.appendChild(reactions.formation[i]);
    }


    tr_title1 = DOMUtil.getTr("", "reactions-table-tr first");
    tr_title1.appendChild(DOMUtil.getTh("fl-left", "Destruction of "+mainspecies,parseInt(colnbr/2)), "");
    tr_title1.appendChild(DOMUtil.getTh("fl-right", "Total destruction rate "+str_to_exp_number(total_dest_rate, 2)+" "+rates_unit,parseInt(colnbr/2)), "");

    tr_title2 = DOMUtil.getTr("", "reactions-table-tr second");
    tr_title2.appendChild(DOMUtil.getTh("left-align", "A + B &#8594; C + D", ""));
    tr_title2.appendChild(DOMUtil.getTh("", "n(A)<br/>(cm-3)", ""));
    tr_title2.appendChild(DOMUtil.getTh("", "n(B)<br/>(cm-3)", ""));

    tr_title2_th = DOMUtil.getTh("", "", "");
    tr_title2_th = DOMUtil.getTh("", "", "").appendChild(DOMUtil.getSpan("", "reaction constant", "k"));
    tr_title2.appendChild(tr_title2_th);

    tr_title2.appendChild(DOMUtil.getTh("", "rate<br/>(cm-3 s-1)", ""));
    tr_title2.appendChild(DOMUtil.getTh("", "proportion", ""));

    let destruction_table = DOMUtil.getTable("destruction-table", "reactions-table");
    destruction_table.appendChild(tr_title1);
    destruction_table.appendChild(tr_title2);
    for (let i=0; i < reactions.destruction.length; i++){
      destruction_table.appendChild(reactions.destruction[i]);
    }

    let result = document.createElement("div");
    result.appendChild(formation_table);
    result.appendChild(destruction_table);
    return result;
  },

  /**
   * Replaces all occurences of target by replace in string
   */
  replaceAll : function(string, target, replace){
    return string.split(target).join(replace);
  },

  /**
   * Add <sup> markup around + characters in species names
   */
  formatReactionHtml : function(reaction){
    var result = " "+reaction;

    result = result.replace("e-", "e<sup>-</sup>");

    var reg=new RegExp("_(..)", "g");
    result = result.replace(reg, "<sup>$1</sup>");

    // numbers after element name
    var reg=new RegExp("([^0-9\s])([0-9]+)", "g");
    result = result.replace(reg, "$1<sub>$2</sub>");

    // numbers before element name
    var reg=new RegExp("(\s)([0-9]+)", "g");
    result = result.replace(reg, "<sup>$1</sup>$2");

    result = this.replaceAll(result, " + ", "#");
    result = this.replaceAll(result, '+', "<sup>+</sup>");
    result = this.replaceAll(result, "#", " + ");
    result = this.replaceAll(result, " > ", " &#8594; ");
    return result;
  },

  formatReaction : function(reaction){
    var result = "";
    var sep = " + ";

    var parts = reaction.split(' > ');

    var reactants = parts[0].split(sep);
    for( var i =0; i<reactants.length; i++){
      reactants[i] = pad(new Array(8).join(" "), reactants[i].trim() );
    }


    var products = parts[1].split(sep);
    for( var i =0; i<products.length; i++){
      products[i] = pad( new Array(8).join(" "), products[i].trim() );
    }

    result = reactants.join(sep) + ' > ' + products.join(sep);
    return pad(new Array(35).join(" "),result);
  },

  dataAsText : function(nodes, abundance){
    var reactions = {formation : '', destruction : ''};
    var node = null;
    var ids = nodes.getIds();
    var str = "";
    var pad_short = new Array(20).join(" ");
    var separator = " ";
    for (var i = 0, len = ids.length; i < len; i++ ){
      str = "";
      node = nodes.get(ids[i]);
      if(node.metadata.reaction !== undefined && node.group !== node_groups.legend){
        str += this.formatReaction(node.metadata.reaction)+separator;
        str += pad( pad_short, str_to_exp_number(node.metadata.abundances[0], 2)+separator, true);
        str += pad( pad_short, str_to_exp_number(node.metadata.abundances[1], 2)+separator, true);
        str += pad( pad_short, str_to_exp_number(get_constant(node.metadata.rate, node.metadata.abundances), 2)+separator, true);
        str += pad( pad_short, str_to_exp_number(node.metadata.rate, 2)+separator, true);
        str += pad( pad_short, str_to_exp_number(node.metadata.rate/abundance, 2), true);
        str += pad( pad_short, node.metadata.proportion, true);
        str += "\n";
      }

      if(node.group === node_groups.product)
        reactions.destruction = reactions.destruction + str;

      if(node.group === node_groups.reactant)
        reactions.formation = reactions.formation + str;
    }


    var result = 'Formation reactions \n';
    result += pad( new Array(36).join(' '), 'A + B -> C + D'+separator )+ pad( pad_short, 'n(A) (cm-3)'+separator, true );
    result += pad( pad_short, 'n(B) (cm-3)'+separator, true ) + pad( pad_short, 'constant'+separator, true );
    result += pad( pad_short, 'rate (cm-3 s-1)'+separator, true ) + pad( pad_short, 'freq (s-1)', true ) + pad( pad_short, 'proportion', true );
    result += '\n';
    result += reactions.formation;
    result += '\n';
    result += 'Destruction reactions\n';
    result += pad( new Array(36).join(' '), 'A + B -> C + D'+separator )+ pad( pad_short, 'n(A) (cm-3)'+separator, true );
    result += pad( pad_short, 'n(B) (cm-3)'+separator, true ) + pad( pad_short, 'constant'+separator, true );
    result += pad( pad_short, 'rate (cm-3 s-1)'+separator, true ) + pad( pad_short, 'freq (s-1)', true )+ pad( pad_short, 'proportion', true );
    result += '\n';
    result += reactions.destruction;
    result += '\n';

    return result;
  },
  data_as_json : function(){}

}

//
// Utility functions
//
function get_constant(rate, abundances){
  var result = parseFloat(rate);

  if (abundances[0] !== "N/A" && abundances[1] !== "N/A"){
    result =  (parseFloat(rate)/(parseFloat(abundances[0])*parseFloat(abundances[1]))).toExponential(2);
  } else if (abundances[0] !== "N/A" && abundances[1] === "N/A"){
    result = (parseFloat(rate)/(parseFloat(abundances[0]))).toExponential(2);
  } else if (abundances[0] === "N/A" && abundances[1] !== "N/A"){
    result = (parseFloat(rate)/(parseFloat(abundances[1]))).toExponential(2);
  }

  return result;
}

function picture(canvas){
  var img    = canvas.toDataURL("image/png");
  document.write('<img src="'+img+'"/>');
}

function map_to_string(map){
  var result = "";
  for (var key in map) {
    result = result + map[key];
  }
  return result;
}

function get_percentage(rate){
  var value = parseFloat(rate.replace('%', ''));
  return value/100;
}

function str_to_exp_number(value, precision){
    if ( value !== "N/A" )
      return  parseFloat(value).toExponential(precision);

    return "";
}

function datalist_to_html(sorted_data, data){
  var result = document.createElement("ul");
  for(var i = 0, len = sorted_data.length; i< len; i++){
    let li = document.createElement("li"); 
    li.appendChild(DOMUtil.getSpan("title", "", sorted_data[i]));
    li.appendChild(DOMUtil.getSpan("value", data[sorted_data[i]].value.toExponential(4), data[sorted_data[i]].value.toExponential(2)));
    li.appendChild(DOMUtil.getSpan("", "", data[sorted_data[i]].unit));
    result.appendChild(li);
  }

  return  result;
}

//
// Global variables
//
var page = {
  thresholdInput : document.getElementById("threshold"),
  simpleThreshold : document.getElementById('minvalue-thresholdtype'),
  cumulatedThreshold : document.getElementById('cumulated-thresholdtype'),
  mainPosVarInput : document.getElementById("main-pos-var"),
  mainPosVarName : document.getElementById("main-pos-var-name"),
  mainPosVarManualInput : document.getElementById("main-pos-var-input-value"),
  controlPanel : document.getElementById('interface'),
  hasControlPanel : document.getElementById('with-control-panel'),
  controlPanelItem : document.getElementById('with-control-panel-item'),
  hasPhysicalConditionPanel : document.getElementById('with-physical-conditions'),
  physicalConditionPanelItem : document.getElementById('with-physical-conditions-item'),
  physicalConditionPanel : document.getElementById('gas-conditions'),
  mainPosVarGraphLegend : document.getElementById("plot-main-pos-var-legend"),
  speciesList : document.getElementById("species-list"),
  filesList : document.getElementById("files-list"),
  isNewFile : false,
  hasGasConditions : document.getElementById('with-gas-conditions'),
  gasConditionsDisplay : document.getElementById("gas-conditions-display"),
  hasGrainConditions : document.getElementById('with-grain-conditions'),
  grainTemperaturesDisplay : document.getElementById("grain-temperatures-display"),
  grainDensitiesDisplay : document.getElementById("grain-densities-display"),
  infosDisplay : document.getElementById("infos"),
  exportPictureButton : document.getElementById("picture-export"),
  exportRatesDisplayedButton : document.getElementById("rates-displayed-export"),
  exportRatesMainPosVarTxtButton : document.getElementById("rates-main-pos-vars-txt-export"),
  exportNetworkButton : document.getElementById("network-export"),
  spinnerArea : document.getElementById('spinner-area'),
  graphDisabler : document.getElementById('graph-disabler'),
  backButton : document.getElementById("go-back"),
  allReactionsDisplay : document.getElementById("all-reactions-display"),
  savedFiles : document.getElementById("saved-files"),
  currentModel : document.getElementById("current-model-name"),
  isLogXPlot : document.getElementById('with-plot-x-log'),
  isLogYPlot : document.getElementById('with-plot-y-log'),
  spinner : null,
  allSpecies : [],
  allMainPosVars : [],

  updateMainPosVarInput : function(main_pos_vars){
    this.allMainPosVars = main_pos_vars;
    var str = '<option value="0">'+main_pos_vars[0]+'</option>';
    this.mainPosVarInput.step = config.av_step;
    for(var i = config.main_pos_var_step, len = main_pos_vars.length; i < len; i = i+config.main_pos_var_step){ // display only 1/main_pos_var_step of values
      str = str + '<option value="'+i+'">'+main_pos_vars[i]+'</option>';
      page.mainPosVarInput.max = i;
    }
    this.mainPosVarInput.innerHTML = str;
  },

  updateGraphVisibility : function(threshold){
    if(threshold == 0 && this.simpleThreshold.checked === true){
      this.disableGraph();
    } else if(threshold == 100 && this.cumulatedThreshold.checked === true){
      this.disableGraph();
    } else{
      this.enableGraph();
    }
  },

  disableGraph : function(){
    this.graphDisabler.style.visibility = "visible";
  },

  enableGraph : function(){
    this.graphDisabler.style.visibility = "hidden";
  },

  getThresholdType : function(){
    if(simpleThreshold.checked === true)
      return;
  },

  goBackSpecies : function(species){
    this.speciesList.selectedIndex = this.allSpecies[species];
  },

  setModelName : function(name){
    this.currentModel.innerHTML = name;
  },

  setMainPositionVarName : function(name, unit){
    this.mainPosVarName.innerHTML = name + ' ( ' + unit + ' ) ';
  },

  resetMainPositionVarValue : function(){
    this.mainPosVarInput.value = 0;
    this.mainPosVarManualInput.value = 0;
  },

  getModelName : function(){
    return this.currentModel.innerHTML;
  },

  switchLeftPanel :function(source){
    if(source === this.hasControlPanel.id ){
      this.displayControlPanel(true);
      this.displayPhysicalConditionPanel(false);
    }else if(source === this.hasPhysicalConditionPanel.id){
      this.displayControlPanel(false);
      this.displayPhysicalConditionPanel(true);
    }

  },

  displayControlPanel : function(is_visible){
    if(is_visible === false){
      this.controlPanel.style.display = "none";
      this.controlPanelItem.className = "pure-menu-item";
    }else{
      this.controlPanel.style.display = "";
      this.controlPanelItem.className = "pure-menu-item pure-menu-selected";
    }

  },

  displayPhysicalConditionPanel : function(is_visible){
    if(this.physicalConditionPanel !== null){
      if(is_visible === false){
        this.physicalConditionPanel.style.display = "none";
        this.physicalConditionPanelItem.className = "pure-menu-item";
      }else{
        this.physicalConditionPanel.style.display = "";
        this.physicalConditionPanelItem.className = "pure-menu-item pure-menu-selected";
      }
    }
  },

  displayGasConditions : function(){
    if(this.hasGasConditions.checked === false){
      this.gasConditionsDisplay.style.display  = "none";
    }else{
      this.gasConditionsDisplay.style.display = "";
    }
  },

  displayGrainConditions : function(){
    if(this.hasGrainConditions.checked === false){
      this.grainTemperaturesDisplay.style.display  = "none";
      this.grainDensitiesDisplay.style.display  = "none";
    }else{
      this.grainTemperaturesDisplay.style.display = "";
      this.grainDensitiesDisplay.style.display  = "";
    }
  },

  /*displayReactionInformations : function(){
    if(this.hasReactionInformations.checked == false){
      this.infosDisplay.style.display  = "none";
    }else{
      this.infosDisplay.style.display = "";
    }
  },*/

  /**
   * updates list of available files
   * @param data array of file names
   */
  updateFilesList : function(data){
    var files_string = '<option value="" selected>-- None --</option>';
    for(var i = 0, len = data.length; i < len; i = i+1){
      files_string = files_string + '<option value="'+data[i]+'">'+data[i]+'</option>';
    }
    this.filesList.innerHTML = files_string;
  },

  /**
   * updates list of available species
   * @param data array of species names
   */
  updateSpeciesList : function(data){
    var species_string = "";//'<option selected>-- None --</option>';
    var selected = "";
    for(var i = 0, len = data.length; i < len; i++){
      if(data[i] === 'H2')
        selected += 'selected="selected"';
      species_string = species_string+'<option value="'+data[i]+'"'+ selected +'>'+data[i]+'</option>';
      page.allSpecies[data[i]] = i;
      if(selected != "")
        selected = "";
    }
    this.speciesList.innerHTML = species_string;
  },

  setInitialGasCondition : function(){
    if (this.hasGasConditions.checked == false)
      this.gasConditionsDisplay.style.display  = "none";
  },

  setInitialGrainConditions : function(){
    if (this.hasGrainConditions.checked == false){
      this.grainTemperaturesDisplay.style.display  = "none";
      this.grainDensitiesDisplay.style.display  = "none";
    }
  },

  updateGrainConditionsDisplay : function(temperatures, densities){

    removeAllChildren(this.grainTemperaturesDisplay);
    removeAllChildren(this.grainDensitiesDisplay);

    let grain_temp = document.createElement("h2");
    grain_temp.innerText = "Grain temperatures";
    this.grainTemperaturesDisplay.appendChild(grain_temp);
    this.grainTemperaturesDisplay.appendChild(temperatures);


    let grain_dens = document.createElement("h2");
    grain_dens.innerText = "Grain densities";
    this.grainDensitiesDisplay.appendChild(grain_dens);
    this.grainDensitiesDisplay.appendChild(densities);
  },

  updateGasConditionsDisplay : function(content){
    removeAllChildren(this.gasConditionsDisplay);

    let gas_cond = document.createElement("h2");
    gas_cond.innerText = "Gas conditions";
    this.gasConditionsDisplay.appendChild(gas_cond);
    this.gasConditionsDisplay.appendChild(content);
  },

  enableButtons : function(){
    this.exportPictureButton.disabled = "";
    this.exportRatesDisplayedButton.disabled = "";
    this.exportRatesMainPosVarTxtButton.disabled = "";
    this.exportNetworkButton.disabled = "";
  },

  toogleBackButton : function(status){
    if(status)
      this.backButton.disabled = "";
    else
      this.backButton.disabled = "disabled";
  },

  updateInfosDisplay : function(nodes, abundance, total_form_rate, total_dest_rate, rates_unit){
    removeAllChildren(this.infosDisplay);
    this.infosDisplay.appendChild(formatter.dataAsHtml(nodes, abundance, total_form_rate, total_dest_rate, rates_unit));
  },

  updateSpeciesDisplay : function(species, abundance, abundance_unit){
    let p = document.createElement("p");
    let strong = document.createElement("strong");
    strong.innerText = "n("+species+") ";
    let span = document.createElement("span");
    span.innerText = parseFloat(abundance).toExponential(2)+" "+abundance_unit;
    p.appendChild(strong);
    p.appendChild(span);

    //final container
    let species_display = document.getElementById("species-display");

    //remove node if it has already been appended before
    removeAllChildren(species_display);
    species_display.appendChild(p);
  },

  highlightReaction : function(element_id){
    var current = document.getElementById(element_id);
    if(current !== null)
      current.style.backgroundColor = mol_network.theme_color;
  },

  initReactionsTable : function(){
    let tables = ["formation-table", "destruction-table"]

    for(var i = 0; i < tables.length; i++){
      let table = document.getElementById(tables[i]);
      for( var j = 0, len = table.rows.length; j < len; j++)
      table.rows[j].style.backgroundColor = "white";
    }
  },

  displayThresholdValue : function(){
    document.getElementById("threshold-value-display").innerHTML = this.thresholdInput.value+" %";
  },

  displayReactionRatesFile : function(url, filename){
    page.savedFiles.innerHTML = page.createLink(url +'/static/' , filename, 'Reaction rates');
  },

  displayFormationDestructionFiles : function(url, filename, formation_file, destruction_file){
    var content = '<p>' + page.createLink(url +'/static/' , formation_file, "Formation data") + '</p>';
    content += '<p>' + page.createLink(url +'/static/' , destruction_file, "Destruction data") + '</p>';
    page.savedFiles.innerHTML = content;
  },

  displayNetworkFile : function(url, filename){
    page.savedFiles.innerHTML = page.createLink(url +'/static/' , filename, "Network");
  },

  createLink : function(url, filename, linkText){
    return  '<a href="' + url + filename + '" target="_blank">' + linkText + '</a> ';
  },

  clearSavedFiles : function(){
    this.savedFiles.innerHTML = "";
  },

  showSpinner : function(){
    this.spinnerArea.style.visibility = "visible";
    this.spinner = new Spinner(opts).spin(this.spinnerArea);

  },

  hideSpinner : function(){
    this.spinner.stop();
    this.spinnerArea.style.visibility = "hidden";
  },

  triggerEvent : function(element, event){
    if (document.createEvent) {
      element.dispatchEvent(event);
    } else {
      element.fireEvent("on" + event.eventType, event);
    }
  }
}
