#!/usr/bin/env python3

import argparse
import subprocess
import os
import sys


parser = argparse.ArgumentParser(prog="idat p|plot")
parser.add_argument("-txt",
                    action='store_const', const='-txt',
                    help="""indicates that <data file> is an ascii .dat file 
                    as generated by `idat extract`""")
parser.add_argument("datafile",
                    help="""either an hdf5 file as produced by the ism tools, 
                    or an ascii .dat file (see the -txt option)""")
parser.add_argument("conf",
                    help="""a .yaml file to configure the plot to be built. 
                    See the documentation for the details and syntax of this 
                    file.
                    If <conf> is a directory, `idat plot` will be called for
                    each .yaml file inside the directory in turn.
                    """)
parser.add_argument("output",
                    nargs='?',
                    help="""optional path to a directory where the computed 
                    data will be stored.
                    Defaults to the directory where the <datafile> is.""")

args = vars(parser.parse_args())

if args['output'] == None:
    args['output'] = os.path.dirname(os.path.realpath(args['datafile']))

conffiles = []
if os.path.isdir(args['conf']):
    conffiles = ['/'.join([args['conf'], fn]) 
            for fn in os.listdir(args['conf'])
            if fn.endswith('yaml')]
else:
    conffiles = [args['conf']]

idat_home = os.getenv("IDAT_HOME")
cmd = [
        sys.executable, 
        idat_home + "/modules/idat-plot/src/idatplot/plot.py"]

for cff in conffiles:
    cmdargs = list(filter(None, 
        [args['txt'], cff, args['datafile'], args['output']]))
    subprocess.Popen(cmd + cmdargs, bufsize = 0).wait()
