- `pdrpostprocs` is a python package used by various parts of the project
- `scripts` contains "intern" scripts not yet validated to be release as user
  scripts.

Usage
-----

The *idat-plot* project is developped as a python package (located in the
`src` folder) that is exposed to user through shell scripts.
The *idat-plot* project is for now compound of 2 modules: *plot* and *extract*

- *extract* is used to extract quantities into a txt data file from PDR hdf5
- *plot* is used to build plots from both a. txt data files produced by
  *extract* and b. directly from a PDR hdf5 file.


### Plot directly from hdf5 file

Example (from the project root directory):

    sh plot.sh \
        <idatpath>/doc/idat-plot/examples/data/PDR152r1809_n_r1e0d1e2A1e0_s_20.hdf5 \
        <idatpath>/conf/idat-plot/conf.yaml \
        .

will generate the plots configured in `conf/idat-plot/conf.yaml` using the data in
`doc/idat-plot/examples/data/PDR152r1809_n_r1e0d1e2A1e0_s_20.hdf5` in the current
directory.

General usage:

    cd <idatpath>
    vim conf/idat-plot/conf.yaml
    sh plot.sh <.hdf5 data file> conf/idat-plot/conf.yaml <output dir>


#### Plot data from extracted .txt data files

This is used to build plot from txt data files produced by the *extract*
postprocessor.

Example (from the project root directory):

    sh plot.sh -txt doc/idat-plot/examples/data/pdr_plot_data_dens.txt \
               conf/idat-plot/conf.yaml \
               .

will generate the plots configured in `conf/idat-plot/conf.yaml` using the data in
`doc/idat-plot/examples/data/pdr_plot_data_dens.txt` in the current directory.

General usage:

    vim conf/idat-plot/conf.yaml
    sh plot.sh  -txt <.txt data file> conf/idat-plot/conf.yaml <output dir>


### Plot options

The configuration of the system is done through YAML files with the format
detailed below.

Text parameters can include embedded latex expressions (plot labels for
example). In this case, the value must be enclosed in simple quotes and be
surrounded by `$$` Additionnaly, only a subset of latex can be used, more
detail here:
https://matplotlib.org/tutorials/text/mathtext.html#sphx-glr-tutorials-text-mathtext-py

Example:

    plot/y1label: '$ abundance of {H}_2 $'


#### extract

    extract/quantities: ['AV', 'Distance', 'nH', 'Tgas', 'Pgas', 'n(H)', 'n(H2)', 'n(C+)', 'n(C)']
    extract/output: "{}.ab"

The *extract* documents configure each one a txt data file containing
quantities extracted from an hdf5 file.

##### special settings

- `extract/output` can contain '{}' string which will be replaced by the root
  of the basename of the hdf5file. The string shouldn't contain any file
extension (like '.jpg') because the image format extension will be added
automatically by the script. The script generates the image in 2 formats: a
`.png` image with limited resolution and a `.pdf` vector image.

Example:

        "my_plot_{}"
        
- `extract/quantitites` can be either a list of quantities to extract or a
  string containing a file path (absolute path) towards a .esf file produced
  by the extractor.


#### plot

    plot/title: 'plot 1'
    plot/x: 'AV'
    plot/y1: ['n(H)', 'n(H2)', 'n(C+)', 'n(C)', 'n(CO)', 'n(e-)']
    plot/y1label: 'Abundance(X)'
    plot/y1normalize: 'n(H)'
    plot/y2: ['Tgas']
    plot/y2label: Gas temperature [K]
    plot/log: ['y1']
    plot/output: "plot1.png"
    ---
    plot/title: 'n(C+) n(C)'
    plot/x: 'Av'
    plot/y1: ['n(C+)', 'n(C)']
    plot/y2: ['Tgas']
    plot/output: "toto.png"

The *plot* documents configure each one a different plot using data in a data
file, extracted from the hdf5 file.


##### special settings

- `plot/y1`, `plot/y2` are lists of the quantities to plot.
    In case one wants to customise some plotting aspects of the quantities each
quantity can be given as a list of options, as follow:

    [<quantity id>, <quantity label>, <pyplot linestyle>, <multiplicator>]

If one wants to set the nth element of the list, all the preceding
elements must be set too. That is, if one wants to set the `pyplot linestyle`
for a quantity, he must set the precedings element (id and label).

If one wants to just use the quantity id he can do so using just a string (no
list of options needed).

The elements not set have the following default values:

        quantity label:   <quantity id>
        pyplot linestyle: '-'

ex:
        plot/y1: ['n(H2)', 'n(H)']
        plot/y1: [['n(h2)', 'n of H2', '-.', -1], 'n(H)', ['n(H2O)', 'int of H20']]

In the example above, the opposite of n(H2) will be plotted (multiplicator is
-1)

- `plot/x` is a list [<quantity id>, <quantity label>], with the same
  facility than above, one can just use a plain string instead of a list if
  the wanted <quantity label> is the same string as <quantity id>.
- `plot/log` is a list of the axis to apply log scale on. If the variation
  range of the values is higher than an internal thresold (for now 10^10), we
  automatically log10() the quantity before plotting it to avoid overcrowded
  minor ticks on logscale axis.
  The automatic log10 of large range y is disabbled if the user configure
  either `ymin` or `ymax`.
- `plot/y1normalize` is a string containing a quantity that will be used to
  normalize all the quantities in `plot/y1`. Same for `plot/y2` and
  `plot/y2normalize`. Note that `plot/x` is never normalized.
- `plot/output` is a string, the path where to save the plot inside the
  `outputdir` passed as command line argument to `plot.sh`. If not provided,
  the name of the input data file is used (with `png` extension). If the
  special sequence '{}' is present in the value of `plot/output` it will be
  replaced with the input data filename.
- `plot/title`: has the same facility involving {} than `plot/output`
- `plot/xmin` can be either:

    - 'auto' in this case the xmin is chosen as to be the most relevant (ie
      skipping parts of plot where nothing happens).
    - <float value> manually set the xmin value.
- plot/xmin,plot/xmax if not defined, use the min and max of the x axis.




### Extract

This component is set through a configuration file, in particular to list what
values to extract during this part of the workflow, to be used later by other
postprocessing (for example ploting).

For now, this module is used internally and not provided with a user
interface, thus it has been placed in `src/scripts`.


#### Usage (produce .txt data files from .hdf5)

Example (from the `src/scripts/extract` directory):

    ksh extract.ksh <hdf5 file> <conf.yaml file> .

General usage:

    cd src/scripts/extract
    vim conf/conf.yaml    <--- do the conf you want
    ksh extract.ksh <hdf5 file> conf/conf.yaml <output dir>



