# -*- coding: utf-8 -*-

import argparse
import os
import re
import datetime as dt

from lib import hdf
from lib import hdfutil as hdfu

from idatplot import utils as iu
# from idatplot.vendors.extractor import hdf

# DEFAULT_SEP = ' # '
DEFAULT_SEP = '   '
cfdb = {
        "idatplot/scommand_processors": {
            "lines": "_process_pdr_lines_scommand"
            },
        "idatplot/pdr_lines_db": "resources/pdr_lines.sqlite"
        }

def _load_esf(ep, sep = '\n'):
    with open(ep) as ef:
        quantities = [e.strip() for e in ef.read().split(sep)]
    return list(filter(None, quantities))

def _model_id(h5filename):
    h5fn_re = r'^(.+)_[asc]_[1-9][0-9]*.hdf5$'
    mo = re.match(h5fn_re, h5filename)
    if mo != None:
        return mo.group(1)
    return h5filename

def extract(edf, quantities, eof, header, sep = DEFAULT_SEP):
    hdfu.export_dat_as_file(hdf.PdrHDF(edf), 
            header, 
            quantities, 
            eof,
            separator = sep)

def cmdline_client(ecffile, h5file, outdir, simple_format = False):

    def _ecf_quantities(ecf):
        """ if quantites are provided as an .esf file, the separtor is newline
        (1 quantitiy per line)"""

        ecfq = ecf['quantities']
        quantities = []
        if isinstance(ecfq, str):
            quantities = _load_esf(ecfq)
        else:
            quantities = ecfq
        return quantities

    dfroot, _ = os.path.splitext(os.path.basename(h5file))

    modelid = _model_id(dfroot)
    header = {
            'hdffile': {
                'label': 'Model name',
                'value': modelid
                },
            'date': {
                'label': 'Extraction date',
                'value': dt.datetime.now().strftime('%d-%m-%Y')
                }
            }

    if simple_format:
        # in this case *outdir* is actually the output file path
        extract(h5file, 
                iu.process_ecf_quantities(cfdb, _load_esf(ecffile)), 
                outdir,
                header,
                DEFAULT_SEP)
    else:
        ecf = dict(iu.load_conf(ecffile))
        for eecf in ecf['extract']:
            ofpath = outdir + '/' + eecf['output'].format(dfroot)
            print('creating {}...'.format(ofpath))
            extract(h5file, 
                    iu.process_ecf_quantities(cfdb, _ecf_quantities(eecf)), 
                    ofpath, 
                    header,
                    eecf.get('separator', DEFAULT_SEP))


def main():
    
    parser = argparse.ArgumentParser()
    parser.add_argument("-t", action='store_true', 
            help = "parse cffile as a simple 1 quantity per line txt file")
    parser.add_argument('hdf5file')
    parser.add_argument('cffile')
    parser.add_argument('outdir')
   
    args = parser.parse_args()
    cmdline_client(args.cffile, args.hdf5file, args.outdir, args.t)

if __name__ == '__main__':
    main()
