"use strict";

const QTY_SEPARATOR = "\n";
const SIMPLE_QTY = "integrated";
const ARRAY_QTY = "local";
const MAX_PLOT_QTY = 15;
var sessionId = null;

// ids of dom elements
var components = {

    setNavBar: function(project, model) {
        const links = '<a href="' + GLOBAL_HOST + 'ismdb/">Projects</a> &gt;' +
            '<a href="' + GLOBAL_HOST + 'ismdb-is/' + project + '/plot">' + project + '</a> &gt;' +
            '<a href="' + GLOBAL_HOST + 'ismdb-mv/projects/' + project + '/models/' + model + '">' + model + '</a> &gt;' +
            'Idat';
        $('#nav-links').html(links);
    },

    setPageType: function(text) {
        $('#page_type').html($('#page_type').html() + text);
    },

    setPageTitle: function(text) {
        $('#title').html(text);
    },

    setFileName: function(filename) {
        $('#filename').html(`${filename}`);
    },

    getTreeContainer: function() {
        return $('#tree_container');
    },

    /**
    Returns fancytree object
    */
    getTree: function() {
        return this.getTreeContainer().fancytree("getTree");
    },

    getSearchBar: function() {
        return $("#searchbar");
    },

    getUserSelection: function(type) {
        if (type === SIMPLE_QTY)
            return $("#user_integrated_selection");
        else if (type === ARRAY_QTY)
            return $("#user_local_selection");
        else throw "Unknown type";
    },

    getUserSelectionError: function(type) {
        if (type === SIMPLE_QTY)
            return $("#user_integrated_selection_error");
        else if (type === ARRAY_QTY)
            return $("#user_local_selection_error");
        else throw "Unknown type";
    },

    getDataTextButton: function(type) {
        if (type === SIMPLE_QTY)
            return $("#integrated_data_text_button");
        else if (type === ARRAY_QTY)
            return $("#local_data_text_button");
        else throw "Unknown type";
    },

    getDataVotableButton: function(type) {
        if (type === SIMPLE_QTY)
            return $("#integrated_data_votable_button");
        else if (type === ARRAY_QTY)
            return $("#local_data_votable_button");
        else throw "Unknown type";
    },

    getGoToPlotButton: function() {
        return $("#local_data_gotoplot_button");
    },

    getPlotButton: function() {
        return $("#plot_data_button");
    },

    getClearTagsButton: function() {
        return $("#clear_tag_buttons");
    },

    showResetButton: function() {
        $(".tags-reset").css("visibility", "visible");
    },

    setTags: function(html) {
        document.getElementById("tags").innerHTML = html;
    },

    setActionButtonsDisabled: function(value) {
        if (value === true || value === false) {
            this.getDataTextButton(SIMPLE_QTY).prop('disabled', value);
            this.getDataTextButton(ARRAY_QTY).prop('disabled', value);
            this.getDataVotableButton(SIMPLE_QTY).prop('disabled', value);
            this.getDataVotableButton(ARRAY_QTY).prop('disabled', value);
            this.getGoToPlotButton().prop('disabled', value);
        }
    },

    getDataStatus: function(type) {
        if (type === SIMPLE_QTY)
            return $("#integrated_data_status");
        else if (type === ARRAY_QTY)
            return $("#local_data_status");
        else throw "Unknown type";
    },

    setPrepareDataStatus: function(type) {
        this.getDataStatus(type).html(`<span>
        We are preparing your data.
        </span>`);
    },

    /**
     * returns valid element ids by forbidden characters
     */
    getTagId: function(id) {
        if ($.trim(id) !== "") {
            return id.replace(/\s+/g, "").replace(/[\(\)]+/g, "prt").replace(/\++/g, "pl");
        }
        return "";
    },

    plot: {
        show: function() {
            $("#plot_container").removeClass("hidden");
            $("#plot_container").addClass("visible");
        },
        hide: function() {
            $("#plot_container").removeClass("visible");
            $("#plot_container").addClass("hidden");
        },
        getTitle: function() {
            return $("#graph_title").val();
        },
        x: {
            getArea: function() {
                return $("#x_plot_area");
            },

            resetArea: function() {
                $("#graph_title").val('');
                $("#x_label").val('');
                $("#x_min").val('');
                $("#x_max").val('');
            },

            getValue: function() {
                return $("#x_data").val();
            },
            getValues: function() {
                let result = [];
                let values = document.getElementById("x_data").options;
                for (const v of values) {
                    result.push(v.value);
                }
                return result;
            },
            getLog: function() {
                return $("input[name=x_log]").is(':checked');
            },
            getLabel: function() {
                return $.trim($("#x_label").val());
            },
            getMin: function() {
                const val = $.trim($("#x_min").val());
                if (val !== '' && isNaN(Number.parseFloat(val)))
                    throw "xmin is not  valid number.";
                return val;
            },
            toggleMinDisabled: function() {
                if ($("#x_min").prop('disabled') === true) {
                    $("#x_min").prop('disabled', false);
                } else {
                    $("#x_min").prop('disabled', true);
                }
            },
            getAutoMin: function() {
                return $("input[name='auto_x_min']").is(':checked');
            },
            setAutoMinOnClick: function(callback) {
                $("input[name='auto_x_min']").on("click", callback);
            },
            getMax: function() {
                const val = $.trim($("#x_max").val());
                if (val !== '' && isNaN(Number.parseFloat(val)))
                    throw "xmax is not  valid number.";
                return val;
            }
        },
        y1: {
            getArea: function() {
                return $("#y1_plot_area");
            },

            resetArea: function() {
                $('#y1_plot_area').html('');
            },

            getShowConfigArea: function() {
                return $("#y1_show");
            },

            setShowConfigAreaOnClick: function(callback) {
                $("#y1_show").on("click", function() { callback(); });
            },

            toggleVisibilityConfigArea: function() {
                if ($("#y1_config").hasClass('hidden')) {
                    $("#y1_config").removeClass("hidden");
                    $("#y1_config").addClass("visible");
                    $("#y1_show").html('-');
                } else {
                    $("#y1_config").removeClass("visble");
                    $("#y1_config").addClass("hidden");
                    $("#y1_show").html('+');
                }
            },

            getValues: function() {
                let result = [];
                $("input:checkbox[name='y1_data']:checked").each(function() {
                    // get user defined label for selected quantity
                    const id = ($(this).attr('id')).replace("y1_data", "y1_label");
                    const label = $.trim($("#" + id).val()) === '' ? $(this).val() : $.trim($("#" + id).val());
                    result.push([$(this).val(), label]);
                });
                return result;
            },
            getLog: function() {
                return $("input[name=y1_log]").is(':checked');
            },
            getLabel: function() {
                return $.trim($("#y1_label").val());
            },
            getMin: function() {
                const val = $.trim($("#y1_min").val());
                if (val !== '' && isNaN(Number.parseFloat(val)))
                    throw "y1 min is not  valid number.";
                return val;
            },
            getMax: function() {
                const val = $.trim($("#y1_max").val());
                if (val !== '' && isNaN(Number.parseFloat(val)))
                    throw "y1 max is not  valid number.";
                return val;
            }
        },
        y2: {
            getArea: function() {
                return $("#y2_plot_area");
            },

            resetArea: function() {
                $('#y2_plot_area').html('');
            },

            getShowConfigArea: function() {
                return $("#y2_show");
            },

            setShowConfigAreaOnClick: function(callback) {
                $("#y2_show").on("click", callback);
            },

            toggleVisibilityConfigArea: function() {
                if ($("#y2_config").hasClass('hidden')) {
                    $("#y2_config").removeClass("hidden");
                    $("#y2_config").addClass("visible");
                    $("#y2_show").html('-');
                } else {
                    $("#y2_config").removeClass("visble");
                    $("#y2_config").addClass("hidden");
                    $("#y2_show").html('+');
                }
            },

            getValues: function() {
                let result = [];
                $("input:checkbox[name='y2_data']:checked").each(function() {
                    // get user defined label for selected quantity
                    const id = ($(this).attr('id')).replace("y2_data", "y2_label");
                    const label = $.trim($("#" + id).val()) === '' ? $(this).val() : $.trim($("#" + id).val());
                    result.push([$(this).val(), label]);
                });
                return result;
            },
            getLog: function() {
                return $("input[name=y2_log]").is(':checked');
            },
            getLabel: function() {
                return $.trim($("#y2_label").val());
            },
            getMin: function() {
                const val = $.trim($("#y2_min").val());
                if (val !== '' && isNaN(Number.parseFloat(val)))
                    throw "y2 min is not  valid number.";
                return val;
            },
            getMax: function() {
                const val = $.trim($("#y2_max").val());
                if (val !== '' && isNaN(Number.parseFloat(val)))
                    throw "y2 max is not  valid number.";
                return val;
            }
        },
        result: {
            getArea: function() {
                return $("#plot_result");
            },
            resetArea: function() {
                $('#plot_result').html('');
            },
            enableSpinner: function() {
                $("#plot_result").html("<img class='center-content' src='static/img/spinner.gif' alt='loading'></img>");
            },
            disableSpinner: function() {
                $("#plot_result").html("");
            },
        }
    }
};


Selection.create({
    // Class for the selection-area
    class: 'selection',

    // All elements in this container can be selected
    selectables: ['.tags > .selectable'],
    singleClick: true,

    // The container is also the boundary in this case
    boundaries: ['.tags']
}).on('move', ({ inst, changed: { removed, added } }) => {
    /*
      add/remove the selected class from a tag
      according to its current selection status

      the selection state is managed externally by the selector 
      object so it does not matter if
      the selection event is added or removed
    */
    function updateTag(node, tag) {
        if (node !== null) {
            if (selector.isSelected(node)) {
                tag.classList.remove('selected');
            } else {
                tag.classList.add('selected');
            }
            selector.updateUserRequest([node]);
        }
    }

    // Add a custom class to the elements that where selected.
    for (const tag of added) {
        const node = components.getTree().getNodeByKey(tag.innerText);
        updateTag(node, tag);
    }

    // since the last selection
    for (const tag of removed) {
        const node = components.getTree().getNodeByKey(tag.innerText);
        updateTag(node, tag);
    }
}).on('stop', ({ inst }) => {
    inst.keepSelection();
});

/**
 * Wait for n milliseconds
 */
const sleep = (milliseconds) => {
    return new Promise(resolve => setTimeout(resolve, milliseconds));
};

/**
  Returns name of file to read on server 
  contained in GET parameter name file
*/
function getParameter() {
    let query = location.search.replace('?', '');
    let params = query.split("&");
    let result = { "model": false, "project": false };

    for (let i = 0; i < params.length; i++) {
        let param = params[i].split('=');
        if (result[param[0]] !== null) {
            result[param[0]] = param[1];
        }
    }

    if (result.model === false) {
        throw new Error('model parameter is required');
    }

    return result;
}

function sendExtractRequest(quantities, sessionId, mode, type) {
    components.setActionButtonsDisabled(true);
    components.setPrepareDataStatus(type);

    $.ajax({
        url: GLOBAL_SERVICE + "export/data/" + mode,
        type: "POST",
        data: JSON.stringify({
            "quantities": quantities,
            "sessionId": sessionId,
            "datatype": type,
            "dataformat": mode
        }),
        dataType: "json",
        contentType: "application/json",
        success: function(result) {
            if (result.status === "ERROR") {

                let errorText = result.error;
                // Special case when user want to plot a spectrum (ex : Wavelegnth + H2 OP ratio)
                if (errorText === 'all the columns do not have the same size') {
                    errorText = 'To extract the spectrum, select only quantities from the "Spectrum" sub-folder.';
                }
                components.getDataStatus(type).html(`<span>job status : ${errorText}</span>`);

                window.alert(errorText);
            } else {
                let message = "";
                if(result.format === "text"){
                    message = `<a  href='${GLOBAL_GENERATE}text/${sessionId}'>
                                    Download your data file
                             </a>`;
                                   

                }else if(result.format === "votable"){
                    message = `<a  href='${GLOBAL_GENERATE}votable/${sessionId}'>
                                Download your data file
                             </a>`; 
                }else{
                    message = "<span>Unknown result format.</span>"; 
                }

                components.getDataStatus(type).html(`<span>${message}</span>`);

            }
            components.setActionButtonsDisabled(false);
        }
    });
}

function sendPlotRequest(parameters, sessionId) {
    if (parameters.y1.length > MAX_PLOT_QTY || parameters.y2.length > MAX_PLOT_QTY) {
        alert(`The plot tool allows a maximum of ${MAX_PLOT_QTY} quantities.`);
    } else {
        components.setActionButtonsDisabled(true);
        components.setPrepareDataStatus(ARRAY_QTY);
        components.plot.result.enableSpinner();

        $.ajax({
            url: GLOBAL_GENERATE + "plot",
            type: "POST",
            data: JSON.stringify({
                "parameters": parameters,
                "sessionId": sessionId
            }),
            dataType: "json",
            contentType: "application/json",
            success: function(result) {
                components.plot.result.disableSpinner();
                if (result.status === "ERROR") {
                    window.alert(result.message);
                    components.getDataStatus(ARRAY_QTY).html("<span>Your plot has failed.</span>");
                } else {
                    components.getDataStatus(ARRAY_QTY).html("<span>Your plot is done.</span>");
                    components.plot.result.getArea().html(
                        `<p>
                            <a 
                            href="${GLOBAL_EXPORT}${result.pdf}" 
                                   target="plot">
                                 <img src="data:image/png;base64,${result.bimg}"
                                      alt="graph" 
                                      width="605" 
                                      height="518"/>
                             </a>                                                                             
                        </p>            
                        <p class="plot-legend">To save the figure: right click and select Save</p>                                    
                        <div >  
                            <p class="plot-legend">
                                <a href="${GLOBAL_EXPORT}${result.pdf}" 
                                       target="plot">
                                    Download figure
                                </a>
                            </p>
                            <p class="plot-legend">
                                <a  href="${GLOBAL_GENERATE}yaml/${sessionId}">
                                    Save plot configuration file
                                </a>
                            </p>
                        </div>`);
                }
                components.setActionButtonsDisabled(false);
            }
        });
    }
}

function checkData(data) {
    let result = { "accepted": [], "refused": [] };
    let set = new Set();
    for (const val of data) {
        const clean_val = $.trim(val);
        if (!set.has(clean_val) && components.getTree().getNodeByKey(clean_val) !== null) {
            result.accepted.push(clean_val);
        } else {
            result.refused.push(clean_val);
        }
    }

    return result;
}

$.ajaxSetup({
    error: function(xhr) {
        let result = JSON.parse(xhr.responseText);
        window.alert(result.message);
    }
});

/**
Manages the relation between the selected tags, the tree and the textarea listing
selected quantities
*/
let Selector = class Selector {
    constructor() {
        this.selected = {
            [SIMPLE_QTY]: new Set(),
            [ARRAY_QTY]: new Set()
        };
    }

    /**
    Completeley reset the selection, clearing textarea and deselecting all tags
    */
    reset() {
        this.selected = {
            [SIMPLE_QTY]: new Set(),
            [ARRAY_QTY]: new Set()
        };
        components.getUserSelection(SIMPLE_QTY).val(Array.from(""));
        components.getUserSelection(ARRAY_QTY).val(Array.from(""));
    }

    /**
    Update content of the textarea according to selected quantities
    */
    updateUserRequest(nodes) {
        for (let node of nodes) {
            if (!this.selected[node.type].has(node.title)) {
                this.selected[node.type].add(node.title);
            } else {
                this.selected[node.type].delete(node.title);
            }
        }
        components.getUserSelection(SIMPLE_QTY).val(Array.from(this.selected[SIMPLE_QTY]).join(QTY_SEPARATOR));
        components.getUserSelection(ARRAY_QTY).val(Array.from(this.selected[ARRAY_QTY]).join(QTY_SEPARATOR));
    }

    updateSelected(nodes) {
        let result = new Set();
        const tree = components.getTree();
        const type = nodes[0].type;
        if (nodes[0] !== '') {
            for (let node of nodes) {
                const id = components.getTagId(node.title);
                if (id !== '') {
                    $("#" + id).addClass("selected");
                    result.add(node.title);
                }
            }
        }

        for (let tag of this.selected[type]) {
            tag = components.getTagId(tag);
            if (!result.has(tag)) {
                const id = tag;
                if (id !== '') {
                    $("#" + id).removeClass("selected");
                }
            }
        }
        this.selected[type] = result;
    }

    isSelected(node) {
        return this.selected[node.type].has(node.title);
    }

}
var selector = new Selector();


function initPage(data) {
    let tree = data.tree;
    sessionId = data.sessionId;
    $(function() { // on page load
        components.getTreeContainer().fancytree({
            source: tree,
            extensions: ["filter"],
            tooltip: true,
            checkbox: false,
            focusOnSelect: true,
            currentDirectory: '',
            clickFolderMode: 3,
            // sort all nodes alphabetically
            init: function(event, data) {
                data.tree.getRootNode().sortChildren(null, true);
            },
            selectMode: 2, //1:single, 2:multi, 3:multi-hier (default: 2)
            //clickFolderMode : 3, // 1:activate, 2:expand, 3:activate and expand, 4:activate/dblclick expands (default: 4)
            quicksearch: true,
            activate: function(event, data) {
                if (!data.node.getFirstChild().hasChildren() &&
                    data.node.key !== this.currentDirectory) {
                    let html = '';
                    let prev = '';
                    let style = build_style();
                    this.currentDirectory = data.node.key;
                    for (const node of data.node.getChildren()) {
                        let selected = selector.isSelected(node) === true ? "selected" : "";
                        if (node.key.length < 5 || (node.key.length >= 5 && prev.length >= 5 &&
                                node.key.substr(0, 5) != prev.substr(0, 5))) {
                            style = build_style();
                        }
                        let select = `<span class="concept-link selectable ${selected}" 
                                            style="background-color:${style.backgroundColor};color:${style.color}"
                                            id="${components.getTagId(node.key)}">
                                      ${node.key}
                                      </span>`;
                        html += select;
                        prev = node.key;
                    }

                    components.showResetButton();
                    components.setTags(html);
                }
            },
            filter: { // override default settings
                counter: false, // No counter badges
                mode: "hide" // "dimm": Grayout unmatched nodes, "hide": remove unmatched nodes
            }
        });

        function show_refused_values(data, type) {
            // quantities that do not exist in tree
            if (data.refused.length > 0 && $.trim(data.refused[0]) !== '') {
                let refused_list = "";
                let conjugated_text = data.refused.length > 1 ? "quantities do" : "quantity does";
                for (const val of data.refused) {
                    refused_list = refused_list + `<li>${val}</li>`;
                }
                components.getUserSelectionError(type).html(`The following ${conjugated_text} not exist : <ul>${refused_list}</ul>`);
            }
        }

        function sendExtractRequestAction(format, type) {
            let selection = components.getUserSelection(type).val().split(QTY_SEPARATOR);
            if (($.trim(selection)).length > 0) {
                const sorted_data = checkData(selection);
                // quantities that do not exist in tree
                show_refused_values(sorted_data, type);
                if (sorted_data.accepted.length > 0) {
                    sendExtractRequest(sorted_data.accepted, sessionId, format, type);
                } else {
                    window.alert("Please select at least one quantity.");
                }
            }
        }


        /**
         * Top-level plot button is clicked
         */
        function initializePlotArea() {
            let selection = checkData(components.getUserSelection(ARRAY_QTY).val().split(QTY_SEPARATOR));
            let selection_length = selection.accepted.length;
            if (selection_length > 0) {
                const sorted_data = selection;
                // quantities that do not exist in tree
                show_refused_values(sorted_data, ARRAY_QTY);
                if (sorted_data.accepted.length > 0) {
                    components.plot.show();
                    updatePlotArea(sorted_data.accepted);
                } else {
                    throw ("No quantity selected");
                }
            } else if (selection_length == 0){
                throw new Error("No quantity selected");
            } else {
                throw new Error("No more than 10 quantities can be plotted at the same time");
            }
        }

        /**
         * Plot button is clicked
         */
        function processPlotRequest(sessionId) {
            let request = {};
            request.log = [];
            if (components.plot.x.getLog() === true) {
                request.log.push("x");
            }
            if (components.plot.y1.getLog() === true) {
                request.log.push("y1");
            }
            if (components.plot.y2.getLog() === true) {
                request.log.push("y2");
            }

            request.title = components.plot.getTitle();
            request.x = components.plot.x.getValue();
            request.y1 = components.plot.y1.getValues();
            request.y2 = components.plot.y2.getValues();
            request.xlabel = components.plot.x.getLabel() === '' ? request.x : components.plot.x.getLabel();
            request.y1label = components.plot.y1.getLabel() === '' ? '' : components.plot.y1.getLabel();
            request.y2label = components.plot.y2.getLabel() === '' ? '' : components.plot.y2.getLabel();

            //check that min/max values are float or empty
            try {
                if (components.plot.x.getAutoMin() === true) {
                    request.xmin = "auto";
                } else {
                    request.xmin = components.plot.x.getMin() === '' ? '' : components.plot.x.getMin();
                }

                request.xmax = components.plot.x.getMax() === '' ? '' : components.plot.x.getMax();
                request.y1min = components.plot.y1.getMin() === '' ? '' : components.plot.y1.getMin();
                request.y1max = components.plot.y1.getMax() === '' ? '' : components.plot.y1.getMax();
                request.y2min = components.plot.y2.getMin() === '' ? '' : components.plot.y2.getMin();
                request.y2max = components.plot.y2.getMax() === '' ? '' : components.plot.y2.getMax();

                if (request.y1.length === 0) {
                    throw ("You must select at least one y1 value to plot.");
                } else {
                    sendPlotRequest(request, sessionId);
                }
            } catch (e) {
                throw (e);
            }
        }


        function updatePlotArea(values) {
            function addYList(axis_name, checked, hiddable, hidden) {
                const hide_button = hiddable === true ? '<button id="axis_name_show" class="link">+</button>' : '';
                const hide_class = hidden === true ? 'hidden' : '';
                let select = `<p>
                              <fieldset>
                                <legend>axis_name axis  ${hide_button}</legend>
                                <div id="axis_name_config" class="${hide_class}">
                                  <p>
                                    <label for="axis_name_label">Label</label>
                                    <input type="text" id="axis_name_label" name="axis_name_label">
                                    <label for="axis_name_log" class="small-label">log</label>
                                    <input type="checkbox" id="axis_name_log" name="axis_name_log">
                                    <label for="axis_name__min" class="small-label">Min</label>
                                    <input type="text" id="axis_name_min" name="axis_name_min" class="small-input">
                                    <label for="axis_name_max" class="small-label">Max</label>
                                    <input type="text" id="axis_name_max" name="axis_name_max" class="small-input">
                                  </p>
                                  <ul class="property-list">`;
                let i = 0;
                let x_values = components.plot.x.getValues();
                for (const value of values) {
                    let unit = components.getTree().getNodeByKey(value).data.unit;
                    if (unit === undefined) {
                        unit = "";
                    }
                    // uncheck default x values (AV, distance, tauV) if user chose them as y value     
                    // no more than MAX_PLOT_QTY checked quantities
                    let checked_attr;
                    if (i >= MAX_PLOT_QTY) {
                        checked_attr = false;
                    } else {
                        checked_attr = x_values.indexOf(value) === -1 ? checked : "";
                    }
                    select += `<li>
                                <input type="checkbox" id="axis_name_data${i}" name="axis_name_data" value="${value}" ${checked_attr}>
                                <input type="text" id="axis_name_label${i}" name="axis_name_label${i}" placeholder="${value}" >
                                <label for='axis_name_data${i}' class="property-unit">${unit}</label>
                                <label for='axis_name_data${i}' class="property-name">${value}</label>
                              </li>`;
                    i += 1;
                }
                select += `   </ul>
                            </div>
                            </fieldset>
                           </p>`;

                return select.replace(/axis_name/g, axis_name);
            }

            components.plot.y1.getArea().html(addYList("y1", "checked", false, false));
            components.plot.y2.getArea().html(addYList("y2", "", true, true));
            components.plot.y2.setShowConfigAreaOnClick(components.plot.y2.toggleVisibilityConfigArea);
        }

        // only shows floders in tree, nodes are hidden
        components.getTree().filterNodes(function(node) {
            return node.isFolder();
        });

        components.getDataTextButton(SIMPLE_QTY).on("click", function() {
            sendExtractRequestAction("text", SIMPLE_QTY);
        });

        components.getDataTextButton(ARRAY_QTY).on("click", function() {
            sendExtractRequestAction("text", ARRAY_QTY);
        });

        components.getDataVotableButton(SIMPLE_QTY).on("click", function() {
            sendExtractRequestAction("votable", SIMPLE_QTY);
        });

        components.getDataVotableButton(ARRAY_QTY).on("click", function() {
            sendExtractRequestAction("votable", ARRAY_QTY);
        });

        components.getGoToPlotButton().on("click", function() {
            try {
                initializePlotArea();
                window.location.hash = 'plot_result';
                processPlotRequest(sessionId);
            } catch (e) {
                window.alert(e);
            }
        });

        components.getPlotButton().on("click", function() {
            processPlotRequest(sessionId);
        });

        components.getClearTagsButton().on("click", function() {
            // clear selected entries
            selector.reset();
            // clear error display areas
            components.getUserSelectionError(SIMPLE_QTY).html('');
            components.getUserSelectionError(ARRAY_QTY).html('');
            components.plot.x.resetArea();
            components.plot.y1.resetArea();
            components.plot.y2.resetArea();
            components.plot.result.resetArea();
            components.plot.hide();
            components.getDataStatus(SIMPLE_QTY).html("");
            components.getDataStatus(ARRAY_QTY).html("");

            $(".concept-link").each(function() {
                $(this).removeClass("selected");
            });
        });

        // data manually added in text area
        components.getUserSelection(ARRAY_QTY).on("input", () => {
            let entries = $.trim(components.getUserSelection(ARRAY_QTY).val()).split("\n");
            let nodes = [];
            for (const e of entries) {
                nodes.push({ "title": e, "type": ARRAY_QTY });
            }
            selector.updateSelected(nodes);
        });
    });
}

function initPlotArea(data) {
    let graphconfig = `<p>
                          <fieldset>
                            <legend>graph</legend>
                            <p>
                              <label for="graph_title">Title</label>
                              <input type="text"
                                     id="graph_title"
                                     name="graph_title"
                                     class="long-input"
                                     placeholder="(text fields accept LaTeX expressions enclosed in $ $)">
                            </p>
                          </fieldset>
                        </p>`;


    let xselect = ` <p>
                      <fieldset>
                        <legend>x axis</legend>
                        <p>
                          <label for="x_data">Quantity</label>
                          <select id="x_data">`;


    for (const x of data.xQuantities) {
        const selected = x === data.xDefault ? "selected" : "";
        xselect += `<option value="${x}" ${selected}>${x}</option>`;
    }
    xselect += `          </select>
                          <label for="x_label">Label</label>
                          <input type="text" id="x_label" name="x_label">
                          <label for="x_log" class="small-label">log</label>
                          <input type="checkbox" id="x_log" name="x_log"></p>
                          <p>
                            <label for="x_min">Min</label>
                            <input type="text" id="x_min" name="x_min" disabled>
                            <label for="x_max">Max</label>
                            <input type="text" id="x_max" name="x_max">
                          </p>
                          <label for="auto_x_min">Min auto</label>
                          <input type="checkbox" id="auto_x_min" name="auto_x_min" checked>
                        </p>
                      </fieldset>
                    </p>`;

    components.plot.x.getArea().html(graphconfig + xselect);
    components.plot.x.setAutoMinOnClick(components.plot.x.toggleMinDisabled);
}
/**
 * what happens in autocompletion bar
 */
function initAutoCompletion() {
    $(".content").css("visibility", "visible");
    $(".loading").css("display", "none");
    new autoComplete({
        selector: "#searchbar",
        minChars: 2,
        delay: 1000,
        selection: [],
        source: function(term, response) {
            if ($.trim(term).length > 0) {
                $.ajax({
                    url: GLOBAL_SERVICE + "autocomplete/" + $.trim(term),
                    type: "GET",
                }).done(function(suggestions) {
                    response(suggestions);
                });
            }
        },
        renderItem: function(item, search) {
            var ssearch = search.replace(/[-\/\\^$*+?.()|[\]{}]/g, '\\$&'),
                re = new RegExp("(" + ssearch.split(' ').join('|') + ")", "gi"),
                suggestion_l = ['<div class="autocomplete-suggestion"',
                    'data-val="' + item[0] + '"',
                    'data-pref="' + item[1] + '">',
                    item[0].replace(re, "<b>$1</b>") + '</div>'
                ];

            return suggestion_l.join('');
        },
        // term = selected value, item = DOM object
        onSelect: function(e, term, item) {
            var prefterm = item.getAttribute('data-pref'),
                node = $('#tree_container').fancytree("getTree").getNodeByKey(prefterm);
            if (node !== null) {
                let chain = $.trim(components.getUserSelection(node.type).val() +
                    QTY_SEPARATOR + components.getTree().getNodeByKey(node.key).title);
                let entries = $.trim(components.getUserSelection(ARRAY_QTY).val()).split("\n");
                // do not add term if already in list
                if (!entries.includes(prefterm)) {
                    // add text in textarea
                    components.getUserSelection(node.type).val(chain);
                    // trigger input event in modified area
                    components.getUserSelection(node.type).trigger("input");
                }
            } else {
                window.alert("This quantity is not available in this model.");
            }
        }
    });
}

try {
    let is_local = false;
    if (window.location.href.includes("localhost") || window.location.href.includes("127.0.0.1")) {
        components.setPageType('Local version');
        is_local = true;
    }
    const parameters = getParameter();
    let path = '';
    if (parameters.project !== false && parameters.model !== false) {
        path = GLOBAL_SERVICE + "projectload/" + parameters.project + "/" + parameters.model;
    } else {
        if (parameters.model !== false) {
            path = GLOBAL_SERVICE + "localload/" + parameters.model;
        }
    }
    $.get(path, function(data) {
        components.setFileName("Opened model: " + parameters.model);
        components.setPageTitle(parameters.model);
        //display navigation bar if online version
        if (!is_local)
            components.setNavBar(parameters.project, parameters.model);

        initPage(data);
        initPlotArea(data);
        initAutoCompletion();
        selector.reset();
    });
} catch (e) {
    window.alert(e.message);
}
