var RGBMIN = 0, // min/max value for rgb channels of the random colors
    RGBMAX = 160,

    HSLH = [30, 180], // blue
    HSLS = [30, 70],
    HSLL = [20, 50];

var sformat = function() {
        var args = Array.apply(null, arguments),
            s = args[0];

        return s.replace(/{(\d+)}/g, function(match, number) { 
            return typeof args[number] != 'undefined'
                ? args[number]
                : match      ;
            }
        );
    },

    _contains = function (str, substr) {
        return !!~('' + str).indexOf(substr);
    },

    create_element = function() {
        if (typeof document.createElement !== 'function') {
            // This is the case in IE7, where the type of createElement is "object".
            // For this reason, we cannot call apply() as Object is not a Function.
            return document.createElement(arguments[0]);
        } else {
            return document.createElement.apply(document, arguments);
        }
    },

    _is_array = function( obj ) {
        // stolen from jQuery source :-)
        return toString.call(obj) === "[object Array]";
    },

    _support_hsl = function() {
        // inspired by modernizr src
        var style = create_element('a').style;
        style.cssText = 'background-color:hsla(120,40%,100%,0.5);';
        return (_contains(style.backgroundColor, 'rgba') 
            || _contains(style.backgroundColor, 'hsla')); 
    },

    _map = function(f, l){
        var mapped_l = [];

        for(var i = 0; i < l.length; i++){
            mapped_l[i] = f(l[i]);
        }

        return mapped_l;
    },

    mix_rgbcolor = function(basec, mixc, bprop, mprop){
       var r = basec[0], g = basec[1], b = basec[2],
           t = bprop + mprop,
           red = (bprop * r + mprop * mixc[0]) / t,
           green = (bprop * g + mprop * mixc[1]) / t,
           blue = (bprop * b + mprop * mixc[2]) / t;

       return _map(parseInt, [red, green, blue]);
    },

    devhslcolor = function(color, msaturation, mlightness){
        // deviate hsl color by applying coeff to saturation and lightness
        var h = color[0],
            s = color[1],
            l = color[2];
      return [h, s + (100 - s) * msaturation, l + (100 - l) * mlightness];
    },

    _rand = function(inf_, sup_){
        if (inf_ >= sup_) {return sup_;}
        return Math.random() * (sup_ - inf_) + inf_;
    },

    randhslcolor = function(hue, saturation, lightness){
        var h = _is_array(hue) ? _rand(hue[0], hue[1]) : hue,
            s = _is_array(saturation) ? _rand(saturation[0], saturation[1]) : saturation,
            l = _is_array(lightness) ? _rand(lightness[0], lightness[1]) : lightness;
        return [h, s, l];
    },

    randrgbcolor = function(vmin, vmax){
        // vmin/vmax are there to tune the generated color so that we don't
        // have to light/dark base color
        return _map(parseInt, 
            [_rand(vmin, vmax), _rand(vmin, vmax), _rand(vmin, vmax)]);
    },

    _rgb2hexstr = function(rgb){
        var _dec2hex = function(d){
            var hexstr = d.toString(16);
            return hexstr.length == 1 ? "0" + hexstr : hexstr;
        };
        return _map(_dec2hex, rgb).join("");
    },

   _hsl2str = function(hsl){
        return sformat('hsl({1}, {2}%, {3}%)', 
            hsl[0], Math.floor(hsl[1]), 
            Math.floor(hsl[2]));
    },

    assign_colors = function(mode){        
  
          var  _make_colors = function(mode){
                var c = null, 
                    bc = null, 
                    strf = null,
                    white = [255, 255, 255];

                if (mode === undefined || mode == 'rgb'){$
                    console.log("rgb");
                    c = randrgbcolor(RGBMIN, RGBMAX);
                    bc = mix_rgbcolor(c, white, 1, 14);
                    strf = _rgb2hexstr;
                }else{
                    console.log("hsl");
                    c = randhslcolor(HSLH, HSLS, HSLL);
                    bc = devhslcolor(c, 0, 0.95);
                    strf = _hsl2str;
                }

                return [[c, strf(c)], [bc, strf(bc)]];
            };
        var result = _make_colors(mode);        
        return _make_colors(result);
    },

    build_tag = function(elems){                   
      return {backgroundColor:elems[1][1], color:elems[0][1]};   
    },
    
    build_style = function(){
      if (HSL){
          return build_tag(assign_colors('hsl'));
      }else{
          return build_tag(assign_colors());
      }    
    };
    
var HSL = _support_hsl();

