import logging
import os.path
from server.lib import config
from logging.handlers import RotatingFileHandler
from datetime import datetime

# création de l'objet logger qui va nous servir à écrire dans les logs
logger = logging.getLogger()
# on met le niveau du logger à DEBUG, comme ça il écrit tout
logger.setLevel(logging.INFO)

# création d'un formateur qui va ajouter le temps, le niveau
# de chaque message quand on écrira un message dans le log
formatter = logging.Formatter('%(asctime)s :: %(levelname)s :: %(message)s')
# création d'un handler qui va rediriger une écriture du log vers
# un fichier en mode 'append', avec 1 backup et une taille max de 1Mo
file_handler = RotatingFileHandler(os.path.join(config.SERVER_DIRECTORY,
                                                'activity.log'),
                                   'a', 1000000, 1)
# on lui met le niveau sur DEBUG, on lui dit qu'il doit utiliser le formateur
# créé précédement et on ajoute ce handler au logger
file_handler.setLevel(logging.DEBUG)
file_handler.setFormatter(formatter)
logger.addHandler(file_handler)


def append_info(text):
    now = datetime.now()
    logger.info("%s :  %s"%(now.strftime("%m/%d/%Y, %H:%M:%S"), text))


def append_exception(message):
    now = datetime.now()
    logger.exception("%s :  %s" % (now.strftime("%m/%d/%Y, %H:%M:%S"), message))
