import json 
import yaml

def add_quotes(value):
    return "'%s'" % str(value)


class YamlWriter(object):
    def __init__(self):
        self._content = ''

    def withX(self, prop):
        if prop is not None:
            self._content += "plot/x: '{}' \n".format(prop)
        else:
            raise Exception('x axis is missing')

    def withY1(self, props):
        if len(props) > 0:
            self._content += "plot/y1: "+json.dumps(props).replace('\\', '\\\\')\
                                                          .replace('],', '],\n')+" \n"
        else:
            raise Exception("y1 values list is empty")

    def withY2(self, props):        
        if len(props) > 0:
            self._content += "plot/y2: "+json.dumps(props).replace('\\', '\\\\')\
                                                          .replace('],', '],\n')+" \n"

    def withOutput(self, value):
        self._content += "plot/output: '{}' \n".format(value)

    def withXLabel(self, value):
        self._content += "plot/xlabel: '{}' \n".format(value)

    def withY1Label(self, value):
        self._content += "plot/y1label: '{}' \n".format(value)

    def withY2Label(self, value):
        self._content += "plot/y2label: '{}' \n".format(value)
        
    def withXMin(self, value):
        if value == "auto":
            self._content += "plot/xmin: '{}' \n".format(value)
        else:
            self._content += "plot/xmin: {} \n".format(value)

    def withY1Min(self, value):
        self._content += "plot/y1min: '{}' \n".format(value)

    def withY2Min(self, value):
        self._content += "plot/y2min: '{}' \n".format(value)
        
    def withXMax(self, value):
        self._content += "plot/xmax: {} \n".format(value)

    def withY1Max(self, value):
        self._content += "plot/y1max: '{}' \n".format(value)

    def withY2Max(self, value):
        self._content += "plot/y2max: '{}' \n".format(value)

    def withLog(self, axes):
        axes2 = [add_quotes(v) for v in axes]
        if len(axes) > 0:
            self._content += "plot/log: [ %s ] \n" % (','.join(axes2))
        else:
            self._content += "plot/log: [] \n"

    def withTitle(self, value):
        self._content += "plot/title: '{}' \n".format(value)

    def getContent(self):
        return self._content


def generate_yaml(params):
    writer = YamlWriter()

    for key, value in params.items():
        if key == 'x':
            writer.withX(value)
        elif key == 'y1':
            writer.withY1(value)
        elif key == 'y2':
            writer.withY2(value)
        elif key == 'output':
            writer.withOutput(value)
        elif key == 'xlabel':
            writer.withXLabel(value)
        elif key == 'y1label':
            writer.withY1Label(value)
        elif key == 'y2label':
            writer.withY2Label(value)
        elif key == 'xmin':
            writer.withXMin(value)
        elif key == 'y1min':
            writer.withY1Min(value)
        elif key == 'y2min':
            writer.withY2Min(value)
        elif key == 'xmax':
            writer.withXMax(value)
        elif key == 'y1max':
            writer.withY1Max(value)
        elif key == 'y2max':
            writer.withY2Max(value)
        elif key == 'log':
            writer.withLog(value)
        elif key == 'title':
            writer.withTitle(value)

    return writer.getContent()

def main():
    pass


if __name__ == '__main__':
    main()
