! Copyright (C) 2002 - Observatoire de Paris/Meudon
!                      Universite Paris 7
!                      CNRS (France)
!                      Contact : Jacques.Lebourlot@obspm.fr
!
! This program is free software; you can redistribute it and/or
! modify it under the terms of the GNU General Public License version 2
! as published by the Free Software Foundation.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this program; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

! JLB - 24 May 2008

MODULE CHEM_ANALYS

  PRIVATE

  PUBLIC :: ANALYS

CONTAINS

!%%%%%%%%%%%%%%%%
SUBROUTINE ANALYS
!%%%%%%%%%%%%%%%%

   USE PXDR_CONSTANTES
   USE PXDR_CHEM_DATA
   USE PXDR_PROFIL
   USE PXDR_STR_DATA
   USE PREP_VAR
   USE PREP_AUTRE

   IMPLICIT NONE

   REAL (KIND=dp), DIMENSION(nqm) :: ab1, ab2, ab3
   REAL (KIND=dp), DIMENSION(nqm) :: fabric, destru

   CHARACTER (LEN=lenSpeci) :: yspec
   REAL (KIND=dp)           :: dmax, fmax
   REAL (KIND=dp)           :: dpc, fpc, toto, dtot, ftot, eq_rel
   REAL (KIND=dp)           :: seuil, answ
   INTEGER                  :: ii, jj, jmax
   INTEGER                  :: j1, j2, j3, mm
   INTEGER                  :: iff, id, k
   INTEGER                  :: nrxi, isp
   INTEGER                  :: ii0, iopt
   INTEGER                  :: skip_h2p

   abnua(n_var+1,0:npo) = 1.0_dp
   abnua(n_var+2,0:npo) = 1.0_dp
   abnua(n_var+3,0:npo) = 1.0_dp
   abnua(n_var+4,0:npo) = 1.0_dp
   abnua(n_var+5,0:npo) = 1.0_dp
   abnua(n_var+6,0:npo) = 1.0_dp
   abnua(n_var+7,0:npo) = 1.0_dp

   PRINT *, '*****************************************'
   PRINT *, '           Chemistry Analysis            '
   PRINT *, '*****************************************'
   PRINT *, ''
   PRINT *,'Do you want reaction rates profiles as a function of Av?'
   PRINT *,'[1] - yes, [0] - no (recommended choice)'
   READ *,ii0
   IF (ii0 == 1) THEN
      CALL EVOLREACT
   ENDIF

   PRINT *,'Do you want chemistry heating rates by reactions ? '
   PRINT *,'[1] - yes, [0] - no (recommended choice)'
   READ *,ii0
   IF (ii0 == 1) THEN
      CALL HEATCOOLCHEM
   ENDIF

   PRINT *, ''
   PRINT *, '-----------------------------------------'
   PRINT *, '  Start chemistry analysis at a fixed Av '
   PRINT *, '-----------------------------------------'

   answ = 0
   DO WHILE (answ /= 1)
      PRINT *, 'Select the position in the cloud (position index)'
      PRINT *, '- Maximum value is: ', npo
      PRINT *, '- To end analysis enter: -1'
      READ *, iopt

      IF (iopt == -1) THEN
         RETURN
      ENDIF

      PRINT *
      av = tau(iopt) * tautav
      PRINT *, 'Point ', iopt, ', Av =', av, ", T = ", tgaz(iopt),&
              "  nH = ", densh(iopt)
      PRINT *, 'Keep that position ?'
      PRINT *, '[1] - yes,  [0] - no'
      READ  *, answ
   ENDDO

   PRINT *, 'Select threshold'
   PRINT *, '- High value => more reactions'
   PRINT *, '- All reactions:  0'
   PRINT *, '- Typical value: 10 to 100'
   READ *, seuil

   temp = tgaz(iopt)

   DO
      PRINT *
      PRINT *, 'Select species? (End: -1)'
      READ (*,'(a8)') yspec
      IF (yspec == '-1') THEN
         EXIT
      ENDIF

      DO jj = 1, n_var
         IF (speci(jj) == yspec) THEN
            isp = jj
            EXIT
         ENDIF
      ENDDO
      IF (jj > nsp_n1 .AND. jj <= nsp_n1+nsp_g1) THEN
         PRINT *, yspec, " is not computed directly"
         PRINT *, "      Use and check the corresponding moment equations."
         CYCLE
      ELSE IF (jj > n_var) THEN
         PRINT *, yspec, " is not in chemistry"
         CYCLE
      ELSE
         nrxi = xmatk(isp)%nbt
         PRINT *, "  Number of reactions: ", nrxi
      ENDIF

      ! Consider only true reactions of H2, without adsorption and desorption
      IF (yspec == "h2" .OR. yspec == "hd") THEN
         PRINT *, ' Do you want to skip adsorption / desorption of: ', yspec,'? (yes:1, no: 0)'
         READ *, skip_h2p
      ENDIF
      IF (yspec == "h" .OR. yspec == "d") THEN
         PRINT *, " WARNING !"
         PRINT *, " Full informations on formation and destruction of ", yspec
         PRINT *, "      are not available here."
         PRINT *, " Partial informations on some reactions are provided, but"
         PRINT *, "         full balance cannot be checked"
         PRINT *, " Use with care."
      ENDIF

      fabric(1:neqt) = 0.0_dp
      destru(1:neqt) = 0.0_dp

      DO k = 1, nrxi
         j1 = xmatk(isp)%fdt(k)%r1
         j2 = xmatk(isp)%fdt(k)%r2
         j3 = xmatk(isp)%fdt(k)%r3
         ab1(k) = abnua(j1,iopt)
         ab2(k) = abnua(j2,iopt)
         ab3(k) = abnua(j3,iopt)
      ENDDO

!    Formation and destruction

      iff = 0
      id = 0
      ftot = 0.0_dp
      dtot = 0.0_dp
      mm = 0
      DO k = 1, nrxi
         IF ((yspec == "h2" .OR. yspec == "hd") .AND. skip_h2p == 1 .AND. &
             (xmatk(isp)%fdt(k)%it == 113 .OR.  &
              xmatk(isp)%fdt(k)%it == 117 .OR. xmatk(isp)%fdt(k)%it == 118)) THEN
            CYCLE
         ENDIF
         IF (xmatk(isp)%fdt(k)%it == 111 .OR. xmatk(isp)%fdt(k)%it == 211) THEN
            mm = mm + 1
            toto = SIGN(1.0_dp,xmatk(isp)%fdt(k)%k) * chem_source_gr(mm,xmatk(isp)%fdt(k)%nb,iopt)
            !print *, k, mm, xmatk(isp)%fdt(k)%k, chem_source_gr(mm,xmatk(isp)%fdt(k)%nb,iopt)
            mm = MOD(mm,npg)
         ELSE
            toto = SIGN(1.0_dp,xmatk(isp)%fdt(k)%k) * chem_source(xmatk(isp)%fdt(k)%nb,iopt)
         ENDIF
         IF (toto > 0.0_dp) THEN
            iff = iff + 1
            fabric(k) = toto
            ftot = ftot + fabric(k)
         ELSE IF (toto < 0.0_dp) THEN
            id = id + 1
            destru(k) = -toto
            dtot = dtot + destru(k)
         ENDIF
      ENDDO

      PRINT *
      PRINT *, '-----------------------------------------------'
      PRINT *, '        Chemistry of:  ', yspec
      PRINT *, '-----------------------------------------------'
      PRINT *
      PRINT *, 'Point ', iopt, ', Av =', av, ", T = ", tgaz(iopt),&
              "  nH = ", densh(iopt)
      IF (isp == nspec) THEN
         WRITE (6,'(" Photo-ejection of electrons by grains:",1p,E18.6," (cm-3 s-1)")') relgr(1,iopt)
         WRITE (6,'(" Recombination of electrons on grains: ",1p,E18.6," (cm-3 s-1)")') relgr(2,iopt)*abnua(isp,iopt)
      ENDIF
      IF (isp <= nspec) THEN
         WRITE (6,1003) ftot, dtot, abnua(isp,iopt)
      ELSE
         WRITE (6,1004) ftot, dtot, abnua(isp,iopt)
      ENDIF
      eq_rel = ABS(ftot - dtot) / (ftot + dtot)
      IF (eq_rel > 1.0e-2_dp) THEN
         WRITE (6,'("  WARNING: poor chemical balance:",1P,E15.6," %")') 100.0_dp * ABS(ftot - dtot) / (ftot + dtot)
      ENDIF
      PRINT *
      fpc = ftot
      dpc = dtot
      IF (seuil /= 0.0_dp) THEN
         ftot = ftot / seuil
         dtot = dtot / seuil
      ELSE
         ftot = 0.0_dp
         dtot = 0.0_dp
      ENDIF

      PRINT *, '-- Formation of ', yspec,':'
      WRITE (*,1005)
      DO ii = 1, nrxi
         fmax = 0.0_dp
         DO jj = 1, nrxi
            IF (fabric(jj) > fmax) THEN
               fmax = fabric(jj)
               jmax = jj
            ENDIF
         ENDDO

         IF (fmax < ftot .OR. fmax == 0.0_dp) EXIT
         toto = SIGN(1.0_dp,xmatk(isp)%fdt(jmax)%k) * chem_source(xmatk(isp)%fdt(jmax)%nb,iopt) &
              / (ab1(jmax) * ab2(jmax) * ab3(jmax))

         WRITE (6,1002) xmatk(isp)%fdt(jmax)%it &
                   , speci(xmatk(isp)%fdt(jmax)%r1), speci(xmatk(isp)%fdt(jmax)%r2) &
                   , speci(xmatk(isp)%fdt(jmax)%p1), speci(xmatk(isp)%fdt(jmax)%p2) &
                   , speci(xmatk(isp)%fdt(jmax)%p3), speci(xmatk(isp)%fdt(jmax)%p4) &
                   , fmax, fmax / abnua(isp,iopt),   toto, 100.0_dp*fmax/fpc
         fabric(jmax) = 0.0_dp
      ENDDO

      PRINT *, ' '
      PRINT *, '-- Destruction of ', yspec,':'
      WRITE (*,1005)
      DO ii = 1, nrxi
         dmax = 0.0_dp
         DO jj = 1, nrxi
           IF (destru(jj) > dmax) THEN
             dmax = destru(jj)
             jmax = jj
           ENDIF
         ENDDO

         IF (dmax < dtot .OR. dmax == 0.0_dp) EXIT
         toto = SIGN(1.0_dp,xmatk(isp)%fdt(jmax)%k) * chem_source(xmatk(isp)%fdt(jmax)%nb,iopt) &
              / (ab1(jmax) * ab2(jmax) * ab3(jmax))

         WRITE (6,1002) xmatk(isp)%fdt(jmax)%it &
                    , speci(xmatk(isp)%fdt(jmax)%r1), speci(xmatk(isp)%fdt(jmax)%r2) &
                    , speci(xmatk(isp)%fdt(jmax)%p1), speci(xmatk(isp)%fdt(jmax)%p2) &
                    , speci(xmatk(isp)%fdt(jmax)%p3), speci(xmatk(isp)%fdt(jmax)%p4) &
                    , dmax, dmax / abnua(isp,iopt),   toto, 100.0_dp*dmax/dpc
         destru(jmax) = 0.0_dp
      ENDDO

      PRINT *, '-----------------------------------------------'

   ENDDO

 1002 format (1x,i4,1x,a8,' + ',a8,' = ',4a8,5x,1pe12.4,'  cm-3 s-1',e12.4,' s-1' &
             ,1pe12.4,' cm3 s-1 or s-1',0pf8.2,' %',i4)
 1003 format (1x,'fabric:',1pe15.6,' (cm-3 s-1), destru:',e15.6,' (cm-3 s-1), abond:',e15.6,' (cm-3)')
 1004 format (1x,'fabric:',1pe15.6,' (cm-3 s-1), destru:',e15.6,' (cm-3 s-1), abond:',e15.6,' (part / grain)')
 1005 format (67x,'Reaction rate (R)     R/n(X)          k                           %' )

END SUBROUTINE ANALYS

!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
SUBROUTINE HEATCOOLCHEM
!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

   USE PXDR_CONSTANTES
   USE PXDR_CHEM_DATA
   USE PXDR_PROFIL
   USE PREP_VAR
   USE PREP_AUTRE

   IMPLICIT NONE

   CHARACTER(len=lenfilename) :: file_HeatCoolChem
   CHARACTER(len=lenSpeci)    :: speci1, speci2, speci3
   CHARACTER(len=lenSpeci)    :: speci4, speci5, speci6, speci7
   INTEGER                    :: answ
   INTEGER                    :: iopt
   INTEGER                    :: ideb, ifin
   INTEGER                    :: ir1, ir2, ir3
   INTEGER                    :: ip1, ip2
   INTEGER                    :: i

   REAL(KIND=dp)              :: ab1, ab2, ab3
   REAL(KIND=dp)              :: rate
   REAL(KIND=dp)              :: xkchim, thresh
   REAL(KIND=dp)              :: xkchim_tot

   !----------------------------------------------------------------------------
   ! SELECT AN AV
   !----------------------------------------------------------------------------
   PRINT *, '-----------------------------------------'
   PRINT *, '             Select and AV               '
   PRINT *, '-----------------------------------------'

   answ = 0
   DO WHILE (answ /= 1)
      PRINT *, 'Give Av position index'
      PRINT *, 'max value: (iopt):', npo
      PRINT *, 'End analysis: -1'
      READ *, iopt

      IF (iopt == -1) THEN
         RETURN
      ENDIF

      PRINT *
      av = tau(iopt) * tautav
      PRINT *, 'Point ', iopt, ', Av =', av, ", T = ", tgaz(iopt), "  nH = ", densh(iopt)
      PRINT *, 'Keep that point?'
      PRINT *, '1- yes   0- no'
      READ  *, answ
   ENDDO

   PRINT *, "Give a threshold (Fraction of total chemical heting/cooling)"
   PRINT *, "     between 0 (no reaction) and 1 (all)"
   READ *, thresh

   !----------------------------------------------------------------------------
   ! SELECT AN OUTPUT FILE
   !----------------------------------------------------------------------------
   PRINT *,""
   PRINT *,"Enter output file name"
   READ  *, file_HeatCoolChem
   OPEN(70, file=file_HeatCoolChem, status="UNKNOWN", action="WRITE")
   WRITE(70, 1072)
   WRITE(70, 1073) fichier
   WRITE(70, 1074) iopt, av, tgaz(iopt), densh(iopt), chochi(iopt)
   WRITE(70, 1075) ! separator
   WRITE(70, 1076) ! columns titles
   WRITE(70, 1077) ! columns units

   xkchim_tot = 0.0_dp

   !----------------------------------------------------------------------------
   ! Compute heating / cooling rate for each reaction [erg cm-3 s-1]
   !----------------------------------------------------------------------------
   ideb = nh2for + ncrion + npsion + nasrad + 1
   ifin = ideb   + nrest  - 1

   DO i = ideb, ifin

      !--- Get abundances of reactants -----------------------------------------
      ir1 = react(i,1)
      ir2 = react(i,2)
      ir3 = react(i,3)

      ab1 = abnua(ir1,iopt)
      ab2 = abnua(ir2,iopt)
      ab3 = abnua(ir3,iopt)

      rate = chem_source(i,iopt) / (ab1 * ab2 * ab3)
      xkchim = chem_source(i,iopt) * de(i) * everg

      speci1 = speci(react(i,1))
      speci2 = speci(react(i,2))
      speci3 = speci(react(i,3))
      speci4 = speci(react(i,4))
      speci5 = speci(react(i,5))
      speci6 = speci(react(i,6))
      speci7 = speci(react(i,7))

      IF (ABS(xkchim) > thresh * ABS(chochi(iopt))) THEN
         WRITE(70,1070) speci1, speci2, speci3, speci4, speci5, speci6, speci7, &
                        rate, xkchim, de(i), itype(i)
      ENDIF
      xkchim_tot = xkchim_tot + xkchim
   ENDDO

   !----------------------------------------------------------------------------
   ! Add here endothermical reactions with H2 (wrong above)
   !----------------------------------------------------------------------------
   ideb = nh2for + ncrion  + npsion + nasrad + nrest + 1
   ifin = ideb   + nh2endo - 1

   DO i = ideb, ifin

      !--- Get abundances of reactants -----------------------------------------
      ir1 = react(i,1)
      ir2 = react(i,2)
      ir3 = react(i,3)
      ip1 = react(i,4)
      ip2 = react(i,5)

      ab1 = abnua(ir1,iopt)
      ab2 = abnua(ir2,iopt)
      ab3 = abnua(ir3,iopt)

      rate = chem_source(i,iopt) / (ab1 * ab2 * ab3)
      xkchim = chem_source(i,iopt) * de(i) * everg

      speci1 = speci(react(i,1))
      speci2 = speci(react(i,2))
      speci3 = speci(react(i,3))
      speci4 = speci(react(i,4))
      speci5 = speci(react(i,5))
      speci6 = speci(react(i,6))
      speci7 = speci(react(i,7))

      IF (ABS(xkchim) > thresh * ABS(chochi(iopt))) THEN
         WRITE(70,1070) speci1, speci2, speci3, speci4, speci5, speci6, speci7, &
                        rate, xkchim, de(i), itype(i)
      ENDIF
      xkchim_tot = xkchim_tot + xkchim
   ENDDO

   !----------------------------------------------------------------------------
   ! Add special reactions (type 101 to 110)
   !----------------------------------------------------------------------------
   ideb = nh2for + ncrion + npsion + nasrad + nrest + nh2endo + nphot_f + nphot_i + n3body + 1
   ifin = ideb   + ncpart - 1
   DO i = ideb, ifin

      !--- Get abundances of reactants -----------------------------------------
      ir1 = react(i,1)
      ir2 = react(i,2)
      ir3 = react(i,3)
      ip1 = react(i,4)
      ip2 = react(i,5)

      ab1 = abnua(ir1,iopt)
      ab2 = abnua(ir2,iopt)
      ab3 = abnua(ir3,iopt)

      rate = chem_source(i,iopt) / (ab1 * ab2 * ab3)
      xkchim = chem_source(i,iopt) * de(i) * everg

      speci1 = speci(react(i,1))
      speci2 = speci(react(i,2))
      speci3 = speci(react(i,3))
      speci4 = speci(react(i,4))
      speci5 = speci(react(i,5))
      speci6 = speci(react(i,6))
      speci7 = speci(react(i,7))

      IF (ABS(xkchim) > thresh * ABS(chochi(iopt))) THEN
         WRITE(70,1070) speci1, speci2, speci3, speci4, speci5, speci6, speci7, &
                        rate, xkchim, de(i), itype(i)
      ENDIF
      xkchim_tot = xkchim_tot + xkchim
   ENDDO

   WRITE(70,1071) xkchim_tot

   PRINT *,"Data stored in output file"
   PRINT *,"Continue with chemistry analyzis"
   PRINT *,""

   CLOSE(70)

!1070 FORMAT(3(A8," , "),"-> , ",4(A8," , "),1pE12.4E3," , ",1pE12.4E3," , ",1pE12.4E3,i8)
1070 FORMAT(3(A8,"   "),"->   ",4(A8,"   "),1pE12.4E3,"   ",1pE12.4E3,"   ",1pE12.4E3,i8)
1071 FORMAT("#Total heating rate by chemistry = ",1pE12.4E3,"  [erg cm-3 s-1]")
1072 FORMAT("#--- Heating rates by chemical reactions ----")
1073 FORMAT("#- Model: ",A255)
1074 FORMAT("#- Position:",I4,1p," , AV= ",E16.8E3,", T= ",E12.4,", nH= ",E12.4 &
           , ", ChoChi= ",E12.4)
1075 FORMAT("#--------------------------------------------")
1076 FORMAT("#React1  , React2   , React3   , -> , Prod1    ,  Prod2   ,&
            & Prod3    , Prod4    ,  Rate        ,  Heating rate,     Enthalpy,    Type")
1077 FORMAT("#        ,          ,          ,    ,          ,          ,&
            &          ,          ,  [cm-3 s-1]  ,  [erg cm-3 s-1],       [eV]")

END SUBROUTINE HEATCOOLCHEM

!%%%%%%%%%%%%%%%%%%%
SUBROUTINE EVOLREACT
!%%%%%%%%%%%%%%%%%%%

   USE PXDR_CONSTANTES
   USE PXDR_CHEM_DATA
   USE PXDR_PROFIL
   USE PREP_VAR
   USE PREP_AUTRE

   IMPLICIT NONE

!-------------------------------------------------------------
! DECLARATIONS
!-------------------------------------------------------------

   REAL (KIND=dp)             :: seuil, toto
   INTEGER                    :: numdest, numfabr
   INTEGER                    :: numdest2, numfabr2
   INTEGER                    :: iopt, k, isp
   INTEGER                    :: jj, nrxi
   INTEGER, DIMENSION(1:neqt) :: flagreact

   CHARACTER (LEN=75)         :: fichier1, fichier2, fichier3
   CHARACTER (LEN=lenSpeci)   :: yspec

   REAL (KIND=dp), DIMENSION(1:neqt,0:npo) :: fabri, destr
   REAL (KIND=dp), DIMENSION(0:npo)        :: fabtot, destot
   REAL (KIND=dp)                          :: taudest, taufabr
   REAL (KIND=dp), DIMENSION(0:npo,1:neqt) :: pourfabr
   REAL (KIND=dp), DIMENSION(0:npo,1:neqt) :: pourdest
   REAL (KIND=dp), DIMENSION(0:npo,1:neqt) :: pourfabr2
   REAL (KIND=dp), DIMENSION(0:npo,1:neqt) :: pourdest2

!------------------------------------------------------------
! xmatk(isp)%fdt(k)%k : cste de vitesse de form (si >0), de dest (si <0)
! de l'espece isp dans la k ieme reaction ou elle intervient.

!------------------------------------------------------------
! ESPECE
!------------------------------------------------------------
   PRINT *,'Which species do you want?'
   READ (*,'(a8)') yspec
   DO jj = 1, nspec
      IF (speci(jj) == yspec) THEN
         isp = jj               !num de l'esp choisie
      ENDIF
   ENDDO

   PRINT *,'Enter a threshold (%)'
   READ(*,*) seuil

!------------------------------------------------------------
! INITIALISATION
!------------------------------------------------------------
   nrxi = xmatk(isp)%nbt      ! nbre de reaction concernant isp
   DO k = 1, nrxi
      flagreact(k) = 0        ! on met un flag a zero a toutes ces reaction
   ENDDO                      ! qd flag vaut 1, on garde la reaction

   numfabr = 0                ! nbre de reactions de fabr qu'il faudra sortir
   numdest = 0                ! idem pour destruction

!----------------------------------------------------------
! CALCUL DES TAUX DE FABRICATION ET DE DESTRUCTION TOTAL.
!----------------------------------------------------------
   DO iopt = 0, npo
      temp = tgaz(iopt)
      av = tau(iopt) * tautav

      fabri(1:neqt,iopt) = 0.0_dp
      destr(1:neqt,iopt) = 0.0_dp
      fabtot(iopt)       = 0.0_dp
      destot(iopt)       = 0.0_dp

      DO k = 1, nrxi
         toto = SIGN(1.0_dp,xmatk(isp)%fdt(k)%k) * chem_source(xmatk(isp)%fdt(k)%nb,iopt) !Source de la kieme reaction
         IF (toto > 0.0_dp) THEN
            fabri(k,iopt) = toto
            fabtot(iopt) = fabtot(iopt) + fabri(k,iopt)
         ELSE IF (toto < 0.0_dp) THEN
            destr(k,iopt) = -toto
            destot(iopt) = destot(iopt) + destr(k,iopt)
         ENDIF
      ENDDO

   ENDDO

!--------------------------------------------------------------------
! RECHERCHE DES REACTIONS A CONSERVER
!--------------------------------------------------------------------
! En chaque point iopt, on calcul le pourcentage que represente chaque
! reaction et si ce pourcentage est superieur au seuil on la garde.
! Les reaction qu'on garde sont associees a un indice different de 0
! dans le tableau flagreact. Si cet indice vaut 1, c'est une reaction
! de formation, s'il vaut -1, de destruction.
! Que la reaction soit conservee ou non, les pourcentages sont conserves
! dans deux tableaux: pourdest et pourfabr.

   DO iopt = 0, npo
      DO k = 1, nrxi
         toto = xmatk(isp)%fdt(k)%k
         IF (toto < 0.0_dp) THEN
            taudest = destr(k, iopt)
            pourdest(iopt,k) = 100.0_dp * taudest / destot(iopt)
            IF (pourdest(iopt,k) > seuil) THEN
               IF (flagreact(k) == 0) THEN
                  flagreact(k)=-1
                  numdest = numdest+1
               ENDIF
            ENDIF
         ELSE IF (toto > 0.0_dp) THEN
            taufabr = fabri(k, iopt)
            pourfabr(iopt,k) = 100.0_dp * taufabr / fabtot(iopt)
            IF (pourfabr(iopt,k) > seuil) THEN
               IF (flagreact(k) == 0) THEN
                  flagreact(k)=1
                  numfabr = numfabr +1
               ENDIF
            ENDIF
         ENDIF
      ENDDO
   ENDDO


! Pour les sorties, il ne faut garder dans deux tableaux que les pourcentages
! des reactions conservees. On stocke ces pourcentages dans deux nouveaux
! tableaux pourdest2 et pourfabr2.
! On releve egalement a cette occasion les nombre de reaction de formation
! et le nombre de reaction de destruction.

   DO iopt = 0, npo
      numfabr2 = 0
      numdest2 = 0
      DO k = 1, nrxi
         IF (flagreact(k) == 1) THEN
            numfabr2 = numfabr2 +1
            pourfabr2(iopt,numfabr2)  = pourfabr(iopt,k)
         ENDIF
         IF (flagreact(k) == -1) THEN
            numdest2 = numdest2 +1
            pourdest2(iopt,numdest2) = pourdest(iopt,k)
         ENDIF
      ENDDO
   ENDDO

!---------------------------------------------------------------------
! SORTIE DES RESULTATS
!---------------------------------------------------------------------
   fichier1 = 'chimevold'
   fichier2 = 'chimevolf'
   fichier3 = 'chimevolr'

   OPEN(10,file=fichier1,status='unknown')
   OPEN(20,file=fichier2,status='unknown')
   OPEN(30,file=fichier3,status='unknown')

   WRITE(30,*) 'Reactions de fabrication'
   DO k = 1, nrxi
      IF (flagreact(k) == 1) THEN
         WRITE(30,*) speci(xmatk(isp)%fdt(k)%r1), speci(xmatk(isp)%fdt(k)%r2) &
                   , speci(xmatk(isp)%fdt(k)%r3), speci(xmatk(isp)%fdt(k)%p1) &
                   , speci(xmatk(isp)%fdt(k)%p2), speci(xmatk(isp)%fdt(k)%p3) &
                   , speci(xmatk(isp)%fdt(k)%p4)
      ENDIF
   ENDDO
   WRITE(30,*)
   WRITE(30,*) 'Reactions de destruction'
   DO k = 1, nrxi
      IF (flagreact(k) == -1) THEN
         WRITE(30,*) speci(xmatk(isp)%fdt(k)%r1), speci(xmatk(isp)%fdt(k)%r2) &
                   , speci(xmatk(isp)%fdt(k)%r3), speci(xmatk(isp)%fdt(k)%p1) &
                   , speci(xmatk(isp)%fdt(k)%p2), speci(xmatk(isp)%fdt(k)%p3) &
                   , speci(xmatk(isp)%fdt(k)%p4)
      ENDIF
   ENDDO

   PRINT *,'Nbre of formation reaction kept:',numfabr
   PRINT *,'Nbre of destruction reaction kept:',numdest
   PRINT *,'Results are in files:'
   PRINT *,' - chimevolr:    list of reactions'
   PRINT *,' - chimevolf:    formation rates'
   PRINT *,' - chimevold:    destructionrates '
   PRINT *,'Rem: Columns order are the same in all files'
   PRINT *

   DO iopt = 0, npo
      av = tau(iopt) * 1.085736
      WRITE(10,'(1p,100e12.5)') av, (pourdest2(iopt,k), k=1,numdest)
      WRITE(20,'(1p,100e12.5)') av, (pourfabr2(iopt,k), k=1,numfabr)
   ENDDO

   CLOSE(10)
   CLOSE(20)
   CLOSE(30)

!-------------------------------------------------------------------
! FIN
!-------------------------------------------------------------------

END SUBROUTINE EVOLREACT

END MODULE CHEM_ANALYS
