PROGRAM read_charge

   IMPLICIT NONE
   INTEGER, PARAMETER                            :: dp = SELECTED_REAL_KIND(15)

   CHARACTER (LEN=200)                           :: fichier
   INTEGER,        PARAMETER                     :: noptm  = 2000
   INTEGER,        PARAMETER                     :: npg = 12
   INTEGER,        PARAMETER                     :: izm0 = 1000       ! Maximum negative charge
   INTEGER,        PARAMETER                     :: izp0 = 10000      ! Maximum positive charge
   REAL (KIND=dp), ALLOCATABLE, DIMENSION(:)     :: abelec, Av
   REAL (KIND=dp), ALLOCATABLE, DIMENSION(:,:,:) :: frac              ! Charge distribution
   INTEGER,        ALLOCATABLE, DIMENSION(:,:)   :: jzmin, jzmax
   REAL (KIND=dp), ALLOCATABLE, DIMENSION(:,:,:) :: xjpe
   REAL (KIND=dp), ALLOCATABLE, DIMENSION(:,:,:) :: xje
   REAL (KIND=dp), ALLOCATABLE, DIMENSION(:,:,:) :: xjit

   INTEGER                                       :: npo, iopt
   INTEGER                                       :: mm, iz
   INTEGER                                       :: izmin, izmax

   ALLOCATE (Av(0:noptm))
   ALLOCATE (abelec(0:noptm))
   ALLOCATE (frac(-izm0:izp0,npg,0:noptm))
   ALLOCATE (jzmin(npg,0:noptm), jzmax(npg,0:noptm))
   ALLOCATE (xjpe(-izm0:izp0,npg,0:noptm))
   ALLOCATE (xje(-izm0:izp0,npg,0:noptm))
   ALLOCATE (xjit(-izm0:izp0,npg,0:noptm))
   frac  = 0.0_dp
   frac  = 0.0_dp
   jzmin = 0
   jzmax = 0
   xjpe  = 0.0_dp
   xje   = 0.0_dp
   xjit  = 0.0_dp

   PRINT *, " Enter file name"
   READ *, fichier
   PRINT *, " Opening:", fichier
   OPEN (46, file = fichier, status = 'old', form = 'unformatted')

   DO
      READ (46,END=100) iopt     
      READ (46) Av(iopt)
      READ (46) abelec(iopt)
      READ (46) (jzmin(mm,iopt), mm = 1, npg)
      READ (46) (jzmax(mm,iopt), mm = 1, npg)
      READ (46) (xjit(jzmin(mm,iopt)-1:jzmax(mm,iopt),mm,iopt), mm = 1, npg)
      READ (46) (xjpe(jzmin(mm,iopt)-1:jzmax(mm,iopt),mm,iopt), mm = 1, npg)
      READ (46) (xje(jzmin(mm,iopt):jzmax(mm,iopt),mm,iopt), mm = 1, npg)
      READ (46) (frac(jzmin(mm,iopt):jzmax(mm,iopt),mm,iopt), mm = 1, npg)
   ENDDO

 100 CONTINUE
   npo = iopt
   PRINT *, " Read", npo, " lines"

   DO mm = 1, npg
      fichier = "Chg_map.dat"//TRIM(ADJUSTL(ii2c(mm)))
      OPEN (10, FILE=fichier, STATUS="unknown")
      izmin = MINVAL(jzmin(mm,:))
      izmax = MAXVAL(jzmax(mm,:))
      DO iopt = 0, npo
         DO iz = izmin, izmax
            WRITE (10,*) Av(iopt), iz, xjpe(iz,mm,iopt), xjit(iz,mm,iopt), xje(iz,mm,iopt) &
                       , xje(iz,mm,iopt)*abelec(iopt), frac(iz,mm,iopt)
         ENDDO
         WRITE (10,'(" ")')
      ENDDO
      CLOSE (10)
   ENDDO

CONTAINS

!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
CHARACTER (LEN=4) FUNCTION ii2c(ii)
!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

   IMPLICIT NONE
   INTEGER, INTENT (IN) :: ii
   INTEGER              :: icm, icc, icd, icu
   INTEGER              :: jj

   icm = 48 + ii/1000
   IF (icm == 48) THEN
      icm = 32
   ENDIF
   jj  = ii - 1000 * (ii/1000)
   icc = 48 + jj/100
   IF (icm == 32 .AND. icc == 48) THEN
      icc = 32
   ENDIF
   jj  = jj - 100 * (jj/100)
   icd = 48 + jj/10
   jj  = jj - 10 * (jj/10)
   icu = 48 + jj
   ii2c = CHAR(icm)//CHAR(icc)//CHAR(icd)//CHAR(icu)

END FUNCTION ii2c

END PROGRAM read_charge
