! Copyright (C) 2002 - Observatoire de Paris/Meudon
!                      Universite Paris 7
!                      CNRS (France)
!                      Contact : Jacques.Lebourlot@obspm.fr
!
! This program is free software; you can redistribute it and/or
! modify it under the terms of the GNU General Public License version 2
! as published by the Free Software Foundation.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this program; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

! JLB - 24 May 2008

MODULE PREP_AUTRE

  PRIVATE

  PUBLIC :: AUTRE

CONTAINS

!%%%%%%%%%%%%%%%
SUBROUTINE AUTRE
!%%%%%%%%%%%%%%%

   USE PXDR_CONSTANTES
   USE PXDR_CHEM_DATA
   USE PXDR_PROFIL
   USE PXDR_AUXILIAR
   USE PXDR_RF_TOOLS
   USE PREP_VAR

   IMPLICIT NONE

   CHARACTER (LEN=lenSpeci)                   :: yspec
   INTEGER                                    :: voie_disso
   INTEGER                                    :: mm, jj, i, j, k
   INTEGER                                    :: ii2, ii3
   REAL (KIND=dp)                             :: lam1, sect_C, sect_S
   REAL (KIND=dp), DIMENSION (:), ALLOCATABLE :: dtau_D
   REAL (KIND=dp), DIMENSION (:), ALLOCATABLE :: chg_p, chg_m

!--------+---------+---------+---------+---------+---------+---------+-*-------+

   ii2 = 0
   DO WHILE (ii2 /= -1)
      PRINT *, '  1: Gas Temperature'
      PRINT *, '  2: Dust Temperatures'
      PRINT *, '  3: Proton Density (n(H) + 2 n(H2) + n(H+) + ...)'
      PRINT *, '  4: Total gas density (n(H) + n(H2) + n(He) + ...)'
      PRINT *, '  5: Ionisation degree'
      PRINT *, '  6: Pressure'
      PRINT *, '  7: l (distance in cm)'
      PRINT *, '  8: UV photodestruction from integration of cross sections'
      PRINT *, '  9: formation destruction of H2 and HD'
      PRINT *, ' 10: Dust charge'
      PRINT *, ' 11: Number of ice layers'
      PRINT *, ' 12: Radiation (photon and energy densities from Lyman to Habing)'
      PRINT *, ' 13: Integrated molecular fraction'
      PRINT *, ' 14: Species characteristic times'
      PRINT *, ' 15: Optical depth at 1000 A'
      PRINT *, ' 16: Grain electr interaction'
      PRINT *, ' 17: Dust properties (du_prop)'
      PRINT *, ' -1: end'
      READ *, ii2
      IF (irec == 1) THEN
         WRITE (8,*) ii2
      ENDIF

      IF (ii2 == -1) THEN
         EXIT
      ENDIF

      !--- Gas temperature -----------------------------------------------------
      IF (ii2 == 1) THEN
         iimax = iimax + 1
         y(iimax,0:npo) = tgaz(0:npo)
         texty(iimax) = 'Temp.Gaz      '

      !--- Dust temperature ----------------------------------------------------
      ELSE IF (ii2 == 2) THEN
         DO mm = 1, npg
            iimax = iimax + 1
            y(iimax,0:npo) = tdust(mm,0:npo)
!           texty(iimax) = 'Td.'//TRIM(ADJUSTL(ii2c(mm)))
            write (texty(iimax),'("Td",a2,"-",1pe7.1)') TRIM(ADJUSTL(ii2c(mm))), adust(mm)
         ENDDO

      !--- Proton density ------------------------------------------------------
      ELSE IF (ii2 == 3) THEN
         iimax = iimax + 1
         y(iimax,0:npo) = densh(0:npo)
         texty(iimax) = 'Dens.H        '

      !--- Total density --------------------------------------------------------
      ELSE IF (ii2 == 4) THEN
         iimax = iimax + 1
         y(iimax,0:npo) = totdensity(0:npo)
         texty(iimax) = 'Dens.Tot      '

      !--- Ionization ----------------------------------------------------------
      ELSE IF (ii2 == 5) THEN
         iimax = iimax + 1
         y(iimax,0:npo) = ionidegree(0:npo)
         texty(iimax) = 'Ionisation    '

         ! Check charge conservation
         ALLOCATE (chg_p(0:npo))
         ALLOCATE (chg_m(0:npo))
         chg_p = 0.0_dp
         chg_m = 0.0_dp
         do j = 0, npo
            if (grcharge(j) >= 0.0_dp) then
               chg_p(j) = grcharge(j)
            else
               chg_m(j) = grcharge(j)
            endif
            DO jj = nsp_ne+1, nspec
               if (abnua(jj,j) * ael(jj,natom) >= 0.0_dp) then
                  chg_p(j) = chg_p(j) + abnua(jj,j) * ael(jj,natom)
               else
                  chg_m(j) = chg_m(j) + abnua(jj,j) * ael(jj,natom)
               endif
            ENDDO
         ENDDO

         iimax = iimax + 1
         y(iimax,0:npo) = 50.0_dp * (chg_p + chg_m) / (chg_p - chg_m)
         texty(iimax) = 'Charge-conv-% '
         iimax = iimax + 1
         y(iimax,0:npo) = chg_p + chg_m
         texty(iimax) = 'Charge-tot    '
         !iimax = iimax + 1
         !y(iimax,0:npo) = chg_p
         !texty(iimax) = 'Charge-p      '
         !iimax = iimax + 1
         !y(iimax,0:npo) = chg_m
         !texty(iimax) = 'Charge-m      '
         DEALLOCATE (chg_p, chg_m)

      !--- Pressure ------------------------------------------------------------
      ELSE IF (ii2 == 6) THEN
         iimax = iimax + 1
         y(iimax,0:npo) = gaspressure(0:npo)
         texty(iimax) = 'Pres.[K.cm-3]'

      !--- Distance ------------------------------------------------------------
      ELSE IF (ii2 == 7) THEN
         iimax = iimax + 1
         y(iimax,0) = 0.0_dp
         y(iimax,0:npo) = d_cm(0:npo)
         texty(iimax) = 'Position[cm]  '

      !--- UV photodestruction from integration of cross sections --------------
      ELSE IF (ii2 == 8) THEN
2006  CONTINUE
         PRINT *, ' which species ?'
         PRINT *, '-1: fin'
         READ (*,'(a8)') yspec
         IF (irec == 1) THEN
            WRITE (8,'(a8)') yspec
         ENDIF

         IF (yspec == "-1") THEN
            CYCLE
         ELSE
            voie_disso = 0
            DO jj = 1, nphot_i
               IF (ndiesp(jj) == yspec) THEN
                  voie_disso = voie_disso + 1
               ENDIF
            ENDDO
            IF (voie_disso == 0) THEN
               PRINT *, '  No cross-section for ', yspec
               GOTO 2006
            ELSE
               PRINT *, 'nbre de voies de disso: ', voie_disso
            ENDIF
            ii3 = 0
            voie_disso = 0
            DO jj = 1, nphot_i
               IF (ndiesp(jj) == yspec) THEN
                  voie_disso = voie_disso + 1
                  iimax = iimax + 1
                  y(iimax,:) = 0.0_dp
                  DO j = 0, npo
                     y(iimax,j) = pdiesp(jj,j)
                  ENDDO
                  texty(iimax) = 'Pd'//TRIM(ADJUSTL(ii2c(voie_disso)))//'.'//yspec
               ENDIF
            ENDDO
            GOTO 2006
         ENDIF

      !--- Formation and destruction of H2 and HD ------------------------------
      ELSE IF (ii2 == 9) THEN

         PRINT *, ' 1: H2, 2: CO, 3: HD'
         READ *, ii3
         IF (irec == 1) THEN
            WRITE (8,*) ii3
         ENDIF

         IF (ii3 == 1) THEN
            iimax = iimax + 1
            y(iimax,0) = 0.0_dp
            y(iimax,0:npo) = timh21(0:npo)
            texty(iimax) = 'FH2chi        '
            iimax = iimax + 1
            y(iimax,0) = 0.0_dp
            y(iimax,0:npo) = timh22(0:npo)
            texty(iimax) = 'DH2chi        '
            iimax = iimax + 1
            y(iimax,0) = 0.0_dp
            y(iimax,0:npo) = timh23(0:npo)
            texty(iimax) = 'FH2-g-LH      '
            iimax = iimax + 1
            y(iimax,0) = 0.0_dp
            y(iimax,0:npo) = timh24(0:npo)
            texty(iimax) = 'DH2phot       '
            iimax = iimax + 1
            y(iimax,0) = 0.0_dp
            y(iimax,0:npo) = timh25(0:npo)
            texty(iimax) = 'FH2-g-ER      '
            iimax = iimax + 1
            y(iimax,0:npo) = (timh23(0:npo) + timh25(0:npo)) &
                           / (vmaxw * SQRT(tgaz(0:npo)) * s_gr_t * densh(0:npo) * abnua(i_h,0:npo) / 8.0_dp)
            texty(iimax) = 'FH2-gr-eff   '                   ! Grain efficiency
         ELSE IF (ii3 == 2) THEN
            PRINT *, " To be upgraded (not done yet)"
         ELSE  IF (ii3 == 3) THEN
            iimax = iimax + 1
            y(iimax,0) = 0.0_dp
            y(iimax,0:npo) = timhd1(0:npo)
            texty(iimax) = 'FHDchi        '
            iimax = iimax + 1
            y(iimax,0) = 0.0_dp
            y(iimax,0:npo) = timhd2(0:npo)
            texty(iimax) = 'DHDchi        '
            iimax = iimax + 1
            y(iimax,0) = 0.0_dp
            y(iimax,0:npo) = timhd3(0:npo)
            texty(iimax) = 'FHDgrain      '
            iimax = iimax + 1
            y(iimax,0) = 0.0_dp
            y(iimax,0:npo) = timhd4(0:npo)
            texty(iimax) = 'DHDphot       '
         ELSE
            PRINT *, 'WARNING: illegal value for ii3:', ii3
            CYCLE
         ENDIF

      !--- Grain charge  -------------------------------------------------------
      ELSE IF (ii2 == 10) THEN
         iimax = iimax + 1
         y(iimax,0:npo) = abnua(nspec,0:npo)
         texty(iimax) = 'n(e-)[cm-3]   '
         iimax = iimax + 1
         y(iimax,0:npo) = grcharge(0:npo)
         texty(iimax) = 'Gr.Char[cm-3] '
         DO mm = 1, npg
            iimax = iimax + 1
            y(iimax,0:npo) = zmoy(mm,0:npo)
            texty(iimax) = 'R.gr.'//TRIM(ADJUSTL(ii2c(mm)))
         ENDDO
         PRINT *, " Do you want min and max charges?"
         PRINT *, "    yes: 1, No: 0"
         READ *, ii3
         IF (irec == 1) THEN
            WRITE (8,*) ii3
         ENDIF
         IF (ii3 == 1) THEN
            DO mm = 1, npg
               iimax = iimax + 1
               y(iimax,0:npo) = jzmin(mm,0:npo)
               texty(iimax) = 'zmin.'//TRIM(ADJUSTL(ii2c(mm)))
            ENDDO
            DO mm = 1, npg
               iimax = iimax + 1
               y(iimax,0:npo) = jzmax(mm,0:npo)
               texty(iimax) = 'zmax.'//TRIM(ADJUSTL(ii2c(mm)))
            ENDDO
         ENDIF

      !--- Number of layers ----------------------------------------------------
      ELSE IF (ii2 == 11) THEN
         iimax = iimax + 1
         y(iimax,0:npo) = x_lay(0:npo)               ! Real value
         texty(iimax) = 'Nb_Ice_Layers '

      !--- Radiation field  ----------------------------------------------------
      ELSE IF (ii2 == 12) THEN
         ! 7 I 2013 - JLB - Stop integration at 2400 Angstrom (Habing limit)
         ! BG 04/2013 - Start the integration at 911 Angstrom (Lyman limit)
         iimax = iimax + 1
         y(iimax,0:npo) = fluph(0:npo)
         texty(iimax) = 'Flu.PHOT      '
         iimax = iimax + 1
         y(iimax,0:npo) = int_rad(0:npo)
         texty(iimax) = 'U.[erg.cm-3]  '

      !--- Integrated molecular fraction ---------------------------------------
      ELSE IF (ii2 == 13) THEN
         iimax = iimax + 1
         y(iimax,0) = 2.0_dp * abnua(i_h2,0) / (2.0_dp*abnua(i_h2,0) + abnua(i_h,0))
         IF (i_hp /= 0) THEN
            y(iimax,1:npo) = 2.0_dp * codesp(i_h2,1:npo) / (2.0_dp*codesp(i_h2,1:npo) &
                           + codesp(i_h,1:npo) + codesp(i_hp,1:npo))
         ELSE
            y(iimax,1:npo) = 2.0_dp * codesp(i_h2,1:npo) / (2.0_dp*codesp(i_h2,1:npo) &
                           + codesp(i_h,1:npo))
         ENDIF
         texty(iimax) = ' f.int '

      !--- Species characteristic times ----------------------------------------
      ELSE IF (ii2 == 14) THEN
         CALL CAR_TIME
      !--- Optical depth at 1000 A ---------------------------------------------
      ELSE IF (ii2 == 15) THEN
         lam1 = 1000.0_dp
         sect_C = 1.6e-17_dp   ! Constant from 912 to 1101 A
         sect_S = 3.1e-17_dp   ! Typical value around 1000 A (without resonnances)
         CALL RF_LOCATE (rf_wl, lam1, jj)
         IF (.NOT. ALLOCATED(dtau_D)) THEN
            ALLOCATE (dtau_D(0:npo))
         ENDIF
         IF (F_DU_PROP == 0) THEN
            dtau_D = (du_prop%abso(jj,0) + du_prop%scat(jj,0)) * densh(0:npo) &
                   + sect_C * abnua(i_c,0:npo) + sect_S * abnua(i_s,0:npo)
         ELSE
            dtau_D = (du_prop%abso(jj,:) + du_prop%scat(jj,:)) * densh(0:npo) &
                   + sect_C * abnua(i_c,0:npo) + sect_S * abnua(i_s,0:npo)
         ENDIF
         iimax = iimax + 1
         y(iimax,0) = 0.0_dp
         DO jj = 1, npo
            y(iimax,jj) = y(iimax,jj-1) + 0.5_dp * (dtau_D(jj-1) + dtau_D(jj)) * (d_cm(jj) - d_cm(jj-1))
         ENDDO
         texty(iimax) = 'tau-1000      '
         DEALLOCATE(dtau_D)
      ELSE IF (ii2 == 16) THEN
         iimax = iimax + 1
         y(iimax,0:npo) = relgr(1,0:npo)               ! Real value
         texty(iimax) = 'el-eject'
         iimax = iimax + 1
         y(iimax,0:npo) = relgr(2,0:npo)               ! Real value
         texty(iimax) = 'el-recom'
      ELSE IF (ii2 == 17) THEN
         fichier = TRIM(out_dir)//"DU_PROP_"//c_ifaf//".dat"
         OPEN (26, FILE=fichier, STATUS="unknown")
         WRITE (26,'("#wl abso scat gcos emis(Av)")')
         WRITE (26,'("#",1p,100e10.3)') (tautav*tau(i), i = 1, npo, 10)
         DO j = 1, nwlg
            WRITE (26,*) rf_wl%Val(j), du_prop%abso(j,0), du_prop%scat(j,0), du_prop%gcos(j,0) &
                       , (du_prop%emis(j,i), i = 1, npo, 10)
         ENDDO
         CLOSE (26)
      ELSE
         PRINT *, 'WARNING: Illegal value for ii2:', ii2
         CYCLE
      ENDIF
   ENDDO

END SUBROUTINE AUTRE

!%%%%%%%%%%%%%%%%%%
SUBROUTINE CAR_TIME
!%%%%%%%%%%%%%%%%%%

   USE PXDR_CONSTANTES
   USE PXDR_CHEM_DATA
   USE PXDR_PROFIL
   USE PXDR_BILTHERM
   USE PREP_VAR

   IMPLICIT NONE

   INTEGER                                     :: iopt
   INTEGER                                     :: i, j, k
   REAL (KIND=dp)                              :: duma
   REAL (KIND=dp), DIMENSION(:,:), ALLOCATABLE :: form_sp
   CHARACTER (LEN=lenSpeci)                    :: spe

   ALLOCATE (form_sp(nspec,0:npo))

   DO iopt = 0, npo
      DO i = 1, nspec-nsp_g1-1
         duma = 0.0_dp
         DO k = 1, xmatk(i)%nbt
            IF (xmatk(i)%fdt(k)%k > 0.0_dp) THEN
               duma = duma + chem_source(xmatk(i)%fdt(k)%nb,iopt)
            ENDIF
         ENDDO
         form_sp(i,iopt) = duma
      ENDDO
   ENDDO

!  DO iopt = 0, npo
!     av = tau(iopt) * tautav
!     DO j = 1, nspec
!        WRITE(10,'(i5,1p,e20.8,i5,a10,3e15.6)') &
!             iopt, av, j, speci(j), form_sp(j,iopt), abnua(j,iopt), abnua(j,iopt) / form_sp(j,iopt)
!     ENDDO
!     WRITE (10,'(" ")')
!  ENDDO

   PRINT *, " Do you want also the densities and formation rates? (Yes: 1, No: 0)"
   READ *, k
   IF (irec == 1) THEN
      WRITE (8,*) k
   ENDIF

   i = 1
   DO WHILE (i == 1)
      PRINT *, " Which species (end: 0)?"
      READ *, spe
      IF (irec == 1) THEN
         WRITE (8,'(a8)') spe
      ENDIF
      IF (spe == "0") THEN
         i = 0
         CYCLE
      ENDIF
      DO j = 1, nspec
         IF (spe == speci(j)) EXIT
      ENDDO
      IF (j > nspec) THEN
         PRINT *, spe, " not found"
      ELSE
         PRINT *, j, speci(j)
      ENDIF
      iimax = iimax + 1
      y(iimax,0:npo) = (abnua(j,0:npo) / form_sp(j,0:npo)) / (Myr * 1.0e-6_dp)
      texty(iimax) = ' t.ca.'//speci(j)
      IF (k == 1) THEN
         iimax = iimax + 1
         y(iimax,0:npo) = abnua(j,0:npo)
         texty(iimax) = ' Ab.'//speci(j)
         iimax = iimax + 1
         y(iimax,0:npo) = form_sp(j,0:npo)
         texty(iimax) = ' FR.'//speci(j)
      ENDIF
   ENDDO

END SUBROUTINE CAR_TIME

END MODULE PREP_AUTRE
