! Copyright (C) 2002 - Observatoire de Paris/Meudon
!                      Universite Paris 7
!                      CNRS (France)
!                      Contact : Jacques.Lebourlot@obspm.fr
!
! This program is free software; you can redistribute it and/or
! modify it under the terms of the GNU General Public License version 2
! as published by the Free Software Foundation.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this program; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

! JLB - 24 May 2008

MODULE PREP_EMISSI

  PRIVATE

  PUBLIC :: EMISSI, VIEW_LINES, EMIINT, FSPECTRUM

CONTAINS

!%%%%%%%%%%%%%%%%
SUBROUTINE EMISSI
!%%%%%%%%%%%%%%%%

   USE PXDR_CONSTANTES
   USE PXDR_CHEM_DATA
   USE PXDR_PROFIL
   USE PXDR_AUXILIAR
   USE PXDR_STR_DATA
   USE PXDR_INTENSITIES
   USE PXDR_RF_TOOLS
   USE PREP_VAR

   IMPLICIT NONE

   CHARACTER (LEN=lenSpeci) :: yspec
   INTEGER                  :: ii, ju, jl, jj
   INTEGER                  :: iloc, i, ind_sp
   REAL (KIND=dp)                             :: ccc1, ccc2, ccc3, ccc4
   REAL (kind=dp), ALLOCATABLE, DIMENSION (:) :: wlr
   INTEGER                                    :: nwl, n
   CHARACTER (LEN=50)                         :: fichout

!--------+---------+---------+---------+---------+---------+---------+-*-------+

   PRINT *, '   Specify which problem:'
   PRINT *, '      0: Local emissivities'
   PRINT *, '      1: Line intensities'
   PRINT *, '      2: Detailed H2 analysis'
   IF (F_W_HDF5_STD == 1) THEN
      PRINT *, '   3: All lines intensities'
   ENDIF
   PRINT *, '      4: Full spectrum in a range'
   READ *, iloc
   IF (irec == 1) THEN
      WRITE (8,*) iloc
   ENDIF

   IF ((F_W_RF_ALB == 1) .AND. (iloc == 1 .OR. iloc == 3 .OR. iloc ==4)) THEN
      ! Read gas absorption coefficient from *.rf file
      IF (.NOT. ALLOCATED(g_absL)) THEN
         fichout = TRIM(ADJUSTL(modele))//".rf_alb"//c_ifaf
         fichier = TRIM(out_dir)//fichout
         OPEN(44, FILE=fichier, STATUS="unknown", FORM = 'unformatted')
         READ (44) nwl
         PRINT *, nwl, " wavelength points"

         ALLOCATE (wlr(nwl))
         ALLOCATE (g_absL(0:npo,nwl))

         READ (44) wlr
         DO n = 0, npo
            READ (44) ccc1, ccc2, g_absL(n,:), ccc3, ccc4
         END DO
         CLOSE (44)
         DEALLOCATE (wlr)
      ENDIF
   ENDIF

   IF (iloc == 1) THEN
      PRINT *, ' Compute integrated emissivities'
      CALL EMIINT
   ELSE IF (iloc == 2) THEN
      PRINT *, ' Enter H2 module'
      CALL EMIH2
   ELSE IF (F_W_HDF5_STD == 1 .AND. iloc ==3) THEN
      CALL PRINT_INTENSITIES
   ELSE IF (iloc == 4) THEN
      PRINT *, ' Compute full spectrum in a range'
      CALL FSPECTRUM
   ELSE IF (iloc == 0) THEN
      PRINT *, ' Compute local emissivities'
      PRINT *, " "
      PRINT *, " Possible species:"
      i = 1
      DO WHILE (i <= nspl)
         IF (MOD(i,3) /= 0) THEN
            WRITE (*,'(a10)',ADVANCE="NO") spec_lin(i)%nam
         ELSE
            WRITE (*,'(a10)') spec_lin(i)%nam
         ENDIF
         i = i + 1
      ENDDO
      WRITE (*,*)

      DO
         PRINT *
         PRINT *, ' Which species?  (end: -1)'
         READ (*,'(a8)') yspec
         yspec = TRIM(ADJUSTL(yspec))
         IF (irec == 1) THEN
            WRITE (8,'(a8)') yspec
         ENDIF
         IF (yspec == "-1") THEN
            PRINT *, "  End lines"
            EXIT
         ENDIF
         DO i = 1, nspec
            IF (yspec /= TRIM(ADJUSTL(speci(i)))) THEN
               CYCLE
            ELSE
               EXIT
            ENDIF
         ENDDO
         IF (i > nspec) THEN
            SELECT CASE (yspec)
            CASE ("c*o")
               i = nspec + 1
            CASE ("co*")
               i = nspec + 2
            CASE ("c*o*")
               i = nspec + 3
            CASE DEFAULT
               i = 0
            END SELECT
         ENDIF
         IF (i == 0 .AND. yspec /= '-1') THEN
            PRINT *, "  WARNING:", yspec, " is not in chemistry!"
            CYCLE
         ELSE IF (in_sp(i) == 0) THEN
            PRINT *, "  WARNING: excitation of ", yspec &
                   , " was not computed"
         ELSE
            ind_sp = in_sp(i)
         ENDIF

         ii = 999
         PRINT *, ' Emission lines of: ', yspec
         CALL VIEW_LINES(spec_lin(ind_sp), irec)
         DO WHILE (ii /= 0)
            PRINT *,"Enter line number (0 end)"
            READ *, ii
            IF (ii > spec_lin(ind_sp)%ntr) THEN
               PRINT *, ii, ' > ', spec_lin(ind_sp)%ntr, ' ! - END'
               ii = 0
               EXIT
            ENDIF
            IF (irec == 1) THEN
               WRITE (8,*) ii
            ENDIF

            IF (ii /= 0) THEN
               PRINT *, 'Collect line number:', ii
               iimax = iimax + 1
               y(iimax,0:npo) = spec_lin(ind_sp)%emi(ii,0:npo)
               texty(iimax) = 'EmL.'//TRIM(spec_lin(ind_sp)%nam)//'.'//TRIM(ADJUSTL(ii2c(ii)))
            ELSE
               PRINT *, 'End lines of: ', yspec
               EXIT
            ENDIF

            PRINT *, "Full line properties? (1: yes, 0: no)"
            READ *, jj
            IF (irec == 1) THEN
               WRITE (8,*) jj
            ENDIF
            IF (jj == 1) THEN
               ju = spec_lin(ind_sp)%iup(ii)
               jl = spec_lin(ind_sp)%jlo(ii)

               iimax = iimax + 1
               y(iimax,0:npo) = spec_lin(ind_sp)%abu(0:npo)
               texty(iimax) = TRIM(spec_lin(ind_sp)%nam)//'.abun.'//TRIM(ADJUSTL(ii2c(ii)))
               iimax = iimax + 1
               y(iimax,0:npo) = spec_lin(ind_sp)%vel(0:npo)
               texty(iimax) = TRIM(spec_lin(ind_sp)%nam)//'.vel.'//TRIM(ADJUSTL(ii2c(ii)))
               iimax = iimax + 1
               y(iimax,0:npo) = spec_lin(ind_sp)%xre(ju,0:npo)
               texty(iimax) = TRIM(spec_lin(ind_sp)%nam)//'.xreu.'//TRIM(ADJUSTL(ii2c(ii)))
               iimax = iimax + 1
               y(iimax,0:npo) = spec_lin(ind_sp)%xre(jl,0:npo)
               texty(iimax) = TRIM(spec_lin(ind_sp)%nam)//'.xrel.'//TRIM(ADJUSTL(ii2c(ii)))
               iimax = iimax + 1
               y(iimax,0:npo) = spec_lin(ind_sp)%XXl(ju,0:npo)
               texty(iimax) = TRIM(spec_lin(ind_sp)%nam)//'.XXlu.'//TRIM(ADJUSTL(ii2c(ii)))
               iimax = iimax + 1
               y(iimax,0:npo) = spec_lin(ind_sp)%XXl(jl,0:npo)
               texty(iimax) = TRIM(spec_lin(ind_sp)%nam)//'.XXll.'//TRIM(ADJUSTL(ii2c(ii)))
               iimax = iimax + 1
               y(iimax,0:npo) = spec_lin(ind_sp)%XXr(ju,0:npo)
               texty(iimax) = TRIM(spec_lin(ind_sp)%nam)//'.XXru.'//TRIM(ADJUSTL(ii2c(ii)))
               iimax = iimax + 1
               y(iimax,0:npo) = spec_lin(ind_sp)%XXr(jl,0:npo)
               texty(iimax) = TRIM(spec_lin(ind_sp)%nam)//'.XXrl.'//TRIM(ADJUSTL(ii2c(ii)))
               iimax = iimax + 1
               y(iimax,0:npo) = spec_lin(ind_sp)%odl(ii,0:npo)
               texty(iimax) = TRIM(spec_lin(ind_sp)%nam)//'.odl.'//TRIM(ADJUSTL(ii2c(ii)))
               iimax = iimax + 1
               y(iimax,0:npo) = spec_lin(ind_sp)%odr(ii,0:npo)
               texty(iimax) = TRIM(spec_lin(ind_sp)%nam)//'.odr.'//TRIM(ADJUSTL(ii2c(ii)))
               iimax = iimax + 1
               y(iimax,0:npo) = spec_lin(ind_sp)%dnu(ii,0:npo)
               texty(iimax) = TRIM(spec_lin(ind_sp)%nam)//'.dnu.'//TRIM(ADJUSTL(ii2c(ii)))
               iimax = iimax + 1
               y(iimax,0:npo) = spec_lin(ind_sp)%emi(ii,0:npo)
               texty(iimax) = TRIM(spec_lin(ind_sp)%nam)//'.emi.'//TRIM(ADJUSTL(ii2c(ii)))
               iimax = iimax + 1
               y(iimax,0:npo) = spec_lin(ind_sp)%pum(0:npo,ii)
               texty(iimax) = TRIM(spec_lin(ind_sp)%nam)//'.pum.'//TRIM(ADJUSTL(ii2c(ii)))
               iimax = iimax + 1
               y(iimax,0:npo) = spec_lin(ind_sp)%jiD(ii,0:npo)
               texty(iimax) = TRIM(spec_lin(ind_sp)%nam)//'.jiD.'//TRIM(ADJUSTL(ii2c(ii)))
               iimax = iimax + 1
               y(iimax,0:npo) = spec_lin(ind_sp)%bel(ii,0:npo)
               texty(iimax) = TRIM(spec_lin(ind_sp)%nam)//'.bel.'//TRIM(ADJUSTL(ii2c(ii)))
               iimax = iimax + 1
               y(iimax,0:npo) = spec_lin(ind_sp)%ber(ii,0:npo)
               texty(iimax) = TRIM(spec_lin(ind_sp)%nam)//'.ber.'//TRIM(ADJUSTL(ii2c(ii)))
               iimax = iimax + 1
               y(iimax,0:npo) = spec_lin(ind_sp)%exl(ii)
               texty(iimax) = TRIM(spec_lin(ind_sp)%nam)//'.exl.'//TRIM(ADJUSTL(ii2c(ii)))
               iimax = iimax + 1
               y(iimax,0:npo) = spec_lin(ind_sp)%exr(ii)
               texty(iimax) = TRIM(spec_lin(ind_sp)%nam)//'.exr.'//TRIM(ADJUSTL(ii2c(ii)))
            ENDIF

         ENDDO
      ENDDO

   ENDIF

END SUBROUTINE EMISSI

!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
SUBROUTINE VIEW_LINES(spelin, irec)
!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

   USE PXDR_CONSTANTES
   USE PXDR_STR_DATA

   IMPLICIT NONE
   TYPE(spectra), INTENT (IN) :: spelin
   INTEGER, INTENT (IN)       :: irec
   INTEGER                    :: ntr, nqn, ntrp
   INTEGER                    :: i, j
   CHARACTER (LEN=40)         :: frmt
   CHARACTER (LEN=1)          :: ld='(', rd=')'

   ntr = spelin%ntr
   nqn = spelin%nqn
   PRINT *, " List length ? (Max:", ntr, ")"
   READ *, ntrp
   IF (irec == 1) THEN
      WRITE (8,*) ntrp
   ENDIF

   PRINT *," num  - ",(spelin%qnam(j),j=1,nqn)   &
          ,"-> ",(spelin%qnam(j),j=1,nqn)

   DO i = 1, MIN(ntrp,spelin%ntr)
      WRITE (frmt,'(a1,"i4,5x,a",i2,",3x,a",i2,a1)') ld, LEN_TRIM(ADJUSTL(spelin%LiHname(i))) &
            , LEN_TRIM(spelin%inftr(i)), rd
      WRITE (*,frmt) i, TRIM(ADJUSTL(spelin%LiHname(i))),TRIM(spelin%inftr(i))
   ENDDO

END SUBROUTINE VIEW_LINES

! That routine recomputes the full detailed processes of H2 excitation at each point in the cloud
!%%%%%%%%%%%%%%%
SUBROUTINE EMIH2
!%%%%%%%%%%%%%%%

   USE PXDR_CONSTANTES
   USE PXDR_AUXILIAR
   USE PXDR_PROFIL
   USE PXDR_INITIAL
   USE PXDR_STR_DATA
   USE PXDR_CHEM_DATA
   USE PXDR_EXCITH2
   USE PREP_VAR

   IMPLICIT NONE

   REAL (KIND=dp), DIMENSION(:,:,:), ALLOCATABLE :: eqradh2
   REAL (KIND=dp), DIMENSION(:,:,:), ALLOCATABLE :: eqcolh2
!  REAL (KIND=dp), DIMENSION(:,:,:), ALLOCATABLE :: eqcash2
   REAL (KIND=dp), DIMENSION(:,:,:), ALLOCATABLE   :: forh2

   INTEGER                                       :: jh2, nl
   INTEGER                                       :: iopt
   INTEGER                                       :: ijl, iju, ivl, ivu
   INTEGER                                       :: lev1, lev2, levl, levu
   REAL (KIND=dp)                                :: q1, q2, q3, q4
   INTEGER                                       :: i, j, ii
   INTEGER                                       :: nqc, ntnr
   REAL (KIND=dp)                                :: gusgl, bet_ijl, bet_ijr, jintD
   REAL (KIND=dp)                                :: xx_ij, yy_ij
   REAL (KIND=dp), DIMENSION(:), ALLOCATABLE     :: provi, opd_GG
   LOGICAL                                       :: isthere

   jh2 = in_sp(i_h2)
   nl = spec_lin(jh2)%use

   ALLOCATE (eqradh2(nl,nl,0:npo))
   ALLOCATE (eqcolh2(nl,nl,0:npo))
   ALLOCATE (eqcash2(nl,nl,0:npo))
   eqradh2 = 0.0_dp
   eqcolh2 = 0.0_dp
   eqcash2 = 0.0_dp

   !---------------------------------------------------------------------
   !- H2 Radiative transitions
   !---------------------------------------------------------------------

   DO iopt = 0, npo
      DO i = 1, spec_lin(jh2)%ntr
         iju = spec_lin(jh2)%iup(i)
         IF (iju > nl) CYCLE
         ijl = spec_lin(jh2)%jlo(i)
         gusgl = spec_lin(jh2)%gst(iju) / spec_lin(jh2)%gst(ijl)

         bet_ijl = spec_lin(jh2)%bel(i,iopt)
         bet_ijr = spec_lin(jh2)%ber(i,iopt)
         jintD = spec_lin(jh2)%jiD(i,iopt)

         yy_ij = jintD
         yy_ij = yy_ij + (bet_ijl * spec_lin(jh2)%exl(i) + bet_ijr * spec_lin(jh2)%exr(i))
         xx_ij = (bet_ijl + bet_ijr) + yy_ij
         xx_ij = xx_ij * spec_lin(jh2)%aij(i)
         yy_ij = yy_ij * spec_lin(jh2)%aij(i) * gusgl

         eqradh2(iju,iju,iopt) = eqradh2(iju,iju,iopt) + xx_ij
         eqradh2(iju,ijl,iopt) = eqradh2(iju,ijl,iopt) - yy_ij
         eqradh2(ijl,ijl,iopt) = eqradh2(ijl,ijl,iopt) + yy_ij
         eqradh2(ijl,iju,iopt) = eqradh2(ijl,iju,iopt) - xx_ij
      ENDDO
   ENDDO


   !---------------------------------------------------------------------
   !- H2 Collision rates
   !---------------------------------------------------------------------

   ! ichh2 is a flag to choose a model for H + H2 collision rates
   IF (ichh2 == 0) THEN
      nqc = 4
   ELSE
      nqc = 3
   ENDIF
   IF (.NOT. ALLOCATED(q_h2_h_1)) THEN
      ALLOCATE (q_h2_h_1(l0h2,l0h2))
      ALLOCATE (q_h2_h_2(l0h2,l0h2))
      ALLOCATE (q_h2_h_3(l0h2,l0h2))
      ALLOCATE (q_h2_h_4(l0h2,l0h2))
      ALLOCATE (q_h2_he_1(l0h2,l0h2))
      ALLOCATE (q_h2_he_2(l0h2,l0h2))
      ALLOCATE (q_h2_he_3(l0h2,l0h2))
      ALLOCATE (q_h2_ph2_1(l0h2,l0h2))
      ALLOCATE (q_h2_ph2_2(l0h2,l0h2))
      ALLOCATE (q_h2_ph2_3(l0h2,l0h2))
      ALLOCATE (q_h2_oh2_1(l0h2,l0h2))
      ALLOCATE (q_h2_oh2_2(l0h2,l0h2))
      ALLOCATE (q_h2_oh2_3(l0h2,l0h2))
   ENDIF
   q_h2_h_1(:,:)   = -50.0_dp
   q_h2_h_2(:,:)   = 0.0_dp
   q_h2_h_3(:,:)   = 0.0_dp
   q_h2_h_4(:,:)   = 0.0_dp
   q_h2_he_1(:,:)  = -50.0_dp
   q_h2_he_2(:,:)  = 0.0_dp
   q_h2_he_3(:,:)  = 0.0_dp
   q_h2_ph2_1(:,:) = -50.0_dp
   q_h2_ph2_2(:,:) = 0.0_dp
   q_h2_ph2_3(:,:) = 0.0_dp
   q_h2_oh2_1(:,:) = -50.0_dp
   q_h2_oh2_2(:,:) = 0.0_dp
   q_h2_oh2_3(:,:) = 0.0_dp

   PRINT *
   PRINT *,'-------------- H2 levels ----------------'
   CALL LLEV_H2 (l0h2)
   PRINT *,'Nb of H2 levels (l0h2):', l0h2

   !- H + H2 collision rates --------------------------------------------
   IF (ichh2 == 0) THEN
      fichier = TRIM(data_dir)//'Collisions/q_h1_h2_MM.dat'
   ELSE
      !fichier = TRIM(data_dir)//'Collisions/q_h1_h2_DRF.dat'
      fichier = TRIM(data_dir)//'Collisions/q_h1_h2_DRFn.dat'
   ENDIF
   WRITE(iscrn,*) "opening file : ", TRIM(fichier)
   OPEN(iread2, file = fichier, status = 'old')
   READ (iread2,*) ntnr

   DO j = 1, ntnr
      READ (iread2,*) levu, levl, q1, q2, q3, q4
      IF ((levu <= l0h2) .AND. (levl <= l0h2)) THEN
         q_h2_h_1(levl,levu) = q1
         q_h2_h_2(levl,levu) = q2
         q_h2_h_3(levl,levu) = q3
         IF (nqc == 4) THEN
            q_h2_h_4(levl,levu) = q4
         ENDIF
      ENDIF
   ENDDO
   CLOSE (iread2)

   ! 5 jan 2000 - Coarse patch for missing transitions
   DO levu = 3, l0h2
      iju = njl_h2(levu)
      ivu = nvl_h2(levu)
      DO levl = 1, levu-1
         ijl = njl_h2(levl)
         IF (MOD(iju-ijl,2) /= 0) CYCLE
         ivl = nvl_h2(levl)

         IF (q_h2_h_1(levl,levu) == -50.0d0) THEN
            IF (ivl == ivu .OR. ivu == 1) THEN
               IF (ivl >= 4) THEN
                  lev2 = lev_h2(ivu-1,iju)
                  lev1 = lev_h2(ivl-1,ijl)
               ELSE IF (ijl <= 1) THEN
                  lev2 = lev_h2(ivu,iju-2)
                  lev1 = levl
               ELSE
                  lev2 = lev_h2(ivu,iju-2)
                  lev1 = lev_h2(ivl,ijl-2)
               ENDIF
            ELSE IF (ivu > ivl .AND.  (iju /= 0 .OR. ijl /= 0)) THEN
               lev2 = lev_h2(ivu-1,iju)
               lev1 = lev_h2(MAX(ivl-1,0),ijl)
            ELSE
               lev2 = levu
               lev1 = levl
            ENDIF
            q_h2_h_1(levl,levu) = q_h2_h_1(lev1,lev2)
            q_h2_h_2(levl,levu) = q_h2_h_2(lev1,lev2)
            q_h2_h_3(levl,levu) = q_h2_h_3(lev1,lev2)
            IF (nqc == 4) THEN
               q_h2_h_4(levl,levu) = q_h2_h_4(lev1,lev2)
            ENDIF
         ENDIF
      ENDDO
   ENDDO

   !- H2 + He collision rates -------------------------------------------

   fichier = TRIM(data_dir)//'Collisions/q_he_h2.dat'
   WRITE(iscrn,*) "opening file : ", TRIM(fichier)
   OPEN(iread2, file = fichier, status = 'old')
   READ (iread2,*) ntnr

   DO j = 1, ntnr
      READ (iread2,*) levu, levl, q1, q2, q3
      IF ((levu <= l0h2) .AND. (levl <= l0h2)) THEN
         q_h2_he_1(levl,levu) = q1
         q_h2_he_2(levl,levu) = q2
         q_h2_he_3(levl,levu) = q3
      ENDIF
   ENDDO
   CLOSE (iread2)

   !- ortho-H2 + H2 collision rates -------------------------------------

   fichier = TRIM(data_dir)//'Collisions/q_oh2_h2.dat'
   WRITE(iscrn,*) "opening file : ", TRIM(fichier)
   OPEN(iread2, file = fichier, status = 'old')
   READ (iread2,*) ntnr

   DO j = 1, ntnr
      READ (iread2,*) levu, levl, q1, q2, q3
      IF ((levu <= l0h2) .AND. (levl <= l0h2)) THEN
         q_h2_oh2_1(levl,levu) = q1
         q_h2_oh2_2(levl,levu) = q2
         q_h2_oh2_3(levl,levu) = q3
      ENDIF
   ENDDO
   CLOSE (iread2)

   !- para-H2 + H2 collision rates --------------------------------------

   fichier = TRIM(data_dir)//'Collisions/q_ph2_h2.dat'
   WRITE(iscrn,*) "opening file : ", TRIM(fichier)
   OPEN(iread2, file = fichier, status = 'old')
   READ (iread2,*) ntnr

   DO j = 1, ntnr
      READ (iread2,*) levu, levl, q1, q2, q3
      IF ((levu <= l0h2) .AND. (levl <= l0h2)) THEN
         q_h2_ph2_1(levl,levu) = q1
         q_h2_ph2_2(levl,levu) = q2
         q_h2_ph2_3(levl,levu) = q3
      ENDIF
   ENDDO
   CLOSE (iread2)

   ALLOCATE (eqcol(nl,nl))
   DO iopt = 0, npo
      CALL COLH2(tgaz(iopt), iopt)
      eqcolh2(:,:,iopt) = eqcol(:,:)
   ENDDO

   !- UV radiative pumping

   fichier = TRIM(out_dir)//TRIM(modele)//'_eqcash2.out'
   INQUIRE (FILE=fichier, EXIST=isthere)
   IF (.NOT. isthere) THEN
      PRINT *, " file:", fichier, " does not exist!"
      PRINT *, " Modify PXDR to save eqcash2"
      PRINT *, " Set flag F_W_H2_EXC to 1, and rerun PDR"
      STOP
   ENDIF
   OPEN (43, file=fichier, status="unknown", form="unformatted")
   READ (43) eqcash2
   CLOSE (43)
   PRINT *, "eqrcash2 read"

   CALL MAKE_POPHM

   ! Build formation process
   !       Reminder: timh23 is formation on grains
   ALLOCATE (forh2(8,l0h2,0:npo))
   !         forh2_1: Radiative transitions in X (quadrupolar transitions) from below
   !         forh2_2: Radiative transitions in X (quadrupolar transitions) from above
   !         forh2_3: Collisions from below
   !         forh2_4: Collisions from above
   !         forh2_5: UV cascades from below
   !         forh2_6: UV cascades from above
   !         forh2_7: Formation on grains
   !         forh2_8: Formation by chemistry
   forh2 = 0.0_dp
   DO iopt = 0, npo
      DO levu = 1, l0h2
         iju = njl_h2(levu)
         ivu = nvl_h2(levu)
         DO levl = 1, levu-1
            forh2(1,levu,iopt) = forh2(1,levu,iopt) - eqradh2(levu,levl,iopt) * spec_lin(jh2)%xre(levl,iopt)
         ENDDO
         DO levl = levu+1, l0h2
            forh2(2,levu,iopt) = forh2(2,levu,iopt) - eqradh2(levu,levl,iopt) * spec_lin(jh2)%xre(levl,iopt)
         ENDDO
         DO levl = 1, levu-1
            forh2(3,levu,iopt) = forh2(3,levu,iopt) - eqcolh2(levu,levl,iopt) * spec_lin(jh2)%xre(levl,iopt)
         ENDDO
         DO levl = levu+1, l0h2
            forh2(4,levu,iopt) = forh2(4,levu,iopt) - eqcolh2(levu,levl,iopt) * spec_lin(jh2)%xre(levl,iopt)
         ENDDO
         DO levl = 1, levu-1
            forh2(5,levu,iopt) = forh2(5,levu,iopt) - eqcash2(levu,levl,iopt) * spec_lin(jh2)%xre(levl,iopt)
         ENDDO
         DO levl = levu+1, l0h2
            forh2(6,levu,iopt) = forh2(6,levu,iopt) - eqcash2(levu,levl,iopt) * spec_lin(jh2)%xre(levl,iopt)
         ENDDO
         forh2(7,levu,iopt) = forh2(7,levu,iopt) + timh23(iopt) * pophm1(ivu,iju) &
                                                 + timh25(iopt) * pophm2(ivu,iju)
      ENDDO
      forh2(8,1,iopt) = forh2(8,1,iopt) + timh21(iopt) * 2.0_dp / 3.0_dp
      forh2(8,2,iopt) = forh2(8,2,iopt) + timh21(iopt) * 1.0_dp / 3.0_dp
      forh2(:,:,iopt) = forh2(:,:,iopt) * abnua(jh2,iopt)
   ENDDO

   iimax = iimax + 1
   y(iimax,0:npo) = spec_lin(jh2)%abu(0:npo)
   texty(iimax) = 'Ab.'//TRIM(spec_lin(jh2)%nam)
   ALLOCATE (provi(0:npo))
   ALLOCATE (opd_GG(0:npo))
   print *, " jh2:", jh2
   print *, spec_lin(jh2)%nam
   CALL VIEW_LINES(spec_lin(jh2), irec)
   ii = 999
   DO WHILE (ii /= 0)
      PRINT *,"Enter line number (0 end)"
      READ *, ii
      IF (ii > spec_lin(jh2)%ntr) THEN
         PRINT *, ii, ' > ', spec_lin(jh2)%ntr, ' ! - END'
         ii = 0
         EXIT
      ENDIF
      IF (irec == 1) THEN
         WRITE (8,*) ii
      ENDIF

      IF (ii /= 0) THEN
         PRINT *, 'Collect line number:', ii
         levu = spec_lin(jh2)%iup(ii)
         opd_GG = spec_lin(jh2)%odl(ii,0:npo) + tau(0:npo) * spec_lin(jh2)%siD(0:npo,ii)
         provi = xk * spec_lin(jh2)%wlk(ii) * spec_lin(jh2)%abu(0:npo) * spec_lin(jh2)%xre(levu,0:npo) &
               * EXP(-opd_GG(0:npo)) * spec_lin(jh2)%aij(ii)
         iimax = iimax + 1
         y(iimax,0:npo) = spec_lin(jh2)%xre(levu,0:npo)
         texty(iimax) = 'Xre.'//TRIM(spec_lin(jh2)%nam)//'.'//TRIM(ADJUSTL(ii2c(levu)))
         iimax = iimax + 1
         y(iimax,0:npo) = spec_lin(jh2)%odl(ii,0:npo)
         texty(iimax) = 'OpD-L.'//TRIM(spec_lin(jh2)%nam)//'.'//TRIM(ADJUSTL(ii2c(ii)))
         iimax = iimax + 1
         y(iimax,0:npo) = tau(0:npo) * spec_lin(jh2)%siD(0:npo,ii)
         texty(iimax) = 'OpD-G.'//TRIM(spec_lin(jh2)%nam)//'.'//TRIM(ADJUSTL(ii2c(ii)))
         iimax = iimax + 1
         y(iimax,0:npo) = opd_GG(0:npo)
         texty(iimax) = 'OpD.'//TRIM(spec_lin(jh2)%nam)//'.'//TRIM(ADJUSTL(ii2c(ii)))
         iimax = iimax + 1
         y(iimax,0:npo) = provi
         texty(iimax) = 'EmL.'//TRIM(spec_lin(jh2)%nam)//'.'//TRIM(ADJUSTL(ii2c(ii)))
         iimax = iimax + 1
         y(iimax,0:npo) = provi * forh2(7,levu,0:npo) &
                        / (forh2(1,levu,0:npo) + forh2(2,levu,0:npo) + forh2(3,levu,0:npo) + forh2(4,levu,0:npo) &
                        + forh2(5,levu,0:npo) + forh2(6,levu,0:npo) + forh2(7,levu,0:npo) + forh2(8,levu,0:npo))
         texty(iimax) = 'EmLG.'//TRIM(spec_lin(jh2)%nam)//'.'//TRIM(ADJUSTL(ii2c(ii)))
         iimax = iimax + 1
         y(iimax,0:npo) = forh2(1,levu,0:npo)
         texty(iimax) = 'QTX-l.'//TRIM(spec_lin(jh2)%nam)//'.'//TRIM(ADJUSTL(ii2c(ii)))
         iimax = iimax + 1
         y(iimax,0:npo) = forh2(2,levu,0:npo)
         texty(iimax) = 'QTX-u.'//TRIM(spec_lin(jh2)%nam)//'.'//TRIM(ADJUSTL(ii2c(ii)))
         iimax = iimax + 1
         y(iimax,0:npo) = forh2(3,levu,0:npo)
         texty(iimax) = 'C-l.'//TRIM(spec_lin(jh2)%nam)//'.'//TRIM(ADJUSTL(ii2c(ii)))
         iimax = iimax + 1
         y(iimax,0:npo) = forh2(4,levu,0:npo)
         texty(iimax) = 'C-u.'//TRIM(spec_lin(jh2)%nam)//'.'//TRIM(ADJUSTL(ii2c(ii)))
         iimax = iimax + 1
         y(iimax,0:npo) = forh2(5,levu,0:npo)
         texty(iimax) = 'UV-l.'//TRIM(spec_lin(jh2)%nam)//'.'//TRIM(ADJUSTL(ii2c(ii)))
         iimax = iimax + 1
         y(iimax,0:npo) = forh2(6,levu,0:npo)
         texty(iimax) = 'UV-u.'//TRIM(spec_lin(jh2)%nam)//'.'//TRIM(ADJUSTL(ii2c(ii)))
         iimax = iimax + 1
         y(iimax,0:npo) = forh2(7,levu,0:npo)
         texty(iimax) = 'Gr-f.'//TRIM(spec_lin(jh2)%nam)//'.'//TRIM(ADJUSTL(ii2c(ii)))
         IF (MAXVAL(forh2(8,levu,0:npo)) > 0.0_dp) THEN
            iimax = iimax + 1
            y(iimax,0:npo) = forh2(8,levu,0:npo)
            texty(iimax) = 'Ch-f.'//TRIM(spec_lin(jh2)%nam)//'.'//TRIM(ADJUSTL(ii2c(ii)))
         ENDIF
      ELSE
         PRINT *, 'End lines: '
         EXIT
      ENDIF
   ENDDO

   DEALLOCATE (eqradh2)
   DEALLOCATE (eqcolh2)
   DEALLOCATE (eqcash2)
   DEALLOCATE (forh2)
   DEALLOCATE (provi)
   DEALLOCATE (opd_GG)

END SUBROUTINE EMIH2

!%%%%%%%%%%%%%%%%
SUBROUTINE EMIINT
!%%%%%%%%%%%%%%%%

   USE PXDR_CONSTANTES
   USE PXDR_CHEM_DATA
   USE PXDR_PROFIL
   USE PXDR_AUXILIAR
   USE PXDR_STR_DATA
   USE PXDR_INTENSITIES
   USE PREP_VAR

   IMPLICIT NONE

   CHARACTER (LEN=lenSpeci) :: yspec
   CHARACTER (LEN=lenIDLIN) :: tampon
   CHARACTER (LEN=50)       :: fichout
   REAL (KIND=dp)           :: xmu
   INTEGER                  :: i, ii
   INTEGER                  :: iupl, jlol
   INTEGER                  :: ind_sp
   INTEGER                  :: flag_AllOutputs  ! Flag to call LDVInt - 1 is used in prep to get lines properties

!--------+---------+---------+---------+---------+---------+---------+-*-------+

   PRINT *
   PRINT *, ' Give lines file name:'
   READ (*,'(a50)') fichout
   fichout = TRIM(ADJUSTL(fichout))
   IF (irec == 1) THEN
      WRITE (8,'(a50)') fichout
   ENDIF

   fichier = TRIM(out_dir)//fichout
   PRINT *," Lines written in:", fichier
   OPEN (35, file = fichier, status = 'new')

   PRINT *
   PRINT *, ' Do want to save the full line profiles (1: yes, 0: no)'
   READ (*,*) flag_WriteLineProf
   IF (irec == 1) THEN
      WRITE (8,*) flag_WriteLineProf
   ENDIF

   PRINT *
   PRINT *, ' Give angle of line of sight (radian) [wrt normal]'
   READ *, xmu
   IF (irec == 1) THEN
      WRITE (8,*) xmu
   ENDIF
   xmu = cos(xmu)
   WRITE (35,'("# mu:",1pe14.6)') xmu
   IF (flag_WriteLineProf == 0) THEN
      WRITE (35,'("# lev_up-lev_down Freq(GHz) Wave_l(micron) I_int(erg_s-1_cm-2_ster) I_int(K_km_s-1)")')
   ENDIF

   PRINT *, " "
   PRINT *, " Possible species:"
   i = 1
   DO WHILE (i <= nspl)
      IF (MOD(i,3) /= 0) THEN
         WRITE (*,'(a10)',ADVANCE="NO") spec_lin(i)%nam
      ELSE
         WRITE (*,'(a10)') spec_lin(i)%nam
      ENDIF
      i = i + 1
   ENDDO
   WRITE (*,*)

   DO
      PRINT *
      PRINT *, ' Which species?  (end: -1)'
      READ (*,'(a8)') yspec
      yspec = TRIM(ADJUSTL(yspec))
      IF (irec == 1) THEN
         WRITE (8,'(a8)') yspec
      ENDIF
      IF (yspec == '-1') THEN
         PRINT *, "End lines"
         EXIT
      ENDIF

      ind_sp = 0
      DO i = 1, nspl
        IF ( TRIM(ADJUSTL(yspec)) == TRIM(ADJUSTL(spec_lin(i)%nam))) THEN
            ind_sp = i
            EXIT
         ENDIF
      ENDDO
      print *, ind_sp, yspec, spec_lin(ind_sp)%nam
      IF (ind_sp == 0) THEN
         PRINT *, "  WARNING: excitation of ", yspec, " was not computed"
         CYCLE
      ENDIF

      ii = 999
      PRINT *, ' Emission lines of: ', yspec
      IF (flag_WriteLineProf == 0) THEN
         WRITE (35,'(/,"# ",a8)') yspec
      ENDIF
      CALL VIEW_LINES(spec_lin(ind_sp), irec)
      DO WHILE (ii /= 0)
         PRINT *, 'Transition?  (0: end)'
         READ *, ii
         IF (ii > spec_lin(ind_sp)%ntr) THEN
            ii = 0
         ENDIF
         IF (irec == 1) THEN
            WRITE (8,*) ii
         ENDIF

         IF (ii /= 0) THEN
            iupl = spec_lin(ind_sp)%iup(ii)
            jlol = spec_lin(ind_sp)%jlo(ii)
            tampon = ""
            IF (flag_WriteLineProf == 1) THEN
               tampon = 'Emis.'//TRIM(spec_lin(ind_sp)%nam)//'.'
            ENDIF
            tampon = TRIM(ADJUSTL(tampon))//TRIM(ADJUSTL(spec_lin(ind_sp)%LiHname(ii)))
            flag_AllOutputs = 1
            CALL LDVInt (xmu, spec_lin(ind_sp), ii, tampon, flag_AllOutputs, 0)
         ENDIF
      ENDDO
   ENDDO

   CLOSE (35)

END SUBROUTINE EMIINT

!%%%%%%%%%%%%%%%%%%%%%%%%%%%%
SUBROUTINE COLH2 (ttry, iopt)
!%%%%%%%%%%%%%%%%%%%%%%%%%%%%

   USE PXDR_CONSTANTES
   USE PXDR_CHEM_DATA
   USE PXDR_PROFIL
   USE PXDR_AUXILIAR
   USE PXDR_STR_DATA

!---------------------------------------------------------------------
!      xnprot is no. density of protons (cm-3)
!      ttry is temperature of neutral fluid

!      called from sub. EQLH2V
!
!      collision rates for H2 + H come from Martin & Mandy.
!      New fits by Jacques Le Bourlot (july 1996)
!
!          or
!
!      from D. Flower (cf MNRAS)
!
!      Collision rates for H2 + He, H2 + H2o, H2 + H2p
!      come from D. Flower
!
!---------------------------------------------------------------------

   IMPLICIT NONE

!  Data from Gerlich for ortho-para transfer with h+.

   REAL (KIND=dp), INTENT (IN)                      :: ttry
   INTEGER, INTENT (IN)                             :: iopt

   REAL (KIND=dp), DIMENSION (30)                   :: gerlic
   INTEGER                                          :: nqc
   REAL (KIND=dp)                                   :: xchh2, fac, ctod
   REAL (KIND=dp)                                   :: decal, tdd, tdd2
   REAL (KIND=dp)                                   :: gmm, ga1, ga2
   REAL (KIND=dp)                                   :: ch2h_nr, ch2h_r, ch2he, ch2h
   REAL (KIND=dp)                                   :: ch2ph2, ch2oh2, cophp, c, d
   INTEGER                                          :: levu, levl, ndej, lev1, lev2
   INTEGER                                          :: ju, jl, ivu, ivl

!  With David's prescription for O/P balance (3/1)

   data gerlic / 2.024e-10_dp, 9.575e-10_dp, 2.729e-10_dp, 7.910e-10_dp &
         , 2.246e-10_dp, 6.223e-10_dp, 1.828e-10_dp, 4.975e-10_dp, 1.484e-10_dp &
         , 3.000e-10_dp, 1.000e-10_dp, 3.000e-10_dp, 1.000e-10_dp, 3.000e-10_dp &
         , 1.000e-10_dp, 3.000e-10_dp, 1.000e-10_dp, 3.000e-10_dp, 1.000e-10_dp &
         , 3.000e-10_dp, 1.000e-10_dp, 3.000e-10_dp, 1.000e-10_dp, 3.000e-10_dp &
         , 1.000e-10_dp, 3.000e-10_dp, 1.000e-10_dp, 3.000e-10_dp, 1.000e-10_dp &
         , 3.000e-10_dp /

!  Up to now collisions with H2 and He are used only for v = 0
!  Educated guesses may obtain from collisions with H
!  WARNING: c: desexcitation, d: excitation
!            h2h: H2 + H
!            h2h2: H2 + H2
!            h2he: H2 + He
!            ophp: H2 + H+ (ortho - para)

   eqcol = 0.0_dp

   h2para = 0.0_dp
   h2orth = 0.0_dp
   DO levu = 1, l0h2
      ju = njl_h2(levu)
      IF (MOD(ju,2) == 0) THEN
         h2para = h2para + spec_lin(in_sp(i_h2))%xre(levu,iopt)
      ELSE
         h2orth = h2orth + spec_lin(in_sp(i_h2))%xre(levu,iopt)
      ENDIF
   ENDDO

   IF (ichh2 == 0) THEN
      nqc = 4
   ELSE
      nqc = 3
   ENDIF
   IF (ichh2 <= 1) THEN
      xchh2 = 0.0_dp
   ELSE
      xchh2 = 1.0_dp
   ENDIF

   DO levu = 2, l0h2
      ju = njl_h2(levu)
      ivu = nvl_h2(levu)

      DO levl = 1, levu-1
         jl = njl_h2(levl)
         ivl = nvl_h2(levl)

         ! ctod is the Boltzmann factor used to compute
         !    excitation rates from desexcitation rates
         ndej = ABS(ju - jl)
         fac = (spec_lin(in_sp(i_h2))%elk(levu) - spec_lin(in_sp(i_h2))%elk(levl)) / ttry
         ctod = EXP(-fac) * (2.0_dp*ju+1.0_dp) / (2.0_dp*jl+1.0_dp)
         IF (MOD(ju,2) == 1) THEN
            ctod = ctod * 3.0_dp
         ENDIF
         IF (MOD(jl,2) == 1) THEN
            ctod = ctod / 3.0_dp
         ENDIF

         ! Fit of H2 + H rates. The reduced variable tdd is offset by a
         ! different value for reactive (resp. non-reactive) collisions.
         IF (MOD(ndej,2) == 0) THEN
            decal = 1.0_dp
         ELSE
            decal = 0.3_dp
         ENDIF
         tdd = decal + ttry * 1.0e-3_dp
         tdd2 = tdd*tdd

         ! Reactive and non-reactive collisions are mixed in Martin and Mandy
         gmm = q_h2_h_1(levl,levu) + q_h2_h_2(levl,levu) / tdd + q_h2_h_3(levl,levu) / tdd2
         IF (nqc == 4) THEN
            gmm = gmm +  q_h2_h_4(levl,levu) * tdd
         ENDIF

         ch2h_nr = 10.0_dp**gmm
         ch2h_r  = 0.0_dp

         ! Add reactive collisions (David's rule) if ichh2 = 3
         IF (ichh2 == 3 .AND. tau(iopt) <= 1.0_dp) THEN

            IF (MOD(ndej,2) == 0) THEN
               ch2h_r = 10.0_dp**gmm * EXP(-MAX(0.0_dp,(3900.0_dp - spec_lin(in_sp(i_h2))%elk(levu) &
                      + spec_lin(in_sp(i_h2))%elk(levl))/ttry))
            ELSE
               IF (jl > 0) THEN
                  lev1 = lev_h2(ivl,jl-1)
               ELSE
                  lev1 = lev_h2(ivl,jl+1)
               ENDIF
               lev2 = lev_h2(ivl,jl+1)
               IF (lev2 == 0) THEN
                  lev2 = lev_h2(ivl,jl-1)
               ENDIF
               ga1 = q_h2_h_1(lev1,levu) + q_h2_h_2(lev1,levu) / tdd + q_h2_h_3(lev1,levu) / tdd2
               IF (nqc == 4) THEN
                  ga1 = ga1 + q_h2_h_4(lev1,levu) * tdd
               ENDIF
               ga2 = q_h2_h_1(lev2,levu) + q_h2_h_2(lev2,levu) / tdd + q_h2_h_3(lev2,levu) / tdd2
               IF (nqc == 4) THEN
                  ga2 = ga2 + q_h2_h_4(lev2,levu) * tdd
               ENDIF
               ch2h_r = (10.0_dp**ga1 + 10.0_dp**ga2) * 0.5 * EXP(-MAX(0.0_dp,(3900.0_dp - spec_lin(in_sp(i_h2))%elk(levu) &
                      + spec_lin(in_sp(i_h2))%elk(levl))/ttry))

               IF (MOD(ju,2) == 1) THEN
                  ch2h_r = ch2h_r / 3.0_dp
               ENDIF
            ENDIF
         ENDIF

         ! Ortho/Para transfer with H from Schofield
         ! plus David's rule (for a few delta v = 0)
         IF (ichh2 >= 2 .AND.  ivu == ivl .AND. ndej < 3) THEN
            IF (MOD(ju,2) == 1 .AND. MOD(ndej,2) == 1) THEN
               ch2h_r = 8.0e-11_dp * EXP(-3900.0_dp/ttry) / 3.0_dp
            ELSE
               ch2h_r = 8.0e-11_dp * EXP(-3900.0_dp/ttry)
            ENDIF
         ENDIF

         ch2h = ch2h_nr + xchh2 * ch2h_r

         ! He - H2
         gmm = q_h2_he_1(levl,levu) + q_h2_he_2(levl,levu) / tdd + q_h2_he_3(levl,levu) / tdd2
         ch2he = 10.0_dp**gmm

         ! pH2 - H2
         gmm = q_h2_ph2_1(levl,levu) + q_h2_ph2_2(levl,levu) / tdd + q_h2_ph2_3(levl,levu) / tdd2
         ch2ph2 = 10.0_dp**gmm

         ! oH2 - H2
         gmm = q_h2_oh2_1(levl,levu) + q_h2_oh2_2(levl,levu) / tdd + q_h2_oh2_3(levl,levu) / tdd2
         ch2oh2 = 10.0_dp**gmm

         ! c- Ortho to Para transfer withions (Gerlich), only for v = 0
         ! IF (ivu == 0 .AND. ndej == 1) THEN
         IF (jl == ju-1) THEN
            cophp = gerlic(ju)
         ELSE
            cophp = 0.0_dp
         ENDIF

         ! Bringing it all back home
         c = ch2ph2 * abnua(i_h2,iopt) * h2para + ch2oh2 * abnua(i_h2,iopt) * h2orth &
           + ch2he * abnua(i_he,iopt) + ch2h * abnua(i_h,iopt)
         IF (i_hp /= 0) THEN
            c = c + cophp * abnua(i_hp,iopt)
         ENDIF
         IF (i_h3p /= 0) THEN
            c = c + cophp * abnua(i_h3p,iopt)
         ENDIF
         d = c * ctod

         eqcol(levl,levu) = eqcol(levl,levu) - c
         eqcol(levu,levu) = eqcol(levu,levu) + c
         eqcol(levu,levl) = eqcol(levu,levl) - d
         eqcol(levl,levl) = eqcol(levl,levl) + d
      ENDDO
   ENDDO

END SUBROUTINE COLH2

!========================================================================================================
! SUBROUTINE PRINT_INTENSITIES
!========================================================================================================
SUBROUTINE PRINT_INTENSITIES

   USE PXDR_CONSTANTES
   USE PXDR_CHEM_DATA
   USE PXDR_PROFIL
   USE PXDR_AUXILIAR
   USE PXDR_STR_DATA
   USE PXDR_INTENSITIES

   IMPLICIT NONE

   CHARACTER(len=lenfilename) :: fichout
   INTEGER                    :: i, j, k

      fichout = TRIM(out_dir)//TRIM(modele)//".AllLines"
      OPEN(81,file=fichout,status="unknown",action="write")
      WRITE (81,1080)
      DO i = 1, nspl
         DO j = 1, spec_lin(i)%ntr
            IF (spec_lin(i)%iup(j) <= spec_lin(i)%use) THEN
               WRITE (81,1081) TRIM(spec_lin(i)%spHname), TRIM(spec_lin(i)%LiHname(j)) &
                              , (spec_lin(i)%int(j,k), k = 1, nangle)
            ENDIF
         ENDDO
      ENDDO
      CLOSE(81)
      WRITE(*,*) "All line intensities have been written in : "
      WRITE(*,*) TRIM(fichout)

1080  FORMAT("#  Species                      Transition        &
      &Intensities [erg cm-2 s-1 sr-1] at 0, 10, 20, 30, 40, 50, 60 degrees")
1081  FORMAT(A10,2x,A30,7(2x,1p,E18.6))

END SUBROUTINE PRINT_INTENSITIES

!---------------------------------------------------------
! Computes the full spectrum including continuum and lines
!---------------------------------------------------------
!%%%%%%%%%%%%%%%%%%%
SUBROUTINE FSPECTRUM
!%%%%%%%%%%%%%%%%%%%

   USE PXDR_CONSTANTES
   USE PXDR_CHEM_DATA
   USE PXDR_PROFIL
   USE PXDR_AUXILIAR
   USE PXDR_STR_DATA
   USE PXDR_RF_TOOLS
   USE PXDR_INTENSITIES
   USE PREP_VAR

   IMPLICIT NONE

   CHARACTER (LEN=lenIDLIN)                     :: tampon
   CHARACTER (LEN=50)                           :: fichout
   REAL (KIND=dp)                               :: xmu
   REAL (KIND=dp)                               :: lam1, lam2, wl1, wl2
   REAL (KIND=dp)                               :: spec_D, spec_BG
   INTEGER                                      :: j, j1, j2
   INTEGER                                      :: i, itr
   INTEGER                                      :: i_line
   REAL (kind=dp)                               :: wl

!--------+---------+---------+---------+---------+---------+---------+-*-------+

   ! Recover the incident radiation field
   ! Read from file *.flin (see ETIN90 for creation)
   ! This avoids recomputing it from scratch
   IF (.NOT. ALLOCATED(I_incr_m)) THEN                ! Read file only on first call
      ALLOCATE (I_incr_m(0:nwlg))
      ALLOCATE (I_isot_m(0:nwlg))
      ALLOCATE (I_incr_p(0:nwlg))
      ALLOCATE (I_isot_p(0:nwlg))

      fichier = TRIM(out_dir)//TRIM(modele)//'.flin'
      OPEN (60, file=fichier, status="unknown")
      READ (60,*)
      DO j = 0, nwlg
         READ (60, '(1p,9e15.6E3)') wl, I_incr_m(j), I_isot_m(j) &
                                  , I_incr_p(j), I_isot_p(j)
      ENDDO
      CLOSE (60)
   ENDIF

   ! Output file names built as usual
   PRINT *
   ! Full spectrum
   fichout = TRIM(ADJUSTL(modele))//".spec"
   fichier = TRIM(out_dir)//fichout
   PRINT *," Spectrum written in:", fichier
   OPEN (35, file = fichier, status = 'new')
   WRITE (35, '("Lambda-A  Spec-tot  Emis-Dust  Extern-Trans")')

   ! Integrated line intensities
   fichout = TRIM(ADJUSTL(modele))//".lines"
   fichier = TRIM(out_dir)//fichout
   PRINT *," Lines integrated intensities written in:", fichier
   OPEN (36, file = fichier, status = 'new')

   ! Note: A potential high background source for mu = 1 is not yet included
   PRINT *
   PRINT *, ' Give angle of line of sight (radian) [wrt normal]'
   READ *, xmu
   IF (irec == 1) THEN
      WRITE (8,*) xmu
   ENDIF
   xmu = cos(xmu)
   WRITE (36,'("# mu:",1pe14.6)') xmu

   PRINT *, " "
   PRINT *, " Give lambda min (Angstrom):"
   READ *, lam1
   IF (irec == 1) THEN
      WRITE (8,*) lam1
   ENDIF
   CALL RF_LOCATE (rf_wl, lam1, j1)
   PRINT *, " Give lambda max (Angstrom):"
   READ *, lam2
   IF (irec == 1) THEN
      WRITE (8,*) lam2
   ENDIF
   CALL RF_LOCATE (rf_wl, lam2, j2)
   WRITE (36,'("# Spectrum from ",1pe14.6," to ",e14.6," Angstrom")') lam1, lam2
   WRITE (36,'("# Species_Line   Nu_GHz   Lambda_micron   Int_erg_cm-2_s-1_ster-1   Int_K_km_s-1")')

   ! We scan wavelength in increasing order
   ! If at line center, then a line profile is computed
   ! The line is kept if strong enough. Else use pure continuum contribution
   ! For each wavelength, the intensity is followed by a flag:
   !     1: part of a line profile
   !     0: computed with continuum only
   wl2 = 0.0_dp
   DO j = j1, j2
      IF (rf_wl%Val(j) < wl2) CYCLE
      i_line = 0
 SPC: DO i = 1, nspl
         DO itr = 1, spec_lin(i)%ntr
            IF ((spec_lin(i)%iup(itr) <= spec_lin(i)%use) .AND. &
                (ABS((rf_wl%Val(j) - spec_lin(i)%lac(itr) * 1.0e8_dp) / rf_wl%Val(j)) < 1.0e-8_dp)) THEN
               tampon = TRIM(ADJUSTL(spec_lin(i)%nam))//"_"//TRIM(ADJUSTL(spec_lin(i)%LiHname(itr)))
               CALL LDVInt_l (xmu, spec_lin(i), itr, tampon, wl1, wl2)
               i_line = 1
               EXIT SPC
            ENDIF
         ENDDO
      ENDDO SPC
      IF (i_line == 0) THEN
         CALL LDVInt_c (xmu, rf_wl%Val(j), spec_D, spec_BG)
         WRITE (35, "(1p,4e16.8E3,a2)") rf_wl%Val(j), spec_D+spec_BG, spec_D, spec_BG, " 0"
      ENDIF
   ENDDO

   CLOSE (35)
   CLOSE (36)

END SUBROUTINE FSPECTRUM

END MODULE PREP_EMISSI
