! Copyright (C) 2002 - Observatoire de Paris/Meudon
!                      Universite Paris 7
!                      CNRS (France)
!                      Contact : Jacques.Lebourlot@obspm.fr
!
! This program is free software; you can redistribute it and/or
! modify it under the terms of the GNU General Public License version 2
! as published by the Free Software Foundation.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this program; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

!--------+---------+---------+---------+---------+---------+---------+-*-------+

MODULE PREP_VAR

  USE PXDR_CONSTANTES
  USE PXDR_STR_DATA

  IMPLICIT NONE

  ! JLB - 4 Sept 2019 - Declare flags here to avoid "PARAMETER"
  ! They are now read from binary files
  !--- Special excitation for some species ----------------------------------------------
  INTEGER, PUBLIC            :: F_CO_UV_pum     = 0 ! (1) CO UV pumping
  INTEGER, PUBLIC            :: F_CO_CH_pum     = 1 ! (1) CO excitation at formation
  INTEGER, PUBLIC            :: F_H3p_CH_pum    = 1 ! (1) H3+ excitation at formation
  INTEGER, PUBLIC            :: F_COLL_H3p      = 0 ! (0) H3+ collision rates
  INTEGER, PUBLIC            :: F_Oat_CH_pum    = 1 ! (1) O excitation at formation

  !--- Chemistry and reactions related --------------------------------------------------
  INTEGER, PUBLIC            :: F_H2_COLL_D     = 1 ! (1) Choice of H2 collisional dissociation

  !--- Radiation field & radiative transfer ---------------------------------------------
  INTEGER, PUBLIC            :: F_ISRF          = 1 ! (1) Mathis or Draine ISRF    (defined in pdr.in)
  INTEGER, PUBLIC            :: F_Norm_RF       = 0 ! (0) Normal rad. field type   (modified by pdr.in parameters)
  INTEGER, PUBLIC            :: F_RF_PP         = 0 ! (0) Isotropic or normal ISRF (modified by pdr.in parameters)
  INTEGER, PUBLIC            :: F_haardt96      = 0 ! (0) Activate Haardt and Madau radiation field
  INTEGER, PUBLIC            :: F_haardt96_only = 0 ! (0) Activate Harrdt and Madau and remove Galactic ISRF
  INTEGER, PUBLIC            :: F_LINE_TRSF     = 1 ! (1) Consider induced transitions
  INTEGER, PUBLIC            :: F_COMP_RF_SCAL  = 1 ! (1) Consider gas continuum absorption (1 = consider absorption)

  !--- Grains associated processes ------------------------------------------------------
  INTEGER, PUBLIC            :: F_DUST_P        = 0 ! (0) DustEM / Fredholm activation (defined in pdr.in)
  INTEGER, PUBLIC            :: F_QABS_SCALED   = 0 ! (0) Scaling of Qabs and Qsca
  INTEGER, PUBLIC            :: F_DU_PROP       = 0 ! (0) 0: Grain properties constant in cloud, 1: can depend on positions

  !--- Main output files ----------------------------------------------------------------
  INTEGER, PUBLIC            :: F_W_ALL_IFAF        ! (0) Write output files at all iterations or not (defined in pdr.in)
  INTEGER, PUBLIC            :: F_W_BINARY      = 1 ! (0) Write .bin files
  INTEGER, PUBLIC            :: F_W_HDF5_STD    = 0 ! (1) Write main HDF5 file
  INTEGER, PUBLIC            :: F_W_HDF5_CHE    = 0 ! (1) Write chemistry HDF5 file
  INTEGER, PUBLIC            :: F_W_HDF5_ANA    = 0 ! (1) Write line analysis HDF5 file

  !--- Optional outputs -----------------------------------------------------------------
  INTEGER, PUBLIC            :: F_W_H2_EXC      = 0 ! (0) Write detailes H2 excitation informations
  INTEGER, PUBLIC            :: F_W_LINE_MAP    = 0 ! (0) Write full line informations (in AV and dnu)
  INTEGER, PUBLIC            :: F_W_RF_ALB      = 1 ! (0) Write energy density as a function of AV and lambda
  INTEGER, PUBLIC            :: F_W_CHARGE      = 0 ! (0) Write .chg files

  !--- Debug, tests and in development -------------------------------------------------
  INTEGER, PUBLIC            :: F_PHE_TWEAK     = 0 ! (0) Test prescriptions for P.E. heating (test)
  INTEGER, PUBLIC            :: F_DUST_T_CTE    = 0 ! (0) If 1 force dust temperatures to a predefined value (see FGK_TR)
  INTEGER, PUBLIC            :: F_CHECK_THBAL   = 0 ! (0) Write informations about thermal balance during execution
  INTEGER, PUBLIC            :: F_CHECK_INIPROF = 0 ! (0) Write initial temperature and density profile

  INTEGER, PUBLIC, PARAMETER :: ncolf  = 400  ! Max number of columns in output file
  INTEGER, PUBLIC, PARAMETER :: lenmax = 128  ! Max length of character string

  REAL (KIND=dp), PUBLIC, DIMENSION (0:nvxhd,0:noptm)             :: codhdv
  REAL (KIND=dp), PUBLIC, ALLOCATABLE, DIMENSION(:,:)             :: codeh3p
  REAL (KIND=dp), PUBLIC, DIMENSION (3,0:noptm)                   :: cdoph2

  CHARACTER(LEN=10), PUBLIC, DIMENSION (:), ALLOCATABLE           :: ndiesp
  CHARACTER(LEN=2), PUBLIC                                        :: c_ifaf

  REAL (KIND=dp), PUBLIC, DIMENSION(0:nvxh2,0:noptm)              :: h2v

  ! Quantites associees a des especes particulieres
  REAL (KIND=dp), PUBLIC                                          :: Emaxh3p_lu
  INTEGER,        PUBLIC, DIMENSION(:,:),   ALLOCATABLE           :: levh3p(:,:)
  INTEGER,        PUBLIC, DIMENSION(:,:,:), ALLOCATABLE           :: levh2o(:,:,:)

  ! grains:
  REAL (KIND=dp), PUBLIC                                          :: dustio
  REAL (KIND=dp), PUBLIC                                          :: depl
  REAL (KIND=dp), PUBLIC                                          :: rath2g

  ! output file
  REAL (KIND=dp), PUBLIC, DIMENSION (0:noptm)                     :: x
  REAL (KIND=dp), PUBLIC, DIMENSION (ncolf,0:noptm)               :: y
  CHARACTER (LEN=14), PUBLIC                                      :: textx
  CHARACTER (LEN=14), PUBLIC, DIMENSION (ncolf)                   :: texty
  INTEGER                                                         :: irec
  INTEGER                                                         :: iimax

  ! Convenient, but not used (see Franck)
  INTEGER, PUBLIC                                                 :: i_nvbc, i_njbc, i_natom

CONTAINS

!-------------------------------------------------------------------------
! SUBROUTINE : CASTIS
!-------------------------------------------------------------------------
SUBROUTINE CASTIS(n,sn,lsn)

INTEGER,           INTENT (IN)  :: n   ! INTEGER to transform in string
CHARACTER(len=10), INTENT (OUT) :: sn  ! string corresponding to the INTEGER
INTEGER,           INTENT (OUT) :: lsn ! number of figures in n

INTEGER :: ntemp ! temporary INTEGER
INTEGER :: div   ! division factor
INTEGER :: nn    ! modulo of ntemp by 10
INTEGER :: i

!--- Number of string to get
lsn = 1
div = 1
DO WHILE (n/div >= 10)
   div = div * 10
   lsn = lsn + 1
ENDDO

IF (lsn > 10) THEN
   PRINT *,"PROBLEM : Increase size of sn in CASTIS"
   STOP
ENDIF

!--- Cast from INTEGER to STRING
ntemp = n
DO i = 1, lsn
   nn = MOD(ntemp,10)
   sn(lsn-i+1:lsn-i+1) = ACHAR(nn+48)
   ntemp = (ntemp-nn)/10
ENDDO

END SUBROUTINE CASTIS

!-----------------------------------------------------------------------------
! SUBROUTINE : BUILD_LEVH2O
!-----------------------------------------------------------------------------
SUBROUTINE BUILD_LEVH2O

  USE PXDR_CHEM_DATA
  USE PXDR_AUXILIAR

  IMPLICIT NONE

  INTEGER :: nlvu
  INTEGER :: jmax, kamax, kcmax
  INTEGER :: qj, qka, qkc ! quantum numbers J, Ka, Kc
  INTEGER :: i, j_sp

   j_sp = in_sp(i_h2o)
   nlvu = spec_lin(j_sp)%use

   jmax = 0; kamax = 0; kcmax = 0
   DO i = 1, nlvu
      CALL STR_INT(spec_lin(j_sp)%quant(i,1),qj)
      CALL STR_INT(spec_lin(j_sp)%quant(i,2),qka)
      CALL STR_INT(spec_lin(j_sp)%quant(i,3),qkc)
      IF (qj  > jmax)  THEN
         CALL STR_INT(spec_lin(j_sp)%quant(i,1),jmax)
      ENDIF
      IF (qka > kamax) THEN
         CALL STR_INT(spec_lin(j_sp)%quant(i,2),kamax)
      ENDIF
      IF (qkc > kcmax) THEN
         CALL STR_INT(spec_lin(j_sp)%quant(i,3),kcmax)
      ENDIF
   ENDDO

   IF (.NOT. ALLOCATED(levh2o)) ALLOCATE(levh2o(0:jmax,0:kamax,0:kcmax))
   levh2o(:,:,:) = -1 ! A stupid value

   DO i = 1, spec_lin(j_sp)%use
      CALL STR_INT(spec_lin(j_sp)%quant(i,1),qj)
      CALL STR_INT(spec_lin(j_sp)%quant(i,2),qka)
      CALL STR_INT(spec_lin(j_sp)%quant(i,3),qkc)
      levh2o(qj, qka, qkc) = i
      PRINT *,"H2O - lev=",i,qj, qka, qkc
   ENDDO

END SUBROUTINE BUILD_LEVH2O


!-------------------------------------------------------------------------------
! SUBROUTINE : BUILD_LEVH3p
!-------------------------------------------------------------------------------

SUBROUTINE BUILD_LEVH3p

  USE PXDR_CHEM_DATA
  USE PXDR_AUXILIAR

  IMPLICIT NONE

  INTEGER :: nlvu
  INTEGER :: jmax, kmax
  INTEGER :: qj, qk ! quantum numbers J, K
  INTEGER :: i, j_sp

   j_sp = in_sp(i_h3p)
   nlvu  = spec_lin(j_sp)%use

   jmax = 0; kmax = 0
   DO i = 1, nlvu
      CALL STR_INT(spec_lin(j_sp)%quant(i,1), qj)
      CALL STR_INT(spec_lin(j_sp)%quant(i,2), qk)
      IF (qj > jmax) THEN
         CALL STR_INT(spec_lin(j_sp)%quant(i,1), jmax)
      ENDIF
      IF (qk > kmax) THEN
         CALL STR_INT(spec_lin(j_sp)%quant(i,2),kmax)
      ENDIF
   ENDDO

   IF (.NOT. ALLOCATED(levh3p)) ALLOCATE(levh3p(0:jmax,0:kmax))
   levh3p(:,:) = -1 ! A stupid value

   DO i = 1, spec_lin(j_sp)%use
      CALL STR_INT(spec_lin(j_sp)%quant(i,1),qj)
      CALL STR_INT(spec_lin(j_sp)%quant(i,2),qk)
      levh3p(qj, qk) = i
   ENDDO

END SUBROUTINE BUILD_LEVH3p

END MODULE PREP_VAR
