! Copyright (C) 2002 - Observatoire de Paris/Meudon
!                      Universite Paris 7
!                      CNRS (France)
!                      Contact : Jacques.Lebourlot@obspm.fr
!
! This program is free software; you can redistribute it and/or
! modify it under the terms of the GNU General Public License version 2
! as published by the Free Software Foundation.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this program; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

! JLB - 23 May 2008

MODULE PXDR_EXCITHD

  PRIVATE

  PUBLIC :: EQLHDV, COLHD

CONTAINS

!%%%%%%%%%%%%%%%%%%%%%%%
SUBROUTINE EQLHDV (iopt)
!%%%%%%%%%%%%%%%%%%%%%%%

   USE PXDR_FLAGS
   USE PXDR_CONSTANTES
   USE PXDR_CHEM_DATA
   USE PXDR_PROFIL
   USE PXDR_AUXILIAR
   USE PXDR_STR_DATA
   USE PXDR_RF_TOOLS

!  HD detailed balance

   IMPLICIT NONE

   INTEGER, INTENT (IN)                         :: iopt

   REAL (KIND=dp), DIMENSION (:,:), ALLOCATABLE :: fr_des
   REAL (KIND=dp), DIMENSION (:),   ALLOCATABLE :: fr_des_m
   INTEGER                                      :: jsp
   INTEGER                                      :: nused ! number of HD levels used in the computation
   INTEGER                                      :: nrxi  ! number of chemical reactions with HD
   INTEGER                                      :: mm    ! index of grain size
   INTEGER                                      :: if_ac
   INTEGER                                      :: lev
   INTEGER                                      :: levJ0
   REAL (KIND=dp)                               :: toto
   REAL (KIND=dp)                               :: ratefLH ! HD formation rate on grains by Langmuir-Hinshelwood
   REAL (KIND=dp)                               :: ratefER ! HD formation rate on grains by Eley-Rideal
   REAL (KIND=dp)                               :: ratefga ! HD formation rate in gas phase
   REAL (KIND=dp)                               :: ratedga ! HD destruction rate in gas phase
   REAL (KIND=dp)                               :: ratedes ! HD desorption rate from grains (thermal + non-thermal)
   REAL (KIND=dp)                               :: rateads ! HD adsorption rate on grains
   REAL (KIND=dp)                               :: ratedph ! HD photo-destruction rate in gas phase

   REAL (KIND=dp)                               :: taux, dumy
   INTEGER                                      :: k, k1, k2, k3
   INTEGER                                      :: ij, iv, iju, ivu, levu

   !-----------------------------------------------------------------------------
   ! Prepare some quantities
   !-----------------------------------------------------------------------------
   jsp   = in_sp(i_hd)
   nused = spec_lin(jsp)%use
   levJ0 = lev_hd(0,0)

   IF ( .NOT. ALLOCATED(fr_des))   ALLOCATE (fr_des(1:nused,npbgr))
   IF ( .NOT. ALLOCATED(fr_des_m)) ALLOCATE (fr_des_m(1:nused))

   if_ac    = 0        ! 0 for accretion, 1 for rejection
   !--- Rates of various processes in cm-3 s-1 ---
   ratefLH = 0.0_dp   ! HD formation rate on grains by Langmuir-Hinshelwood
   ratefER = 0.0_dp   ! HD formation rate on grains by Eley-Rideal
   ratefga = 0.0_dp   ! HD formation rate in gas phase
   ratedga = 0.0_dp   ! HD destruction rate in gas phase
   ratedes = 0.0_dp   ! HD desorption rate from grains (thermal + non-thermal)
   rateads = 0.0_dp   ! HD adsorption rate on grains
   ratedph = 0.0_dp   ! HD photo-destruction rate in gas phase

   create(1:nused) = 0.0_dp  ! formation   rate in each level      [cm-3 s-1]
   remove(1:nused) = 0.0_dp  ! destruction rate in each level      [cm-3 s-1]

   !----------------------------------------------------------------------------
   ! HD desorption from grains
   ! * controls HD excitation after thermal desorption from grains
   ! * follows a Boltzman distribution at grain temperatures
   !----------------------------------------------------------------------------
   DO mm = 1, npbgr
      toto = 0.0_dp
      DO lev = 1, nused
         fr_des(lev,mm) = spec_lin(jsp)%gst(lev) &
                        * EXP(-spec_lin(jsp)%elk(lev) / tdust(mm,iopt))
         toto = toto + fr_des(lev,mm)
      ENDDO
      fr_des(:,mm) = fr_des(:,mm) / toto ! normalization
   ENDDO
   ! Sum on grain sizes
   toto = 0.0_dp
   DO lev = 1, nused
      fr_des_m(lev) = SUM(fr_des(lev,1:npbgr) * wwmg(1:npbgr)) / SUM(wwmg(1:npbgr))
      toto = toto + fr_des_m(lev)
   ENDDO
   fr_des_m(:) = fr_des_m(:) / toto

   !----------------------------------------------------------------------------
   ! Fill create and remove matrixes
   !----------------------------------------------------------------------------

   nrxi = xmatk(i_hd)%nbt ! number of chemical reactions with HD

   DO k = 1, nrxi
      k1 = xmatk(i_hd)%fdt(k)%r1  ! Index of reactif 1
      k2 = xmatk(i_hd)%fdt(k)%r2  ! Index of reactif 2
      k3 = xmatk(i_hd)%fdt(k)%r3  ! Index of reactif 3

      SELECT CASE (xmatk(i_hd)%fdt(k)%it) ! number of chemical reactions with HD

      !===================================================================================
      !=== HD coming from grains (formation and desorption)
      !===================================================================================

      !.........................................................................
      ! (0) HD formation on grains - Old style/ R = 3*10-17 * sqrt(T/100)
      !     Not implemented
      !.........................................................................
      CASE (0)  ! Formation on grains
         PRINT *,"HD excitation with old style formation on grains is not implemented"
         STOP

      !.........................................................................
      ! (211) HD formation on grains - Langmuir-Hinshelwood
      !       Note: chemistry file must not have both type 0 and type 211
      !.........................................................................
      CASE (211) ! Formation on grains
         ratefLH = ratefLH + xmatk(i_hd)%fdt(k)%k * ab(k1) * ab(k2)
         !--- Excitation: use pophd
         DO lev = 1, nused
            ij = njl_hd(lev)
            iv = nvl_hd(lev)
            dumy = 0.0_dp
           DO levu = nused+1, nlevhd
              iju = njl_hd(levu)
              ivu = nvl_hd(levu)
              dumy = dumy + pophd(ivu,iju) * cashd(levu,lev)
           ENDDO
           create(lev) = create(lev) + (dumy + pophd(iv,ij)) * ratefLH
         ENDDO

      !.........................................................................
      ! (20) HD formation on grains - Eley-Rideal with chemisorbed species
      !.........................................................................
      CASE (20) ! example: D + H:: -> H2 (ER with chemisorbed species)
         ratefER = ratefER + xmatk(i_hd)%fdt(k)%k * ab(k1) * ab(k2)       ![cm-3 s-1]
         !--- Excitation ---
         ! create is filled after ratefER has been completed with ER formation
         ! involving physisorbed species (type 113)

      !.........................................................................
      ! (116 / 117) HD desorption from grains by cosmic rays (116) and photons (117)
      !.........................................................................
      CASE (116, 117)  ! Various desorption processes
         taux = xmatk(i_hd)%fdt(k)%k * ab(k1)                             ![cm-3 s-1]
         DO lev = 1, nused
            iju = njl_hd(lev)
            ivu = nvl_hd(lev)
            create(lev) = create(lev) + taux * pophd(ivu,iju)
         ENDDO
         ratedes = ratedes + taux                                         ![cm-3 s-1]

      !.........................................................................
      ! (118) HD evaporation from grains
      !.........................................................................
      CASE (118)  ! Various desorption processes
         taux = xmatk(i_hd)%fdt(k)%k * ab(k1)                             ![cm-3 s-1]
         DO lev = 1, nused
            create(lev) = create(lev) + taux * fr_des_m(lev)
         ENDDO
         ratedes = ratedes + taux                                         ![cm-3 s-1]

      !===================================================================================
      !=== HD destruction
      !===================================================================================

      !.........................................................................
      ! (7) H2 photodissociation
      !.........................................................................
      CASE (7)  ! Photo-dissociation
         ! Matrix evrdhd is added below (WHERE ?? - check EQMAT)
         ! We do nothing

      !.........................................................................
      ! (113) HD sticking on grains
      !.........................................................................
       CASE (113)  ! sticking on grains and Eley-Rideal on physisorbed sites
         ! ATTENTION: besoin de verifier les signes dans les + / - taux
         taux = xmatk(i_hd)%fdt(k)%k                  ! xmatk is negative or positive
         IF (k1 == i_hd) THEN
            IF (if_ac == 0) THEN                      ! adsorption
               ! taux is negative
               ! * use -taux to add the contribution to rateads and remove
               rateads   = rateads   - taux * ab(k1)
               remove(:) = remove(:) - taux * ab(k1)
               if_ac = 1
            ELSE                                      ! rejection
               ! taux is positive
               ! * use -taux because it is a correction term to the previous IF
               rateads   = rateads   - taux * ab(k1) * ab(k2)
               remove(:) = remove(:) - taux * ab(k1) * ab(k2)
            ENDIF
         ELSE     ! Eley-Rideal formation on physisorbed species
            ! taux is positive
            ratefER = ratefER + taux * ab(k1) * ab(k2)
            !--- Excitation ---
            ! create is filled after ratefer has been completed with ER formation
            ! involving chemisorbed species (type 20)
         ENDIF

      !.........................................................................
      ! DEFAULT - HD formation and destruction in gas chemical reactions
      !           other than photo-dissociation
      !.........................................................................
      CASE DEFAULT
         taux = xmatk(i_hd)%fdt(k)%k * ab(k1) * ab(k2) * ab(k3)
         IF (taux > 0.0_dp) THEN
            ! Scenarios : all in J = 0
            create(levJ0) = create(levJ0) + taux
            ratefga       = ratefga + taux
         ELSE ! taux is negative
            remove(:) = remove(:) - taux
            ratedga   = ratedga   - taux
         ENDIF
      END SELECT
   ENDDO


   !======================================================================================
   ! Additionnal terms in create(:)
   !======================================================================================
   !--- Excitation by Eley-Rideal mechanisms (chemisorbed + physisorbed) : pophd
   DO lev = 1, nused
      ij = njl_hd(lev)
      iv = nvl_hd(lev)
      dumy = 0.0_dp
      DO levu = nused+1, nlevhd
         iju = njl_hd(levu)
         ivu = nvl_hd(levu)
         dumy = dumy + pophd(ivu,iju) * cashd(levu,lev)
      ENDDO
      create(lev) = create(lev) + (dumy + pophd(iv,ij)) * ratefER
   ENDDO

   !--------------------------------------------------------------------------------------
   ! Store some useful data for PREP concerning H2 formation / destruction processes
   !  timhd1(iopt) : HD formation by gas phase chemistry                        [cm-3 s-1]
   !  timhd2(iopt) : HD destruction by gas phase chemistry                      [cm-3 s-1]
   !  timhd3(iopt) : HD formation by Langmuir-Hinshelwood process or old style  [cm-3 s-1]
   !  timhd4(iopt) : HD destruction by photons                                  [cm-3 s-1]
   !  timhd5(iopt) : HD formation by Eley-Rideal process                        [cm-3 s-1]
   !--------------------------------------------------------------------------------------
   timhd1(iopt) = ratefga + ratedes
   timhd2(iopt) = ratedga
   timhd3(iopt) = ratefLH ! + ratefgr (ratefgr could be old style - not implemented)
   timhd4(iopt) = pdiesp(kdeph0+1,iopt) * ab(i_hd)


END SUBROUTINE EQLHDV

!-------------------------------------------------------------

!%%%%%%%%%%%%%%%%%%%%%%
SUBROUTINE COLHD (ttry)
!%%%%%%%%%%%%%%%%%%%%%%

   USE PXDR_FLAGS
   USE PXDR_CONSTANTES
   USE PXDR_CHEM_DATA
   USE PXDR_PROFIL
   USE PXDR_AUXILIAR
   USE PXDR_STR_DATA

!---------------------------------------------
!        ttry  : cloud temperature in the slab
!---------------------------------------------

   IMPLICIT NONE

   REAL (KIND=dp), INTENT (IN)                  :: ttry

   REAL (KIND=dp)                               :: deche, dech1, dech2
   REAL (KIND=dp)                               :: tddhe, tddh1, tddh2
   REAL (KIND=dp)                               :: ctod, fac
   REAL (KIND=dp)                               :: chdhe, chdh, chdph2, chdoh2, c, d
   REAL (KIND=dp)                               :: dip, betas, b0, deltaE, taux
   INTEGER                                      :: ju, jl, levu, levl
   INTEGER                                      :: nj, nv
   INTEGER                                      :: jsp_hd

   jsp_hd = in_sp(i_hd)

   ! New fits (Jacques (1999))
   deche = 1.0_dp / 3.0_dp
   tddhe = deche + ttry * 1.0e-3_dp
   dech1 = 2.0_dp / 3.0_dp
   tddh1 = dech1 + ttry * 1.0e-3_dp
   dech2 = 0.5_dp
   tddh2 = dech2 + ttry * 1.0e-3_dp

   DO levu = 2, l0hd
      ju = njl_hd(levu)
      DO levl = 1, levu-1
         jl = njl_hd(levl)
         fac = (spec_lin(jsp_hd)%elk(levu) - spec_lin(jsp_hd)%elk(levl)) / ttry
         ctod = EXP(-fac) * (2.0_dp*ju+1.0_dp) / (2.0_dp*jl+1.0_dp)

         ! Collision HD - He
         chdhe = 10.0_dp**(q_hd_he(1,levu,levl) + q_hd_he(2,levu,levl) / tddhe &
                        + q_hd_he(3,levu,levl) / (tddhe*tddhe))

         ! Collision HD - H
         chdh = 10.0_dp**(q_hd_h(1,levu,levl) + q_hd_h(2,levu,levl) / tddh1 &
                       + q_hd_h(3,levu,levl) / (tddh1*tddh1))

         ! Collision HD - H2 para
         chdph2 = 10.0_dp**(q_hd_ph2(1,levu,levl) + q_hd_ph2(2,levu,levl) / tddh2 &
                       + q_hd_ph2(3,levu,levl) / (tddh2*tddh2))

         ! Collisions HD - H2 ortho
         chdoh2 = 10.0_dp**(q_hd_oh2(1,levu,levl) + q_hd_oh2(2,levu,levl) / tddh2 &
                         + q_hd_oh2(3,levu,levl) / (tddh2*tddh2))

         c = chdhe * ab(i_he) + chdh * ab(i_h) + chdph2 * ab(i_h2) * h2para + chdoh2 * ab(i_h2) * h2orth
         d = c * ctod

         eqcol(levl,levu) = eqcol(levl,levu) - c
         eqcol(levu,levu) = eqcol(levu,levu) + c
         eqcol(levu,levl) = eqcol(levu,levl) - d
         eqcol(levl,levl) = eqcol(levl,levl) + d
      ENDDO
   ENDDO

!  Collisions with electrons -------------------------------------
!  cf Dickinson et al.  A&A 54,645(1977)

   dip = 8.4e-4_dp
   betas = 11600.0_dp / ttry
   b0 = 45.655_dp
   DO levl = 1, l0hd-1
      nj = njl_hd(levl)
      nv = nvl_hd(levl)
      IF (nv == 0) THEN
         levu = lev_hd(nv,nj+1)
         deltaE = 2.48e-04_dp * b0 * (nj+1)
         c = 9.08e3_dp / (b0*(nj+1))
         taux = (3.56e-06_dp * (dip**2) / SQRT(ttry)) * (nj+1) / (2*nj+1)
         taux = taux * EXP(-betas*deltaE)
         taux = taux * LOG(c*deltaE+(c/betas) * EXP(-0.577_dp / (1.0_dp+2.0_dp*betas*deltaE)))
         taux = taux * ab(nspec)
         eqcol(levu,levl) = eqcol(levu,levl) - taux
         eqcol(levl,levl) = eqcol(levl,levl) + taux
         fac = (spec_lin(jsp_hd)%elk(levu) - spec_lin(jsp_hd)%elk(levl)) / ttry
         ctod = EXP(fac) * (2.0_dp*nj+1.0_dp) / (2.0_dp*nj+3.0_dp)
         taux = taux * ctod
         eqcol(levl,levu) = eqcol(levl,levu) - taux
         eqcol(levu,levu) = eqcol(levu,levu) + taux
      ENDIF
   ENDDO

END SUBROUTINE COLHD

END MODULE PXDR_EXCITHD
