! Copyright (C) 2002 - Observatoire de Paris/Meudon
!                      Universite Paris 7
!                      CNRS (France)
!                      Contact : Jacques.Lebourlot@obspm.fr
!
! This program is free software; you can redistribute it and/or
! modify it under the terms of the GNU General Public License version 2
! as published by the Free Software Foundation.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this program; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

! JLB 21 - May 2008

MODULE PXDR_ISOTCO

PRIVATE

PUBLIC :: ISOTC, ISOTO

CONTAINS

!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
SUBROUTINE ISOTC (nreact,idiff)
!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  USE PXDR_FLAGS
  USE PXDR_CONSTANTES
  USE PXDR_CHEM_DATA
  USE PXDR_AUXILIAR
  USE PXDR_STR_DATA

! *************************************************************************
!          CONSTRUCTION DE LA CHIMIE ISOTOPEE DU CARBONE
! *************************************************************************
!
!         NEQT   = NOMBRE TOTAL DE REACTIONS SANS ISOTOPE
!         NREACT = NUMERO DE LA REACTION ISOTOPEE
!         IDIFF  = NUMERO DE LA DERNIERE REACTION DIFFERENTIELLE
!         JJ     = NUMERO DE LA REACTION CORRESPONDANTE (SS ISOTOPE)
!
! *************************************************************************
!
!   CONVENTION :    DANS LA LISTE DES ESPECES, CHAQUE ESPECE ISOTOPEE DOIT
!   ----------      SUIVRE IMMEDIATEMENT l ESPECE MERE
!                   DANS LE CAS OU l ESPECE CONTIENT DE l OXYGENE ET DU
!                   CARBONE (C OU C*) ==> l ESPECE ISOTOPEE: O* SUIT C*
!                   EX :    HCO2
!                           HC*O2
!                           HCO*O
!                           HC*O*O
! *************************************************************************

  IMPLICIT NONE

  INTEGER, INTENT (INOUT) :: nreact
  INTEGER, INTENT (OUT)   :: idiff
  INTEGER                 :: ir0, ir1, ir2, ir3c, ir3, ir4, ir5, ir6
  INTEGER                 :: ncr1, ncr2, ncp1, ncp2, ncp3, ncp4
  INTEGER                 :: itest, jj
  INTEGER, DIMENSION(4)   :: prod_isot1 ! Table indicating in which of the 4 products is the 1st C
  INTEGER, DIMENSION(4)   :: prod_isot2 ! Table indicating in which of the 4 products is the 2nd C

  PRINT *,"Build C13 chemistry"

  !=======================================================================================
  ! Loop on all chemical reactions
  !=======================================================================================
  DO jj = 1, neqt

     !--- Skip photo-destruction reactions -----------------------------------------------
     IF (itype(jj) == 7) CYCLE

     !--- Get index of chemical species in reactions -------------------------------------
     ir0  = react(jj,0)    !
     ir1  = react(jj,1)    ! reactant 1
     ir2  = react(jj,2)    ! reactant 2
     ir3c = react(jj,3)    ! reactant 3 (3 bodies reactions)
     ir3  = react(jj,4)    ! product  1
     ir4  = react(jj,5)    ! product  2
     ir5  = react(jj,6)    ! product  3
     ir6  = react(jj,7)    ! product  4

     !--- Special: reactions with photons or cosmic rays ---------------------------------
     !--- IL N'Y A PAS DE CARBONE DANS LES PHOTONS OU LES CRP
     !    ou les photons secondaires ou les grains
     !            (POUR LESQUELS AEL N'EST PAS DEFINI)
     IF (ir2 == n_var+1 .OR. ir2 == n_var+2 .OR. ir2 == n_var+3 .OR. ir2 == n_var+4) THEN
        ncr2 = 0
        go to 550
     ENDIF

     !------------------------------------------------------------------------------------
     ! Composition of elements in the chemical reaction
     ! Y-A-t IL DEJA DES ESPECES ISOTOPEES (C* OU O*) DANS LA REACTION ?
     ! SI OUI, ON PASSE A LA REACTION SUIVANTE
     !------------------------------------------------------------------------------------
     itest = ael(ir1,7) + ael(ir2,7) + ael(ir1,9) + ael(ir2,9)
     IF (itest > 0)    THEN
        idiff = jj
        CYCLE
     ENDIF

     !------------------------------------------------------------------------------------
     !--- Y-A-t IL DU CARBONE DANS LA REACTION CONSIDEREE ?
     !                                        ET COMBIEN ?
     !
     !          NCR1, NCR2 = NOMBRE DE C DANS LES REACTANTS R1 ET R2
     !          NCP1, NCP2, NCP3, NCP4 = NOMBRE DE C DANS LES REACTANTS P1, P2, P3 et P4
     !------------------------------------------------------------------------------------

     !====================================================================================
     ! Get the number of C atoms in the second reactant
     !====================================================================================
     IF (ael(ir2,1) /= ael(ir2+1,1)   .OR.  ael(ir2,2) /= ael(ir2+1,2)+1 .OR. &
         ael(ir2,7) /= ael(ir2+1,7)-1) THEN
        ncr2 = 0
     ELSE
        ncr2 = ael(ir2,2)
     ENDIF

  550 CONTINUE

     !====================================================================================
     ! Get the number of C atoms in the chemical reaction species
     !====================================================================================
     !--- Number of C in reactant 1 ------------------------------------------------------
     IF (ael(ir1,1) /= ael(ir1+1,1)   .OR.  ael(ir1,2) /= ael(ir1+1,2)+1 .OR. &
        ael(ir1,7) /= ael(ir1+1,7)-1) THEN
        ncr1 = 0
     ELSE
        ncr1 = ael(ir1,2)
     ENDIF

     !--- Number of C in product 1 ------------------------------------------------------
     IF (ael(ir3,1) /= ael(ir3+1,1)   .OR.  ael(ir3,2) /= ael(ir3+1,2)+1 .OR. &
        ael(ir3,7) /= ael(ir3+1,7)-1) THEN
        ncp1 = 0
     ELSE
        ncp1 = ael(ir3,2)
     ENDIF

     !--- Number of C in product 2 ------------------------------------------------------
     IF (ir4 /= 0) THEN
        IF (ael(ir4,1) /= ael(ir4+1,1)   .OR.  ael(ir4,2) /= ael(ir4+1,2)+1 .OR. &
           ael(ir4,7) /= ael(ir4+1,7)-1) THEN
           ncp2 = 0
        ELSE
           ncp2 = ael(ir4,2)
        ENDIF
     ELSE
        ncp2 = 0
     ENDIF

     !--- Number of C in product 3 ------------------------------------------------------
     IF (ir5 /= 0) THEN
        IF (ael(ir5,1) /= ael(ir5+1,1)   .OR.  ael(ir5,2) /= ael(ir5+1,2)+1 .OR. &
           ael(ir5,7) /= ael(ir5+1,7)-1) THEN
           ncp3 = 0
        ELSE
           ncp3 = ael(ir5,2)
        ENDIF
     ELSE
        ncp3 = 0
     ENDIF

     !--- Number of C in product 4 ------------------------------------------------------
     IF (ir6 /= 0) THEN
        IF (ael(ir6,1) /= ael(ir6+1,1)   .OR.  ael(ir6,2) /= ael(ir6+1,2)+1 .OR. &
           ael(ir6,7) /= ael(ir6+1,7)-1) THEN
           ncp4 = 0
        ELSE
           ncp4 = ael(ir6,2)
        ENDIF
     ELSE
        ncp4 = 0
     ENDIF

     !--- Special: second product is photon or cosmic ------------------------------------
     IF (ir4 == n_var+1 .OR. ir4 == n_var+2 .OR. ir4 == n_var+3) THEN
        ncp2 = 0
     ENDIF

     !====================================================================================
     ! Create tables indicating the positions of the 2 potential C in products
     !====================================================================================
     ! First carbon:
     prod_isot1(:) = 0
     IF (ncp1 > 0) THEN
        prod_isot1(1) = 1
     ELSE IF(ncp2 > 0) THEN
        prod_isot1(2) = 1
     ELSE IF(ncp3 > 0) THEN
        prod_isot1(3) = 1
     ELSE IF (ncp4 > 0) THEN
        prod_isot1(4) = 1
     ENDIF
     ! Second carbon:
     prod_isot2(:) = 0
     IF(ncp2 > 0       .AND. prod_isot1(2) == 0) THEN
       prod_isot2(2) = 1
     ELSE IF(ncp3 > 0  .AND. prod_isot1(3) == 0) THEN
       prod_isot2(3) = 1
     ELSE IF (ncp4 > 0 .AND. prod_isot1(4) == 0) THEN
       prod_isot2(4) = 1
     ENDIF

     !===================================================================================
     ! 1 - FIRST CASE: C is present in two reactants
     !     We considerer 3 possibilites:
     !     1.1: C is present in 2 products => create two reactions
     !     1.2: C is present in 1 product  => create one reaction
     !     1.3: The reaction cannot be isotoped => do not create a reaction
     !===================================================================================
     IF (ncr1 > 0 .AND. ncr2 > 0) THEN
        !--------------------------------------------------------------------------------
        ! 1.1: C is present in 2 products
        !      CONVENTION: the reaction must be written in the order:
        !      AC + BC -> A'C + B'C + D'  + E'
        !      AC + BC -> A'  + B'C + D'C + E
        !      AC + BC -> A'C + B'  + D'C + E
        !      AC + BC -> A'  + B'  + D'C + E'C
        !      so that: =>  AC* + BC  -> A'C* + B'C  + D' + E'
        !                             -> A'C  + B'C* + D' + E'
        !          and: =>  AC  + BC* -> A'C* + B'C  + D' + E'
        !                             -> A'C  + B'C* + D' + E'
        !      => creation of 4 reactions
        !--------------------------------------------------------------------------------
        IF (      (ncp1 > 0 .AND. ncp2 > 0) .OR. (ncp1 > 0 .AND. ncp3 > 0) &
             .OR. (ncp1 > 0 .AND. ncp4 > 0)                                &
             .OR. (ncp2 > 0 .AND. ncp3 > 0) .OR. (ncp2 > 0 .AND. ncp4 > 0) &
             .OR. (ncp3 > 0 .AND. ncp4 > 0) &
           ) THEN

           !PRINT *, speci(ir1), speci(ir2), speci(ir3), speci(ir4), speci(ir5), speci(ir6)

           !--- PREMIERE REACTION ISOTOPEE --------------------------------------------------
           nreact = nreact + 1

           react (nreact,0) = ir0
           react (nreact,1) = ir1 + 1
           react (nreact,2) = ir2
           react (nreact,3) = ir3c
           
           react (nreact,4) = ir3 + prod_isot1(1)
           react (nreact,5) = ir4 + prod_isot1(2)
           react (nreact,6) = ir5 + prod_isot1(3)
           react (nreact,7) = ir6 + prod_isot1(4)

           IF (ir1 /= ir2) gamm(nreact)  = gamm(jj) / 2.0_dp
           IF (ir1 == ir2) gamm(nreact)  = gamm(jj) / 4.0_dp
           alpha(nreact) = alpha(jj)
           bet(nreact)   = bet(jj)
           itype(nreact) = itype(jj)
           !PRINT *, speci(react(nreact,1)), speci(react(nreact,2)), speci(react(nreact,4)) &
           !       , speci(react(nreact,5)), speci(react(nreact,6)), speci(react(nreact,7))

           !--- SECONDE REACTION ISOTOPEE --------------------------------------------------
           nreact = nreact + 1

           react (nreact,0) = ir0
           react (nreact,1) = ir1 + 1
           react (nreact,2) = ir2
           react (nreact,3) = ir3c
           
           react (nreact,4) = ir3 + prod_isot2(1)
           react (nreact,5) = ir4 + prod_isot2(2)
           react (nreact,6) = ir5 + prod_isot2(3)
           react (nreact,7) = ir6 + prod_isot2(4)

           IF (ir1 /= ir2) gamm(nreact)  = gamm(jj) / 2.0_dp
           IF (ir1 == ir2) gamm(nreact)  = gamm(jj) / 4.0_dp
           alpha(nreact) = alpha(jj)
           bet(nreact)   = bet(jj)
           itype(nreact) = itype(jj)
           !PRINT *, speci(react(nreact,1)), speci(react(nreact,2)), speci(react(nreact,4)) &
           !       , speci(react(nreact,5)), speci(react(nreact,6)), speci(react(nreact,7))

           !--- TROSIEME REACTION ISOTOPEE -------------------------------------------------
           nreact = nreact + 1
   
           react (nreact,0) = ir0
           react (nreact,1) = ir1
           react (nreact,2) = ir2 + 1
           react (nreact,3) = ir3c
           
           react (nreact,4) = ir3 + prod_isot1(1)
           react (nreact,5) = ir4 + prod_isot1(2)
           react (nreact,6) = ir5 + prod_isot1(3)
           react (nreact,7) = ir6 + prod_isot1(4)
   
           IF (ir1 /= ir2) gamm(nreact)  = gamm(jj) / 2.0_dp
           IF (ir1 == ir2) gamm(nreact)  = gamm(jj) / 4.0_dp
           alpha(nreact) = alpha(jj)
           bet(nreact)   = bet(jj)
           itype(nreact) = itype(jj)
           !PRINT *, speci(react(nreact,1)), speci(react(nreact,2)), speci(react(nreact,4)) &
           !       , speci(react(nreact,5)), speci(react(nreact,6)), speci(react(nreact,7))

           !--- QUATRIEME REACTION ISOTOPEE -------------------------------------------------
           nreact = nreact + 1
   
           react (nreact,0) = ir0
           react (nreact,1) = ir1
           react (nreact,2) = ir2 + 1
           react (nreact,3) = ir3c
           
           react (nreact,4) = ir3 + prod_isot2(1)
           react (nreact,5) = ir4 + prod_isot2(2)
           react (nreact,6) = ir5 + prod_isot2(3)
           react (nreact,7) = ir6 + prod_isot2(4)
   
           IF (ir1 /= ir2) gamm(nreact)  = gamm(jj) / 2.0_dp
           IF (ir1 == ir2) gamm(nreact)  = gamm(jj) / 4.0_dp
           alpha(nreact) = alpha(jj)
           bet(nreact)   = bet(jj)
           itype(nreact) = itype(jj)
           !PRINT *, speci(react(nreact,1)), speci(react(nreact,2)), speci(react(nreact,4)) &
           !       , speci(react(nreact,5)), speci(react(nreact,6)), speci(react(nreact,7))
           !PRINT *, ""
        !--------------------------------------------------------------------------------
        ! 1.2: C is present in 1 product
        !
        !      AC  +  BC  -->  A'C2  +  B'    + D'   + E'
        !  OU  AC  +  BC  -->  A'    +  B'C2  + D'   + E'
        !  OU  AC  +  BC  -->  A'    +  B'    + D'C2 + E'
        !  OU  AC  +  BC  -->  A'    +  B'    + D'   + E'C2
        !
        !                            =====>      AC* + BC  -->  A'C* +  B'    + D'   + E'
        !                                    OU  AC* + BC  -->  A'   +  B'C*  + D'   + E'
        !                                    OU  AC* + BC  -->  A'   +  B'    + D'C* + E'
        !                                    OU  AC* + BC  -->  A'   +  B'    + D'   + E'C*
        !
        !                        ET  =====>      AC  + BC* -->  A'C* +  B'    + D'   + E'
        !                                    OU  AC  + BC* -->  A'   +  B'C*  + D'   + E'
        !                                    OU  AC  + BC* -->  A'   +  B'    + D'C* + E'
        !                                    OU  AC  + BC* -->  A'   +  B'    + D'   + E'C*
        !
        !                       ==> 2 REACTIONS ISOTOPEES
        !--------------------------------------------------------------------------------
        ELSE IF (ncp1 > 0 .OR. ncp2 > 0 .OR. ncp3 > 0 .OR. ncp4 > 0) THEN

           !--- PREMIERE REACTION ISOTOPEE :
           nreact = nreact + 1

           react (nreact,0) = ir0
           react (nreact,1) = ir1 + 1
           react (nreact,2) = ir2
           react (nreact,3) = ir3c

           react (nreact,4) = ir3 + prod_isot1(1)
           react (nreact,5) = ir4 + prod_isot1(2)
           react (nreact,6) = ir5 + prod_isot1(3)
           react (nreact,7) = ir6 + prod_isot1(4)

           IF (ir1 /= ir2) gamm(nreact)  = gamm(jj)
           IF (ir1 == ir2) gamm(nreact)  = gamm(jj) / 2.0_dp
           alpha(nreact) = alpha(jj)
           bet(nreact)   = bet(jj)
           itype(nreact) = itype(jj)

           !--- DEUXIEME REACTION ISOTOPEE :
           nreact = nreact + 1

           react (nreact,0) = ir0
           react (nreact,1) = ir1
           react (nreact,2) = ir2 + 1
           react (nreact,3) = ir3c

           react (nreact,4) = ir3 + prod_isot1(1)
           react (nreact,5) = ir4 + prod_isot1(2)
           react (nreact,6) = ir5 + prod_isot1(3)
           react (nreact,7) = ir6 + prod_isot1(4)
   
           IF (ir1 /= ir2) gamm(nreact)  = gamm(jj)
           IF (ir1 == ir2) gamm(nreact)  = gamm(jj) / 2.0_dp
           alpha(nreact) = alpha(jj)
           bet(nreact)   = bet(jj)
           itype(nreact) = itype(jj)
        ENDIF
        !--- ON PASSE A l EXAMEN DE LA REACTION SUIVANTE :
        CYCLE
     ENDIF

     !===================================================================================
     ! 2 - SECOND CASE: C is present in one reactant only
     !     We considerer 2 possibilites:
     !     2.1: C is present in 2 products => create two reactions
     !     2.2: C is present in 1 product  => create one reaction
     !     2.3: The reaction cannot be isotoped => do not create a reaction
     !===================================================================================
     IF (ncr1 > 0 .OR. ncr2 > 0) THEN

        !---------------------------------------------------------------------------------
        ! 2.1: C is present in 2 products
        !      AC  +  B   -->  A'C  +  B'C + D' + E'        
        !  OU  A   +  BC  -->  A'C  +  B'C + D' + E'
        !
        !      =====>      AC*  +  B    -->  A'C*  +  B'C + ...
        !              OU  A    +  BC   -->  A'C*  +  B'C + ...
        !
        !  ET  =====>      AC*  +  B    -->  A'C   +  B'C* + ...
        !              OU  A    +  BC   -->  A'C   +  B'C* + ...
        !
        ! ==> 2 REACTIONS ISOTOPEES with gamma / 2.0_dp
        !---------------------------------------------------------------------------------
        IF (      (ncp1 > 0 .AND. ncp2 > 0) .OR. (ncp1 > 0 .AND. ncp3 > 0) &
             .OR. (ncp1 > 0 .AND. ncp4 > 0)                                &
             .OR. (ncp2 > 0 .AND. ncp3 > 0) .OR. (ncp2 > 0 .AND. ncp4 > 0) &
             .OR. (ncp3 > 0 .AND. ncp4 > 0) &
            ) THEN

           !--- PREMIERE REACTION ISOTOPEE:
           nreact = nreact + 1

           react (nreact,0) = ir0
           IF (ncr1 == 0) THEN
              react (nreact,1) = ir1
              react (nreact,2) = ir2 + 1
           ELSE
              react (nreact,1) = ir1 + 1
              react (nreact,2) = ir2
           ENDIF
           react (nreact,3) = ir3c

           react (nreact,4) = ir3 + prod_isot1(1)
           react (nreact,5) = ir4 + prod_isot1(2)
           react (nreact,6) = ir5 + prod_isot1(3)
           react (nreact,7) = ir6 + prod_isot1(4)

           gamm(nreact)  = gamm(jj)/2.0_dp
           alpha(nreact) = alpha(jj)
           bet(nreact)   = bet(jj)
           itype(nreact) = itype(jj)

           !--- DEUXIEME REACTION ISOTOPEE:
           nreact = nreact + 1

           react (nreact,0) = ir0
           IF (ncr1 == 0) THEN
              react (nreact,1) = ir1
              react (nreact,2) = ir2 + 1
           ELSE
              react (nreact,1) = ir1 + 1
              react (nreact,2) = ir2
           ENDIF
           react (nreact,3) = ir3c

           react (nreact,4) = ir3 + prod_isot2(1)
           react (nreact,5) = ir4 + prod_isot2(2)
           react (nreact,6) = ir5 + prod_isot2(3)
           react (nreact,7) = ir6 + prod_isot2(4)
           
           gamm(nreact)  = gamm(jj)/2.0
           alpha(nreact) = alpha(jj)
           bet(nreact)   = bet(jj)
           itype(nreact) = itype(jj)

        !---------------------------------------------------------------------------------
        ! 2.2: C is present in 1 product
        !      AC  +  B   -->  A'C  +  B'
        !  OU  A   +  BC  -->  A'C  +  B'
        !  OU  AC  +  B   -->  A'   +  B'C
        !  OU  A   +  BC  -->  A'   +  B'C
        !  OU  A   +  BC  -->  A'   +  B'     + C
        !  OU  A   +  BC  -->  A'   +  B'     + D + C
        !
        !                            =====>      AC*  +  B    -->  A'C*  +  B'
        !                                    OU  A    +  BC*  -->  A'C*  +  B'
        !                                    OU  AC*  +  B    -->  A'    +  B'C*
        !                                    OU  A    +  BC*  -->  A'    +  B'C*
        !                                    OU  A    +  BC*  -->  A'    +  B'      + C*
        !                                    OU  A    +  BC*  -->  A'    +  B'      + D + C*
        !                       ==> 1 REACTION ISOTOPEE
        !---------------------------------------------------------------------------------
        ELSE IF (ncp1 > 0 .OR. ncp2 > 0  .OR. ncp3 > 0 .OR. ncp4 > 0) THEN
           !--- UNE SEULE REACTION ISOTOPEE :
           nreact = nreact + 1

           react (nreact,0) = ir0
           IF (ncr1 == 0) THEN
              react (nreact,1) = ir1
              react (nreact,2) = ir2 + 1
           ELSE
              react (nreact,1) = ir1 + 1
              react (nreact,2) = ir2
           ENDIF
           react (nreact,3) = ir3c

           react (nreact,4) = ir3 + prod_isot1(1)
           react (nreact,5) = ir4 + prod_isot1(2)
           react (nreact,6) = ir5 + prod_isot1(3)
           react (nreact,7) = ir6 + prod_isot1(4)

           gamm(nreact)  = gamm(jj)
           alpha(nreact) = alpha(jj)
           bet(nreact)   = bet(jj)
           itype(nreact) = itype(jj)

        ENDIF
     ENDIF !--- End of SECOND CASE: C is present in one reactant only
  ENDDO    !--- End of loop on chemical reactions

!----------------------------------------------------------------------------
!                      FIN DE l EXAMEN DES REACTIONS
!          ET DE LA CONSTRUCTION DE LA CHIMIE ISOTOPEE DU CARBONE
!----------------------------------------------------------------------------

END SUBROUTINE ISOTC

!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
SUBROUTINE ISOTO (nreact,idiff)
!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

  USE PXDR_FLAGS
  USE PXDR_CONSTANTES
  USE PXDR_CHEM_DATA
  USE PXDR_AUXILIAR
  USE PXDR_STR_DATA

! *************************************************************************
!          CONSTRUCTION DE LA CHIMIE ISOTOPEE DE l OXYGENE
! *************************************************************************
!
!         NEQT   = NOMBRE TOTAL DE REACTIONS SANS ISOTOPE
!         IRC    = NOMBRE TOTAL DE REACTIONS SANS ISOTOPE DE l OXYGENE
!         NREACT = NUMERO DE LA REACTION ISOTOPEE
!         IDIFF  = NUMERO DE LA DERNIERE REACTION DIFFERENTIELLE
!         JJ     = NUMERO DE LA REACTION CORRESPONDANTE (SS ISOTOPE)
!
! *************************************************************************
!
!   CONVENTION :    DANS LA LISTE DES ESPECES, CHAQUE ESPECE ISOTOPEE DOIT
!   ----------      SUIVRE IMMEDIATEMENT l ESPECE MERE
!                   DANS LE CAS OU l ESPECE CONTIENT DE l OXYGENE ET DU
!                   CARBONE (C OU C*) ==> l ESPECE ISOTOPEE: O* SUIT C*
!                   EX :    HCO2
!                           HC*O2
!                           HCO*O
!                           HC*O*O
! *************************************************************************

  IMPLICIT NONE

  INTEGER, INTENT (INOUT) :: nreact
  INTEGER, INTENT (IN)    :: idiff
  INTEGER                 :: ir0, ir1, ir2, ir3c, ir3, ir4, ir5, ir6
  INTEGER                 :: jj
  INTEGER                 :: nor1, nor2, nop1, nop2, nop3, nop4, is1, is2, is3, is4, is5, is6
  INTEGER, DIMENSION(4)   :: prod_isot1 ! Table indicating in which of the 4 products is the 1st O
  INTEGER, DIMENSION(4)   :: prod_isot2 ! Table indicating in which of the 4 products is the 2nd O

  PRINT *,"Build O18 chemistry"

  !=======================================================================================
  ! Loop on all chemical reactions
  !=======================================================================================
  DO jj = idiff+1, nreact

     !--- Skip photo-destruction reactions -----------------------------------------------
     IF (itype(jj) == 7) CYCLE

     !--- Get index of chemical species in reactions -------------------------------------
     ir0 = react(jj,0)
     ir1 = react(jj,1)    ! reactant 1
     ir2 = react(jj,2)    ! reactant 2
     ir3c = react(jj,3)   ! reactant 3 (3 bodies reactions)
     ir3 = react(jj,4)    ! product 1
     ir4 = react(jj,5)    ! product 2
     ir5 = react(jj,6)    ! product 3
     ir6 = react(jj,7)    ! product 4

     !--- Initialization : flag to indicate if we need to jump over a 13C specy ----------
     is1 = 0
     is2 = 0
     is3 = 0
     is4 = 0
     is5 = 0
     is6 = 0

     !--- Special: reactions with photons or cosmic rays ---------------------------------
     !--- IL N'Y A PAS D'OXYGENE DANS LES PHOTONS OU LES CRP
     !    ou les photons secondaires ni dans les grains
     !            (POUR LESQUELS AEL N'EST PAS DEFINI)
     !-----------------------------------------------------------------------------
     !---Y-A-t IL DE l OXYGENE DANS LA REACTION CONSIDEREE ?
     !                                          ET COMBIEN ?
     !
     !          NOR1, NOR2 = NOMBRE DE O DANS LES REACTANTS R1 ET R2
     !          NOP1, NOP2 = NOMBRE DE O DANS LES REACTANTS P1 ET P2
     !
     !---Y-A-t IL un isotope carbone a sauter? si oui, is123456 = 1
     !
     !          NCR1, NCR2 = NOMBRE DE C (OU C*) DANS LES REACTANTS R1 ET R2
     !          NCP1, NCP2 = NOMBRE DE C (OU C*) DANS LES REACTANTS P1 ET P2
     !-----------------------------------------------------------------------------

     !====================================================================================
     ! Get the number of O atoms in the second reactant
     !====================================================================================
     !--- Second reactant is an electron, a photon or a cosmic ray or a grain
     IF (ir2 == n_var+1 .OR. ir2 == n_var+2 .OR. ir2 == n_var+3 .OR. ir2 == n_var+4) THEN
        nor2 = 0
     ELSE IF ( ael(ir2,1) == ael(ir2+1,1)   .AND.  ael(ir2,2) == ael(ir2+1,2)   &
         .AND. ael(ir2,4) == ael(ir2+1,4)+1 .AND.  ael(ir2,9) == ael(ir2+1,9)-1 ) THEN
        nor2 = ael(ir2,4)
        is2  = 0
     ELSE IF ( ael(ir2,1) == ael(ir2+2,1)   .AND.  ael(ir2,2) == ael(ir2+2,2)      &
         .AND. ael(ir2,4) == ael(ir2+2,4)+1 .AND.  ael(ir2,9) == ael(ir2+2,9)-1) THEN
        nor2 = ael(ir2,4)
        is2 = 1
     ELSE
        nor2 = 0
     ENDIF

     !====================================================================================
     ! Get the number of O atoms in the chemical reaction species
     !====================================================================================
     !--- Number of O in reactant 1 ------------------------------------------------------
     IF (ael(ir1,1) == ael(ir1+1,1)   .AND. ael(ir1,2) == ael(ir1+1,2)   .AND. &
         ael(ir1,4) == ael(ir1+1,4)+1 .AND. ael(ir1,9) == ael(ir1+1,9)-1) THEN
        nor1 = ael(ir1,4)
        is1 = 0
     ELSE IF (ael(ir1,1) == ael(ir1+2,1)   .AND. ael(ir1,2) == ael(ir1+2,2) .AND. &
              ael(ir1,4) == ael(ir1+2,4)+1 .AND. ael(ir1,9) == ael(ir1+2,9)-1) THEN
        nor1 = ael(ir1,4)
        is1 = 1
     ELSE
        nor1 = 0
     ENDIF

     !--- Number of O in product 1 -------------------------------------------------------
     IF (ael(ir3,1) == ael(ir3+1,1)   .AND. ael(ir3,2) == ael(ir3+1,2)    .AND. &
         ael(ir3,4) == ael(ir3+1,4)+1 .AND. ael(ir3,9) == ael(ir3+1,9)-1) THEN
        nop1 = ael(ir3,4)
        is3 = 0
     ELSE IF (ael(ir3,1) == ael(ir3+2,1)   .AND. ael(ir3,2) == ael(ir3+2,2)    .AND. &
              ael(ir3,4) == ael(ir3+2,4)+1 .AND. ael(ir3,9) == ael(ir3+2,9)-1) THEN
        nop1 = ael(ir3,4)
        is3 = 1
     ELSE
        nop1 = 0
     ENDIF

     !--- Number of O in product 2 -------------------------------------------------------
     IF (ir4 /= 0) THEN
        IF (ael(ir4,1) == ael(ir4+1,1)   .AND. ael(ir4,2) == ael(ir4+1,2)    .AND. &
            ael(ir4,4) == ael(ir4+1,4)+1 .AND. ael(ir4,9) == ael(ir4+1,9)-1) THEN
           nop2 = ael(ir4,4)
           is4 = 0
        ELSE IF (ael(ir4,1) == ael(ir4+2,1)   .AND. ael(ir4,2) == ael(ir4+2,2)    .AND. &
                 ael(ir4,4) == ael(ir4+2,4)+1 .AND. ael(ir4,9) == ael(ir4+2,9)-1) THEN
           nop2 = ael(ir4,4)
           is4 = 1
        ELSE
           nop2 = 0
        ENDIF
     ELSE
        nop2 = 0
     ENDIF

     !--- Number of O in product 3 ------------------------------------------------------
     IF (ir5 /= 0) THEN
        IF (ael(ir5,1) == ael(ir5+1,1)   .AND. ael(ir5,2) == ael(ir5+1,2)    .AND. &
            ael(ir5,4) == ael(ir5+1,4)+1 .AND. ael(ir5,9) == ael(ir5+1,9)-1) THEN
           nop3 = ael(ir5,4)
           is5 = 0
        ELSE IF (ael(ir5,1) == ael(ir5+2,1)   .AND. ael(ir5,2) == ael(ir5+2,2)    .AND. &
                 ael(ir5,4) == ael(ir5+2,4)+1 .AND. ael(ir5,9) == ael(ir5+2,9)-1) THEN
           nop3 = ael(ir5,4)
           is5 = 1
        ELSE
          nop3 = 0
        ENDIF
     ELSE
        nop3 = 0
     ENDIF

     !--- Number of O in product 4 ------------------------------------------------------
     IF (ir6 /= 0) THEN
        IF (ael(ir6,1) == ael(ir6+1,1)   .AND. ael(ir6,2) == ael(ir6+1,2)    .AND. &
            ael(ir6,4) == ael(ir6+1,4)+1 .AND. ael(ir6,9) == ael(ir6+1,9)-1) THEN
           nop3 = ael(ir6,4)
           is6 = 0
        ELSE IF (ael(ir6,1) == ael(ir6+2,1)   .AND. ael(ir6,2) == ael(ir6+2,2)    .AND. &
                 ael(ir6,4) == ael(ir6+2,4)+1 .AND. ael(ir6,9) == ael(ir6+2,9)-1) THEN
           nop4 = ael(ir6,4)
           is6 = 1
        ELSE
           nop4 = 0
        ENDIF
     ELSE
        nop4 = 0
     ENDIF

     !--- Special: second product is photon or cosmic ------------------------------------
     IF (ir4 == n_var+1 .OR.  ir4 == n_var+2 .OR. ir4 == n_var+3) THEN
        nop2 = 0
     ENDIF

     !====================================================================================
     ! Create tables indicating the positions of the 2 potential O in products
     !====================================================================================
      ! First oxygen:
      prod_isot1(:) = 0
      IF (nop1 > 0) THEN
         prod_isot1(1) = 1
      ELSE IF(nop2 > 0) THEN
         prod_isot1(2) = 1
      ELSE IF(nop3 > 0) THEN
         prod_isot1(3) = 1
      ELSE IF (nop4 > 0) THEN
         prod_isot1(4) = 1
      ENDIF
      ! Second oxygen:
      prod_isot2(:) = 0
      IF(nop2 > 0       .AND. prod_isot1(2) == 0) THEN
         prod_isot2(2) = 1
      ELSE IF(nop3 > 0  .AND. prod_isot1(3) == 0) THEN
         prod_isot2(3) = 1
      ELSE IF (nop4 > 0 .AND. prod_isot1(4) == 0) THEN
         prod_isot2(4) = 1
      ENDIF

     !===================================================================================
     ! 1 - FIRST CASE: O is present in two reactants
     !     We considerer 3 possibilites:
     !     1.1: O is present in 2 products => create four reactions
     !     1.2: O is present in 1 product  => create two reactions
     !     1.3: The reaction cannot be isotoped => do not create a reaction
     !     => no branching ratios
     !===================================================================================
     IF (nor1 > 0 .AND. nor2 > 0) THEN
        !--------------------------------------------------------------------------------
        ! 1.1: O is present in 2 products
        !      OONVENTION: the reaction must be written in the order:
        !      AO + BO -> A'O + B'O + D'  + E'
        !      AO + BO -> A'  + B'O + D'O + E
        !      AO + BO -> A'O + B'  + D'O + E
        !      AO + BO -> A'  + B'  + D'O + E'O
        !      so that: =>  AO* + BO  -> A'O* + B'O  + D' + E'
        !                             -> A'O  + B'O* + D' + E'
        !          and: =>  AO  + BO* -> A'O* + B'O  + D' + E'
        !                             -> A'O  + B'O* + D' + E'
        !      => creation of 4 reactions
        !--------------------------------------------------------------------------------
        IF (      (nop1 > 0 .AND. nop2 > 0) .OR. (nop1 > 0 .AND. nop3 > 0) &
             .OR. (nop1 > 0 .AND. nop4 > 0)                                &
             .OR. (nop2 > 0 .AND. nop3 > 0) .OR. (nop2 > 0 .AND. nop4 > 0) &
             .OR. (nop3 > 0 .AND. nop4 > 0) &
           ) THEN

           !--- PREMIERE REACTION ISOTOPEE -----------------------------------------------
           !PRINT *,"----------"
           !PRINT *, speci(ir1), speci(ir2), speci(ir3), speci(ir4), speci(ir5), speci(ir6)
           nreact = nreact + 1

           react (nreact,0) = ir0
           react (nreact,1) = ir1 + 1 + is1
           react (nreact,2) = ir2
           react (nreact,3) = ir3c
          
           react (nreact,4) = ir3 + prod_isot1(1) + is3 * prod_isot1(1)
           react (nreact,5) = ir4 + prod_isot1(2) + is4 * prod_isot1(2)
           react (nreact,6) = ir5 + prod_isot1(3) + is5 * prod_isot1(3)
           react (nreact,7) = ir6 + prod_isot1(4) + is6 * prod_isot1(4)

           IF (ir1 /= ir2) gamm(nreact)  = gamm(jj) / 2.0_dp
           IF (ir1 == ir2) gamm(nreact)  = gamm(jj) / 4.0_dp
           alpha(nreact) = alpha(jj)
           bet(nreact)   = bet(jj)
           itype(nreact) = itype(jj)

           !PRINT *, speci(react(nreact,1)), speci(react(nreact,2)), speci(react(nreact,4)) &
           !       , speci(react(nreact,5)), speci(react(nreact,6)), speci(react(nreact,7))

           !--- SECONDE REACTION ISOTOPEE -----------------------------------------------
           nreact = nreact + 1

           react (nreact,0) = ir0
           react (nreact,1) = ir1 + 1 + is1
           react (nreact,2) = ir2
           react (nreact,3) = ir3c
          
           react (nreact,4) = ir3 + prod_isot2(1) + is3 * prod_isot2(1)
           react (nreact,5) = ir4 + prod_isot2(2) + is4 * prod_isot2(2)
           react (nreact,6) = ir5 + prod_isot2(3) + is5 * prod_isot2(3)
           react (nreact,7) = ir6 + prod_isot2(4) + is6 * prod_isot2(4)

           IF (ir1 /= ir2) gamm(nreact)  = gamm(jj) / 2.0_dp
           IF (ir1 == ir2) gamm(nreact)  = gamm(jj) / 4.0_dp
           alpha(nreact) = alpha(jj)
           bet(nreact)   = bet(jj)
           itype(nreact) = itype(jj)

           !PRINT *, speci(react(nreact,1)), speci(react(nreact,2)), speci(react(nreact,4)) &
           !       , speci(react(nreact,5)), speci(react(nreact,6)), speci(react(nreact,7))

           !--- TROISIEME REACTION ISOTOPEE ----------------------------------------------
           nreact = nreact + 1
  
           react (nreact,0) = ir0
           react (nreact,1) = ir1
           react (nreact,2) = ir2 + 1 + is2
           react (nreact,3) = ir3c
          
           react (nreact,4) = ir3 + prod_isot1(1) + is3 * prod_isot1(1)
           react (nreact,5) = ir4 + prod_isot1(2) + is4 * prod_isot1(2)
           react (nreact,6) = ir5 + prod_isot1(3) + is5 * prod_isot1(3)
           react (nreact,7) = ir6 + prod_isot1(4) + is6 * prod_isot1(4)

           IF (ir1 /= ir2) gamm(nreact)  = gamm(jj) / 2.0_dp
           IF (ir1 == ir2) gamm(nreact)  = gamm(jj) / 4.0_dp
           alpha(nreact) = alpha(jj)
           bet(nreact)   = bet(jj)
           itype(nreact) = itype(jj)

           !PRINT *, speci(react(nreact,1)), speci(react(nreact,2)), speci(react(nreact,4)) &
           !       , speci(react(nreact,5)), speci(react(nreact,6)), speci(react(nreact,7))

           !--- QUATRIEME REACTION ISOTOPEE ----------------------------------------------
           nreact = nreact + 1
  
           react (nreact,0) = ir0
           react (nreact,1) = ir1
           react (nreact,2) = ir2 + 1 + is2
           react (nreact,3) = ir3c
          
           react (nreact,4) = ir3 + prod_isot2(1) + is3 * prod_isot2(1)
           react (nreact,5) = ir4 + prod_isot2(2) + is4 * prod_isot2(2)
           react (nreact,6) = ir5 + prod_isot2(3) + is5 * prod_isot2(3)
           react (nreact,7) = ir6 + prod_isot2(4) + is6 * prod_isot2(4)

           IF (ir1 /= ir2) gamm(nreact)  = gamm(jj) / 2.0_dp
           IF (ir1 == ir2) gamm(nreact)  = gamm(jj) / 4.0_dp
           alpha(nreact) = alpha(jj)
           bet(nreact)   = bet(jj)
           itype(nreact) = itype(jj)

           !PRINT *, speci(react(nreact,1)), speci(react(nreact,2)), speci(react(nreact,4)) &
           !       , speci(react(nreact,5)), speci(react(nreact,6)), speci(react(nreact,7))
           !PRINT *, ""

        !--------------------------------------------------------------------------------
        ! 1.2: O is present in 1 product
        !
        !      AO  +  BO  -->  A'O2  +  B'    + D'   + E'
        !  OU  AO  +  BO  -->  A'    +  B'O2  + D'   + E'
        !  OU  AO  +  BO  -->  A'    +  B'    + D'O2 + E'
        !  OU  AO  +  BO  -->  A'    +  B'    + D'   + E'O2
        !
        !                            =====>      AO* + BO  -->  A'O* +  B'    + D'   + E'
        !                                    OU  AO* + BO  -->  A'   +  B'O*  + D'   + E'
        !                                    OU  AO* + BO  -->  A'   +  B'    + D'O* + E'
        !                                    OU  AO* + BO  -->  A'   +  B'    + D'   + E'O*
        !
        !                        ET  =====>      AO  + BO* -->  A'O* +  B'    + D'   + E'
        !                                    OU  AO  + BO* -->  A'   +  B'O*  + D'   + E'
        !                                    OU  AO  + BO* -->  A'   +  B'    + D'O* + E'
        !                                    OU  AO  + BO* -->  A'   +  B'    + D'   + E'O*
        !
        !                       ==> 2 REACTIONS ISOTOPEES
        !--------------------------------------------------------------------------------
        ELSE IF (nop1 > 0 .OR. nop2 > 0 .OR. nop3 >0 .OR. nop4>0 )  THEN
           !PRINT *, speci(ir1), speci(ir2), speci(ir3), speci(ir4), speci(ir5), speci(ir6)

           !--- PREMIERE REACTION ISOTOPEE :
           nreact = nreact + 1

           react (nreact,0) = ir0
           react (nreact,1) = ir1 + 1 + is1
           react (nreact,2) = ir2
           react (nreact,3) = ir3c

           react (nreact,4) = ir3 + prod_isot1(1) + is3 * prod_isot1(1)
           react (nreact,5) = ir4 + prod_isot1(2) + is4 * prod_isot1(2)
           react (nreact,6) = ir5 + prod_isot1(3) + is5 * prod_isot1(3)
           react (nreact,7) = ir6 + prod_isot1(4) + is6 * prod_isot1(4)

           IF (ir1 /= ir2) gamm(nreact)  = gamm(jj)
           IF (ir1 == ir2) gamm(nreact)  = gamm(jj) / 2.0_dp
           alpha(nreact) = alpha(jj)
           bet(nreact)   = bet(jj)
           itype(nreact) = itype(jj)
           !PRINT *, speci(react(nreact,1)), speci(react(nreact,2)), speci(react(nreact,4)) &
           !       , speci(react(nreact,5)), speci(react(nreact,6)), speci(react(nreact,7))

           !--- DEUXIEME REACTION ISOTOPEE :
           nreact = nreact + 1

           react (nreact,0) = ir0
           react (nreact,1) = ir1
           react (nreact,2) = ir2 + 1 + is2
           react (nreact,3) = ir3c

           react (nreact,4) = ir3 + prod_isot1(1) + is3 * prod_isot1(1)
           react (nreact,5) = ir4 + prod_isot1(2) + is4 * prod_isot1(2)
           react (nreact,6) = ir5 + prod_isot1(3) + is5 * prod_isot1(3)
           react (nreact,7) = ir6 + prod_isot1(4) + is6 * prod_isot1(4)
  
           IF (ir1 /= ir2) gamm(nreact)  = gamm(jj)
           IF (ir1 == ir2) gamm(nreact)  = gamm(jj) / 2.0_dp
           alpha(nreact) = alpha(jj)
           bet(nreact)   = bet(jj)
           itype(nreact) = itype(jj)
           !PRINT *, speci(react(nreact,1)), speci(react(nreact,2)), speci(react(nreact,4)) &
           !       , speci(react(nreact,5)), speci(react(nreact,6)), speci(react(nreact,7))
           !PRINT *, ""

        ENDIF
        !--- ON PASSE A l EXAMEN DE LA REACTION SUIVANTE :
        CYCLE
     ENDIF

     !===================================================================================
     ! 2 - SECOND CASE: O is present in one reactant only
     !     We considerer 2 possibilites:
     !     2.1: O is present in 2 products => create two reactions
     !     2.2: O is present in 1 product  => create one reaction
     !     2.3: The reaction cannot be isotoped => do not create a reaction
     !===================================================================================
     IF (nor1 > 0 .OR. nor2 > 0) THEN

        !---------------------------------------------------------------------------------
        ! 2.1: O is present in 2 products
        !      AO  +  B   -->  A'O  +  B'O + D' + E'
        !  OU  A   +  BO  -->  A'O  +  B'O + D' + E'
        !
        !      =====>      AO*  +  B    -->  A'O*  +  B'O + ...
        !              OU  A    +  BO   -->  A'O*  +  B'O + ...
        !
        !  ET  =====>      AO*  +  B    -->  A'O   +  B'O* + ...
        !              OU  A    +  BO   -->  A'O   +  B'O* + ...
        !
        ! ==> 2 REAOTIONS ISOTOPEES with gamma / 2.0_dp
        !---------------------------------------------------------------------------------
        IF (     (nop1 > 0 .AND. nop2 > 0) .OR. (nop1 > 0 .AND. nop3 > 0) &
            .OR. (nop1 > 0 .AND. nop4 > 0)                                &
            .OR. (nop2 > 0 .AND. nop3 > 0) .OR. (nop2 > 0 .AND. nop4 > 0) &
            .OR. (nop3 > 0 .AND. nop4 > 0)                                &
           ) THEN

           !PRINT *,"----------"
           !PRINT *, speci(ir1), speci(ir2), speci(ir3), speci(ir4), speci(ir5), speci(ir6)

           !--- PREMIERE REACTION ISOTOPEE:
           nreact = nreact + 1

           react (nreact,0) = ir0
           IF (nor1 == 0) THEN
              react (nreact,1) = ir1
              react (nreact,2) = ir2 + 1 + is2
           ELSE
              react (nreact,1) = ir1 + 1 + is1
              react (nreact,2) = ir2
           ENDIF
           react (nreact,3) = ir3c

           react (nreact,4) = ir3 + prod_isot1(1) + is3 * prod_isot1(1)
           react (nreact,5) = ir4 + prod_isot1(2) + is4 * prod_isot1(2)
           react (nreact,6) = ir5 + prod_isot1(3) + is5 * prod_isot1(3)
           react (nreact,7) = ir6 + prod_isot1(4) + is6 * prod_isot1(4)

           gamm(nreact)  = gamm(jj)/2.0_dp
           alpha(nreact) = alpha(jj)
           bet(nreact)   = bet(jj)
           itype(nreact) = itype(jj)
           !PRINT *, speci(react(nreact,1)), speci(react(nreact,2)), speci(react(nreact,4)) &
           !       , speci(react(nreact,5)), speci(react(nreact,6)), speci(react(nreact,7))

           !--- SECONDE REACTION ISOTOPEE:
           nreact = nreact + 1

           react (nreact,0) = ir0
           IF (nor1 == 0) THEN
              react (nreact,1) = ir1
              react (nreact,2) = ir2 + 1 + is2
           ELSE
              react (nreact,1) = ir1 + 1 + is1
              react (nreact,2) = ir2
           ENDIF
           react (nreact,3) = ir3c

           react (nreact,4) = ir3 + prod_isot2(1) + is3 * prod_isot2(1)
           react (nreact,5) = ir4 + prod_isot2(2) + is4 * prod_isot2(2)
           react (nreact,6) = ir5 + prod_isot2(3) + is5 * prod_isot2(3)
           react (nreact,7) = ir6 + prod_isot2(4) + is6 * prod_isot2(4)

           gamm(nreact)  = gamm(jj)/2.0_dp
           alpha(nreact) = alpha(jj)
           bet(nreact)   = bet(jj)
           itype(nreact) = itype(jj)
           !PRINT *, speci(react(nreact,1)), speci(react(nreact,2)), speci(react(nreact,4)) &
           !       , speci(react(nreact,5)), speci(react(nreact,6)), speci(react(nreact,7))

        !---------------------------------------------------------------------------------
        ! 2.2: O is present in 1 product
        !      AO  +  B   -->  A'O  +  B'
        !  OU  A   +  BO  -->  A'O  +  B'
        !  OU  AO  +  B   -->  A'   +  B'O
        !  OU  A   +  BO  -->  A'   +  B'O
        !  OU  A   +  BO  -->  A'   +  B'     + O
        !  OU  A   +  BO  -->  A'   +  B'     + D + O
        !
        !                            =====>      AO*  +  B    -->  A'O*  +  B'
        !                                    OU  A    +  BO*  -->  A'O*  +  B'
        !                                    OU  AO*  +  B    -->  A'    +  B'O*
        !                                    OU  A    +  BO*  -->  A'    +  B'O*
        !                                    OU  A    +  BO*  -->  A'    +  B'      + O*
        !                                    OU  A    +  BO*  -->  A'    +  B'      + D + O*
        !                       ==> 1 REAOTION ISOTOPEE
        !---------------------------------------------------------------------------------
        ELSE IF (nop1 > 0 .OR. nop2 > 0 .OR. nop3 > 0 .OR. nop4 > 0) THEN
           !PRINT *,"----------"
           !PRINT *, speci(ir1), speci(ir2), speci(ir3), speci(ir4), speci(ir5), speci(ir6)

           !--- UNE SEULE REACTION ISOTOPEE :
           nreact = nreact + 1

           react (nreact,0) = ir0
           IF (nor1 == 0) THEN
              react (nreact,1) = ir1
              react (nreact,2) = ir2 + 1 + is2
           ELSE
              react (nreact,1) = ir1 + 1 + is1
              react (nreact,2) = ir2
           ENDIF
           react (nreact,3) = ir3c

           react (nreact,4) = ir3 + prod_isot1(1) + is3 * prod_isot1(1)
           react (nreact,5) = ir4 + prod_isot1(2) + is4 * prod_isot1(2)
           react (nreact,6) = ir5 + prod_isot1(3) + is5 * prod_isot1(3)
           react (nreact,7) = ir6 + prod_isot1(4) + is6 * prod_isot1(4)

           gamm(nreact)  = gamm(jj)
           alpha(nreact) = alpha(jj)
           bet(nreact)   = bet(jj)
           itype(nreact) = itype(jj)

           !PRINT *, speci(react(nreact,1)), speci(react(nreact,2)), speci(react(nreact,4)) &
           !       , speci(react(nreact,5)), speci(react(nreact,6)), speci(react(nreact,7))
           !PRINT *, ""

       ENDIF
    ENDIF !--- End of SECOND CASE: O is present in one reactant only
  ENDDO   !--- End of loop on chemical reactions

!----------------------------------------------------------------------------
!                     FIN DE l EXAMEN DES REACTIONS
!          ET DE LA CONSTRUCTION DE LA CHIMIE ISOTOPEE DE l OXYGENE
!----------------------------------------------------------------------------

  PRINT *,"End of construction of O18 chemistry"

END SUBROUTINE ISOTO

END MODULE PXDR_ISOTCO
