! Copyright (C) 2002 - Observatoire de Paris/Meudon
!                      Universite Paris 7
!                      CNRS (France)
!                      Contact : Jacques.Lebourlot@obspm.fr
!
! This program is free software; you can redistribute it and/or
! modify it under the terms of the GNU General Public License version 2
! as published by the Free Software Foundation.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this program; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.


MODULE PXDR_TRANSFER

! +new formulation for the non-embedded case
!  but incuding R' effects   (spring 2005) jlb
! +theory & code revision (june-july 2005) jrg
! +embedded sources preliminary coding (oct 2005) jlb+jrg

   USE PXDR_FLAGS
   USE PXDR_CONSTANTES

   IMPLICIT NONE

   REAL (KIND=dp), PUBLIC                            :: flu_rad0
   REAL (KIND=dp), PUBLIC                            :: flu_rad1
   REAL (KIND=dp), PUBLIC                            :: flu_rad2
   REAL (KIND=dp), PUBLIC                            :: int_rad0
   REAL (KIND=dp), PUBLIC                            :: int_rad1
   REAL (KIND=dp), PUBLIC                            :: int_rad2
   REAL (KIND=dp), PUBLIC                            :: alpha_m
   REAL (KIND=dp), PUBLIC                            :: alpha_p

   REAL (KIND=dp), PUBLIC                             :: G0m
   REAL (KIND=dp), PUBLIC                             :: G0p

   ! wlrinc is used to increase the WL step within a line profile
!  REAL(KIND=dp), PUBLIC, PARAMETER                   :: wlrinc = 1.2_dp
   REAL(KIND=dp), PUBLIC, PARAMETER                   :: wlrinc = 1.5_dp
   ! dopl_scal is used to set the line resolution in unit of Doppler width
!  REAL(KIND=dp), PUBLIC, PARAMETER                   :: dopl_scal = 0.3_dp
   REAL(KIND=dp), PUBLIC, PARAMETER                   :: dopl_scal = 1.0_dp

   PRIVATE

   PUBLIC :: ETIN90_CteG, ETIN90, ETIN901S, CO_UV_EXC

!------------------------------------------------------------------------

CONTAINS

SUBROUTINE ETIN90 (nwlu)

 USE PXDR_FLAGS
 USE PXDR_CONSTANTES
 USE PXDR_CHEM_DATA
 USE PXDR_PROFIL
 USE PXDR_AUXILIAR
 USE PXDR_STR_DATA
 USE PXDR_RF_TOOLS

! Radiative transfer computation. See file transfer_S.pdf for definitions
! and comments. Index conventions:
!       l goes from 0 to lleg (first Legendre Pol. index)
!       m goes from +/- 1 to +/- mleg (second Legendre Pol. index)
!       i goes from +/- 1 to +/- nmui (discrete angles for boundary conditions)
!       n goes from 0 to npo (optical depth grid)
! 2 sides version: tmax finite and radiation are impiging from mu_i > 0 and mu_i < 0

 IMPLICIT NONE

 INTEGER, INTENT (IN)                          :: nwlu     ! Number of wavelength computed

 INTEGER                                       :: l, m, n, n1, n3

 ! Radiative quantities
 INTEGER                                       :: iwlg
 REAL (KIND=dp), DIMENSION(:,:),   ALLOCATABLE :: tl       ! Full optical depth
 REAL (KIND=dp), DIMENSION(:),     ALLOCATABLE :: dtl      ! Full optical depth increment
 REAL (KIND=dp), DIMENSION(:,:),   ALLOCATABLE :: sigl     ! Phase function

 REAL (KIND=dp), DIMENSION(:,:),   ALLOCATABLE :: albed    ! Reduced albedo (including gas)
 REAL (KIND=dp)                                :: ggHG     ! Henyey-Greenstein g = < cos theta > coefficient

 REAL (KIND=dp), DIMENSION(:),     ALLOCATABLE :: kap_gas
 REAL (KIND=dp), DIMENSION(:,:),   ALLOCATABLE :: kap_dus
 REAL (KIND=dp), DIMENSION(:,:),   ALLOCATABLE :: sca_dus
 REAL (KIND=dp), DIMENSION(:),     ALLOCATABLE :: emm_dus
 REAL (KIND=dp), DIMENSION(:),     ALLOCATABLE :: Camp     ! Legendre amplitudes

 type(rf_type)                                 :: rf_tmp, rf_lamI

 ! Legendre polynomial related variables
 REAL (KIND=dp), DIMENSION(:),     ALLOCATABLE :: p_0
 REAL (KIND=dp), DIMENSION(:),     ALLOCATABLE :: q_0

 ! Matrix A coefficients
 REAL (KIND=dp), DIMENSION(:,:),   ALLOCATABLE :: hl
 REAL (KIND=dp), DIMENSION(:),     ALLOCATABLE :: wr, wi
 REAL (KIND=dp), DIMENSION(:,:),   ALLOCATABLE :: xkm      ! Legendre absissa
 REAL (KIND=dp), DIMENSION(:,:),   ALLOCATABLE :: Smk_m    ! integrated xkm
 REAL (KIND=dp), DIMENSION(:,:),   ALLOCATABLE :: Smk_p    ! integrated xkm
 ! Smk_m = Emk_m_mag * grand + s => Exp(-Smk_m) = G**Emk_m_mag * Emk_m_cmp
 !       with G = exp(-grand) and Emk_m_cmp = exp(-s)
 INTEGER (KIND=ilk), DIMENSION(:,:), ALLOCATABLE :: Emk_m_mag! Exp(-Smk_m) 0 -> t - power of G
 REAL (KIND=dp), DIMENSION(:,:),   ALLOCATABLE :: Emk_m_cmp! Exp(-Smk_m) 0 -> t - remaining part
 REAL (KIND=dp), DIMENSION(:,:),   ALLOCATABLE :: Emk_m_val! Exp(-Smk_m) 0 -> t - Total value
 INTEGER (KIND=ilk), DIMENSION(:,:), ALLOCATABLE :: Emk_p_mag! Exp(-Smk_m) 0 -> t - power of G
 REAL (KIND=dp), DIMENSION(:,:),   ALLOCATABLE :: Emk_p_cmp! Exp(-Smk_m) 0 -> t - remaining part
 REAL (KIND=dp), DIMENSION(:,:),   ALLOCATABLE :: Emk_p_val! Exp(-Smk_m) 0 -> t - Total value
 INTEGER (KIND=ilk), DIMENSION(:), ALLOCATABLE :: md_m, md_p
 INTEGER,        DIMENSION(:,:),   ALLOCATABLE :: ndp_m, ndq_m
 INTEGER,        DIMENSION(:,:),   ALLOCATABLE :: ndp_p, ndq_p
 INTEGER (KIND=ilk), DIMENSION(:,:), ALLOCATABLE :: nn_m, nn_p
 REAL (KIND=dp), DIMENSION(:,:),   ALLOCATABLE :: T_m, T_p
 REAL (KIND=dp), DIMENSION(:,:),   ALLOCATABLE :: R_m, R_p
 INTEGER (KIND=ilk)                            :: k_ilk, i_ilk

 REAL (KIND=dp), DIMENSION(:),     ALLOCATABLE :: QmG
 REAL (KIND=dp), DIMENSION(:,:,:), ALLOCATABLE :: rlm      ! Eigenvectors matrix
 REAL (KIND=dp), DIMENSION(:,:,:), ALLOCATABLE :: rlmp     ! derivative
 REAL (KIND=dp), DIMENSION(:,:,:), ALLOCATABLE :: rlmi     ! inverse of Rlm(t)

 ! Matrix product of R^-1 x R'
 REAL (KIND=dp), DIMENSION(:,:,:), ALLOCATABLE :: rlmip
 REAL (KIND=dp), DIMENSION(:,:),   ALLOCATABLE :: Srm

! Source function contributions

 ! Source function expansion coeficients (as function of l and m respectively)
 REAL (KIND=dp), DIMENSION(:,:),   ALLOCATABLE :: sfl, sfm, sfl_g, sfl_d
 REAL (KIND=dp), DIMENSION(:,:),   ALLOCATABLE :: ggm      ! Inhomogeneous contribution (g vector in notes)
 REAL (KIND=dp), DIMENSION(:,:),   ALLOCATABLE :: Git      ! Product R^-1*A^-1*g vector, same DIMENSIONs as Qit

 ! Boundary conditions related variables
 REAL (KIND=dp), DIMENSION(:,:),   ALLOCATABLE :: Tim      ! Auxiliary
 REAL (KIND=dp), DIMENSION(:,:),   ALLOCATABLE :: Wim      ! BC matrix
 REAL (KIND=dp), DIMENSION(:),     ALLOCATABLE :: Iinc     ! Incident radiation field

 ! Iteration related variables
 REAL (KIND=dp), DIMENSION(:,:),   ALLOCATABLE :: Qit
 REAL (KIND=dp), DIMENSION(:,:),   ALLOCATABLE :: Qit_o
 REAL (KIND=dp), DIMENSION(:,:),   ALLOCATABLE :: Dit
 REAL (KIND=dp), DIMENSION(:,:),   ALLOCATABLE :: Yit
 REAL (KIND=dp), DIMENSION(:),     ALLOCATABLE :: Hit

! LAPACK related variables

 ! Eigenvalues only
 CHARACTER (LEN=1)                             :: JOBZ  = 'N'
 REAL (KIND=dp), DIMENSION(:),     ALLOCATABLE :: WORK         ! Real work array
 INTEGER,        DIMENSION(:),     ALLOCATABLE :: IWORK        ! Work array
 REAL (KIND=dp), DIMENSION(:,:),   ALLOCATABLE :: aam, bbm
 REAL (KIND=dp), DIMENSION(:,:),   ALLOCATABLE :: aamf, bbmf
 INTEGER                                       :: info
 CHARACTER (LEN=1)                             :: FACT  = 'E'  ! Equilibrate system
 CHARACTER (LEN=1)                             :: TRANS = 'N'  ! Transpose
 INTEGER                                       :: nrhs
 CHARACTER (LEN=1)                             :: EQUED        ! Kind of equilibration
 REAL (KIND=dp), DIMENSION(:), ALLOCATABLE     :: R, C         ! Equilibration factors
 REAL (KIND=dp), DIMENSION(:), ALLOCATABLE     :: FERR, BERR   ! Errors (estimates)
 REAL (KIND=dp)                                :: RCOND        ! Condition number
 INTEGER, DIMENSION(:), ALLOCATABLE            :: indx

 ! Auxiliaries
 INTEGER                                       :: nl, nnit
 INTEGER                                       :: it, i, i0
 INTEGER                                       :: nb_m, nb_p
 INTEGER (KIND=ilk)                            :: n0, le_n
 INTEGER                                       :: i1, i2
 REAL (KIND=dp)                                :: ht
 REAL (KIND=dp)                                :: taux1, taux2
 REAL (KIND=dp)                                :: taux3, taux4
 REAL (KIND=dp)                                :: wlth, wlthcm
 REAL (KIND=dp)                                :: deriv_d
 REAL (KIND=dp), DIMENSION(:), ALLOCATABLE     :: argu
 REAL (KIND=dp), DIMENSION(-mleg:mleg)         :: auxil
 INTEGER, DIMENSION(:,:), ALLOCATABLE          :: indmi, indmj

 ! Thermalized (black body) emission of dust and gas
 REAL (KIND=dp)                                :: bbgas

 ! Computation of emerging specific intensities
 REAL (KIND=dp), DIMENSION(:), ALLOCATABLE     :: fl_0, fl_max

    REAL(KIND=dp)                              :: ff1, ff2, ff3, ff4, denom
    REAL(KIND=dp), DIMENSION(:,:), ALLOCATABLE :: tmp_Jmp
    INTEGER                                    :: itmp, ntmp, ii

 REAL(KIND=dp)                                 :: lambda1, lambda2
!-----------------------------------------------------------------------------

  ! Number of Legendre coefficient (used for loops that run from 1 to lleg+1)
  nl = lleg + 1

  IF (ALLOCATED(g_absL)) THEN
     DEALLOCATE (g_absL)
  ENDIF

  IF (ALLOCATED(intJm)) THEN
     DEALLOCATE (intJm, intJp)
  ENDIF

  IF (ALLOCATED(I_esc_0)) DEALLOCATE(I_esc_0)
  IF (ALLOCATED(I_esc_max)) DEALLOCATE(I_esc_max)

  ALLOCATE (p_0(0:lleg), q_0(0:lleg))
  ALLOCATE (hl(0:lleg,0:npo))
  ALLOCATE (wr(lleg+1), wi(lleg+1))
  ALLOCATE (xkm(-mleg:mleg,0:npo))
  ALLOCATE (Smk_m(0:npo,mleg))
  ALLOCATE (Smk_p(0:npo,mleg))
  ALLOCATE (Emk_m_mag(0:npo,mleg))
  ALLOCATE (Emk_m_cmp(0:npo,mleg))
  ALLOCATE (Emk_m_val(0:npo,mleg))
  ALLOCATE (Emk_p_mag(0:npo,mleg))
  ALLOCATE (Emk_p_cmp(0:npo,mleg))
  ALLOCATE (Emk_p_val(0:npo,mleg))
  ALLOCATE (T_m(0:npo,mleg))
  ALLOCATE (T_p(0:npo,mleg))

  ALLOCATE (QmG(0:npo))
  ALLOCATE (rlm(0:lleg,-mleg:mleg,0:npo))
  ALLOCATE (rlmp(0:lleg,-mleg:mleg,0:npo))
  ALLOCATE (rlmi(0:lleg,-mleg:mleg,0:npo))
  ALLOCATE (rlmip(0:lleg,-mleg:mleg,0:npo))
  ALLOCATE (sfm(-mleg:mleg,0:npo))
  ALLOCATE (sfl(0:lleg,0:npo))
  ALLOCATE (sfl_g(0:lleg,0:npo),sfl_d(0:lleg,0:npo))
  ALLOCATE (ggm(-mleg:mleg,0:npo))
  ALLOCATE (Srm(mleg,0:npo))
  ALLOCATE (sigl(0:lleg,0:npo))
  ALLOCATE (dtl(0:npo))
  ALLOCATE (tl(0:npo,0:nwlu))
  ALLOCATE (albed(0:npo,0:nwlu))
  ALLOCATE (kap_gas(0:npo))
  ALLOCATE (kap_dus(0:npo,0:nwlu))
  ALLOCATE (sca_dus(0:npo,0:nwlu))
  ALLOCATE (emm_dus(0:npo))
  ALLOCATE (g_absL(0:npo,0:nwlu))
  ALLOCATE (intJm(0:npo,0:nwlu))
  ALLOCATE (intJp(0:npo,0:nwlu))
  ALLOCATE (R(nl), C(nl))
  ALLOCATE (FERR(1), BERR(1))
  ALLOCATE (WORK(4*nl))
  ALLOCATE (IWORK(nl))
  ALLOCATE (aam(nl,nl), bbm(nl,1))
  ALLOCATE (aamf(nl,nl), bbmf(nl,1))
  ALLOCATE (indx(nl))
  ALLOCATE (Tim(-nmui:nmui,-mleg:mleg))
  ALLOCATE (Wim(-nmui:nmui,-mleg:mleg))
  ALLOCATE (Iinc(-nmui:nmui))
  ALLOCATE (Qit(-mleg:mleg,0:npo))
  ALLOCATE (Qit_o(-mleg:mleg,0:npo))
  ALLOCATE (Git(-mleg:mleg,0:npo))
  ALLOCATE (Dit(-mleg:mleg,0:npo))
  ALLOCATE (Yit(-mleg:mleg,0:npo))
  ALLOCATE (Hit(-nmui:nmui))
  ALLOCATE (Camp(-nmui:nmui))
  ALLOCATE (argu(0:npo))
  ALLOCATE (indmi(mleg,0:npo), indmj(mleg,0:npo))
  ALLOCATE (fl_0(0:lleg))
  ALLOCATE (fl_max(0:lleg))
  ALLOCATE (I_esc_0(0:nwlu,-nmui:nmui))
  ALLOCATE (I_esc_max(0:nwlu,-nmui:nmui))

  p_0(:)         = 0.0_dp
  q_0(:)         = 0.0_dp
  sigl(:,:)      = 0.0_dp
  Tim(:,:)       = 0.0_dp
  Wim(:,:)       = 0.0_dp
  Iinc(:)        = 0.0_dp
  xkm(:,:)       = 0.0_dp
  Qit(:,:)       = 0.0_dp
  Qit_o(:,:)     = 0.0_dp
  Yit(0,0:npo)   = 0.0_dp
  I_esc_0(:,:)   = 0.0_dp
  I_esc_max(:,:) = 0.0_dp

! STEP 0 - Initialise radiative quantities
!  (boundary conditions and optical depth grid)

  !--- PREPARE SOME FILES TO STORE QUANTITIES RELATIVE TO RADIATION FIELD ---

  ! uvspec.out is the total absorption through the cloud. multiply by
  !            incident radition field to compute spectrum.

  !--- Emission spectra (specific intensity) at Legendre decomposition angles
  fichier = TRIM(out_dir)//TRIM(modele)//'.Iesc'
  PRINT *, " Escaping intensities in:", TRIM(fichier)
  OPEN (36, file=fichier, status="unknown")
  ! Write comment lines of .Iesc file
  ! --------------------------------------------------------------------------------------
  ! The angles written correspond to "theta" in the document transfer_v6.pdf put in
  ! increasing order from 6.7 to 173.3 degrees (limits of the Legende decomposition
  !  0 < theta <  90 correspond to intensity radiated toward decreasing tau (increasing s)
  ! 90 < theta < 180 correspond to intensity radiated toward increasing tau (decreasing s)
  ! The resulting intensities are computed at the end of the function. BG 09/2013
  ! --------------------------------------------------------------------------------------
  WRITE (36,'("# Emission spectra (specific intensity in erg cm-2 s-1 sr-1 Ang-1)")')
  WRITE (36,'("# provided on observer and back side of the cloud")')
  WRITE (36,'("# at each angle theta = acos(mu) of the Legendre decomposition in degrees (see transfer_v6.pdf)")')
  WRITE (36,'("# On comment line below, ob means observer side")')
  WRITE (36,'("#                        ba means back side")')
  WRITE (36,'("#                        numeric values are angles in degrees")')
  WRITE (36,'("# Lambda[A]       ")',ADVANCE="NO")
  DO i = 1, nmui
     WRITE (36,'("I_ob:",F7.3,3x)',ADVANCE="NO") AngleLeg_deg(i)
  ENDDO
  DO i = nmui, 1, -1
     WRITE (36,'("I_ob:",F7.3,3x)',ADVANCE="NO") 180.0_dp - AngleLeg_deg(i)
  ENDDO
  DO i = 1, nmui
     WRITE (36,'("I_ba:",F7.3,3x)',ADVANCE="NO") AngleLeg_deg(i)
  ENDDO
  DO i = nmui, 1, -1
     WRITE (36,'("I_ba:",F7.3,3x)',ADVANCE="NO") 180.0_dp - AngleLeg_deg(i)
  ENDDO
  WRITE (36,*)

  !--- Extinction ------------------------------------
  fichier = TRIM(out_dir)//TRIM(modele)//'.uv'
  PRINT *, " Absorption in:", TRIM(fichier)
  OPEN (37, file=fichier, status="unknown")
  WRITE (37,'("#  W-L   EXP(-tau)  tau")')

  !--- Incident radiation field ----------------------
  fichier = TRIM(out_dir)//TRIM(modele)//'.flin'
  PRINT *, " External field:", TRIM(fichier)
  OPEN (60, file=fichier, status="unknown")
  WRITE(60,*) "9      # Number of columns"

  ! Interpolate intJmp on new grid (tau - lambda) to use in ABSGAZ before new radiation field is known
  ! Moved from ADD_WL
  IF (SIZE(intJmp,dim=1) /= npo .OR. SIZE(intJmp,dim=2) /= nwlg) THEN
     ALLOCATE (tmp_Jmp(0:npo,0:nwlg))
     ntmp = 0
     ! JLB - 13 May 2020
     tmp_Jmp(0,0) = intJmp(0,0)
     itmp = 0
     DO ii = 0, nwlu
        CALL LOC_IND (itmp, 0, SIZE(tmp_wl)-1, rf_wl%Val(ii), tmp_wl)
        denom = tmp_wl(itmp+1) - tmp_wl(itmp)
        ff1   = (tmp_wl(itmp+1) - rf_wl%Val(ii)) / denom
        ff3   = (rf_wl%Val(ii)  - tmp_wl(itmp)) / denom
        tmp_Jmp(0,ii) = intJmp(0,itmp) * ff1 + intJmp(0,itmp+1) * ff3
     ENDDO

     DO n = 1, npo
        CALL LOC_IND (ntmp, 0, SIZE(tmptau)-1, tau_o(n), tmptau)
        itmp = 0
        DO ii = 0, nwlu
           CALL LOC_IND (itmp, 0, SIZE(tmp_wl)-1, rf_wl%Val(ii), tmp_wl)
           denom = (tmptau(ntmp+1) - tmptau(ntmp)) * (tmp_wl(itmp+1) - tmp_wl(itmp))
           ff1   = (tmptau(ntmp+1) - tau_o(n)) * (tmp_wl(itmp+1) - rf_wl%Val(ii)) / denom
           ff2   = (tau_o(n)   - tmptau(ntmp)) * (tmp_wl(itmp+1) - rf_wl%Val(ii)) / denom
           ff3   = (tmptau(ntmp+1) - tau_o(n)) * (rf_wl%Val(ii)  - tmp_wl(itmp)) / denom
           ff4   = (tau_o(n)   - tmptau(ntmp)) * (rf_wl%Val(ii)  - tmp_wl(itmp)) / denom
           tmp_Jmp(n,ii) = intJmp(ntmp,itmp) * ff1 + intJmp(ntmp+1,itmp) * ff2 &
                         + intJmp(ntmp,itmp+1) * ff3 + intJmp(ntmp+1,itmp+1) * ff4
        ENDDO
     ENDDO
     DEALLOCATE (intJmp)
     ALLOCATE (intJmp(0:npo,0:nwlg))
     intJmp = tmp_Jmp
     DEALLOCATE (tmp_Jmp)
  ENDIF

  ! Compute absorption by gas
  CALL ABSGAZ (nwlu)
  IF (ALLOCATED(intJmp)) DEALLOCATE(intJmp)
  ALLOCATE (intJmp(0:npo,0:nwlg))

! STEP 1 - Compute P_l(mu_i) - see recurence in Num Rec
  WRITE(iscrn,'("       Point             Angstroms       erg cm-2 s-1 A-1 sr-1")')
  ! Example on screen : iwlg =   100 wlth =        925.259  intJmp: edge =   4.5370E-08 middle =  9.0686E-26

  p_0(0) = 1.0_dp
  q_0(1) = p_0(0) / 2.0_dp
  DO l = 1, (lleg-1)/2
     p_0(2*l) = - (2.0_dp*l-1) * p_0(2*l-2) / (2.0_dp*l)
     q_0(2*l+1) = p_0(2*l) / (2.0_dp*l+2.0_dp)
  ENDDO

  ! All wavelengths are independent
  DO iwlg = 0, nwlu
     ! sigl = gg**l (Formule 48, Henyey-Greenstein phase function)
     DO n = 0, npo
        IF (F_DU_PROP == 0) THEN
           n1 = 0
        ELSE
           n1 = n
        ENDIF
        ggHG = du_prop%gcos(iwlg,n1)
        !ggHG = 1.0_dp - set g = 1 for tests (pure forward scattering)
        sigl(0,n) = 1.0_dp
        DO l = 1, lleg
           sigl(l,n) = sigl(l-1,n) * ggHG
        ENDDO
     ENDDO

     ! Reset to 0 most variables
     hl = 0.0_dp
     wr = 0.0_dp
     wi = 0.0_dp
     tl(:,iwlg) = 0.0_dp
     dtl = 0.0_dp
     albed(:,iwlg)   = 0.0_dp
     kap_dus(:,iwlg) = 0.0_dp
     sca_dus(:,iwlg) = 0.0_dp
     kap_gas(:)      = 0.0_dp
     emm_dus(:)      = 0.0_dp

     wlth = rf_wl%Val(iwlg)
     wlthcm = wlth * 1.0e-8_dp

     ! Compute local values of absorption coef for dust and gas and dust scatering coef
     ! Compute effective albedo !
     DO n = 0, npo
        IF (F_DU_PROP == 0) THEN
           n1 = 0
        ELSE
           n1 = n
        ENDIF
        ! JLB - 30 XI 2016
        !IF (ifaf > 3) THEN
        IF (ifaf > 1) THEN  ! JLB - 1 XII 2018
           kap_gas(n) = g_absL(n,iwlg)
        ELSE
           kap_gas(n) = 0.0_dp
        ENDIF
        kap_dus(n,iwlg) = du_prop%abso(iwlg,n1) * dens_o(n)
        sca_dus(n,iwlg) = du_prop%scat(iwlg,n1) * dens_o(n)
        emm_dus(n)      = du_prop%emis(iwlg,n)  * dens_o(n)
        albed(n,iwlg)   = sca_dus(n,iwlg) / (kap_dus(n,iwlg) + kap_gas(n) + sca_dus(n,iwlg))
     ENDDO

     ! b - Compute optical depth grid
     !     Convert Delta Tau_v (dust only at V) to
     !     Delta tau_lambda (dust + gaz at Lambda)
     !     17 Oct 2011: use dd_cm (ds) for integration
     tl(0,iwlg) = 0.0_dp
     DO n = 0, npo-1
        ht = ((kap_dus(n,iwlg)   + kap_gas(n)   + sca_dus(n,iwlg)) &
           +  (kap_dus(n+1,iwlg) + kap_gas(n+1) + sca_dus(n+1,iwlg))) &
           * 0.5_dp * dd_cm(n+1)
        dtl(n) = ht
        tl(n+1,iwlg) = tl(n,iwlg) + ht
     ENDDO
     ! Compute total attenuation by dust + gas through whole cloud
     WRITE (37,"(1p,3e17.7E3)") wlth, EXP(-tl(npo,iwlg)), tl(npo,iwlg)

! STEP 2a - Compute matrix A(tau)

     DO n = 0, npo
        ! Formulae A2 and A3 of Roberge:
        hl(0,n) = 1.0_dp - albed(n,iwlg)
        DO l = 1, lleg
           hl(l,n) = (2.0_dp*l+1.0_dp) * (1.0_dp - albed(n,iwlg) * sigl(l,n))
        ENDDO
        ! Appendix C of Goicoechea & Le Bourlot (2007)
        DO l = 1, lleg
           wi(l) = DBLE(l) / SQRT(hl(l-1,n)*hl(l,n))
        ENDDO
        wr = 0.0_dp

        ! Compute eigenvalues (tridiagonal symetric matrix) - LAPACK
        CALL DSTEVD(JOBZ, nl, wr, wi, aam, nl, WORK, 4*nl, IWORK, nl, info )

        ! order eigenvalues by decreasing order
        wi = wr
        DO l = 1, nl
           wr(l) = wi(nl+1-l)
        ENDDO

        ! km = 1 / lamda_m in increasing order
        DO m = 1, mleg
           xkm(m,n) = 1.0_dp / wr(m)
           xkm(-m,n) = -xkm(m,n)
        ENDDO
     ENDDO   ! end loop in npo

     ! Compute Rlm: Formule A9 from Roberge 1983
     ! Eigenvectors computed only from recurrence relations.
     DO n = 0, npo
        DO m = 1, mleg
           rlm(0,m,n)  = 1.0_dp
           rlm(0,-m,n) = 1.0_dp
           rlm(1,m,n)  = (1.0_dp - albed(n,iwlg)) / xkm(m,n)
           rlm(1,-m,n) = -rlm(1,m,n)                             !jrg
           DO l = 2, lleg
              rlm(l,m,n) = (hl(l-1,n) * rlm(l-1,m,n) &
                         - (l-1) * xkm(m,n) * rlm(l-2,m,n)) / (l*xkm(m,n))
              rlm(l,-m,n) = xmunl(l) * rlm(l,m,n)
           ENDDO
        ENDDO
     ENDDO   ! end loop in npo

! STEP 2c - Compute Srm - used for mean intensity

     CALL AUX_Srm(q_0, rlm, Srm)

! STEP 3 - Compute Smk_m = Sum (xkm) (left to right) and Smk_p (right to left)
!        - Simple trapezoidal rule

     Smk_m(0,:) = 0.0_dp
     Smk_p(npo,:) = 0.0_dp
     DO m = 1, mleg
        DO n = 1, npo
           Smk_m(n,m) = Smk_m(n-1,m) + 0.5_dp * dtl(n-1) * (xkm(m,n) + xkm(m,n-1))
        ENDDO
        DO n = npo-1, 0, -1
           Smk_p(n,m) = Smk_p(n+1,m) + 0.5_dp * dtl(n) * (xkm(m,n) + xkm(m,n+1))
        ENDDO
     ENDDO

     DO m = 1, mleg
        Emk_m_mag(:,m) = INT(MIN(Smk_m(:,m),1.0e18_dp) / grand, ilk)
        Emk_m_cmp(:,m) = EXP(-(Smk_m(:,m)-grand*Emk_m_mag(:,m)))
        Emk_m_val(:,m) = expgr(MIN(mgr,Emk_m_mag(:,m))) * Emk_m_cmp(:,m)
        Emk_p_mag(:,m) = INT(MIN(Smk_p(:,m),1.0e18_dp) / grand, ilk)
        Emk_p_cmp(:,m) = EXP(-(Smk_p(:,m)-grand*Emk_p_mag(:,m)))
        Emk_p_val(:,m) = expgr(MIN(mgr,Emk_p_mag(:,m))) * Emk_p_cmp(:,m)
     ENDDO
     ALLOCATE (md_m(mleg))
     ALLOCATE (md_p(mleg))
     md_m(:) = Emk_m_mag(npo,:)
     md_p(:) = Emk_p_mag(0,:)
     ALLOCATE (ndp_m(0:npo+1,mleg))
     ALLOCATE (ndq_m(0:npo+1,mleg))
     ALLOCATE (ndp_p(0:npo+1,mleg))
     ALLOCATE (ndq_p(0:npo+1,mleg))
     ALLOCATE (R_m(npo+1,mleg))
     ALLOCATE (R_p(npo+1,mleg))
     ALLOCATE (nn_m(npo+1,mleg))
     ALLOCATE (nn_p(npo+1,mleg))
     ndp_m = 0
     ndq_m = 0
     ndp_p = 0
     ndq_p = 0

     DO m = 1, mleg
        le_n = Emk_m_mag(0,m)
        nb_m = 1
        ndp_m(1,m) = 0
        nn_m(nb_m,m) = le_n
        DO n = 1, npo
           IF (Emk_m_mag(n,m) == le_n) THEN
              CYCLE
           ELSE
              ndq_m(nb_m,m) = n - 1
              nn_m(nb_m,m) = le_n
              le_n = Emk_m_mag(n,m)
              nb_m = nb_m + 1
              ndp_m(nb_m,m) = n
           ENDIF
        ENDDO
        ndq_m(nb_m,m) = npo
        nn_m(nb_m,m) = le_n
        md_m(m) = nb_m
     ENDDO

     DO m = 1, mleg
        le_n = Emk_p_mag(npo,m)
        nb_p = 1
        ndq_p(1,m) = npo
        nn_p(nb_p,m) = le_n
        DO n = npo-1, 0, -1
           IF (Emk_p_mag(n,m) == le_n) THEN
              CYCLE
           ELSE
              ndp_p(nb_p,m) = n + 1
              nn_p(nb_p,m) = le_n
              le_n = Emk_p_mag(n,m)
              nb_p = nb_p + 1
              ndq_p(nb_p,m) = n
           ENDIF
        ENDDO
        ndp_p(nb_p,m) = 0
        nn_p(nb_p,m) = le_n
        md_p(m) = nb_p
     ENDDO

! STEP 4 - Compute R**-1    !Loop to all cloud optical depth points

     CALL AUX_Rlmi(rlm, hl, rlmi)

! STEP 5 - Compute source function contribution g~ = R^-1*A^-1*g (in notes)
!          JLB : Check Javier correction (not yet done here)

     DO n = 0, npo
        ! Ex. Black body emission of thermalized gas
        IF (wlthcm*tgaz(n) > 1.0e-2_dp) THEN
           bbgas = 2.0_dp*xhp*clum**2 / (wlthcm)**5
           bbgas = bbgas * 1.0e-8_dp / (EXP(hcsurk/(wlthcm*tgaz(n))) -1.0_dp)
        ELSE
           bbgas = 0.0_dp
        ENDIF

! 5a - Assign source function, isotropic  (PROVISIONAL)

        ! METHOD B1: approx. to continuum thermal isotropic emission
        sfl_d(0,n) = emm_dus(n) / (kap_dus(n,iwlg) + kap_gas(n) + sca_dus(n,iwlg))

        ! METHOD B2: approx. to line thermal isotropic emission
        sfl_g(0,n) = bbgas * kap_gas(n) / (kap_dus(n,iwlg) + kap_gas(n) + sca_dus(n,iwlg))

        ! Add both contributions
!       sfl(0,n) = sfl_g(0,n) + sfl_d(0,n)
        sfl(0,n) = sfl_d(0,n)
!       sfl(0,n) = 0.0_dp
        ! JLB - 8 III 2013 - Add a low intensity  background source
        sfl(0,n) = sfl(0,n) + 1.0e-40_dp

        sfl(1:lleg,n) = 0.0_dp
     ENDDO

     ! Put them in m,n desired order
     sfm(-mleg:-1,:) = sfl(0:mleg-1,:)
     sfm(1:mleg,:) = sfl(mleg:lleg,:)

! 5b - Compute vector g

     DO n = 0, npo
        DO m = -mleg, -1
           l = mleg + m
           ggm(m,n) = -sfm(m,n) / ( 1.0_dp - albed(n,iwlg) * sigl(l,n) )
        ENDDO
        DO m = 1, mleg
           l = mleg + m - 1
           ggm(m,n) = -sfm(m,n) / ( 1.0_dp - albed(n,iwlg) * sigl(l,n) )
        ENDDO
     ENDDO

! 5d - Compute matrix product  R^-1*A^-1*g = Git

     CALL AUX_GIT(rlmi, ggm, xkm, Git)

! STEP 6 - Compute derivatives of Rlm

! 6c - derivative of Rlm

     DO n = 2, npo-2
        DO m = 1, mleg
           rlmp(0,m,n) = 0.0_dp
           rlmp(0,-m,n) = 0.0_dp
           n1 = n - 2
           n3 = n + 2
           DO l = 1, lleg
              ! Direct variable
              deriv_d = -rlm(l,m,n1) * (dtl(n)+dtl(n+1)) / ((dtl(n-2)+dtl(n-1)) * (dtl(n-2)+dtl(n-1)+dtl(n)+dtl(n+1))) &
                      - rlm(l,m,n)  * (dtl(n-2)+dtl(n-1)-dtl(n)-dtl(n+1)) / ((dtl(n)+dtl(n+1)) * (dtl(n-2)+dtl(n-1))) &
                      + rlm(l,m,n3) * (dtl(n-2)+dtl(n-1)) / ((dtl(n-2)+dtl(n-1)+dtl(n)+dtl(n+1)) * (dtl(n)+dtl(n+1)))
              rlmp(l,m,n) = deriv_d
              rlmp(l,-m,n) = xmunl(l) * rlmp(l,m,n)
           ENDDO
        ENDDO
     ENDDO

     ! Boundary condition. Replace par linear extrapolation (TODO)
     DO m = 1, mleg
        rlmp(0,m,0) = 0.0_dp
        rlmp(0,-m,0) = 0.0_dp
        rlmp(0,m,1) = 0.0_dp
        rlmp(0,-m,1) = 0.0_dp
        rlmp(0,m,npo-1) = 0.0_dp
        rlmp(0,-m,npo-1) = 0.0_dp
        rlmp(0,m,npo) = 0.0_dp
        rlmp(0,-m,npo) = 0.0_dp
        DO l = 1, lleg
           rlmp(l,m,0) = rlmp(l,m,2)
           rlmp(l,-m,0) = xmunl(l) * rlmp(l,m,0)
           rlmp(l,m,1) = rlmp(l,m,2)
           rlmp(l,-m,1) = xmunl(l) * rlmp(l,m,1)
           rlmp(l,m,npo-1) = rlmp(l,m,npo-2)
           rlmp(l,-m,npo-1) = xmunl(l) * rlmp(l,m,npo-1)
           rlmp(l,m,npo) = rlmp(l,m,npo-2)
           rlmp(l,-m,npo) = xmunl(l) * rlmp(l,m,npo-2)
        ENDDO
     ENDDO

! 6d - Multiply by R**-1

     CALL AUX_Rlmip(rlmi, rlmp, rlmip)

! STEP 7 - Compute Tim

     DO m = 1, mleg
        DO i = 1, nmui
           taux1 = 0.0_dp
           taux2 = 0.0_dp
           taux3 = 0.0_dp
           taux4 = 0.0_dp
           DO l = 0, lleg
              taux1 = taux1 + (2.0_dp * l + 1.0_dp) * rlm(l,-m,0)   * pli(l,-i)
              taux2 = taux2 + (2.0_dp * l + 1.0_dp) * rlm(l,m,0)    * pli(l,-i)
              taux3 = taux3 + (2.0_dp * l + 1.0_dp) * rlm(l,-m,npo) * pli(l,i)
              taux4 = taux4 + (2.0_dp * l + 1.0_dp) * rlm(l,m,npo)  * pli(l,i)
           ENDDO
           Tim(-i,-m) = taux1
           Tim(-i,m)  = taux2
           Tim(i,-m)  = taux3
           Tim(i,m)   = taux4
        ENDDO
     ENDDO

! STEP 8 - Compute Wim (Boundary conditions matrix)

     DO m = 1, mleg
        DO i = 1, nmui
           Wim(-i,-m) = Tim(-i,-m)
           Wim(-i,m)  = Tim(-i,m) * Emk_p_val(0,m)
           Wim(i,-m)  = Tim(i,-m) * Emk_m_val(npo,m)
           Wim(i,m)   = Tim(i,m)
        ENDDO
     ENDDO

! STEP 9 - Set boundary conditions

     ! Add potential // source (star)
     WRITE (60, '(1p,9e15.6E3)') wlth, I_incr_m(iwlg), rf_incIm%Val(iwlg) &
                               , I_incr_p(iwlg), rf_incIp%Val(iwlg) &
                               , sfl_d(0,0), sfl_d(0,npo), sfl_g(0,0), sfl_g(0,npo)

     ! 19 XII 2005 - Use "REAL" units (specific intensity)
     Iinc(-nmui:-1) = rf_incIm%Val(iwlg)
     Iinc(1:nmui)   = rf_incIp%Val(iwlg)

     ! Add potential // source (star)
     ! JLB - 4 IX 08 : Checked after outgoing intensities have been computed
     !                 xmui_M(1) = 0.99... so theta = 0 is on +1 and theta = pi on -1
     ! Dont forget to update also subroutine AUX_YIT1 where Iinc(1) is used
     Iinc(-1) = rf_incIm%Val(iwlg) + I_incr_m(iwlg)
     Iinc(1)  = rf_incIp%Val(iwlg) + I_incr_p(iwlg)

! STEP 10 - Start iteration with an appropriate guess for Y and Q

! 10a - Compute Y_o = R^-1 x F_approx
!       F_approx ~ sfm / 1 - albed + Iexp(kt), THEN it IS wavelength and tau dependent

     CALL AUX_YIT1(Iinc, xkm, tl(:,iwlg), albed(:,iwlg), rlmi, sfm, Yit)

! 10b - Compute Q_o = R^-1 x R' x Y_o

     CALL AUX_QIT(rlmip, Yit, Qit)

!    nnit = 200
!    nnit = 20
     nnit = MAX(10, ifaf)

     !Git = 0.0_dp !old procedure if Git = 0 and Qit = 0
     !Qit = 0.0_dp !tests javi

     DO it = 1, nnit ! begin loop :: 101

! STEP 11 - Compute D  - Simple trapezoidal rule

        DO m = 1, mleg
           QmG(0:npo) = Qit(-m,0:npo) - Git(-m,0:npo)
           DO k_ilk = 1, md_m(m)
              n = ndp_m(k_ilk,m) + 1
              IF (n > npo) CYCLE
              T_m(n,m) = (QmG(n-1) / Emk_m_cmp(n-1,m) &
                       + QmG(n) / Emk_m_cmp(n,m)) * dtl(n-1)
              DO n = ndp_m(k_ilk,m)+2, ndq_m(k_ilk,m)
                 T_m(n,m) = T_m(n-1,m) + (QmG(n-1) / Emk_m_cmp(n-1,m) &
                                       + QmG(n) / Emk_m_cmp(n,m)) * dtl(n-1)
              ENDDO
              i0 = ndq_m(k_ilk,m)+1
              IF (i0 > npo) THEN
                 R_m(k_ilk,m) = 0.0_dp
              ELSE
                 T_m(i0,m) = T_m(i0-1,m) + (QmG(i0-1) / Emk_m_cmp(i0-1,m)) * dtl(i0-1)
                 R_m(k_ilk,m) = (QmG(i0) / Emk_m_cmp(i0,m)) * dtl(i0-1)
              ENDIF
           ENDDO
        ENDDO
        T_m(:,:) = T_m(:,:) * 0.5_dp
        R_m(:,:) = R_m(:,:) * 0.5_dp

        Dit(-mleg:-1,0) = 0.0_dp
        DO m = -mleg, -1
           DO k_ilk = 1, md_m(-m)
              taux1 = 0.0_dp
              DO i_ilk = 1, k_ilk-1
                 n0 = nn_m(k_ilk,-m) - nn_m(i_ilk,-m)
                 taux1 = taux1 + T_m(ndq_m(i_ilk,-m)+1,-m) * expgr(MIN(mgr,n0))
                 n0 = nn_m(k_ilk,-m) - nn_m(i_ilk+1,-m)
                 taux1 = taux1 + R_m(i_ilk,-m) * expgr(MIN(mgr,n0))
              ENDDO
              n = ndp_m(k_ilk,-m)
              Dit(m,n) = taux1 * Emk_m_cmp(n,-m)
              DO n = ndp_m(k_ilk,-m)+1, ndq_m(k_ilk,-m)
                 Dit(m,n) = (taux1 + T_m(n,-m)) * Emk_m_cmp(n,-m)
              ENDDO
           ENDDO
        ENDDO

        DO m = 1, mleg
           QmG(0:npo) = Qit(m,0:npo) - Git(m,0:npo)
           DO k_ilk = 1, md_p(m)
              n = ndq_p(k_ilk,m) - 1
              IF (n < 0) CYCLE
              T_p(n,m) = (QmG(n+1) / Emk_p_cmp(n+1,m) &
                       + QmG(n) / Emk_p_cmp(n,m)) * dtl(n)
              DO n = ndq_p(k_ilk,m)-2, ndp_p(k_ilk,m), -1
                 T_p(n,m) = T_p(n+1,m) + (QmG(n+1) / Emk_p_cmp(n+1,m) &
                                       + QmG(n) / Emk_p_cmp(n,m)) * dtl(n)
              ENDDO
              i0 = ndp_p(k_ilk,m) - 1
              IF (i0 == -1) THEN
                 R_p(k_ilk,m) = 0.0_dp
              ELSE
                 T_p(i0,m) = T_p(i0+1,m) + (QmG(i0+1) / Emk_p_cmp(i0+1,m)) * dtl(i0)
                 R_p(k_ilk,m) = (QmG(i0) / Emk_p_cmp(i0,m)) * dtl(i0)
              ENDIF
           ENDDO
        ENDDO
        T_p(:,:) = T_p(:,:) * 0.5_dp
        R_p(:,:) = R_p(:,:) * 0.5_dp

        Dit(1:mleg,npo) = 0.0_dp
        DO m = 1, mleg
           DO k_ilk = 1, md_p(m)
              taux1 = 0.0_dp
              DO i_ilk = 1, k_ilk-1
                 n0 = nn_p(k_ilk,m) - nn_p(i_ilk,m)
                 taux1 = taux1 + T_p(ndp_p(i_ilk,m)-1,m) * expgr(MIN(mgr,n0))
                 n0 = nn_p(k_ilk,m) - nn_p(i_ilk+1,m)
                 taux1 = taux1 + R_p(i_ilk,m) * expgr(MIN(mgr,n0))
              ENDDO
              n = ndq_p(k_ilk,m)
              Dit(m,n) = taux1 * Emk_p_cmp(n,m)
              DO n = ndq_p(k_ilk,m)-1, ndp_p(k_ilk,m), -1
                 Dit(m,n) = (taux1 + T_p(n,m)) * Emk_p_cmp(n,m)
              ENDDO
           ENDDO
        ENDDO

! STEP 12 - Compute Hit

        Hit = 0.0_dp
        DO i = 1, nmui
           taux1 = 0.0_dp
           taux2 = 0.0_dp
           DO m = 1, mleg
              taux1 = taux1 + Dit(-m,npo) * Tim(i,-m)
              taux2 = taux2 + Dit(m,0) * Tim(-i,m)
           ENDDO
           Hit(-i) = Iinc(-i) - taux2
           Hit(i) = Iinc(i) + taux1
        ENDDO

! STEP 13 - Boundary conditions - solve using LAPACK

        aam(1:nmui,1:mleg)        = Wim(-nmui:-1,-mleg:-1)
        aam(1:nmui,mleg+1:2*mleg) = Wim(-nmui:-1,1:mleg)
        aam(nmui+1:nl,1:mleg)     = Wim(1:nmui,-mleg:-1)
        aam(nmui+1:nl,mleg+1:nl)  = Wim(1:nmui,1:mleg)

        bbm(1:nmui,1)    = Hit(-nmui:-1)
        bbm(nmui+1:nl,1) = Hit(1:nmui)

        nrhs = 1
        CALL DGESVX (FACT, TRANS, nl, nrhs, aam, nl, aamf, nl, indx, EQUED, &
             R, C, bbm, nl, bbmf, nl, RCOND, FERR, BERR, WORK, IWORK, info)

        Camp(-nmui:-1) = bbmf(1:nmui,1)
        Camp(1:nmui) = bbmf(nmui+1:nl,1)

! STEP 14 - Compute Y

        DO n = 0, npo
           DO m = 1, mleg
              Yit(-m,n) = Camp(-m) * Emk_m_val(n,m) - Dit(-m,n)
              Yit(m,n)  = Camp(m)  * Emk_p_val(n,m) + Dit(m,n)
           ENDDO
           ! JLB - Only useful within the deepest lines
           if (SUM(Yit(:,n)) < 0.0_dp) then
              Yit(:,n) = 0.0_dp
           endif
        ENDDO

! STEP 15 - Compute new Q

        Qit_o = Qit

        CALL AUX_QIT(rlmip, Yit, Qit)

! STEP 16 - convergence ?

        auxil(0) = 0.0_dp
        taux2 = 0.0_dp
        DO n = 0, npo
           DO m = -mleg, -1
              IF (Qit(m,n) /= 0.0_dp) THEN
                 auxil(m) = ABS((Qit(m,n)-Qit_o(m,n))/Qit(m,n))
              ELSE
                 auxil(m) = 0.0_dp
              ENDIF
           ENDDO
           DO m = 1, mleg
              IF (Qit(m,n) /= 0.0_dp) THEN
                 auxil(m) = ABS((Qit(m,n)-Qit_o(m,n))/Qit(m,n))
              ELSE
                 auxil(m) = 0.0_dp
              ENDIF
           ENDDO
           taux2 = taux2 + SUM(auxil)
        ENDDO

        IF (taux2 < 1.0e-3_dp) THEN
           EXIT      !EXITs sub-iteration loop
        ENDIF

     ENDDO        ! Exits loop to find q  :: 101
     IF (taux2 > 1.0e-3_dp) THEN
        PRINT *, " Pb:", iwlg, wlth, taux2
     ENDIF

! STEP 17 - Compute radiation field - mean intensity
!           intJm is 0.5 SUM(-1->0) of I
!           intJp is 0.5 SUM(0->1) of I
!           intJmp is 0.5 SUM(-1->1) of I
!           Thus intJm + intJp = J in erg cm-2 s-1 A-1 sterad-1
!           one gets u in erg cm-3 A-1 by u = J * 4 pi / c

     DO n = 0, npo
        taux1 = 0.0_dp
        taux2 = 0.0_dp
        taux3 = 0.0_dp
        DO m = 1, mleg
           taux1 = taux1 + Yit(-m,n) * (1.0_dp + Srm(m,n)) &
                         + Yit(m,n) * (1.0_dp - Srm(m,n))
           taux2 = taux2 + Yit(-m,n) * (1.0_dp - Srm(m,n)) &
                         + Yit(m,n) * (1.0_dp + Srm(m,n))
           taux3 = taux3 + Yit(-m,n) +  Yit(m,n)
        ENDDO

        ! 19 XII 2005 - Change normalisation
        intJm(n,iwlg) = 0.5_dp * taux1
        intJp(n,iwlg) = 0.5_dp * taux2
        intJmp(n,iwlg) = taux3
     ENDDO

     IF (MOD(iwlg,100) == 0) THEN
        WRITE(iscrn,'("iwlg = ",I6," wlth[A] = ",F16.3,"  intJmp: edge = ",1p,E12.4," middle =",1p,E12.4)') &
              iwlg, wlth, intJmp(0,iwlg), intJmp(INT(npo/2),iwlg)
     ENDIF

     ! Compute emerging specific intensities
     DO l = 0, lleg
        fl_0(l)   = 0.0_dp
        fl_max(l) = 0.0_dp
        DO m = 1, mleg
           fl_0(l)   = fl_0(l)   + rlm(l,m,0)   * Yit(m,0)   + rlm(l,-m,0)   * Yit(-m,0)
           fl_max(l) = fl_max(l) + rlm(l,m,npo) * Yit(m,npo) + rlm(l,-m,npo) * Yit(-m,npo)
        ENDDO
     ENDDO

     DO i = 1, nmui
        I_esc_0(iwlg,-i)   = 0.0_dp
        I_esc_0(iwlg,i)    = 0.0_dp
        I_esc_max(iwlg,-i) = 0.0_dp
        I_esc_max(iwlg,i)  = 0.0_dp
        DO l = 0, lleg
           I_esc_0(iwlg,-i)   = I_esc_0(iwlg,-i)   + (2.0_dp*l+1.0_dp) * fl_0(l)   * pli(l,-i)
           I_esc_0(iwlg,i)    = I_esc_0(iwlg,i)    + (2.0_dp*l+1.0_dp) * fl_0(l)   * pli(l,i)
           I_esc_max(iwlg,-i) = I_esc_max(iwlg,-i) + (2.0_dp*l+1.0_dp) * fl_max(l) * pli(l,-i)
           I_esc_max(iwlg,i)  = I_esc_max(iwlg,i)  + (2.0_dp*l+1.0_dp) * fl_max(l) * pli(l,i)
        ENDDO
     ENDDO
     WRITE (36,'(1p,163e15.6)') wlth, (I_esc_0(iwlg,i), i=1,nmui), (I_esc_0(iwlg,i), i=-nmui,-1) &
                              , (I_esc_max(iwlg,i), i=1,nmui), (I_esc_max(iwlg,i), i=-nmui,-1)

     DEALLOCATE (md_m)
     DEALLOCATE (md_p)
     DEALLOCATE (ndp_m)
     DEALLOCATE (ndq_m)
     DEALLOCATE (ndp_p)
     DEALLOCATE (ndq_p)
     DEALLOCATE (R_m)
     DEALLOCATE (R_p)
     DEALLOCATE (nn_m)
     DEALLOCATE (nn_p)
  ENDDO    !end wavelenght loop :: 99  !too slow !!

  CLOSE (36)
  CLOSE (37)
  CLOSE (60)

  lambda1 = 3.0e+0_dp
  lambda2 = 1.0e+3_dp
  Icont_3_1000   = INTEG_TAB(I_esc_0(0:nwlu,1),nwlu,lambda1,lambda2)
  lambda1 = 2.5e+1_dp
  lambda2 = 6.0e+1_dp
  Icont_25_60    = INTEG_TAB(I_esc_0(0:nwlu,1),nwlu,lambda1,lambda2)
  lambda1 = 6.0e+1_dp
  lambda2 = 2.0e+2_dp
  Icont_60_200   = INTEG_TAB(I_esc_0(0:nwlu,1),nwlu,lambda1,lambda2)
  lambda1 = 2.0e+2_dp
  lambda2 = 1.0e+3_dp
  Icont_200_1000 = INTEG_TAB(I_esc_0(0:nwlu,1),nwlu,lambda1,lambda2)
  lambda1 = 6.0e+1_dp
  lambda2 = 1.0e+3_dp
  Icont_60_1000  = INTEG_TAB(I_esc_0(0:nwlu,1),nwlu,lambda1,lambda2)

  IF (F_W_RF_ALB == 1) THEN
     ! Write here full radiation field with effective albedo (JLB - 18 Oct 2012)
     fichier = TRIM(out_dir)//TRIM(modele)//'.rf_alb'//TRIM(ADJUSTL(ii2c(ifaf)))
     OPEN(61, file=fichier, status="unknown" , form = 'unformatted')
     WRITE (61) rf_wl%upper-rf_wl%lower+1 ! number of pixels in wavelength grid
     WRITE (61) rf_wl%Val                 ! Wavelength grid in Angstroms
     DO n = 0, npo
        WRITE (61) tau_o(n)*tautav, intJmp(n,0:nwlu) * J_to_u, g_absL(n,:), kap_dus(n,:), sca_dus(n,:)
     ENDDO
     CLOSE (61)
  ENDIF

  ! Free memory
  DEALLOCATE (p_0, q_0)
  DEALLOCATE (hl)
  DEALLOCATE (wr, wi)
  DEALLOCATE (xkm)
  DEALLOCATE (Emk_m_mag)
  DEALLOCATE (Emk_m_cmp)
  DEALLOCATE (Emk_m_val)
  DEALLOCATE (Emk_p_mag)
  DEALLOCATE (Emk_p_cmp)
  DEALLOCATE (Emk_p_val)
  DEALLOCATE (Smk_m, Smk_p)
  DEALLOCATE (QmG)
  DEALLOCATE (rlm, rlmp, rlmi, rlmip)
  DEALLOCATE (sfl, sfm, sfl_g, sfl_d)
  DEALLOCATE (Srm)
  DEALLOCATE (sigl)
  DEALLOCATE (tl, dtl)
  DEALLOCATE (albed)
  DEALLOCATE (kap_gas, kap_dus, sca_dus, emm_dus)
  DEALLOCATE (aam, bbm)
  DEALLOCATE (aamf, bbmf)
  DEALLOCATE (Tim, Wim)
  DEALLOCATE (Iinc)
  DEALLOCATE (ggm)
  DEALLOCATE (Dit)
  DEALLOCATE (Qit, Qit_o, Git, Yit)
  DEALLOCATE (Hit)
  DEALLOCATE (R, C)
  DEALLOCATE (FERR, BERR)
  DEALLOCATE (WORK, IWORK)
  DEALLOCATE (Camp)
  DEALLOCATE (indx)
  DEALLOCATE (argu)
  DEALLOCATE (indmi, indmj)
  DEALLOCATE (fl_0)
  DEALLOCATE (fl_max)

  !====================================================================================
  ! COMPUTE THE RADIATION FIELD INTENSITY AND ENERGY FROM THE RIGHT AND LEFT SIDE
  ! (USE RF_U AS AUXILIARY VARIABLE HERE)
  ! The goal is to obtain mean_rad(1,2),  int_rad(1,2),  rad(m,p),  and flu_rad(1,2)
  ! The goal is to obtain mean_xray(1,2), int_xray(1,2), xray(m,p), and flu_xray(1,2)
  ! note BG 04/2013: is it important to do right first ?
  !====================================================================================
  CALL RF_CREATE(rf_lamI,0,nwlu)
  CALL RF_CREATE(rf_tmp,1,2)

  !--------------------------------------------------------------------------------
  ! Integration of UV radiation field from Lyman up to Habing limit
  ! (wl_lym = 911 Å ; wl_hab = 2400 Å) - BG 04/2013
  !--------------------------------------------------------------------------------
  CALL RF_LOCATE(rf_wl, wl_lym, i1)
  rf_tmp%Val(1) = 1.0_dp
  rf_tmp%Ord(1) = rf_wl%Ord(i1)
  CALL RF_LOCATE(rf_wl, wl_hab, i2)
  rf_tmp%Val(2) = 1.0_dp
  rf_tmp%Ord(2) = rf_wl%Ord(i2)
  ! -------------------
  rf_u%Val(0:nwlu)    = intJp(npo,0:nwlu)
  mean_rad2           = SECT_INT(rf_tmp, rf_u, wl_pointer, wl_hab)
  rf_u%Val(0:nwlu)    = intJm(0  ,0:nwlu)
  mean_rad1           = SECT_INT(rf_tmp, rf_u, wl_pointer, wl_hab)
  ! -------------------
  rf_u%Val(0:nwlu)    = intJmp(npo,0:nwlu) * J_to_u
  int_rad2            = SECT_INT(rf_tmp, rf_u, wl_pointer, wl_hab)
  radp                = int_rad2 / int_rad0
  rf_u%Val(0:nwlu)    = intJmp(0  ,0:nwlu) * J_to_u
  int_rad1            = SECT_INT(rf_tmp, rf_u, wl_pointer, wl_hab)
  radm                = int_rad1 / int_rad0
  int_rad(0)          = SECT_INT(rf_tmp, rf_u, wl_pointer, wl_hab)
  ! -------------------
  rf_lamI%Val(0:nwlu) = intJmp(npo,0:nwlu) * J_to_u* rf_wl%Val(0:nwlu)
  flu_rad2            = SECT_INT(rf_tmp, rf_lamI, wl_pointer, wl_hab) / (1.0e8_dp * xhp)
  rf_lamI%Val(0:nwlu) = intJmp(0  ,0:nwlu) * J_to_u* rf_wl%Val(0:nwlu)
  flu_rad1            = SECT_INT(rf_tmp, rf_lamI, wl_pointer, wl_hab) / (1.0e8_dp * xhp)

  !--------------------------------------------------------------------------------
  ! Write results of these integrations on the screen and in the .def file
  !--------------------------------------------------------------------------------
  WRITE(iscrn, *)
  WRITE(iscrn, '("* Scaled field at cloud edges")')
  WRITE(iscrn, '("- Obs. side, u = ",1pe12.4," erg cm-3,     ","from",0pf9.2," to",0pf9.2," Ang")') int_rad1,  wl_lym,  wl_hab
  WRITE(iscrn, '("             F = ",1pe12.4," phot cm-2 s-1,","from",0pf9.2," to",0pf9.2," Ang")') flu_rad1,  wl_lym,  wl_hab
  WRITE(iscrn, '("- Back side, u = ",1pe12.4," erg cm-3,     ","from",0pf9.2," to",0pf9.2," Ang")') int_rad2,  wl_lym,  wl_hab
  WRITE(iscrn, '("             F = ",1pe12.4," phot cm-2 s-1,","from",0pf9.2," to",0pf9.2," Ang")') flu_rad2,  wl_lym,  wl_hab

  WRITE(iwrit2,*)
  WRITE(iwrit2,'("-------------------------------------------------------------------------------")')
  WRITE(iwrit2,'("---                           Radiation field                               ---")')
  WRITE(iwrit2,'("-------------------------------------------------------------------------------")')
  WRITE(iwrit2,*)
  WRITE(iwrit2,'("* Scaled radiation field at edges at iteration ifaf = ",I2)') ifaf
  WRITE(iwrit2,'("- Obs. side, u = ",1pe12.4," erg cm-3,     ","from",0pf9.2," to",0pf9.2," Ang")') int_rad1,  wl_lym,  wl_hab
  WRITE(iwrit2,'("             F = ",1pe12.4," phot cm-2 s-1,","from",0pf9.2," to",0pf9.2," Ang")') flu_rad1,  wl_lym,  wl_hab
  WRITE(iwrit2,'("- Back side, u = ",1pe12.4," erg cm-3,     ","from",0pf9.2," to",0pf9.2," Ang")') int_rad2,  wl_lym,  wl_hab
  WRITE(iwrit2,'("             F = ",1pe12.4," phot cm-2 s-1,","from",0pf9.2," to",0pf9.2," Ang")') flu_rad2,  wl_lym,  wl_hab

  !--- FUV scaling factors ---
  WRITE(iscrn, *)
  WRITE(iscrn, '("* FUV scaling factors at cloud edges:")')
  WRITE(iscrn, '("- Obs. side: ",1pe12.4," x ",A65)') radm, TRIM(ISRF_name)
  WRITE(iscrn, '("- Back side: ",1pe12.4," x ",A65)') radp, TRIM(ISRF_name)

  WRITE(iwrit2,*)
  WRITE(iwrit2,'("* FUV scaling factors at cloud edges:")')
  WRITE(iwrit2,'("- Obs. side: ",1pe12.4," x ",A65)') radm, TRIM(ISRF_name)
  WRITE(iwrit2,'("- Back side: ",1pe12.4," x ",A65)') radp, TRIM(ISRF_name)
  WRITE(iwrit2,'("ISRF energy density is u0 = ",1pe13.4, " erg cm-3")') int_rad0

  !--- Conversion to G0 ---
  G0m = int_rad1  / u_habing
  G0p = int_rad2  / u_habing
  WRITE(iscrn, *)
  WRITE(iscrn,'("* FUV scaling factors at cloud edges expressed in G0:")')
  WRITE(iscrn,'("- Obs. side, G0 = ",1pe12.4," Habing units")') G0m
  WRITE(iscrn,'("- Back side, G0 = ",1pe12.4," Habing units")') G0p

  WRITE(iwrit2,*)
  WRITE(iwrit2,'("* FUV scaling factors at cloud edges expressed in G0:")')
  WRITE(iwrit2,'("* Obs. side, G0 = ",1pe12.4," Habing units")') G0m
  WRITE(iwrit2,'("* Back side, G0 = ",1pe12.4," Habing units")') G0p
  WRITE(iwrit2,*)
  WRITE(iwrit2,'("Reminder - above values consider:")')
  WRITE(iwrit2,'("- cloud edges see isotropic ISRF from 2 pi steradians")')
  WRITE(iwrit2,'("- potential cloud transparency to UV")')
  WRITE(iwrit2,'("- dust backscattering")')
  WRITE(iwrit2,'("- additionnal optional radiation field (ex: star)")')
  WRITE(iwrit2,'("Habing energy density used for conversion is: ",1pE12.4," erg cm-3 ")') u_habing
  WRITE(iwrit2,'("-------------------------------------------------------------------------------")')

  CALL RF_FREE (rf_tmp)
  CALL RF_FREE (rf_lamI)

END SUBROUTINE ETIN90

!---------------------------------------------------------------------------

SUBROUTINE ETIN901S (nwlu)

 USE PXDR_FLAGS
 USE PXDR_CONSTANTES
 USE PXDR_CHEM_DATA
 USE PXDR_PROFIL
 USE PXDR_AUXILIAR
 USE PXDR_STR_DATA
 USE PXDR_RF_TOOLS

! Radiative transfer computation. See file transfer_S.pdf for definitions
! and comments. Index conventions:
!       l goes from 0 to lleg (first Legendre Pol. index)
!       m goes from +/- 1 to +/- mleg (second Legendre Pol. index)
!       i goes from +/- 1 to +/- nmui (discrete angles for boundary conditions)
!       n goes from 0 to npo (optical depth grid)
! 1 side version: tmax is infinity and radiation is impiging from mu_i > 0 only

 IMPLICIT NONE

 INTEGER, INTENT (IN)                          :: nwlu     ! Number of wavelength computed

 INTEGER                                       :: l, m, n, n1, n3

 ! Radiative quantities
 INTEGER                                       :: iwlg
 REAL (KIND=dp), DIMENSION(:,:),   ALLOCATABLE :: tl       ! Full optical depth
 REAL (KIND=dp), DIMENSION(:),     ALLOCATABLE :: dtl      ! Full optical depth increment
 REAL (KIND=dp), DIMENSION(:,:),   ALLOCATABLE :: sigl     ! Phase function

 REAL (KIND=dp), DIMENSION(:,:),   ALLOCATABLE :: albed    ! Reduced albedo (including gas)
 REAL (KIND=dp)                                :: ggHG     ! Henyey-Greenstein g = < cos theta > coefficient

 REAL (KIND=dp), DIMENSION(:),     ALLOCATABLE :: kap_gas
 REAL (KIND=dp), DIMENSION(:,:),   ALLOCATABLE :: kap_dus
 REAL (KIND=dp), DIMENSION(:,:),   ALLOCATABLE :: sca_dus
 REAL (KIND=dp), DIMENSION(:),     ALLOCATABLE :: emm_dus
 REAL (KIND=dp), DIMENSION(:),     ALLOCATABLE :: Camp     ! Legendre amplitudes

 type(rf_type)                                 :: rf_tmp, rf_lamI

 ! Legendre polynomial related variables
 REAL (KIND=dp), DIMENSION(:),     ALLOCATABLE :: p_0
 REAL (KIND=dp), DIMENSION(:),     ALLOCATABLE :: q_0

 ! Matrix A coefficients
 REAL (KIND=dp), DIMENSION(:,:),   ALLOCATABLE :: hl
 REAL (KIND=dp), DIMENSION(:),     ALLOCATABLE :: wr, wi
 REAL (KIND=dp), DIMENSION(:,:),   ALLOCATABLE :: xkm      ! Legendre absissa
 REAL (KIND=dp), DIMENSION(:,:),   ALLOCATABLE :: Smk_m    ! integrated xkm
 ! Smk_m = Emk_m_mag * grand + s => Exp(-Smk_m) = G**Emk_m_mag * Emk_m_cmp
 !       with G = exp(-grand) and Emk_m_cmp = exp(-s)
 INTEGER (KIND=ilk), DIMENSION(:,:), ALLOCATABLE :: Emk_m_mag! Exp(-Smk_m) 0 -> t - power of G
 REAL (KIND=dp), DIMENSION(:,:),   ALLOCATABLE :: Emk_m_cmp! Exp(-Smk_m) 0 -> t - remaining part
 REAL (KIND=dp), DIMENSION(:,:),   ALLOCATABLE :: Emk_m_val! Exp(-Smk_m) 0 -> t - Total value
 REAL (KIND=dp), DIMENSION(:,:),   ALLOCATABLE :: Emk_p_val! Exp(-Smk_m) 0 -> t - Total value
 INTEGER (KIND=ilk), DIMENSION(:), ALLOCATABLE :: md_m
 INTEGER,        DIMENSION(:,:),   ALLOCATABLE :: ndp_m, ndq_m
 INTEGER (KIND=ilk), DIMENSION(:,:), ALLOCATABLE :: nn_m
 REAL (KIND=dp), DIMENSION(:,:),   ALLOCATABLE :: T_m
 REAL (KIND=dp), DIMENSION(:,:),   ALLOCATABLE :: R_m
 INTEGER (KIND=ilk)                            :: k_ilk, i_ilk

 REAL (KIND=dp), DIMENSION(:),     ALLOCATABLE :: QmG
 REAL (KIND=dp), DIMENSION(:,:,:), ALLOCATABLE :: rlm      ! Eigenvectors matrix
 REAL (KIND=dp), DIMENSION(:,:,:), ALLOCATABLE :: rlmp     ! derivative
 REAL (KIND=dp), DIMENSION(:,:,:), ALLOCATABLE :: rlmi     ! inverse of Rlm(t)

 ! Matrix product of R^-1 x R'
 REAL (KIND=dp), DIMENSION(:,:,:), ALLOCATABLE :: rlmip
 REAL (KIND=dp), DIMENSION(:,:),   ALLOCATABLE :: Srm

! Source function contributions

 ! Source function expansion coeficients (as function of l and m respectively)
 REAL (KIND=dp), DIMENSION(:,:),   ALLOCATABLE :: sfl, sfm, sfl_g, sfl_d
 REAL (KIND=dp), DIMENSION(:,:),   ALLOCATABLE :: ggm      ! Inhomogeneous contribution (g vector in notes)
 REAL (KIND=dp), DIMENSION(:,:),   ALLOCATABLE :: Git      ! Product R^-1*A^-1*g vector, same DIMENSIONs as Qit

 ! Boundary conditions related variables
 REAL (KIND=dp), DIMENSION(:,:),   ALLOCATABLE :: Tim      ! Auxiliary
 REAL (KIND=dp), DIMENSION(:,:),   ALLOCATABLE :: Wim      ! BC matrix
 REAL (KIND=dp), DIMENSION(:),     ALLOCATABLE :: Iinc     ! Incident radiation field

 ! Iteration related variables
 REAL (KIND=dp), DIMENSION(:,:),   ALLOCATABLE :: Qit
 REAL (KIND=dp), DIMENSION(:,:),   ALLOCATABLE :: Qit_o
 REAL (KIND=dp), DIMENSION(:,:),   ALLOCATABLE :: Dit
 REAL (KIND=dp), DIMENSION(:,:),   ALLOCATABLE :: Yit
 REAL (KIND=dp), DIMENSION(:),     ALLOCATABLE :: Hit

! LAPACK related variables

 ! Eigenvalues only
 CHARACTER (LEN=1)                             :: JOBZ  = 'N'
 REAL (KIND=dp), DIMENSION(:),     ALLOCATABLE :: WORK         ! Real work array
 INTEGER,        DIMENSION(:),     ALLOCATABLE :: IWORK        ! Work array
 REAL (KIND=dp), DIMENSION(:,:),   ALLOCATABLE :: aam, bbm
 REAL (KIND=dp), DIMENSION(:,:),   ALLOCATABLE :: aamf, bbmf
 INTEGER                                       :: info
 CHARACTER (LEN=1)                             :: FACT  = 'E'  ! Equilibrate system
 CHARACTER (LEN=1)                             :: TRANS = 'N'  ! Transpose
 INTEGER                                       :: nrhs
 CHARACTER (LEN=1)                             :: EQUED        ! Kind of equilibration
 REAL (KIND=dp), DIMENSION(:), ALLOCATABLE     :: R, C         ! Equilibration factors
 REAL (KIND=dp), DIMENSION(:), ALLOCATABLE     :: FERR, BERR   ! Errors (estimates)
 REAL (KIND=dp)                                :: RCOND        ! Condition number
 INTEGER, DIMENSION(:), ALLOCATABLE            :: indx

 ! Auxiliaries
 INTEGER                                       :: nl, nnit
 INTEGER                                       :: it, i, i0
 INTEGER                                       :: nb_m
 INTEGER (KIND=ilk)                            :: n0, le_n
 INTEGER                                       :: i1, i2
 REAL (KIND=dp)                                :: ht
 REAL (KIND=dp)                                :: taux1, taux2, taux3
 REAL (KIND=dp)                                :: taux4
 REAL (KIND=dp)                                :: wlth, wlthcm
 REAL (KIND=dp)                                :: deriv_d
 REAL (KIND=dp), DIMENSION(:), ALLOCATABLE     :: argu
 REAL (KIND=dp), DIMENSION(-mleg:mleg)         :: auxil
 INTEGER, DIMENSION(:,:), ALLOCATABLE          :: indmi, indmj

 ! Thermalized (black body) emission of dust and gas
 REAL (KIND=dp)                                :: bbgas

 ! Computation of emerging specific intensities
 REAL (KIND=dp), DIMENSION(:), ALLOCATABLE     :: fl_0

    REAL(KIND=dp)                              :: ff1, ff2, ff3, ff4, denom
    REAL(KIND=dp), DIMENSION(:,:), ALLOCATABLE :: tmp_Jmp
    INTEGER                                    :: itmp, ntmp, ii

 REAL(KIND=dp)                                 :: lambda1, lambda2
!-----------------------------------------------------------------------------

  ! Number of Legendred coefficient (used for loops that run from 1 to lleg+1)
  nl = lleg + 1

  IF (ALLOCATED(g_absL)) THEN
     DEALLOCATE (g_absL)
  ENDIF

  IF (ALLOCATED(intJm)) THEN
     DEALLOCATE (intJm, intJp)
  ENDIF

  IF (ALLOCATED(I_esc_0)) DEALLOCATE(I_esc_0)

  ALLOCATE (p_0(0:lleg), q_0(0:lleg))
  ALLOCATE (hl(0:lleg,0:npo))
  ALLOCATE (wr(lleg+1), wi(lleg+1))
  ALLOCATE (xkm(-mleg:mleg,0:npo))
  ALLOCATE (Smk_m(0:npo,mleg))
  ALLOCATE (Emk_m_mag(0:npo,mleg))
  ALLOCATE (Emk_m_cmp(0:npo,mleg))
  ALLOCATE (Emk_m_val(0:npo,mleg))
  ALLOCATE (Emk_p_val(0:npo,mleg))
  ALLOCATE (T_m(0:npo,mleg))

  ALLOCATE (QmG(0:npo))
  ALLOCATE (rlm(0:lleg,-mleg:mleg,0:npo))
  ALLOCATE (rlmp(0:lleg,-mleg:mleg,0:npo))
  ALLOCATE (rlmi(0:lleg,-mleg:mleg,0:npo))
  ALLOCATE (rlmip(0:lleg,-mleg:mleg,0:npo))
  ALLOCATE (sfm(-mleg:mleg,0:npo))
  ALLOCATE (sfl(0:lleg,0:npo))
  ALLOCATE (sfl_g(0:lleg,0:npo),sfl_d(0:lleg,0:npo))
  ALLOCATE (ggm(-mleg:mleg,0:npo))
  ALLOCATE (Srm(mleg,0:npo))
  ALLOCATE (sigl(0:lleg,0:npo))
  ALLOCATE (dtl(0:npo))
  ALLOCATE (tl(0:npo,0:nwlu))
  ALLOCATE (albed(0:npo,0:nwlu))
  ALLOCATE (kap_gas(0:npo))
  ALLOCATE (kap_dus(0:npo,0:nwlu))
  ALLOCATE (sca_dus(0:npo,0:nwlu))
  ALLOCATE (emm_dus(0:npo))
  ALLOCATE (g_absL(0:npo,0:nwlu))
  ALLOCATE (intJm(0:npo,0:nwlu))
  ALLOCATE (intJp(0:npo,0:nwlu))
  ALLOCATE (R(nl), C(nl))
  ALLOCATE (FERR(1), BERR(1))
  ALLOCATE (WORK(4*nl))
  ALLOCATE (IWORK(nl))
  ALLOCATE (aam(nl,nl), bbm(nl,1))
  ALLOCATE (aamf(nl,nl), bbmf(nl,1))
  ALLOCATE (indx(nl))
  ALLOCATE (Tim(-nmui:nmui,-mleg:mleg))
  ALLOCATE (Wim(-nmui:nmui,-mleg:mleg))
  ALLOCATE (Iinc(-nmui:nmui))
  ALLOCATE (Qit(-mleg:mleg,0:npo))
  ALLOCATE (Qit_o(-mleg:mleg,0:npo))
  ALLOCATE (Git(-mleg:mleg,0:npo))
  ALLOCATE (Dit(-mleg:mleg,0:npo))
  ALLOCATE (Yit(-mleg:mleg,0:npo))
  ALLOCATE (Hit(-nmui:nmui))
  ALLOCATE (Camp(-nmui:nmui))
  ALLOCATE (argu(0:npo))
  ALLOCATE (indmi(mleg,0:npo), indmj(mleg,0:npo))
  ALLOCATE (fl_0(0:lleg))
  ALLOCATE (I_esc_0(0:nwlu,-nmui:nmui))

  p_0(:)         = 0.0_dp
  q_0(:)         = 0.0_dp
  sigl(:,:)      = 0.0_dp
  Tim(:,:)       = 0.0_dp
  Wim(:,:)       = 0.0_dp
  Iinc(:)        = 0.0_dp
  xkm(:,:)       = 0.0_dp
  Qit(:,:)       = 0.0_dp
  Qit_o(:,:)     = 0.0_dp
  Yit(0,0:npo)   = 0.0_dp
  I_esc_0(:,:)   = 0.0_dp

! STEP 0 - Initialise radiative quantities
!  (boundary conditions and optical depth grid)

  ! uvspec.out is the total absorption through the cloud. multiply by
  !            incident radition field to compute spectrum.

  !--- Emission spectra (specific intensity) at Legendre decomposition angles
  fichier = TRIM(out_dir)//TRIM(modele)//'.Iesc'
  PRINT *, " Escaping intensities in:", TRIM(fichier)
  OPEN (36, file=fichier, status="unknown")
  ! Write comment lines of .Iesc file
  ! --------------------------------------------------------------------------------------
  ! The angles written correspond to "theta" in the document transfer_v6.pdf put in
  ! increasing order from 6.7 to 173.3 degrees (limits of the Legende decomposition
  !  0 < theta <  90 correspond to intensity radiated toward decreasing tau (increasing s)
  ! 90 < theta < 180 correspond to intensity radiated toward increasing tau (decreasing s)
  ! The resulting intensities are computed at the end of the function. BG 09/2013
  ! --------------------------------------------------------------------------------------
  WRITE (36,'("# Emission spectra (specific intensity in erg cm-2 s-1 sr-1 Ang-1)")')
  WRITE (36,'("# provided on observer and back side of the cloud")')
  WRITE (36,'("# at each angle theta = acos(mu) of the Legendre decomposition in degrees (see transfer_v6.pdf)")')
  WRITE (36,'("# Lambda[A]       ")',ADVANCE="NO")
  DO i = 1, nmui
     WRITE (36,'("I_ob:",F7.3,3x)',ADVANCE="NO") AngleLeg_deg(i)
  ENDDO
  DO i = nmui, 1, -1
     WRITE (36,'("I_ob:",F7.3,3x)',ADVANCE="NO") 180.0_dp - AngleLeg_deg(i)
  ENDDO
  DO i = 1, nmui
     WRITE (36,'("I_ba:",F7.3,3x)',ADVANCE="NO") AngleLeg_deg(i)
  ENDDO
  DO i = nmui, 1, -1
     WRITE (36,'("I_ba:",F7.3,3x)',ADVANCE="NO") 180.0_dp - AngleLeg_deg(i)
  ENDDO
  WRITE (36,*)

  !--- Extinction ------------------------------------
  fichier = TRIM(out_dir)//TRIM(modele)//'.uv'
  PRINT *, " Absorption in:", TRIM(fichier)
  OPEN (37, file=fichier, status="unknown")
  WRITE (37,'("#  W-L   EXP(-tau)  tau")')

  !--- Incident radiation field ----------------------
  fichier = TRIM(out_dir)//TRIM(modele)//'.flin'
  PRINT *, " External field:", TRIM(fichier)
  OPEN (60, file=fichier, status="unknown")
  WRITE(60,*) "4      # Number of columns"

  ! Interpolate intJmp on new grid (tau - lambda) to use in ABSGAZ before new radiation field is known
  ! Moved from ADD_WL
  IF (SIZE(intJmp,dim=1) /= npo .OR. SIZE(intJmp,dim=2) /= nwlg) THEN
     ALLOCATE (tmp_Jmp(0:npo,0:nwlg))
     ntmp = 0
     ! JLB - 13 May 2020
     tmp_Jmp(0,0) = intJmp(0,0)
     itmp = 0
     DO ii = 0, nwlu
        CALL LOC_IND (itmp, 0, SIZE(tmp_wl)-1, rf_wl%Val(ii), tmp_wl)
        denom = tmp_wl(itmp+1) - tmp_wl(itmp)
        ff1   = (tmp_wl(itmp+1) - rf_wl%Val(ii)) / denom
        ff3   = (rf_wl%Val(ii)  - tmp_wl(itmp)) / denom
        tmp_Jmp(0,ii) = intJmp(0,itmp) * ff1 + intJmp(0,itmp+1) * ff3
     ENDDO

     DO n = 1, npo
        CALL LOC_IND (ntmp, 0, SIZE(tmptau)-1, tau_o(n), tmptau)
        itmp = 0
        DO ii = 0, nwlg
           CALL LOC_IND (itmp, 0, SIZE(tmp_wl)-1, rf_wl%Val(ii), tmp_wl)
           denom = (tmptau(ntmp+1) - tmptau(ntmp)) * (tmp_wl(itmp+1) - tmp_wl(itmp))
           ff1   = (tmptau(ntmp+1) - tau_o(n)) * (tmp_wl(itmp+1) - rf_wl%Val(ii)) / denom
           ff2   = (tau_o(n)   - tmptau(ntmp)) * (tmp_wl(itmp+1) - rf_wl%Val(ii)) / denom
           ff3   = (tmptau(ntmp+1) - tau_o(n)) * (rf_wl%Val(ii)  - tmp_wl(itmp)) / denom
           ff4   = (tau_o(n)   - tmptau(ntmp)) * (rf_wl%Val(ii)  - tmp_wl(itmp)) / denom
           tmp_Jmp(n,ii) = intJmp(ntmp,itmp) * ff1 + intJmp(ntmp+1,itmp) * ff2 &
                         + intJmp(ntmp,itmp+1) * ff3 + intJmp(ntmp+1,itmp+1) * ff4
        ENDDO
     ENDDO
     DEALLOCATE (intJmp)
     ALLOCATE (intJmp(0:npo,0:nwlg))
     intJmp = tmp_Jmp
     DEALLOCATE (tmp_Jmp)
  ENDIF

  ! Compute absorption by gas
  CALL ABSGAZ (nwlu)
  IF (ALLOCATED(intJmp)) DEALLOCATE(intJmp)
  ALLOCATE (intJmp(0:npo,0:nwlg))

! STEP 1 - Compute P_l(mu_i) - see recurence in Num Rec
  WRITE(iscrn,'("       Point             Angstroms       erg cm-2 s-1 A-1 sr-1")')
  ! Example on screen : iwlg =   100 wlth =        925.259  intJmp: edge =   4.5370E-08 middle =  9.0686E-26

  p_0(0) = 1.0_dp
  q_0(1) = p_0(0) / 2.0_dp
  DO l = 1, (lleg-1)/2
     p_0(2*l) = - (2.0_dp*l-1) * p_0(2*l-2) / (2.0_dp*l)
     q_0(2*l+1) = p_0(2*l) / (2.0_dp*l+2.0_dp)
  ENDDO

  DO iwlg = 0, nwlu
     ! sigl = gg**l (Formule 48, Henyey-Greenstein phase function)
     DO n = 0, npo
        IF (F_DU_PROP == 0) THEN
           n1 = 0
        ELSE
           n1 = n
        ENDIF
        ggHG = du_prop%gcos(iwlg,n1)
        !ggHG = 1.0_dp - set g = 1 for tests (pure forward scattering)
        sigl(0,n) = 1.0_dp
        DO l = 1, lleg
           sigl(l,n) = sigl(l-1,n) * ggHG
        ENDDO
     ENDDO

     ! Reset to 0 most variables
     hl = 0.0_dp
     wr = 0.0_dp
     wi = 0.0_dp
     tl(:,iwlg) = 0.0_dp
     dtl = 0.0_dp
     albed(:,iwlg)   = 0.0_dp
     kap_dus(:,iwlg) = 0.0_dp
     sca_dus(:,iwlg) = 0.0_dp
     emm_dus(:)      = 0.0_dp
     kap_gas(:)      = 0.0_dp

     wlth = rf_wl%Val(iwlg)
     wlthcm = wlth * 1.0e-8_dp

     ! Compute local values of absorption coef for dust and gas and dust scatering coef
     ! Compute effective albedo !
     DO n = 0, npo
        IF (F_DU_PROP == 0) THEN
           n1 = 0
        ELSE
           n1 = n
        ENDIF
        kap_gas(n)      = g_absL(n,iwlg)
        kap_dus(n,iwlg) = du_prop%abso(iwlg,n1) * dens_o(n)
        sca_dus(n,iwlg) = du_prop%scat(iwlg,n1) * dens_o(n)
        emm_dus(n)      = du_prop%emis(iwlg,n)  * dens_o(n)
        albed(n,iwlg)   = sca_dus(n,iwlg) / (kap_dus(n,iwlg) + kap_gas(n) + sca_dus(n,iwlg))
     ENDDO

     ! b - Compute optical depth grid
     !     Convert Delta Tau_v (dust only at V) to
     !     Delta tau_lambda (dust + gaz at Lambda)
     tl(0,iwlg) = 0.0_dp
     DO n = 0, npo-1
        ht = ((kap_dus(n,iwlg)   + kap_gas(n)   + sca_dus(n,iwlg)) &
           +  (kap_dus(n+1,iwlg) + kap_gas(n+1) + sca_dus(n+1,iwlg))) &
           * 0.5_dp * dd_cm(n+1)
        dtl(n) = ht
        tl(n+1,iwlg) = tl(n,iwlg) + ht
     ENDDO
     ! Compute total attenuation by dust + gas through whole cloud
     WRITE (37,"(1p,3e17.7E3)") wlth, EXP(-tl(npo,iwlg)), tl(npo,iwlg)

! STEP 2a - Compute matrix A(tau)

     DO n = 0, npo
        ! Formulae A2 and A3 of Roberge:
        hl(0,n) = 1.0_dp - albed(n,iwlg)
        DO l = 1, lleg
           hl(l,n) = (2.0_dp*l+1.0_dp) * (1.0_dp - albed(n,iwlg) * sigl(l,n))
        ENDDO
        ! Appendix C of Goicoechea & Le Bourlot (2007)
        DO l = 1, lleg
           wi(l) = DBLE(l) / SQRT(hl(l-1,n)*hl(l,n))
        ENDDO
        wr = 0.0_dp

        ! Compute eigenvalues (tridiagonal symetric matrix) - LAPACK
        CALL DSTEVD(JOBZ, nl, wr, wi, aam, nl, WORK, 4*nl, IWORK, nl, info )

        ! order eigenvalues by decreasing order
        wi = wr
        DO l = 1, nl
           wr(l) = wi(nl+1-l)
        ENDDO

        ! km = 1 / lamda_m in increasing order
        DO m = 1, mleg
           xkm(m,n) = 1.0_dp / wr(m)
           xkm(-m,n) = -xkm(m,n)
        ENDDO
     ENDDO   ! end loop in npo

     ! Compute Rlm: Formule A9 from Roberge 1983
     ! Eigenvectors computed only from recurrence relations.
     DO n = 0, npo
        DO m = 1, mleg
           rlm(0,m,n)  = 1.0_dp
           rlm(0,-m,n) = 1.0_dp
           rlm(1,m,n)  = (1.0_dp - albed(n,iwlg)) / xkm(m,n)
           rlm(1,-m,n) = -rlm(1,m,n)                             !jrg
           DO l = 2, lleg
              rlm(l,m,n) = (hl(l-1,n) * rlm(l-1,m,n) &
                         - (l-1) * xkm(m,n) * rlm(l-2,m,n)) / (l*xkm(m,n))
              rlm(l,-m,n) = xmunl(l) * rlm(l,m,n)
           ENDDO
        ENDDO
     ENDDO   ! end loop in npo

! STEP 2c - Compute Srm - used for mean intensity

     CALL AUX_Srm(q_0, rlm, Srm)

! STEP 3 - Compute Smk_m = Sum (xkm) (left to right)
!        - Simple trapezoidal rule

     Smk_m(0,:) = 0.0_dp
     DO m = 1, mleg
        DO n = 1, npo
           Smk_m(n,m) = Smk_m(n-1,m) + 0.5_dp * dtl(n-1) * (xkm(m,n) + xkm(m,n-1))
        ENDDO
     ENDDO

     DO m = 1, mleg
        Emk_m_mag(:,m) = INT(MIN(Smk_m(:,m),1.0e18_dp) / grand, ilk)
        Emk_m_cmp(:,m) = EXP(-(Smk_m(:,m)-grand*Emk_m_mag(:,m)))
        Emk_m_val(:,m) = expgr(MIN(mgr,Emk_m_mag(:,m))) * Emk_m_cmp(:,m)
     ENDDO
     ALLOCATE (md_m(mleg))
     md_m(:) = Emk_m_mag(npo,:)
     ALLOCATE (ndp_m(0:npo+1,mleg))
     ALLOCATE (ndq_m(0:npo+1,mleg))
     ALLOCATE (R_m(npo+1,mleg))
     ALLOCATE (nn_m(npo+1,mleg))
     ndp_m = 0
     ndq_m = 0

     DO m = 1, mleg
        le_n = Emk_m_mag(0,m)
        nb_m = 1
        ndp_m(1,m) = 0
        nn_m(nb_m,m) = le_n
        DO n = 1, npo
           IF (Emk_m_mag(n,m) == le_n) THEN
              CYCLE
           ELSE
              ndq_m(nb_m,m) = n - 1
              nn_m(nb_m,m) = le_n
              le_n = Emk_m_mag(n,m)
              nb_m = nb_m + 1
              ndp_m(nb_m,m) = n
           ENDIF
        ENDDO
        ndq_m(nb_m,m) = npo
        nn_m(nb_m,m) = le_n
        md_m(m) = nb_m
     ENDDO

     Emk_p_val = 0.0_dp      ! Tau_max = infty

! STEP 4 - Compute R**-1    !Loop to all cloud optical depth points

     CALL AUX_Rlmi(rlm, hl, rlmi)

! STEP 5 - Compute source function contribution g~ = R^-1*A^-1*g (in notes)
!          JLB : Check Javier correction (not yet done here)

     DO n = 0, npo
        ! Ex. Black body emission of thermalized gas
        IF (wlthcm*tgaz(n) > 1.0e-2_dp) THEN
           bbgas = 2.0_dp*xhp*clum**2 / (wlthcm)**5
           bbgas = bbgas * 1.0e-8_dp / (EXP(hcsurk/(wlthcm*tgaz(n))) -1.0_dp)
        ELSE
           bbgas = 0.0_dp
        ENDIF

! 5a - Assign source function, isotropic  (PROVISIONAL)

        ! METHOD B1: approx. to continuum thermal isotropic emission
        sfl_d(0,n) = emm_dus(n) / (kap_dus(n,iwlg) + kap_gas(n) + sca_dus(n,iwlg))

        ! METHOD B2: approx. to line thermal isotropic emission
        sfl_g(0,n) = bbgas * kap_gas(n) / (kap_dus(n,iwlg) + kap_gas(n) + sca_dus(n,iwlg))

        ! Add both contributions
!       sfl(0,n) = sfl_g(0,n) + sfl_d(0,n)
        sfl(0,n) = sfl_d(0,n)
!       sfl(0,n) = 0.0_dp
        ! JLB - 8 III 2013 - Add a low intensity  background source
        sfl(0,n) = sfl(0,n) + 1.0e-40_dp

        sfl(1:lleg,n) = 0.0_dp
     ENDDO

     ! Put them in m,n desired order
     sfm(-mleg:-1,:) = sfl(0:mleg-1,:)
     sfm(1:mleg,:) = sfl(mleg:lleg,:)

! 5b - Compute vector g

     DO n = 0, npo
        DO m = -mleg, -1
           l = mleg + m
           ggm(m,n) = -sfm(m,n) / ( 1.0_dp - albed(n,iwlg) * sigl(l,n) )
        ENDDO
        DO m = 1, mleg
           l = mleg + m - 1
           ggm(m,n) = -sfm(m,n) / ( 1.0_dp - albed(n,iwlg) * sigl(l,n) )
        ENDDO
     ENDDO

! 5d - Compute matrix product  R^-1*A^-1*g = Git

     CALL AUX_GIT(rlmi, ggm, xkm, Git)

! STEP 6 - Compute derivatives of Rlm

! 6c - derivative of Rlm

     DO n = 2, npo-2
        DO m = 1, mleg
           rlmp(0,m,n) = 0.0_dp
           rlmp(0,-m,n) = 0.0_dp
           n1 = n - 2
           n3 = n + 2
           DO l = 1, lleg
              ! Direct variable
              deriv_d = -rlm(l,m,n1) * (dtl(n)+dtl(n+1)) / ((dtl(n-2)+dtl(n-1)) * (dtl(n-2)+dtl(n-1)+dtl(n)+dtl(n+1))) &
                      - rlm(l,m,n)  * (dtl(n-2)+dtl(n-1)-dtl(n)-dtl(n+1)) / ((dtl(n)+dtl(n+1)) * (dtl(n-2)+dtl(n-1))) &
                      + rlm(l,m,n3) * (dtl(n-2)+dtl(n-1)) / ((dtl(n-2)+dtl(n-1)+dtl(n)+dtl(n+1)) * (dtl(n)+dtl(n+1)))
              rlmp(l,m,n) = deriv_d
              rlmp(l,-m,n) = xmunl(l) * rlmp(l,m,n)
           ENDDO
        ENDDO
     ENDDO

     ! Boundary condition. Replace par linear extrapolation (TODO)
     DO m = 1, mleg
        rlmp(0,m,0) = 0.0_dp
        rlmp(0,-m,0) = 0.0_dp
        rlmp(0,m,1) = 0.0_dp
        rlmp(0,-m,1) = 0.0_dp
        rlmp(0,m,npo-1) = 0.0_dp
        rlmp(0,-m,npo-1) = 0.0_dp
        rlmp(0,m,npo) = 0.0_dp
        rlmp(0,-m,npo) = 0.0_dp
        DO l = 1, lleg
           rlmp(l,m,0) = rlmp(l,m,2)
           rlmp(l,-m,0) = xmunl(l) * rlmp(l,m,0)
           rlmp(l,m,1) = rlmp(l,m,2)
           rlmp(l,-m,1) = xmunl(l) * rlmp(l,m,1)
           rlmp(l,m,npo-1) = rlmp(l,m,npo-2)
           rlmp(l,-m,npo-1) = xmunl(l) * rlmp(l,m,npo-1)
           rlmp(l,m,npo) = rlmp(l,m,npo-2)
           rlmp(l,-m,npo) = xmunl(l) * rlmp(l,m,npo-2)
        ENDDO
     ENDDO

! 6d - Multiply by R**-1

     CALL AUX_Rlmip(rlmi, rlmp, rlmip)

! STEP 7 - Compute Tim

     DO m = 1, mleg
        DO i = 1, nmui
           taux1 = 0.0_dp
           taux4 = 0.0_dp
           DO l = 0, lleg
              taux1 = taux1 + (2.0_dp * l + 1.0_dp) * rlm(l,-m,0)  * pli(l,-i)
              taux4 = taux4 + (2.0_dp * l + 1.0_dp) * rlm(l,m,npo) * pli(l,i)
           ENDDO
           Tim(-i,-m) = taux1
           Tim(i,m)   = taux4
        ENDDO
     ENDDO

! STEP 8 - Compute Wim (Boundary conditions matrix)

     DO m = 1, mleg
        DO i = 1, nmui
           Wim(-i,-m) = Tim(-i,-m)
           Wim(-i,m)  = 0.0_dp
           Wim(i,-m)  = 0.0_dp
           Wim(i,m)   = Tim(i,m)
        ENDDO
     ENDDO

! STEP 9 - Set boundary conditions

     ! Add potential // source (star)
     WRITE (60, '(1p,4e15.6E3)') wlth, I_incr_m(iwlg), rf_incIm%Val(iwlg)

     ! 19 XII 2005 - Use "REAL" units (specific intensity)
     Iinc(-nmui:-1) = rf_incIm%Val(iwlg)
     Iinc(1:nmui)   = 0.0_dp

     ! JLB 4 IX 08 - Corrected. See other routine for comment
     Iinc(-1) = rf_incIm%Val(iwlg) + I_incr_m(iwlg)

! STEP 10 - Start iteration with an appropriate guess for Y and Q

! 10a - Compute Y_o = R^-1 x F_approx
!       F_approx ~ sfm / 1 - albed + Iexp(kt), THEN it IS wavelength and tau dependent

     CALL AUX_YIT1(Iinc, xkm, tl(:,iwlg), albed(:,iwlg), rlmi, sfm, Yit)

! 10b - Compute Q_o = R^-1 x R' x Y_o

     CALL AUX_QIT(rlmip, Yit, Qit)

!    nnit = 200
     nnit = MAX(10, ifaf)

     !Git = 0.0_dp !old procedure if Git = 0 and Qit = 0
     !Qit = 0.0_dp !tests javi

     DO it = 1, nnit ! begin loop :: 101

! STEP 11 - Compute D  - Simple trapezoidal rule

        DO m = 1, mleg
           QmG(0:npo) = Qit(-m,0:npo) - Git(-m,0:npo)
           DO k_ilk = 1, md_m(m)
              n = ndp_m(k_ilk,m) + 1
              IF (n > npo) CYCLE
              T_m(n,m) = (QmG(n-1) / Emk_m_cmp(n-1,m) &
                       + QmG(n) / Emk_m_cmp(n,m)) * dtl(n-1)
              DO n = ndp_m(k_ilk,m)+2, ndq_m(k_ilk,m)
                 T_m(n,m) = T_m(n-1,m) + (QmG(n-1) / Emk_m_cmp(n-1,m) &
                                       + QmG(n) / Emk_m_cmp(n,m)) * dtl(n-1)
              ENDDO
              i0 = ndq_m(k_ilk,m)+1
              IF (i0 > npo) THEN
                 R_m(k_ilk,m) = 0.0_dp
              ELSE
                 T_m(i0,m) = T_m(i0-1,m) + (QmG(i0-1) / Emk_m_cmp(i0-1,m)) * dtl(i0-1)
                 R_m(k_ilk,m) = (QmG(i0) / Emk_m_cmp(i0,m)) * dtl(i0-1)
              ENDIF
           ENDDO
        ENDDO
        T_m(:,:) = T_m(:,:) * 0.5_dp
        R_m(:,:) = R_m(:,:) * 0.5_dp

        Dit(-mleg:-1,0) = 0.0_dp
        DO m = -mleg, -1
           DO k_ilk = 1, md_m(-m)
              taux1 = 0.0_dp
              DO i_ilk = 1, k_ilk-1
                 n0 = nn_m(k_ilk,-m) - nn_m(i_ilk,-m)
                 taux1 = taux1 + T_m(ndq_m(i_ilk,-m)+1,-m) * expgr(MIN(mgr,n0))
                 n0 = nn_m(k_ilk,-m) - nn_m(i_ilk+1,-m)
                 taux1 = taux1 + R_m(i_ilk,-m) * expgr(MIN(mgr,n0))
              ENDDO
              n = ndp_m(k_ilk,-m)
              Dit(m,n) = taux1 * Emk_m_cmp(n,-m)
              DO n = ndp_m(k_ilk,-m)+1, ndq_m(k_ilk,-m)
                 Dit(m,n) = (taux1 + T_m(n,-m)) * Emk_m_cmp(n,-m)
              ENDDO
           ENDDO
        ENDDO

        Dit(1:mleg,:) = 0.0_dp

! STEP 12 - Compute Hit

        Hit = 0.0_dp
        DO i = 1, nmui
           taux1 = 0.0_dp
           taux2 = 0.0_dp
           DO m = 1, mleg
              taux1 = taux1 + Dit(-m,npo) * Tim(i,-m)
           ENDDO
           Hit(-i) = Iinc(-i) - taux2
           Hit(i) = 0.0_dp + taux1
        ENDDO

! STEP 13 - Boundary conditions - solve using LAPACK

        aam(1:nmui,1:mleg)        = Wim(-nmui:-1,-mleg:-1)
        aam(1:nmui,mleg+1:2*mleg) = Wim(-nmui:-1,1:mleg)
        aam(nmui+1:nl,1:mleg)     = Wim(1:nmui,-mleg:-1)
        aam(nmui+1:nl,mleg+1:nl)  = Wim(1:nmui,1:mleg)

        bbm(1:nmui,1)    = Hit(-nmui:-1)
        bbm(nmui+1:nl,1) = Hit(1:nmui)

        nrhs = 1
        CALL DGESVX (FACT, TRANS, nl, nrhs, aam, nl, aamf, nl, indx, EQUED, &
             R, C, bbm, nl, bbmf, nl, RCOND, FERR, BERR, WORK, IWORK, info)

        Camp(-nmui:-1) = bbmf(1:nmui,1)
        Camp(1:nmui) = bbmf(nmui+1:nl,1)

! STEP 14 - Compute Y

        DO n = 0, npo
           DO m = 1, mleg
              Yit(-m,n) = Camp(-m) * Emk_m_val(n,m) - Dit(-m,n)
              Yit(m,n)  = 0.0_dp
           ENDDO
        ENDDO

! STEP 15 - Compute new Q

        Qit_o = Qit

        CALL AUX_QIT(rlmip, Yit, Qit)

! STEP 16 - convergence ?

        auxil(0) = 0.0_dp
        taux2 = 0.0_dp
        DO n = 0, npo
           DO m = -mleg, -1
              IF (Qit(m,n) /= 0.0_dp) THEN
                 auxil(m) = ABS((Qit(m,n)-Qit_o(m,n))/Qit(m,n))
              ELSE
                 auxil(m) = 0.0_dp
              ENDIF
           ENDDO
           DO m = 1, mleg
              IF (Qit(m,n) /= 0.0_dp) THEN
                 auxil(m) = ABS((Qit(m,n)-Qit_o(m,n))/Qit(m,n))
              ELSE
                 auxil(m) = 0.0_dp
              ENDIF
           ENDDO
           taux2 = taux2 + SUM(auxil)
        ENDDO

        IF (taux2 < 1.0e-3_dp) THEN
           EXIT      !EXITs sub-iteration loop
        ENDIF

     ENDDO        ! Exits loop to find q  :: 101
     IF (taux2 > 1.0e-3_dp) THEN
        PRINT *, " Pb:", iwlg, wlth, taux2
     ENDIF

! STEP 17 - Compute radiation field - mean intensity
!           intJm is 0.5 SUM(-1->0) of I
!           intJp is 0.5 SUM(0->1) of I
!           Thus intJm + intJp = J in erg cm-2 s-1 A-1 sterad-1
!           one gets u in erg cm-3 A-1 by u = J * 4 pi / c

     DO n = 0, npo
        taux1 = 0.0_dp
        taux2 = 0.0_dp
        taux3 = 0.0_dp
        DO m = 1, mleg
           taux1 = taux1 + Yit(-m,n) * (1.0_dp + Srm(m,n))
           taux2 = taux2 + Yit(-m,n) * (1.0_dp - Srm(m,n))
           taux3 = taux3 + Yit(-m,n)
        ENDDO

        ! 19 XII 2005 - Change normalisation
        intJm(n,iwlg) = 0.5_dp * taux1
        intJp(n,iwlg) = 0.5_dp * taux2
        intJmp(n,iwlg) = taux3
     ENDDO

     IF (MOD(iwlg,100) == 0) THEN
        WRITE(iscrn,'("iwlg = ",I6," wlth[A] = ",F16.3,"  intJmp: edge = ",1p,E12.4," middle =",1p,E12.4)') &
              iwlg, wlth, intJmp(0,iwlg), intJmp(INT(npo/2),iwlg)
     ENDIF

     ! Compute emerging specific intensities
     DO l = 0, lleg
        fl_0(l) = 0.0_dp
        DO m = 1, mleg
           fl_0(l) = fl_0(l) + rlm(l,m,0) * Yit(m,0) + rlm(l,-m,0) * Yit(-m,0)
        ENDDO
     ENDDO

     DO i = 1, nmui
        I_esc_0(iwlg,-i) = 0.0_dp
        I_esc_0(iwlg,i)  = 0.0_dp
        DO l = 0, lleg
           I_esc_0(iwlg,-i) = I_esc_0(iwlg,-i) + (2.0_dp*l+1.0_dp) * fl_0(l) * pli(l,-i)
           I_esc_0(iwlg,i)  = I_esc_0(iwlg,i)  + (2.0_dp*l+1.0_dp) * fl_0(l) * pli(l,i)
        ENDDO
     ENDDO
     WRITE (36,'(1p,32e15.6)') wlth, (I_esc_0(iwlg,i), i=1,nmui), (I_esc_0(iwlg,i), i=-nmui,-1)

     DEALLOCATE (md_m)
     DEALLOCATE (ndp_m)
     DEALLOCATE (ndq_m)
     DEALLOCATE (R_m)
     DEALLOCATE (nn_m)
  ENDDO    !end wavelenght loop :: 99  !too slow !!

  CLOSE (36)
  CLOSE (37)
  CLOSE (60)

  lambda1 = 3.0e+0_dp
  lambda2 = 1.0e+3_dp
  Icont_3_1000   = INTEG_TAB(I_esc_0(0:nwlu,1),nwlu,lambda1,lambda2)
  lambda1 = 2.5e+1_dp
  lambda2 = 6.0e+1_dp
  Icont_25_60    = INTEG_TAB(I_esc_0(0:nwlu,1),nwlu,lambda1,lambda2)
  lambda1 = 6.0e+1_dp
  lambda2 = 2.0e+2_dp
  Icont_60_200   = INTEG_TAB(I_esc_0(0:nwlu,1),nwlu,lambda1,lambda2)
  lambda1 = 2.0e+2_dp
  lambda2 = 1.0e+3_dp
  Icont_200_1000 = INTEG_TAB(I_esc_0(0:nwlu,1),nwlu,lambda1,lambda2)
  lambda1 = 6.0e+1_dp
  lambda2 = 1.0e+3_dp
  Icont_60_1000  = INTEG_TAB(I_esc_0(0:nwlu,1),nwlu,lambda1,lambda2)

  IF (F_W_RF_ALB == 1) THEN
     ! Write here full radiation field with effective albedo (JLB - 18 Oct 2012)
     fichier = TRIM(out_dir)//TRIM(modele)//'.rf_alb'//TRIM(ADJUSTL(ii2c(ifaf)))
     OPEN(61, file=fichier, status="unknown" , form = 'unformatted')
     WRITE (61) rf_wl%upper-rf_wl%lower+1 ! number of pixels in wavelength grid
     WRITE (61) rf_wl%Val                 ! Wavelength grid in Angstroms
     DO n = 0, npo
        WRITE (61) tau_o(n)*tautav, intJmp(n,0:nwlu) * J_to_u, g_absL(n,:), kap_dus(n,:), sca_dus(n,:)
     ENDDO
     CLOSE (61)
  ENDIF

  ! Free memory
  DEALLOCATE (p_0, q_0)
  DEALLOCATE (hl)
  DEALLOCATE (wr, wi)
  DEALLOCATE (xkm)
  DEALLOCATE (Emk_m_mag)
  DEALLOCATE (Emk_m_cmp)
  DEALLOCATE (Emk_m_val)
  DEALLOCATE (Emk_p_val)
  DEALLOCATE (Smk_m)
  DEALLOCATE (QmG)
  DEALLOCATE (rlm, rlmp, rlmi, rlmip)
  DEALLOCATE (sfl, sfm, sfl_g, sfl_d)
  DEALLOCATE (Srm)
  DEALLOCATE (sigl)
  DEALLOCATE (tl, dtl)
  DEALLOCATE (albed)
  DEALLOCATE (kap_gas, kap_dus, sca_dus, emm_dus)
  DEALLOCATE (aam, bbm)
  DEALLOCATE (aamf, bbmf)
  DEALLOCATE (Tim, Wim)
  DEALLOCATE (Iinc)
  DEALLOCATE (ggm)
  DEALLOCATE (Dit)
  DEALLOCATE (Qit, Qit_o, Git, Yit)
  DEALLOCATE (Hit)
  DEALLOCATE (R, C)
  DEALLOCATE (FERR, BERR)
  DEALLOCATE (WORK, IWORK)
  DEALLOCATE (Camp)
  DEALLOCATE (indx)
  DEALLOCATE (argu)
  DEALLOCATE (indmi, indmj)
  DEALLOCATE (fl_0)

  !====================================================================================
  ! COMPUTE THE RADIATION FIELD INTENSITY AND ENERGY FROM THE LEFT SIDE
  ! (USE RF_U AS AUXILIARY VARIABLE HERE)
  ! The goal is to obtain mean_rad(1,2),  int_rad(1,2),  rad(m,p),  and flu_rad(1,2)
  ! The goal is to obtain mean_xray(1,2), int_xray(1,2), xray(m,p), and flu_xray(1,2)
  !====================================================================================
  CALL RF_CREATE(rf_lamI,0,nwlu)
  CALL RF_CREATE(rf_tmp,1,2)

  !--------------------------------------------------------------------------------
  ! Integration of UV radiation field from Lyman up to Habing limit
  ! (wl_lym = 911 Å ; wl_hab = 2400 Å) - BG 04/2013
  !--------------------------------------------------------------------------------
  CALL RF_LOCATE(rf_wl, wl_lym, i1)
  rf_tmp%Val(1) = 1.0_dp
  rf_tmp%Ord(1) = rf_wl%Ord(i1)
  CALL RF_LOCATE(rf_wl, wl_hab, i2)
  rf_tmp%Val(2) = 1.0_dp
  rf_tmp%Ord(2) = rf_wl%Ord(i2)
  ! -------------------
  mean_rad2           = 0.0_dp
  rf_u%Val(0:nwlu)    = intJm(0  ,0:nwlu)
  mean_rad1           = SECT_INT(rf_tmp, rf_u, wl_pointer, wl_hab)
  ! -------------------
  int_rad2            = 0.0_dp
  radp                = 0.0_dp
  rf_u%Val(0:nwlu)    = intJmp(0  ,0:nwlu) * J_to_u
  int_rad1            = SECT_INT(rf_tmp, rf_u, wl_pointer, wl_hab)
  radm                = int_rad1 / int_rad0
  int_rad(0)          = SECT_INT(rf_tmp, rf_u, wl_pointer, wl_hab)
  ! -------------------
  flu_rad2            = 0.0_dp
  rf_lamI%Val(0:nwlu) = intJmp(0  ,0:nwlu) * J_to_u* rf_wl%Val(0:nwlu)
  flu_rad1            = SECT_INT(rf_tmp, rf_lamI, wl_pointer, wl_hab) / (1.0e8_dp * xhp)

  !--------------------------------------------------------------------------------
  ! Write results of these integrations on the screen and in the .def file
  !--------------------------------------------------------------------------------
  PRINT *, " "
  PRINT *, " Scaled field at cloud edge"
  WRITE (iscrn, '("  U1_UV:",1pe12.4," erg cm-3,     ","from",0pf9.2," to",0pf9.2," Angstrom")') int_rad1,  wl_lym,  wl_hab
  WRITE (iscrn, '("  F1_UV:",1pe12.4," phot cm-2 s-1,","from",0pf9.2," to",0pf9.2," Angstrom")') flu_rad1,  wl_lym,  wl_hab
  WRITE (iwrit2,'("  U1_UV:",1pe12.4," erg cm-3,     ","from",0pf9.2," to",0pf9.2," Angstrom")') int_rad1,  wl_lym,  wl_hab
  WRITE (iwrit2,'("  F1_UV:",1pe12.4," phot cm-2 s-1,","from",0pf9.2," to",0pf9.2," Angstrom")') flu_rad1,  wl_lym,  wl_hab

  !---------------------------------------------------------------------
  ! Compute the effective UV scaling factors in units of G0 and FX0/clum
  !---------------------------------------------------------------------
  G0m = int_rad1  / u_habing
  G0p = 0.0_dp
  WRITE(iwrit2,'(/," U0:",1pe13.4, " erg cm-3")') int_rad0
  PRINT *, " "
  WRITE (iscrn,'("  G_0m :",1pe12.4," (in Habing units)")') G0m
  WRITE(iwrit2,'(" G_0m =", 1pe12.3, " in Habing units",/)') G0m

  ! set here wavelength threshold for integrated intensity.
  CALL RF_LOCATE(rf_wl, wl_hab, i2)
  rf_tmp%Val(2) = 1.0_dp
  rf_tmp%Ord(2) = rf_wl%Ord(i2)
  int_rad(0) = SECT_INT(rf_tmp, rf_u, wl_pointer, wl_hab)

  PRINT *, " "
  PRINT *, " Effective Radiation field strength AT the edge:"
  WRITE (iscrn, '("  left :",1pe12.4," ",A65)') radm,  ISRF_name
  WRITE (iwrit2,'("  left :",1pe12.4," ",A65)') radm,  ISRF_name
  PRINT *, " "

  CALL RF_FREE (rf_tmp)
  CALL RF_FREE (rf_lamI)

END SUBROUTINE ETIN901S

!**************************
SUBROUTINE CO_UV_EXC (iopt)
!**************************

    USE PXDR_FLAGS
    USE PXDR_CONSTANTES
    USE PXDR_CHEM_DATA
    USE PXDR_PROFIL
    USE PXDR_AUXILIAR
    USE PXDR_STR_DATA
    USE PXDR_DRAINE_DUST
    USE PXDR_RF_TOOLS

    IMPLICIT NONE

    INTEGER, INTENT (IN) :: iopt

    INTEGER              :: il2, il1
    INTEGER              :: j
    INTEGER              :: ll, levu, lMX
    INTEGER              :: iwlg
    REAL (KIND=dp)       :: coef, beinst, wl
    REAL (KIND=dp)       :: toto
    REAL (KIND=dp)       :: gl, gu
    REAL (KIND=dp)       :: J_rad

    REAL (KIND=dp), ALLOCATABLE, DIMENSION (:,:) :: absorb

    coef = 1.0e-32_dp / (2.0 * xhp * clum**2)

    UV_exc_CO = 0.0_dp
    lMX = spec_lin(in_sp(i_co))%use
    ALLOCATE (absorb(lMX,SIZE(lm_CO_A,2)))

    ! Interpolation location in previous grid
    DO j = 1, npo
       IF (tau(iopt) <= tau_o(j)) THEN
          icd_o = j-1
          EXIT
       ENDIF
    ENDDO

    raptau = (tau(iopt) - tau_o(icd_o)) / dtau_o(icd_o+1)

    DO il2 = 1, lMX
       gl = spec_lin(in_sp(i_co))%gst(il2)
       toto = 0.0_dp
       DO ll = 1, lm_CO_A(il2,0)
          levu = lm_CO_A(il2,ll)
          gu = 2.0_dp * DBLE(J_CO_A(levu)) + 1.0_dp
          wl = lam_CO_A(levu,il2)
          beinst = coef * (gl / gu) * wl**5 * Aij_CO_A(levu,il2)
          CALL RF_LOCATE(rf_wl, wl, iwlg)
          J_rad = intJmp(icd_o,iwlg) + (intJmp(icd_o+1,iwlg) - intJmp(icd_o,iwlg)) * raptau
          absorb(il2,ll) = beinst * J_rad
          toto = toto + absorb(il2,ll)
       ENDDO
       UV_exc_CO(il2,il2) = toto
    ENDDO

    DO il2 = 1, lMX
       DO il1 = 1, lMX
          toto = 0.0_dp
          DO ll = 1, lm_CO_A(il2,0)
             levu = lm_CO_A(il2,ll)
             toto = toto + absorb(il2,ll) * Cij_CO_X(levu,il1)
          ENDDO
          UV_exc_CO(il2,il1) = UV_exc_CO(il2,il1) - toto
       ENDDO
    ENDDO

    DEALLOCATE (absorb)

END SUBROUTINE CO_UV_EXC

!===================================================================================================
! Radiative transfer routine for fix spatial grain properties
!===================================================================================================
SUBROUTINE ETIN90_CteG (nwlu)

 USE PXDR_FLAGS
 USE PXDR_CONSTANTES
 USE PXDR_CHEM_DATA
 USE PXDR_PROFIL
 USE PXDR_AUXILIAR
 USE PXDR_STR_DATA
 USE PXDR_RF_TOOLS

! Radiative transfer computation. See file transfer_S.pdf for definitions
! and comments. Index conventions:
!       l goes from 0 to lleg (first Legendre Pol. index)
!       m goes from +/- 1 to +/- mleg (second Legendre Pol. index)
!       i goes from +/- 1 to +/- nmui (discrete angles for boundary conditions)
!       n goes from 0 to npo (optical depth grid)
! 2 sides version: tmax finite and radiation are impiging from mu_i > 0 and mu_i < 0

 IMPLICIT NONE

 INTEGER, INTENT (IN)                          :: nwlu     ! Number of wavelength computed

 INTEGER                                       :: l, m, n, n1

 ! Radiative quantities
 INTEGER                                       :: iwlg
 REAL (KIND=dp), DIMENSION(0:npo)              :: tlm      ! Full optical depth (left to right)
 REAL (KIND=dp), DIMENSION(0:npo)              :: tlp      ! Full optical depth (right to left)
 REAL (KIND=dp), DIMENSION(0:npo)              :: dtl      ! Full optical depth increment
 REAL (KIND=dp), DIMENSION(0:lleg)             :: sigl     ! Phase function

 ! Here the albedo does not include gas contribution and hence is constant in tau
 REAL (KIND=dp)                                :: ggHG     ! Henyey-Greenstein g = < cos theta > coefficient

 REAL (KIND=dp), DIMENSION(0:npo)              :: A_gim    ! Integrated normalized gas absorption (left to right)
 REAL (KIND=dp), DIMENSION(0:npo)              :: A_gip    ! Integrated normalized gas absorption (right to left)
 REAL (KIND=dp), DIMENSION(0:npo)              :: dta      ! gas absorption increment
 REAL (KIND=dp), DIMENSION(0:npo)              :: alp_gas
 REAL (KIND=dp), DIMENSION(:,:),   ALLOCATABLE :: kap_dus
 REAL (KIND=dp), DIMENSION(:,:),   ALLOCATABLE :: sca_dus
 REAL (KIND=dp), DIMENSION(:),     ALLOCATABLE :: albed    ! Reduced albedo (excluding gas)
 REAL (KIND=dp), DIMENSION(0:npo)              :: emm_dus
 REAL (KIND=dp), DIMENSION(-nmui:nmui)         :: Camp     ! Legendre amplitudes

 type(rf_type)                                 :: rf_tmp, rf_lamI

 ! Legendre polynomial related variables
 REAL (KIND=dp), DIMENSION(:),     ALLOCATABLE :: p_0
 REAL (KIND=dp), DIMENSION(:),     ALLOCATABLE :: q_0

 ! Matrix A & B coefficients
 REAL (KIND=dp), DIMENSION(0:lleg)             :: bsa      ! Diagonal matrix B in notes
 REAL (KIND=dp), DIMENSION(0:lleg)             :: hl
 REAL (KIND=dp), DIMENSION(lleg+1)             :: wr, wi
 REAL (KIND=dp), DIMENSION(-mleg:mleg)         :: xkm      ! Legendre absissa
 REAL (KIND=dp), DIMENSION(-mleg:mleg)         :: gmm      ! Diagonal of Matrix K * R-1 * B * R
 REAL (KIND=dp), DIMENSION(0:npo,mleg)         :: u_im     ! Increment of Ei / Ei-1
 REAL (KIND=dp), DIMENSION(0:npo,mleg)         :: u_ip     ! idem (+)
 REAL (KIND=dp), DIMENSION(0:npo,mleg)         :: EsEm     ! Ei / Ei-1
 REAL (KIND=dp), DIMENSION(0:npo,mleg)         :: EsEp     ! idem (+)

 REAL (KIND=dp), DIMENSION(mleg)               :: Emkt_m   ! Exp(-xkm * tau_max) 0 -> tau
 REAL (KIND=dp), DIMENSION(mleg)               :: Emkt_p   ! Exp(-xkm * tau_max) tau -> tau_max
 REAL (KIND=dp), DIMENSION(mleg,0:npo)         :: Emk_m    ! Exp(-xkm * tau) 0 -> tau
 REAL (KIND=dp), DIMENSION(mleg,0:npo)         :: Emk_p    ! Exp(-xkm * tau) tau -> tau_max

 REAL (KIND=dp), DIMENSION(0:npo,-mleg:mleg)   :: vvmk
 REAL (KIND=dp), DIMENSION(0:npo,-mleg:mleg)   :: vmkg
 REAL (KIND=dp), DIMENSION(0:npo)              :: vmkd
 ! JLB 15 V 2018: bug in the indexation of rlmi
 REAL (KIND=dp), DIMENSION(0:lleg,-mleg:mleg)  :: rlm      ! Eigenvectors matrix
 REAL (KIND=dp), DIMENSION(-mleg:mleg,0:lleg)  :: rlmi     ! inverse of Rlm(t)
 REAL (KIND=dp), DIMENSION(-mleg:mleg,-mleg:mleg) :: krbr     ! Product K R**-1 B R, without alpha
 REAL (KIND=dp), DIMENSION(mleg)               :: Srm

! Source function contributions

 ! Source function expansion coeficients (as function of l and m respectively)
 REAL (KIND=dp), DIMENSION(0:npo)              :: sfl
 REAL (KIND=dp), DIMENSION(0:npo)              :: ssour
 REAL (KIND=dp), DIMENSION(-mleg:mleg,0:npo)   :: Git      ! Product R^-1*A^-1*g vector, same DIMENSIONs as Qit

 ! Boundary conditions related variables
 REAL (KIND=dp), DIMENSION(-nmui:nmui,-mleg:mleg) :: Tim      ! Auxiliary
 REAL (KIND=dp), DIMENSION(-nmui:nmui,-mleg:mleg) :: Wim      ! BC matrix
 REAL (KIND=dp), DIMENSION(-nmui:nmui)            :: Iinc     ! Incident radiation field

 ! Iteration related variables
 REAL (KIND=dp), DIMENSION(-mleg:mleg,0:npo)   :: Qit
 REAL (KIND=dp), DIMENSION(-mleg:mleg,0:npo)   :: Qit_o
 REAL (KIND=dp), DIMENSION(-mleg:mleg,0:npo)   :: Dit
 REAL (KIND=dp), DIMENSION(-mleg:mleg,0:npo)   :: GGit
 REAL (KIND=dp), DIMENSION(-mleg:mleg,0:npo)   :: Yit
 REAL (KIND=dp), DIMENSION(0:npo)              :: f0, f0_o
 REAL (KIND=dp), DIMENSION(-nmui:nmui)         :: Hit

! LAPACK related variables

 ! Eigenvalues only
 CHARACTER (LEN=1)                             :: JOBZ  = 'N'
 REAL (KIND=dp), DIMENSION(4*lleg+4)           :: WORK         ! Real work array
 INTEGER,        DIMENSION(lleg+1)             :: IWORK        ! Work array
 REAL (KIND=dp), DIMENSION(lleg+1,lleg+1)      :: aam, aamf
 REAL (KIND=dp), DIMENSION(lleg+1,1)           :: bbm, bbmf
 INTEGER                                       :: info
 CHARACTER (LEN=1)                             :: FACT  = 'E'  ! Equilibrate system
 CHARACTER (LEN=1)                             :: TRANS = 'N'  ! Transpose
 INTEGER                                       :: nrhs
 CHARACTER (LEN=1)                             :: EQUED        ! Kind of equilibration
 REAL (KIND=dp), DIMENSION(lleg+1)             :: R, C         ! Equilibration factors
 REAL (KIND=dp), DIMENSION(1)                  :: FERR, BERR   ! Errors (estimates)
 REAL (KIND=dp)                                :: RCOND        ! Condition number
 INTEGER, DIMENSION(lleg+1)                    :: indx

 ! Auxiliaries
 INTEGER                                       :: nl, nnit
 INTEGER                                       :: it, i
 INTEGER                                       :: i1, i2
 REAL (KIND=dp)                                :: taux1, taux2
 REAL (KIND=dp)                                :: taux3, taux4
 REAL (KIND=dp)                                :: wlth, wlthcm
 REAL (KIND=dp), DIMENSION(-mleg:mleg)         :: auxil

 ! Computation of emerging specific intensities
 REAL (KIND=dp), DIMENSION(0:lleg)             :: fl_0, fl_max

 REAL(KIND=dp)                                 :: ff1, ff2, ff3, ff4
 REAL(KIND=dp)                                 :: denom
 REAL(KIND=dp), DIMENSION(:,:), ALLOCATABLE    :: tmp_Jmp
 INTEGER                                       :: itmp, ntmp, ii

 REAL(KIND=dp)                                 :: lambda1, lambda2
 REAL(KIND=dp)                                 :: relax

 INTEGER, PARAMETER                            :: F_NEW_TRID = 1
!-----------------------------------------------------------------------------

  ! Number of Legendre coefficient (used for loops that run from 1 to lleg+1)
  nl = lleg + 1

  IF (ALLOCATED(g_absL)) THEN
     DEALLOCATE (g_absL)
  ENDIF
  ALLOCATE (g_absL(0:npo,0:nwlu))

  IF (ALLOCATED(intJm)) THEN
     DEALLOCATE (intJm, intJp)
  ENDIF
  ALLOCATE (intJm(0:npo,0:nwlu))
  ALLOCATE (intJp(0:npo,0:nwlu))

  IF (ALLOCATED(I_esc_0)) THEN
     DEALLOCATE(I_esc_0)
  ENDIF
  ALLOCATE (I_esc_0(0:nwlu,-nmui:nmui))
  IF (ALLOCATED(I_esc_max)) THEN
     DEALLOCATE(I_esc_max)
  ENDIF
  ALLOCATE (I_esc_max(0:nwlu,-nmui:nmui))

  ALLOCATE (albed(0:nwlu))
  ALLOCATE (kap_dus(0:npo,0:nwlu))
  ALLOCATE (sca_dus(0:npo,0:nwlu))
  ALLOCATE (p_0(0:lleg), q_0(0:lleg))

  p_0(:)         = 0.0_dp
  q_0(:)         = 0.0_dp
  sigl(:)        = 0.0_dp
  Tim(:,:)       = 0.0_dp
  Wim(:,:)       = 0.0_dp
  Iinc(:)        = 0.0_dp
  xkm(:)         = 0.0_dp
  gmm(:)         = 0.0_dp
  rlm(:,:)       = 0.0_dp
  rlmi(:,:)      = 0.0_dp
  Qit(:,:)       = 0.0_dp
  Qit_o(:,:)     = 0.0_dp
  Dit(:,:)       = 0.0_dp
  GGit(:,:)      = 0.0_dp
  Yit(:,:)       = 0.0_dp
  I_esc_0(:,:)   = 0.0_dp
  I_esc_max(:,:) = 0.0_dp
  f0(:)          = 0.0_dp
  f0_o(:)        = 0.0_dp

! STEP 0 - Initialise radiative quantities
!  (boundary conditions and optical depth grid)

  !--- PREPARE SOME FILES TO STORE QUANTITIES RELATIVE TO RADIATION FIELD ---

  ! uvspec.out is the total absorption through the cloud. multiply by
  !            incident radition field to compute spectrum.

  !--- Emission spectra (specific intensity) at Legendre decomposition angles
  fichier = TRIM(out_dir)//TRIM(modele)//'.Iesc'
  PRINT *, " Escaping intensities in:", TRIM(fichier)
  OPEN (36, file=fichier, status="unknown")
  ! Write comment lines of .Iesc file
  ! --------------------------------------------------------------------------------------
  ! The angles written correspond to "theta" in the document transfer_v6.pdf put in
  ! increasing order from 6.7 to 173.3 degrees (limits of the Legende decomposition
  !  0 < theta <  90 correspond to intensity radiated toward decreasing tau (increasing s)
  ! 90 < theta < 180 correspond to intensity radiated toward increasing tau (decreasing s)
  ! The resulting intensities are computed at the end of the function. BG 09/2013
  ! --------------------------------------------------------------------------------------
  WRITE (36,'("# Emission spectra (specific intensity in erg cm-2 s-1 sr-1 Ang-1)")')
  WRITE (36,'("# provided on observer and back side of the cloud")')
  WRITE (36,'("# at each angle theta = acos(mu) of the Legendre decomposition in degrees (see transfer_v6.pdf)")')
  WRITE (36,'("# On comment line below, ob means observer side")')
  WRITE (36,'("#                        ba means back side")')
  WRITE (36,'("#                        numeric values are angles in degrees")')
  WRITE (36,'("# Lambda[A]       ")',ADVANCE="NO")
  DO i = 1, nmui
     WRITE (36,'("I_ob:",F7.3,3x)',ADVANCE="NO") AngleLeg_deg(i)
  ENDDO
  DO i = nmui, 1, -1
     WRITE (36,'("I_ob:",F7.3,3x)',ADVANCE="NO") 180.0_dp - AngleLeg_deg(i)
  ENDDO
  DO i = 1, nmui
     WRITE (36,'("I_ba:",F7.3,3x)',ADVANCE="NO") AngleLeg_deg(i)
  ENDDO
  DO i = nmui, 1, -1
     WRITE (36,'("I_ba:",F7.3,3x)',ADVANCE="NO") 180.0_dp - AngleLeg_deg(i)
  ENDDO
  WRITE (36,*)

  !--- Extinction ------------------------------------
  fichier = TRIM(out_dir)//TRIM(modele)//'.uv'
  PRINT *, " Absorption in:", TRIM(fichier)
  OPEN (37, file=fichier, status="unknown")
  WRITE (37,'("#  W-L   EXP(-tau)  tau")')

  !--- Incident radiation field ----------------------
  fichier = TRIM(out_dir)//TRIM(modele)//'.flin'
  PRINT *, " External field:", TRIM(fichier)
  OPEN (60, file=fichier, status="unknown")
  WRITE(60,*) "9      # Number of columns"

  ! Interpolate intJmp on new grid (tau - lambda) to use in ABSGAZ before new radiation field is known
  ! Moved from ADD_WL
  IF (SIZE(intJmp,dim=1) /= npo .OR. SIZE(intJmp,dim=2) /= nwlg) THEN
     ALLOCATE (tmp_Jmp(0:npo,0:nwlg))
     ntmp = 0
     ! JLB - 13 May 2020
     tmp_Jmp(0,0) = intJmp(0,0)
     itmp = 0
     DO ii = 0, nwlu
        CALL LOC_IND (itmp, 0, SIZE(tmp_wl)-1, rf_wl%Val(ii), tmp_wl)
        denom = tmp_wl(itmp+1) - tmp_wl(itmp)
        ff1   = (tmp_wl(itmp+1) - rf_wl%Val(ii)) / denom
        ff3   = (rf_wl%Val(ii)  - tmp_wl(itmp)) / denom
        tmp_Jmp(0,ii) = intJmp(0,itmp) * ff1 + intJmp(0,itmp+1) * ff3
     ENDDO

     DO n = 1, npo
        CALL LOC_IND (ntmp, 0, SIZE(tmptau)-1, tau_o(n), tmptau)
        itmp = 0
        DO ii = 0, nwlu
           CALL LOC_IND (itmp, 0, SIZE(tmp_wl)-1, rf_wl%Val(ii), tmp_wl)
           denom = (tmptau(ntmp+1) - tmptau(ntmp)) * (tmp_wl(itmp+1) - tmp_wl(itmp))
           ff1   = (tmptau(ntmp+1) - tau_o(n)) * (tmp_wl(itmp+1) - rf_wl%Val(ii)) / denom
           ff2   = (tau_o(n)   - tmptau(ntmp)) * (tmp_wl(itmp+1) - rf_wl%Val(ii)) / denom
           ff3   = (tmptau(ntmp+1) - tau_o(n)) * (rf_wl%Val(ii)  - tmp_wl(itmp)) / denom
           ff4   = (tau_o(n)   - tmptau(ntmp)) * (rf_wl%Val(ii)  - tmp_wl(itmp)) / denom
           tmp_Jmp(n,ii) = intJmp(ntmp,itmp) * ff1 + intJmp(ntmp+1,itmp) * ff2 &
                         + intJmp(ntmp,itmp+1) * ff3 + intJmp(ntmp+1,itmp+1) * ff4
        ENDDO
     ENDDO
     DEALLOCATE (intJmp)
     ALLOCATE (intJmp(0:npo,0:nwlg))
     intJmp = tmp_Jmp
     DEALLOCATE (tmp_Jmp)
  ENDIF

! STEP 1a: Initialize absorption and scattering coefficients
  ! Compute absorption by gas
  CALL ABSGAZ (nwlu)

  IF (ALLOCATED(intJmp)) THEN
     DEALLOCATE(intJmp)
  ENDIF
  ALLOCATE (intJmp(0:npo,0:nwlg))

  WRITE(iscrn,'("       Point             Angstroms       erg cm-2 s-1 A-1 sr-1")')
  ! Example on screen : iwlg =   100 wlth =        925.259  intJmp: edge =   4.5370E-08 middle =  9.0686E-26

  p_0(0) = 1.0_dp
  q_0(1) = p_0(0) / 2.0_dp
  DO l = 1, (lleg-1)/2
     p_0(2*l) = - (2.0_dp*l-1) * p_0(2*l-2) / (2.0_dp*l)
     q_0(2*l+1) = p_0(2*l) / (2.0_dp*l+2.0_dp)
  ENDDO

  ! All wavelengths are independent
  DO iwlg = 0, nwlu

     ! Henyey-Greenstein g coefficient:
     !     Use first point arbitrarily
     ggHG = du_prop%gcos(iwlg,0)
     !ggHG = 1.0_dp
     ! b - sigl = gg**l (Formule 48, Henyey-Greenstein phase function)
     sigl(0) = 1.0_dp
     DO l = 1, lleg
        sigl(l) = sigl(l-1) * ggHG
     ENDDO

     ! Reset to 0 most variables
     bsa = 0.0_dp
     hl = 0.0_dp
     wr = 0.0_dp
     wi = 0.0_dp
     tlm = 0.0_dp
     tlp = 0.0_dp
     dtl = 0.0_dp
     dta = 0.0_dp
     u_im = 0.0_dp
     u_ip = 0.0_dp
     EsEm = 0.0_dp
     EsEp = 0.0_dp

     wlth = rf_wl%Val(iwlg)
     wlthcm = wlth * 1.0e-8_dp

     ! Compute local values of absorption coef for dust and gas and dust scatering coef
     ! Compute effective albedo !
     DO n = 0, npo
        IF (F_DU_PROP == 0) THEN
           n1 = 0
        ELSE
           n1 = n
        ENDIF
        kap_dus(n,iwlg) = du_prop%abso(iwlg,n1) * dens_o(n)
        sca_dus(n,iwlg) = du_prop%scat(iwlg,n1) * dens_o(n)
        emm_dus(n)      = du_prop%emis(iwlg,n)  * dens_o(n)
     ENDDO
     ! Using value at n = 1 is arbitrary since it does not depend on depth
     albed(iwlg) = sca_dus(1,iwlg) / (kap_dus(1,iwlg) + sca_dus(1,iwlg))
     ! During first iteration, we do not use any gas contribution.
     ! Then, continuous contribution from atoms and ions are pluged first, then the lines.
     ! This is done in ABSGAZ
     IF (ifaf > 1) THEN
        alp_gas(0:npo) = g_absL(0:npo,iwlg) / (kap_dus(0:npo,iwlg) + sca_dus(0:npo,iwlg))
     ELSE
        alp_gas = 0.0_dp
     ENDIF

     ! Compute optical depth grid
     !      Convert Delta Tau_v (dust at V) to Delta tau_lambda (dust at Lambda)
     !      17 Oct 2011: use dd_cm (ds) for integration
     !      26 VII 2019: Change indexing to right side of interval
!$OMP PARALLEL
!$OMP DO SCHEDULE(DYNAMIC)
     DO n = 0, npo-1
        dtl(n+1) = ((kap_dus(n,iwlg) + sca_dus(n,iwlg)) + (kap_dus(n+1,iwlg) + sca_dus(n+1,iwlg))) * 0.5_dp * dd_cm(n+1)
     ENDDO
!$OMP END DO
!$OMP END PARALLEL
     ! Optical depth is computed from both sides to avoid unnecessary substractions and loss of precision
     tlm(0) = 0.0_dp
     DO n = 0, npo-1
        tlm(n+1) = tlm(n) + dtl(n+1)
     ENDDO
     tlp(npo) = 0.0_dp
     DO n = npo, 1, -1
        tlp(n-1) = tlp(n) + dtl(n)
     ENDDO

     ! d - Compute integrated contribution of gas (lines)
     !     26 VII 2019 - Change indexing to right side of interval
!$OMP PARALLEL
!$OMP DO SCHEDULE(DYNAMIC)
     DO n = 0, npo-1
        dta(n+1) = (alp_gas(n) + alp_gas(n+1)) * 0.5_dp * dtl(n+1)
     ENDDO
!$OMP END DO
!$OMP END PARALLEL
     A_gim(0) = 0.0_dp
     DO n = 0, npo-1
        A_gim(n+1) = A_gim(n) + dta(n+1)
     ENDDO
     A_gip(npo) = 0.0_dp
     DO n = npo, 1, -1
        A_gip(n-1) = A_gip(n) + dta(n)
     ENDDO
     ! output attenuation by dust and gas through whole cloud
     WRITE (37,"(1p,5e17.7E3)") wlth, EXP(-tlm(npo)), tlm(npo), EXP(-A_gim(npo)), A_gim(npo)

! STEP 2a - Compute matrix A

     ! Formulae A2 and A3 of Roberge:
     hl(0)  = 1.0_dp - albed(iwlg)
     DO l = 1, lleg
        hl(l)  = REAL(2*l+1,dp) * (1.0_dp - albed(iwlg) * sigl(l))
     ENDDO
     ! Appendix C of Goicoechea & Le Bourlot (2007)
     DO l = 1, lleg
        wi(l) = REAL(l,dp) / SQRT(hl(l-1)*hl(l))
     ENDDO

! STEP 2b - Diagonalize. Use new routine TRID0 by default.

     IF (F_NEW_TRID == 0) THEN
        wr = 0.0_dp
        ! b - Compute eigenvalues (tridiagonal symetric matrix) - LAPACK
        CALL DSTEVD(JOBZ, nl, wr, wi, aam, nl, WORK, 4*nl, IWORK, nl, info )

        ! order eigenvalues by decreasing order
        wi = wr
        DO l = 1, mleg
           wr(l) = wi(nl+1-l)
        ENDDO
        ! Enforce symetry
        do l = mleg+1,nl
           wr(l) = -wr(2*mleg-l+1)
        enddo
     ELSE
        CALL TRID0 (nl,wi,wr)
        do l = 1, mleg
           wi(l) = wr(mleg+1-l)
        enddo
        do l = mleg+1,nl
           wi(l) = - wr(l-mleg)
        enddo
        wr = wi
     ENDIF

     ! km = 1 / lamda_m in increasing order
     DO m = 1, mleg
        xkm(m) = 1.0_dp / wr(m)
        xkm(-m) = -xkm(m)
     ENDDO

     ! Compute Rlm: Formule A9 from Roberge 1983
     ! Eigenvectors computed only from recurrence relations.
     DO m = 1, mleg
        rlm(0,m)  = 1.0_dp
        rlm(0,-m) = 1.0_dp
        rlm(1,m)  = (1.0_dp - albed(iwlg)) / xkm(m)
        rlm(1,-m) = -rlm(1,m)                             !jrg
        DO l = 2, lleg
           rlm(l,m) = (hl(l-1) * rlm(l-1,m) - REAL(l-1,dp) * xkm(m) * rlm(l-2,m)) &
                    / (REAL(l,dp) * xkm(m))
           rlm(l,-m) = xmunl(l) * rlm(l,m)
        ENDDO
     ENDDO

! STEP 2c - Compute Srm - used for mean intensity

     DO m = 1, mleg
        taux1 = 0.0_dp
        DO l = 1, lleg, 2
           taux1 = taux1 + REAL(2*l+1,dp) * q_0(l) * rlm(l,m)
        ENDDO
        Srm(m) = taux1
     ENDDO

! STEP 3 - Compute R**-1

     CALL AUX_Rlmi_CteG(rlm, hl, rlmi)

! STEP 4 - Compute source function contribution g = R^-1*A^-1*g~ (in notes)

     ! STEP 4a: Compute g tilde:
     ! METHOD B1: approx. to continuum thermal isotropic emission
     ! JLB - 8 III 2013 - Add a low intensity  background source
     sfl(0:npo) = 1.0e-40_dp + emm_dus(0:npo) / (kap_dus(0:npo,iwlg) + sca_dus(0:npo,iwlg))

     ! Initialize source function here for EDDINGTON below
     ssour(0:npo) = sfl(0:npo)
     ! For l = 0, sigl(0) = 1
     sfl(0:npo) = -sfl(0:npo) / (1.0_dp - albed(iwlg))

     ! STEP 4b - Compute vector g = R^-1*A^-1*g~ (this is Git)
!$OMP PARALLEL PRIVATE(l,m)
!$OMP DO SCHEDULE(DYNAMIC)
     DO n = 0, npo
        DO m = -mleg, -1
           Git(m,n) = sfl(n) * xkm(m) * rlmi(m,0)
        ENDDO
        DO m = 1, mleg
           Git(m,n) = sfl(n) * xkm(m) * rlmi(m,0)
        ENDDO
     ENDDO
!$OMP END DO
!$OMP END PARALLEL

! STEP 5 - S = K * R**-1 * B * R

     ! STEP 5a - Diagonal matrix B
     bsa(0) = albed(iwlg) / (1.0_dp - albed(iwlg))
     DO l = 1, lleg
        bsa(l) = albed(iwlg) * sigl(l) / (1.0_dp - albed(iwlg) * sigl(l))
     ENDDO

     ! STEP 5b - Multiply
     CALL AUX_Slm_CteG(xkm, rlmi, bsa, rlm, krbr)

     ! STEP 5c - Put diagonal in Emk (gmm is diagonal of matrix Gamma)
     gmm = 0.0_dp
     DO m = 1, mleg
        gmm(-m) = krbr(-m,-m)
        gmm(m)  = krbr(m,m)
        krbr(-m,-m) = 0.0_dp
        krbr(m,m)   = 0.0_dp
     ENDDO
     DO m = 1, mleg
        vvmk(0:npo,-m)  = 1.0_dp / (xkm(-m) + (xkm(-m) + gmm(-m)) * alp_gas(0:npo))
        vvmk(0:npo,m)   = 1.0_dp / (xkm(m)  + (xkm(m)  + gmm(m))  * alp_gas(0:npo))
     ENDDO
     DO m = 1, mleg
        vmkg(0:npo,-m) =  Git(-m,0:npo) * vvmk(0:npo,-m)
        vmkg(0:npo,m)  =  Git(m,0:npo)  * vvmk(0:npo,m)
     ENDDO

! STEP 6 Compute E_j / E_{j-1}
     ! STEP 6a - u: denominators
     DO m = 1, mleg
        u_im(0,m)       = 0.0_dp
        u_im(1:npo,m)   = -xkm(-m) * dtl(1:npo) - (xkm(-m) + gmm(-m)) * dta(1:npo)
        u_ip(npo,m)     = 0.0_dp
        u_ip(0:npo-1,m) =  xkm(m)  * dtl(1:npo) + (xkm(m)  + gmm(m))  * dta(1:npo)
     ENDDO
     EsEm(:,:) = EXP(-u_im(:,:))
     EsEp(:,:) = EXP(-u_ip(:,:))

     ! STEP 6b - Compute Emkt_m and Emkt_p
     DO m = 1, mleg
        Emk_m(m,:) = EXP(xkm(-m) * tlm(:) + (gmm(-m) + xkm(-m)) * A_gim(:))
        Emk_p(m,:) = EXP(-xkm(m) * tlp(:) - (gmm(m)  + xkm(m))  * A_gip(:))
     ENDDO
     DO m = 1, mleg
        Emkt_m(m)  = Emk_m(m,npo)
        Emkt_p(m)  = Emk_p(m,0)
     ENDDO

! STEP 7 - Compute Tim

!  xx  $OMP PARALLEL PRIVATE(taux1,taux2,taux3,taux4,i)
!  xx  $OMP DO SCHEDULE(DYNAMIC)
     ! Try to split these loops to optimize memory acces (?)
     DO m = 1, mleg
        DO i = 1, nmui
           taux1 = 0.0_dp
           taux2 = 0.0_dp
           taux3 = 0.0_dp
           taux4 = 0.0_dp
           DO l = 0, lleg
              taux1 = taux1 + (2.0_dp * l + 1.0_dp) * rlm(l,-m) * pli(l,-i)
              taux2 = taux2 + (2.0_dp * l + 1.0_dp) * rlm(l,m)  * pli(l,-i)
              taux3 = taux3 + (2.0_dp * l + 1.0_dp) * rlm(l,-m) * pli(l,i)
              taux4 = taux4 + (2.0_dp * l + 1.0_dp) * rlm(l,m)  * pli(l,i)
           ENDDO
           Tim(-i,-m) = taux1
           Tim(-i,m)  = taux2
           Tim(i,-m)  = taux3
           Tim(i,m)   = taux4
        ENDDO
     ENDDO
!  xx  $OMP END DO
!  xx  $OMP END PARALLEL

! STEP 8 - Compute Wim (Boundary conditions matrix)

!  xx  $OMP PARALLEL PRIVATE(i)
!  xx  $OMP DO SCHEDULE(DYNAMIC)
     DO m = 1, mleg
        DO i = 1, nmui
           Wim(-i,-m) = Tim(-i,-m)
           Wim(-i,m)  = Tim(-i,m) * Emkt_p(m)
           Wim(i,-m)  = Tim(i,-m) * Emkt_m(m)
           Wim(i,m)   = Tim(i,m)
        ENDDO
     ENDDO
!  xx  $OMP END DO
!  xx  $OMP END PARALLEL

! STEP 9 - Set boundary conditions

     WRITE (60, '(1p,7e15.6E3)') wlth, I_incr_m(iwlg), rf_incIm%Val(iwlg) &
                               , I_incr_p(iwlg), rf_incIp%Val(iwlg) &
                               , sfl(0), sfl(npo)

     ! 19 XII 2005 - Use "REAL" units (specific intensity)
     Iinc(-nmui:-1) = rf_incIm%Val(iwlg)
     Iinc(1:nmui)   = rf_incIp%Val(iwlg)

     ! Add potential // source (star)
     ! JLB - 4 IX 08 : Checked after outgoing intensities have been computed
     !                 xmui_M(1) = 0.99... so theta = 0 is on +1 and theta = pi on -1
     Iinc(-1) = rf_incIm%Val(iwlg) + I_incr_m(iwlg)
     Iinc(1)  = rf_incIp%Val(iwlg) + I_incr_p(iwlg)

! STEP 10a - Initialize Yit^(0)

     ! This routine computes a first approximation of the radiation field using only L = 1
     ! This is a kind of "Eddington approximation", that can be used for debug
     ! or for initializing the full iteration.
     ! ssour must be initialized before rescale of sfl
     !ssour(0:npo) = sfl(0:npo)
     CALL EDDINGTON (dtl, alp_gas, albed(iwlg), ggHG, ssour &
                  , Iinc(-1), Iinc(1), Yit, rlmi)
     ! Yit = 0 could be used as initialization

! STEP 10b - Compute GG (G in notes) - Does not need to iterate.

     GGit(-mleg:-1,0) = 0.0_dp
     GGit(1:mleg,npo) = 0.0_dp
!!x !$OMP PARALLEL PRIVATE(n,taux1,taux2)
!!x !$OMP DO SCHEDULE(DYNAMIC)
     DO m = 1, mleg
        DO n = 1, npo
           taux1 = vmkg(n-1,-m) * (1.0_dp - EsEm(n,m) - u_im(n,m) * EsEm(n,m))
           taux2 = vmkg(n,-m)   * (1.0_dp - EsEm(n,m) - u_im(n,m))
           GGit(-m,n) = EsEm(n,m) * GGit(-m,n-1) + (taux2 - taux1) / u_im(n,m)
        ENDDO
        DO n = npo-1, 0, -1
           taux1 = vmkg(n+1,m) * (1.0_dp - EsEp(n,m) - u_ip(n,m) * EsEp(n,m))
           taux2 = vmkg(n,m)   * (1.0_dp - EsEp(n,m) - u_ip(n,m))
           GGit(m,n) = EsEp(n,m) * GGit(m,n+1) + (taux2 - taux1) / u_ip(n,m)
        ENDDO
     ENDDO
!!x !$OMP END DO
!!x !$OMP END PARALLEL

! STEP 10c - Compute Q^(0) from Yit^(0)

     !Qit  = 0.0_dp ! If Yit = 0
     Qit_o = 0.0_dp
     Qit   = 0.0_dp
     DO n = 0, npo
        Qit(:,n) = alp_gas(n) * MATMUL(krbr,Yit(:,n))
     ENDDO

     !nnit = 200
     !nnit = 20
     !nnit = 5
     ! How many time should we iterate?
     nnit = MAX(5, ifaf)

     DO it = 1, nnit ! begin loop :: 101

! STEP 11 - Compute D

        Dit(-mleg:-1,0) = 0.0_dp
        Dit(1:mleg,npo) = 0.0_dp
!!x !$OMP PARALLEL PRIVATE(n,taux1,taux2)
!!x !$OMP DO SCHEDULE(DYNAMIC)
        DO m = 1, mleg
           vmkd(0:npo) = Qit(-m,0:npo) * vvmk(0:npo,-m)
           DO n = 1, npo
              taux1 = vmkd(n-1) * (1.0_dp - EsEm(n,m) - u_im(n,m) * EsEm(n,m))
              taux2 = vmkd(n)   * (1.0_dp - EsEm(n,m) - u_im(n,m))
              Dit(-m,n) = EsEm(n,m) * Dit(-m,n-1) + (taux2 - taux1) / u_im(n,m)
           ENDDO
           vmkd(0:npo) = Qit(m,0:npo) * vvmk(0:npo,m)
           DO n = npo-1, 0, -1
              taux1 = vmkd(n+1) * (1.0_dp - EsEp(n,m) - u_ip(n,m) * EsEp(n,m))
              taux2 = vmkd(n)   * (1.0_dp - EsEp(n,m) - u_ip(n,m))
              Dit(m,n) = EsEp(n,m) * Dit(m,n+1) + (taux2 - taux1) / u_ip(n,m)
           ENDDO
        ENDDO
!!x !$OMP END DO
!!x !$OMP END PARALLEL

! STEP 12 - Compute Hit

        Hit = 0.0_dp
!  xx  $OMP PARALLEL PRIVATE(taux1,taux2,m)
!  xx  $OMP DO SCHEDULE(DYNAMIC)
        DO i = 1, nmui
           taux1 = 0.0_dp
           taux2 = 0.0_dp
           DO m = 1, mleg
              taux1 = taux1 + (GGit(-m,npo) + Dit(-m,npo)) * Tim(i,-m)
              taux2 = taux2 + (GGit(m,0)    + Dit(m,0))    * Tim(-i,m)
           ENDDO
           Hit(-i) = Iinc(-i) - taux2
           Hit(i)  = Iinc(i)  - taux1
        ENDDO
!  xx  $OMP END DO
!  xx  $OMP END PARALLEL

! STEP 13 - Boundary conditions - solve using LAPACK

        aam(1:nmui,1:mleg)        = Wim(-nmui:-1,-mleg:-1)
        aam(1:nmui,mleg+1:2*mleg) = Wim(-nmui:-1,1:mleg)
        aam(nmui+1:nl,1:mleg)     = Wim(1:nmui,-mleg:-1)
        aam(nmui+1:nl,mleg+1:nl)  = Wim(1:nmui,1:mleg)

        bbm(1:nmui,1)    = Hit(-nmui:-1)
        bbm(nmui+1:nl,1) = Hit(1:nmui)

        nrhs = 1
        CALL DGESVX (FACT, TRANS, nl, nrhs, aam, nl, aamf, nl, indx, EQUED, &
             R, C, bbm, nl, bbmf, nl, RCOND, FERR, BERR, WORK, IWORK, info)

        Camp(-nmui:-1) = bbmf(1:nmui,1)
        Camp(1:nmui)   = bbmf(nmui+1:nl,1)

! STEP 14 - Compute new Y

        ! 22 III 2018: try to use relaxation to damp oscillations
        ! 4 IV 2018: Is this still necessary? V 2018: Yes !
        IF (ifaf < 2) THEN
           relax = 1.0_dp
        ELSE
           relax = 0.6_dp
        ENDIF
!$OMP PARALLEL PRIVATE(m,taux1)
!$OMP DO SCHEDULE(DYNAMIC)
        DO n = 0, npo
           DO m = 1, mleg
              Yit(-m,n) = (1.0_dp - relax) *  Yit(-m,n) + relax * (Camp(-m) * Emk_m(m,n) + GGit(-m,n) + Dit(-m,n))
              Yit(m,n)  = (1.0_dp - relax) *  Yit(m,n)  + relax * (Camp(m)  * Emk_p(m,n) + GGit(m,n)  + Dit(m,n))
           ENDDO
           ! JLB - Lift sign degeneracy. taux1 is f0
           taux1 = 0.0_dp
           DO m = 1, mleg
              taux1 = taux1 + Yit(-m,n) + Yit(m,n)
           ENDDO
           f0(n) = taux1
           IF (taux1 < 0.0_dp) THEN
              !print *, "  +++ been there!", iwlg, n
              Yit(:,n) = - Yit(:,n)
           ENDIF
        ENDDO
!$OMP END DO
!$OMP END PARALLEL

     !if (iwlg == 100 .AND. ifaf == 2) then
     !   open (63, file="YInt_It_N"//TRIM(ADJUSTL(ii2c(it)))//".dat", status="unknown")
     !   write (63,*) "#", iwlg, "  Lambda:", wlth
     !   write (63,*) "# Camp:", (Camp(m),  m = -mleg, -1), (Camp(m) , m = 1, mleg)
     !   do n = 0, npo
     !      !                                  2 - 11                      12 - 21
     !      !                                 22 - 31                      32 - 41
     !      !                                 42 - 51                      52 - 61
     !      !                                 62 - 71                      72 - 81
     !      !                                   82
     !      write (63,*) tautav*tau_o(n), (Yit(m,n), m = -mleg, -1), (Yit(m,n), m = 1, mleg) &
     !                                  , (Dit(m,n), m = -mleg, -1), (Dit(m,n), m = 1, mleg) &
     !                                  , (GGit(m,n), m = -mleg, -1), (GGit(m,n), m = 1, mleg) &
     !                                  , (Emk_m(-m,n), m = -mleg, -1), (Emk_p(m,n), m = 1, mleg) &
     !                                  , f0(n) &
     !                                  , (Yit_o(m,n), m = -mleg, -1), (Yit_o(m,n), m = 1, mleg)
     !   enddo
     !   close (63)
     !   !stop
     !endif

! STEP 15 - Compute new Q

        Qit_o = Qit
        Qit = 0.0_dp
        DO n = 0, npo
           Qit(:,n) = alp_gas(n) * MATMUL(krbr,Yit(:,n))
        ENDDO

! STEP 16 - convergence ?

  !!    auxil(0) = 0.0_dp
  !!    taux2 = 0.0_dp
  !!    DO n = 0, npo
  !!       DO m = -mleg, -1
  !!          IF (Qit(m,n) /= 0.0_dp) THEN
  !!             auxil(m) = ABS((Qit(m,n)-Qit_o(m,n))/Qit(m,n))
  !!          ELSE
  !!             auxil(m) = 0.0_dp
  !!          ENDIF
  !!       ENDDO
  !!       DO m = 1, mleg
  !!          IF (Qit(m,n) /= 0.0_dp) THEN
  !!             auxil(m) = ABS((Qit(m,n)-Qit_o(m,n))/Qit(m,n))
  !!          ELSE
  !!             auxil(m) = 0.0_dp
  !!          ENDIF
  !!       ENDDO
  !!       taux2 = taux2 + SUM(auxil)
  !!    ENDDO
        ! Convergence is tested only on the variation of the total intensity in the cloud.
        ! This term should be dominated by edges.
        taux2 = SUM(ABS(f0_o-f0)) / SUM(f0)
        f0_o = f0

        IF (ifaf == 0) THEN
           EXIT
        ELSE IF (taux2 < 1.0e-3_dp .AND. it > 1) THEN
           EXIT      !EXITs sub-iteration loop
        ENDIF

     ENDDO        ! Exits loop to find q  :: 101
     IF (taux2 > 1.0e-3_dp .AND. ifaf > 0) THEN
        PRINT *, " Pb:", iwlg, wlth, taux2
     ENDIF

! STEP 17 - Compute radiation field - mean intensity
!           intJm is 0.5 SUM(-1->0) of I
!           intJp is 0.5 SUM(0->1) of I
!           intJmp is 0.5 SUM(-1->1) of I
!           Thus intJm + intJp = J in erg cm-2 s-1 A-1 sterad-1
!           one gets u in erg cm-3 A-1 by u = J * 4 pi / c

!$OMP PARALLEL PRIVATE(taux1,taux2,taux3,m)
!$OMP DO SCHEDULE(DYNAMIC)
     DO n = 0, npo
        taux1 = 0.0_dp
        taux2 = 0.0_dp
        taux3 = 0.0_dp
        DO m = 1, mleg
           taux1 = taux1 + Yit(-m,n) * (1.0_dp + Srm(m)) + Yit(m,n) * (1.0_dp - Srm(m))
           taux2 = taux2 + Yit(-m,n) * (1.0_dp - Srm(m)) + Yit(m,n) * (1.0_dp + Srm(m))
           taux3 = taux3 + Yit(-m,n) + Yit(m,n)
        ENDDO

        ! 19 XII 2005 - Change normalisation
        intJm(n,iwlg) = 0.5_dp * taux1
        intJp(n,iwlg) = 0.5_dp * taux2
        intJmp(n,iwlg) = taux3
     ENDDO
!$OMP END DO
!$OMP END PARALLEL

     IF (MOD(iwlg,100) == 0) THEN
        WRITE(iscrn,'("iwlg = ",I6," wlth[A] = ",F16.3,"  intJmp: edge = ",1p,E12.4," middle =",1p,E12.4)') &
              iwlg, wlth, intJmp(0,iwlg), intJmp(INT(npo/2),iwlg)
     ENDIF

     ! Compute emerging specific intensities
     DO l = 0, lleg
        fl_0(l)   = 0.0_dp
        fl_max(l) = 0.0_dp
        DO m = 1, mleg
           fl_0(l)   = fl_0(l)   + rlm(l,m) * Yit(m,0)   + rlm(l,-m) * Yit(-m,0)
           fl_max(l) = fl_max(l) + rlm(l,m) * Yit(m,npo) + rlm(l,-m) * Yit(-m,npo)
        ENDDO
     ENDDO

     DO i = 1, nmui
        I_esc_0(iwlg,-i)   = 0.0_dp
        I_esc_0(iwlg,i)    = 0.0_dp
        I_esc_max(iwlg,-i) = 0.0_dp
        I_esc_max(iwlg,i)  = 0.0_dp
        DO l = 0, lleg
           I_esc_0(iwlg,-i)   = I_esc_0(iwlg,-i)   + (2.0_dp*l+1.0_dp) * fl_0(l)   * pli(l,-i)
           I_esc_0(iwlg,i)    = I_esc_0(iwlg,i)    + (2.0_dp*l+1.0_dp) * fl_0(l)   * pli(l,i)
           I_esc_max(iwlg,-i) = I_esc_max(iwlg,-i) + (2.0_dp*l+1.0_dp) * fl_max(l) * pli(l,-i)
           I_esc_max(iwlg,i)  = I_esc_max(iwlg,i)  + (2.0_dp*l+1.0_dp) * fl_max(l) * pli(l,i)
        ENDDO
     ENDDO
     WRITE (36,'(1p,163e15.6)') wlth, (I_esc_0(iwlg,i), i=1,nmui), (I_esc_0(iwlg,i), i=-nmui,-1) &
                              , (I_esc_max(iwlg,i), i=1,nmui), (I_esc_max(iwlg,i), i=-nmui,-1)

  ENDDO    !end wavelenght loop :: 99  !too slow !!

  CLOSE (36)
  CLOSE (37)
  CLOSE (60)

  lambda1 = 3.0e+0_dp
  lambda2 = 1.0e+3_dp
  Icont_3_1000   = INTEG_TAB(I_esc_0(0:nwlu,1),nwlu,lambda1,lambda2)
  lambda1 = 2.5e+1_dp
  lambda2 = 6.0e+1_dp
  Icont_25_60    = INTEG_TAB(I_esc_0(0:nwlu,1),nwlu,lambda1,lambda2)
  lambda1 = 6.0e+1_dp
  lambda2 = 2.0e+2_dp
  Icont_60_200   = INTEG_TAB(I_esc_0(0:nwlu,1),nwlu,lambda1,lambda2)
  lambda1 = 2.0e+2_dp
  lambda2 = 1.0e+3_dp
  Icont_200_1000 = INTEG_TAB(I_esc_0(0:nwlu,1),nwlu,lambda1,lambda2)
  lambda1 = 6.0e+1_dp
  lambda2 = 1.0e+3_dp
  Icont_60_1000  = INTEG_TAB(I_esc_0(0:nwlu,1),nwlu,lambda1,lambda2)

  IF (F_W_RF_ALB == 1) THEN
     ! Write here full radiation field with effective albedo (JLB - 18 Oct 2012)
     !fichier = TRIM(out_dir)//TRIM(modele)//'.rf_alb'
     fichier = TRIM(out_dir)//TRIM(modele)//'.rf_alb'//TRIM(ADJUSTL(ii2c(ifaf)))
     OPEN(61, file=fichier, status="unknown" , form = 'unformatted')
     WRITE (61) rf_wl%upper-rf_wl%lower+1 ! number of pixels in wavelength grid
     WRITE (61) rf_wl%Val                 ! Wavelength grid in Angstroms
     DO n = 0, npo
        WRITE (61) tau_o(n)*tautav, intJmp(n,0:nwlu) * J_to_u, g_absL(n,:), kap_dus(n,:), sca_dus(n,:)
     ENDDO
     CLOSE (61)
  ENDIF

  ! Free memory
  DEALLOCATE (albed)
  DEALLOCATE (kap_dus, sca_dus)
  DEALLOCATE (p_0, q_0)

  !====================================================================================
  ! COMPUTE THE RADIATION FIELD INTENSITY AND ENERGY FROM THE RIGHT AND LEFT SIDE
  ! (USE RF_U AS AUXILIARY VARIABLE HERE)
  ! The goal is to obtain mean_rad(1,2),  int_rad(1,2),  rad(m,p),  and flu_rad(1,2)
  ! The goal is to obtain mean_xray(1,2), int_xray(1,2), xray(m,p), and flu_xray(1,2)
  ! note BG 04/2013: is it important to do right first ?
  !====================================================================================
  CALL RF_CREATE(rf_lamI,0,nwlu)
  CALL RF_CREATE(rf_tmp,1,2)

  !--------------------------------------------------------------------------------
  ! Integration of UV radiation field from Lyman up to Habing limit
  ! (wl_lym = 911 Å ; wl_hab = 2400 Å) - BG 04/2013
  !--------------------------------------------------------------------------------
  CALL RF_LOCATE(rf_wl, wl_lym, i1)
  rf_tmp%Val(1) = 1.0_dp
  rf_tmp%Ord(1) = rf_wl%Ord(i1)
  CALL RF_LOCATE(rf_wl, wl_hab, i2)
  rf_tmp%Val(2) = 1.0_dp
  rf_tmp%Ord(2) = rf_wl%Ord(i2)
  ! -------------------
  rf_u%Val(0:nwlu)    = intJp(npo,0:nwlu)
  mean_rad2           = SECT_INT(rf_tmp, rf_u, wl_pointer, wl_hab)
  rf_u%Val(0:nwlu)    = intJm(0  ,0:nwlu)
  mean_rad1           = SECT_INT(rf_tmp, rf_u, wl_pointer, wl_hab)
  ! -------------------
  rf_u%Val(0:nwlu)    = intJmp(npo,0:nwlu) * J_to_u
  int_rad2            = SECT_INT(rf_tmp, rf_u, wl_pointer, wl_hab)
  radp                = int_rad2 / int_rad0
  rf_u%Val(0:nwlu)    = intJmp(0  ,0:nwlu) * J_to_u
  int_rad1            = SECT_INT(rf_tmp, rf_u, wl_pointer, wl_hab)
  radm                = int_rad1 / int_rad0
  int_rad(0)          = SECT_INT(rf_tmp, rf_u, wl_pointer, wl_hab)
  ! -------------------
  rf_lamI%Val(0:nwlu) = intJmp(npo,0:nwlu) * J_to_u* rf_wl%Val(0:nwlu)
  flu_rad2            = SECT_INT(rf_tmp, rf_lamI, wl_pointer, wl_hab) / (1.0e8_dp * xhp)
  rf_lamI%Val(0:nwlu) = intJmp(0  ,0:nwlu) * J_to_u* rf_wl%Val(0:nwlu)
  flu_rad1            = SECT_INT(rf_tmp, rf_lamI, wl_pointer, wl_hab) / (1.0e8_dp * xhp)

  !--------------------------------------------------------------------------------
  ! Write results of these integrations on the screen and in the .def file
  !--------------------------------------------------------------------------------
  PRINT *, " "
  PRINT *, " Scaled field at cloud edge"
  WRITE (iscrn, '("  U2_UV:",1pe12.4," erg cm-3,     ","from",0pf9.2," to",0pf9.2," Angstrom")') int_rad2,  wl_lym,  wl_hab
  WRITE (iscrn, '("  F2_UV:",1pe12.4," phot cm-2 s-1,","from",0pf9.2," to",0pf9.2," Angstrom")') flu_rad2,  wl_lym,  wl_hab
  WRITE (iwrit2,'("  U2_UV:",1pe12.4," erg cm-3,     ","from",0pf9.2," to",0pf9.2," Angstrom")') int_rad2,  wl_lym,  wl_hab
  WRITE (iwrit2,'("  F2_UV:",1pe12.4," phot cm-2 s-1,","from",0pf9.2," to",0pf9.2," Angstrom")') flu_rad2,  wl_lym,  wl_hab
  !--------------------------------------------------------------------------------
  WRITE (iscrn, '("  U1_UV:",1pe12.4," erg cm-3,     ","from",0pf9.2," to",0pf9.2," Angstrom")') int_rad1,  wl_lym,  wl_hab
  WRITE (iscrn, '("  F1_UV:",1pe12.4," phot cm-2 s-1,","from",0pf9.2," to",0pf9.2," Angstrom")') flu_rad1,  wl_lym,  wl_hab
  WRITE (iwrit2,'("  U1_UV:",1pe12.4," erg cm-3,     ","from",0pf9.2," to",0pf9.2," Angstrom")') int_rad1,  wl_lym,  wl_hab
  WRITE (iwrit2,'("  F1_UV:",1pe12.4," phot cm-2 s-1,","from",0pf9.2," to",0pf9.2," Angstrom")') flu_rad1,  wl_lym,  wl_hab

  !---------------------------------------------------------------------
  ! Compute the effective UV scaling factors in units of G0 and FX0/clum
  !---------------------------------------------------------------------
  G0m = int_rad1  / u_habing
  G0p = int_rad2  / u_habing
  WRITE(iwrit2,'(/," U0:",1pe13.4, " erg cm-3")') int_rad0
  PRINT *, " "
  WRITE (iscrn,'("  G_0m :",1pe12.4," (in Habing units)")') G0m
  WRITE (iscrn,'("  G_0p :",1pe12.4," (in Habing units)")') G0p
  WRITE(iwrit2,'(" G_0m =", 1pe12.3, " in Habing units",/)') G0m
  WRITE(iwrit2,'(" G_0p =", 1pe12.3, " in Habing units",/)') G0p
  PRINT *, " "
  PRINT *, " Effective Radiation field strength AT the edge:"
  WRITE (iscrn, '("  left :",1pe12.4," ",A65)') radm,  ISRF_name
  WRITE (iscrn, '(" right :",1pe12.4," ",A65)') radp,  ISRF_name
  WRITE (iwrit2,'("  left :",1pe12.4," ",A65)') radm,  ISRF_name
  WRITE (iwrit2,'(" right :",1pe12.4," ",A65)') radp,  ISRF_name
  PRINT *, " "

  CALL RF_FREE (rf_tmp)
  CALL RF_FREE (rf_lamI)

END SUBROUTINE ETIN90_CteG

!---------------------------------------------------------------------------
SUBROUTINE EDDINGTON (dtl, alphg, albed, ggHG, ssour, Im, Ip, Yit, rlmi)

   USE PXDR_FLAGS
   USE PXDR_CONSTANTES
   USE PXDR_PROFIL
   USE PXDR_AUXILIAR
   USE PXDR_RF_TOOLS

   IMPLICIT NONE

   REAL (KIND=dp), DIMENSION(0:npo), INTENT (IN)       :: dtl
   REAL (KIND=dp), DIMENSION(0:npo), INTENT (IN)       :: alphg
   REAL (KIND=dp),                   INTENT (IN)       :: albed
   REAL (KIND=dp),                   INTENT (IN)       :: ggHG
   REAL (KIND=dp), DIMENSION(0:npo), INTENT (IN)       :: ssour
   REAL (KIND=dp),                   INTENT (IN)       :: Im, Ip
   REAL (KIND=dp), DIMENSION(-mleg:mleg,0:lleg), INTENT (IN) :: rlmi
   REAL (KIND=dp), DIMENSION(-mleg:mleg,0:npo), INTENT (OUT) :: Yit

   REAL (KIND=dp)                                      :: Cm, Cp
   REAL (KIND=dp), DIMENSION(0:npo)                    :: lambd
   REAL (KIND=dp), DIMENSION(0:npo)                    :: phid
   REAL (KIND=dp), DIMENSION(0:npo)                    :: Sphid, ESphid
   REAL (KIND=dp), DIMENSION(0:npo)                    :: psid
   REAL (KIND=dp), DIMENSION(0:npo)                    :: Ephim, Ephip
   REAL (KIND=dp), DIMENSION(0:npo)                    :: Dphim, Dphip
   REAL (KIND=dp), DIMENSION(0:npo)                    :: fm, fp
   REAL (KIND=dp), DIMENSION(0:npo)                    :: f0, f1

   INTEGER                                             :: n, m
   REAL (KIND=dp)                                      :: aa11, aa12, aa21, aa22
   REAL (KIND=dp)                                      :: bb1, bb2
   REAL (KIND=dp)                                      :: det, d1, d2

   lambd(0:npo) = SQRT((1.0_dp - albed + alphg(0:npo)) / (3.0_dp * (1.0_dp - albed * ggHG + alphg(0:npo))))
   phid(0:npo)  = SQRT(3.0_dp * (1.0_dp - albed + alphg(0:npo)) * (1.0_dp - albed * ggHG + alphg(0:npo)))
   psid(0:npo)  = ssour(0:npo) / lambd(0:npo)

   Sphid(0) = 0.0_dp
   DO n = 1, npo
      Sphid(n) = 0.5_dp * (phid(n-1) + phid(n)) * dtl(n)
   ENDDO
   ESphid(:) = EXP(-Sphid(:))

   Ephim(0) = 0.0_dp
   DO n = 1, npo
      Ephim(n) = Ephim(n-1) + Sphid(n)
   ENDDO
   Ephip(npo) = 0.0_dp
   DO n = npo-1, 0, -1
      Ephip(n) = Ephip(n+1) + Sphid(n+1)
   ENDDO
   Ephim(:) = EXP(-Ephim(:))
   Ephip(:) = EXP(-Ephip(:))

   psid(0:npo) = psid(0:npo) / phid(0:npo)
   Dphim(0) = 0.0_dp
   DO n = 1, npo
      d1 = 1.0_dp - ESphid(n) - Sphid(n)
      d2 = 1.0_dp - ESphid(n) - Sphid(n) * ESphid(n)
      Dphim(n) = Esphid(n) * Dphim(n-1) + (psid(n) * d2 - psid(n-1) * d1) / Sphid(n)
   ENDDO

   Dphip(npo) = 0.0_dp
   DO n = npo-1, 0, -1
      d1 = 1.0_dp - ESphid(n) - Sphid(n)
      d2 = 1.0_dp - ESphid(n) - Sphid(n) * ESphid(n)
      Dphip(n) = Esphid(n) * Dphip(n+1) + (psid(n) * d2 - psid(n+1) * d1) / Sphid(n+1)
   ENDDO

   aa11 = (1.0_dp - SQRT(3.0_dp) * lambd(0))     * Ephip(0)
   aa12 =  1.0_dp + SQRT(3.0_dp) * lambd(0)
   aa21 =  1.0_dp + SQRT(3.0_dp) * lambd(npo)
   aa22 = (1.0_dp - SQRT(3.0_dp) * lambd(npo))   * Ephim(npo)
   bb1  =  2.0_dp * Im - ((1.0_dp - SQRT(3.0_dp) * lambd(0))   * Dphip(0))
   bb2  =  2.0_dp * Ip - ((1.0_dp - SQRT(3.0_dp) * lambd(npo)) * Dphim(npo))
   det  = aa11 * aa22 - aa21 * aa12
   d1   = bb1 * aa22 - bb2 * aa12
   d2   = aa11 * bb2 - aa21 * bb1
   Cp   = d1 / det
   Cm   = d2 / det

   fm(0:npo) = Cm * Ephim(0:npo) + Dphim(0:npo)
   fp(0:npo) = Cp * Ephip(0:npo) + Dphip(0:npo)

   f0(0:npo) = 0.5_dp * (fp(0:npo) + fm(0:npo))
   f1(0:npo) = 0.5_dp * (fp(0:npo) - fm(0:npo)) * lambd(0:npo)

   DO m = -mleg,-1
      Yit(m,0:npo) = rlmi(m,0) * f0(0:npo) + rlmi(m,1) * f1(0:npo)
   ENDDO
   DO m = 1,mleg
      Yit(m,0:npo) = rlmi(m,0) * f0(0:npo) + rlmi(m,1) * f1(0:npo)
   ENDDO

END SUBROUTINE EDDINGTON

END MODULE PXDR_TRANSFER
